import java.util.* ;

// -----------------------------------------------------------------------------

class CheckChemistry{

   static String checkValencies( List<Atom> molecule, String type ){

      List<Bond> bonds = new ArrayList<Bond>() ;

      for( Atom atom : molecule ){
         for( Bond bond : atom ) if( bond.atom1 == atom ) bonds.add( bond ) ;
      }

      for( Bond bond : bonds ){
         if( bond.type.equals( "metal" ) ){
            bond.atom1.remove( bond ) ;
            bond.atom2.remove( bond ) ;
         }
         else if( type.equals( "nonRestrictive" ) ){
            if( bond.atom1.symbol.equals( "H" ) ) bond.atom2.remove( bond ) ;
            if( bond.atom2.symbol.equals( "H" ) ) bond.atom1.remove( bond ) ;
         }
      }

      float epsilon = 0.26f ;

      List<Atom> errorList = new ArrayList<Atom>( ) ;
      for( Atom atom : molecule ){

         float bondOrders = 0 ;
         for( Bond bond : atom ) bondOrders += getBondOrder( bond ) ;
         bondOrders -= atom.charge ;

         if( atom.symbol.equals( "H" ) ){
            if( bondOrders != 1 || atom.size( ) > 1 ){
               errorList.add( atom ) ;
            }
         }
         else if(
            atom.symbol.equals( "F" ) ||
            atom.symbol.equals( "Cl" ) ||
            atom.symbol.equals( "Br" ) ||
            atom.symbol.equals( "I" )
         ){
            if( bondOrders > 7 || atom.size( ) > 7 ){
               errorList.add( atom ) ;
            }
         }
         else if( atom.symbol.equals( "O" ) ){
            float value = 2.5f ;
            if( type.equals( "restrictive" ) ) value = 2.0f ;
            value += epsilon ;
            if( bondOrders > value || atom.size( ) > 2 ){
               errorList.add( atom ) ;
            }
         }
         else if( atom.symbol.equals( "C" ) ){
            float value = 5.0f ;
            if( type.equals( "restrictive" ) ) value = 4.5f ;
            value += epsilon ;
            if( bondOrders > value || atom.size( ) > 4 ){
               errorList.add( atom ) ;
            }
         }
         else if( atom.symbol.equals( "N" ) ){
            if( bondOrders > ( 5 + epsilon ) || atom.size( ) > 4 ){
               errorList.add( atom ) ;
            }
         }
         else if( atom.symbol.equals( "P" ) ){
            float value = 6.0f ;
            if( type.equals( "restrictive" ) ) value = 5.0f ;
            value += epsilon ;
            if( bondOrders > value || atom.size( ) > 4 ){
               errorList.add( atom ) ;
            }
         }
         else if( atom.symbol.equals( "S" ) ){
            float value = 7.0f ;
            if( type.equals( "restrictive" ) ) value = 6.0f ;
            value += epsilon ;
            if( bondOrders > value || atom.size( ) > 4 ){
               errorList.add( atom ) ;
            }
         }
         else if( atom.symbol.equals( "Se" ) ){
            // is Se the same as S ???
            if( bondOrders > 6 || atom.size( ) > 4 ){
               errorList.add( atom ) ;
            }
         }
      }

      for( Bond bond : bonds ){
         if( bond.type.equals( "metal" ) ){
            bond.atom1.add( bond ) ;
            bond.atom2.add( bond ) ;
         }
         else if( type.equals( "nonRestrictive" ) ){
            if( bond.atom1.symbol.equals( "H" ) ) bond.atom2.add( bond ) ;
            if( bond.atom2.symbol.equals( "H" ) ) bond.atom1.add( bond ) ;
         }
      }

      if( errorList.size() == 0 ) return null ;

      String bsn = System.getProperty( "line.separator" ) ;
      String exceptionString = "" ;
      for( Atom atom : errorList ){
         exceptionString += bsn + atom.key4 + " has incorrect valency" ;
      }
      return exceptionString ;
   }

   static float getBondOrder( Bond bond ){
      float bondOrder = 0 ;

      if( bond.type.equals( "single" ) ){
         bondOrder = 1.0f ;
      }
      else if( bond.type.equals( "double" ) ){
         bondOrder = 2.0f ;
      }
      else if( bond.type.equals( "triple" ) ){
         bondOrder = 3.0f ;
      }
      else if( bond.type.startsWith( "delo" ) || bond.type.startsWith( "arom" ) ){
         float neighbourCharge = 0 ;

         if( bond.atom1.symbol.equals( "O" ) ){
            neighbourCharge = bond.atom1.charge ;
         }
         else if( bond.atom2.symbol.equals( "O" ) ){
            neighbourCharge = bond.atom2.charge ;
         }
         else{
            neighbourCharge = bond.atom1.charge ;              //:: ????
         }

         if( neighbourCharge < 0 && neighbourCharge > -1 ){
            bondOrder = 2.0f + neighbourCharge ;
         }
         else if( neighbourCharge > 0 && neighbourCharge < 1 ){
            bondOrder += 1.0f + neighbourCharge ;
         }
         else{
            bondOrder += 1.5f + neighbourCharge ;
         }
      }
      return bondOrder ;
   }
}
// -----------------------------------------------------------------------------

class CorrectBonds{

   static String compare( List<Atom> monOld, List<Atom> monNew ){
      String changes = "" ;

      for( Atom atomNew : monNew ){
         atomNew.coupled = atomNew ;
      }
      for( Atom atomOld : monOld ){
         Atom atomNew = atomOld.coupled ;
         if( atomNew == atomOld ) throw new NullPointerException() ;
         atomNew.coupled = atomOld ;
      }
      for( Atom atomNew : monNew ){
         Atom atomOld = atomNew.coupled ;
         if( atomOld == atomNew ) throw new NullPointerException() ;
      }

      for( Atom atomNew : monNew ){
         Atom atomOld = atomNew.coupled ;
         if( ! atomOld.charge.equals( atomNew.charge ) ){
            changes += String.format(
                  "Atom %-4s    charge changed from %7.2f to %7.2f",
                  atomNew.key4, atomOld.charge, atomNew.charge
            ) ;
            changes += Env.bsn ;
         }
      }

      for( Atom atomNew1 : monNew ){
         for( Bond bondNew : atomNew1 ){
            if( bondNew.atom1 == atomNew1 ){
               Atom atomNew2 = bondNew.another( atomNew1 ) ;
               Atom atomOld1 = atomNew1.coupled ;
               Atom atomOld2 = atomNew2.coupled ;
               Bond bondOld = null ;
               for( Bond bond : atomOld1 ){
                  bondOld = bond ;
                  if( bondOld.another( atomOld1 ) == atomOld2 ) break ;
                  bondOld = null ;
               }
               if( bondOld == null ) throw new NullPointerException() ;
            
               if( ! bondOld.type.equals( bondNew.type ) ){
                  changes += String.format(
                        "Bond %-4s-%-4s type changed from %7s to %7s",
                        atomNew1.key4, atomNew2.key4, bondOld.type, bondNew.type
                  ) ;
                  changes += Env.bsn ;
               }
            }
         }
      }

      return changes ;
   }

   static boolean isChange ;

   static boolean correctBondOrders( List<Atom> molecule ){

      List<Bond> bonds = new ArrayList<Bond>() ;

      for( Atom atom : molecule ){
         for( Bond bond : atom ) if( bond.atom1 == atom ) bonds.add( bond ) ;
      }

      for( Bond bond : bonds ){
         if( bond.type.equals( "metal" ) ){
            bond.atom1.remove( bond ) ;
            bond.atom2.remove( bond ) ;
         }
      }

      boolean corrected = false ;
      isChange = true ;
      for( int i = 0 ; i < 100 ; i++ ){
         if( isChange == false ) break ;
         isChange = false ;

         for( Atom atom1 : molecule ){
            if(
               atom1.symbol.equals( "H" ) ||
               atom1.symbol.equals( "F" ) ||
               atom1.symbol.equals( "Cl" ) ||
               atom1.symbol.equals( "Br" ) ||
               atom1.symbol.equals( "I" )
            ){
               for( Bond bond : atom1 ){
                  if( ! bond.type.equals( "single" ) ){
                     bond.type = "single" ;
                     isChange = true ;
                  }
               }
            }

            else if( atom1.symbol.equals( "O" ) ){
               sortOxygens( atom1 ) ;
            }

            else if( atom1.symbol.equals( "C" ) ){
               sortCarbons( atom1 ) ;
            }

//@@@@    /*  /// To avoid problems with Arg
            else if( atom1.symbol.equals( "N" ) ){
               sortNitrogens( atom1 ) ;
            }
//@@@@    */

            else if( atom1.symbol.equals( "P" ) ){
               sortPhosphors( atom1 ) ;
            }

            else if( atom1.symbol.equals( "S" ) ){
               sortSulphurs( atom1 ) ;
            }

            // selenium, etc???
         }
         corrected = corrected || isChange ;
      }

      for( Bond bond : bonds ){
         if( bond.type.equals( "metal" ) ){
            bond.atom1.add( bond ) ;
            bond.atom2.add( bond ) ;
         }
      }

      return corrected ;
   }
   // --------------------------------------------------------------------------

   static void sortOxygens( Atom atom1 ){

      int noNeighbours = atom1.size( ) ;

      if( noNeighbours == 1 ){
         // sorted by other functions?
      }

      else if( noNeighbours == 2 ){

         for( Bond bond1 : atom1 ){
            if( ! bond1.type.equals( "single" ) ){
               bond1.type = "single" ;
               isChange = true ;
            }
         }
      }
   }
   // --------------------------------------------------------------------------

   static void sortCarbons( Atom atom1 ){

      int noNeighbours = atom1.size( ) ;

      if( noNeighbours == 1 ){
         // do nothing
      }

      else if( noNeighbours == 2 ){
         // do nothing
      }

      else if( noNeighbours == 3 ){

         int noSingleBonds = 0 ;
         int noDelocBonds = 0 ;
         int noDoubleBonds = 0 ;

         // calculate the no of single, deloc, double and triple bonds
         for( Bond bond1 : atom1 ){
            if( bond1.type.equals( "single" ) ){
               noSingleBonds++ ;
            }
            else if( bond1.type.startsWith( "delo" ) || bond1.type.startsWith( "arom" ) ){
               noDelocBonds++ ;
            }
            else if( bond1.type.equals( "double" ) ){
               noDoubleBonds++ ;
            }
         }

         // double to deloc
         if( noSingleBonds == 1 ){

            for( Bond bond1 : atom1 ){

               if( bond1.type.startsWith( "double" ) ){
                  bond1.type = "deloc" ;
                  isChange = true ;
               }
            }
         }

         // deloc to double
         else if( noSingleBonds == 2 && noDelocBonds == 1 ){

            for( Bond bond1 : atom1 ){

               if( bond1.type.startsWith( "delo" ) || bond1.type.startsWith( "arom" ) ){
                  bond1.type = "double" ;
                  isChange = true ;
               }
            }
         }

         // ifmore than 1 deloc or double ...
         if( ( noDelocBonds + noDoubleBonds ) > 1 ){

            // -C( O )O case
            int noLoneOxygens = 0 ;
            int noLoneOxygensDoubleOrDeloc = 0 ;

            for( Bond bond1 : atom1 ){
               Atom atom2 = bond1.another( atom1 ) ;

               if( atom2.symbol.equals( "O" ) && atom2.size() == 1 ){
                  noLoneOxygens++ ;
                  if(
                     bond1.type.equals( "double" ) ||
                     bond1.type.startsWith( "delo" ) ||
                     bond1.type.startsWith( "arom" )
                  ){
                     noLoneOxygensDoubleOrDeloc++ ;
                  }
               }
            }

            if( noLoneOxygens == 1 ){
               // do nothing
            }

            else if( noLoneOxygens == 2 ){

               // If one of the double or deloc bonds is with one of these oxygens then
               // convert C-O bonds to deloc and add to oxygens charge -0.5

               if( noLoneOxygensDoubleOrDeloc > 0 ){

                  for( Bond bond1 : atom1 ){
                     Atom atom2 = bond1.another( atom1 ) ;

                     if( atom2.symbol.equals( "O" ) ){

                        if( ! bond1.type.startsWith( "delo" ) ){
                           bond1.type = "deloc" ;
                           isChange = true ;
                        }

                        if( atom2.charge != -0.5f ){
                           atom2.charge = -0.5f ;
                           isChange = true ;
                        }
                     }
                  }
               }
            }

            else if( noLoneOxygens == 3 ){

               // Convert all bonds to deloc and add charge to oxygens -0.66
               for( Bond bond1 : atom1 ){
                  Atom atom2 = bond1.another( atom1 ) ;

                  if( ! bond1.type.startsWith( "delo" ) ){
                     bond1.type = "deloc" ;
                     isChange = true ;
                  }

                  if( atom2.charge != -0.66f ){
                     atom2.charge = -0.66f ;
                     isChange = true ;
                  }
               }
            }

//@@@@    /*  /// To avoid problems with Arg
            int noLoneNitrogensDouble = 0 ;

            for( Bond bond1 : atom1 ){
               Atom atom2 = bond1.another( atom1 ) ;

               if( atom2.symbol.equals( "N" ) && atom2.size() == 1 && bond1.type.equals( "double" ) ){
                  noLoneNitrogensDouble++ ;
               }
            }

            // Arg NC( N )N case
            if( noLoneNitrogensDouble == 2 ){

               for( Bond bond1 : atom1 ){
                  Atom atom2 = bond1.another( atom1 ) ;

                  if( atom2.symbol.equals( "N" ) ){

                     if( ! bond1.type.startsWith( "delo" ) ){
                        bond1.type = "deloc" ;
                        isChange = true ;
                     }

                     if( atom2.charge != 0.5f ){
                        atom2.charge = 0.5f ;
                        isChange = true ;
                     }
                  }
               }
            }
//@@@@    */
         }
      }

      else if( noNeighbours == 4 ){

         for( Bond bond1 : atom1 ){
            if( ! bond1.type.equals( "single" ) ){
               bond1.type = "single" ;
               isChange = true ;
            }
         }
      }
   }
   // --------------------------------------------------------------------------

   static void sortNitrogens( Atom atom1 ){

      int noNeighbours = atom1.size( ) ;

      if( noNeighbours == 1 ){

         for( Bond bond1 : atom1 ){
            Atom atom2 = bond1.another( atom1 ) ;

            if( atom2.symbol.equals( "C" ) && bond1.type.equals( "single" ) ){

               if( atom1.charge != 1.0f ){
                  atom1.charge = 1.0f ;
                  isChange = true ;
               }
            }
         }
      }

      else if( noNeighbours == 2 ){

         Bond bond1A = atom1.get( 0 ) ;
         Bond bond1B = atom1.get( 1 ) ;

//       Atom atom2A = bond1A.another( atom1 ) ;
//       Atom atom2B = bond1B.another( atom1 ) ;

         if( bond1A.type.equals( "single" ) && bond1B.type.equals( "deloc" ) ){
            bond1B.type = "double" ;
            isChange = true ;
         }

         if( bond1B.type.equals( "single" ) && bond1A.type.equals( "deloc" ) ){
            bond1A.type = "double" ;
            isChange = true ;
         }

         if( bond1A.type.equals( "double" ) && bond1B.type.equals( "double" ) ){
            atom1.charge = 1.0f ;
            isChange = true ;
         }
      }

      else if( noNeighbours == 3 ){

         if( atom1.charge <= 0.5 ){

            for( Bond bond1 : atom1 ){
               if( ! bond1.type.equals( "single" ) ){
                  bond1.type = "single" ;
                  isChange = true ;
               }
            }

            if( atom1.charge != 0.0 ){
               atom1.charge = 0.0f ;
               isChange = true ;
            }
         }

         // what happens when atomPartialCharge in between 0.4 and 0.6???
         else if( atom1.charge > 0.5 ){

            if( atom1.charge != 1.0 ){
               atom1.charge = 1.0f ;
               isChange = true ;
            }

            int noSingleBonds = 0 ;
            int noDelocBonds = 0 ;
            int noDoubleBonds = 0 ;

            for( Bond bond1 : atom1 ){
               if( bond1.type.equals( "single" ) ){
                  noSingleBonds++ ;
               }
               else if( bond1.type.startsWith( "delo" ) || bond1.type.startsWith( "arom" ) ){
                  noDelocBonds++ ;
               }
               else if( bond1.type.equals( "double" ) ){
                  noDoubleBonds++ ;
               }
            }

            if( noSingleBonds == 2 && noDelocBonds == 1 ){

               for( Bond bond1 : atom1 ){
                  if( bond1.type.startsWith( "delo" ) || bond1.type.startsWith( "arom" ) ){

                     if( ! bond1.type.equals( "double" ) ){
                        bond1.type = "double" ;
                        isChange = true ;
                     }
                  }
               }
            }

            else if( noSingleBonds == 1 && noDoubleBonds == 2 ){

               for( Bond bond1 : atom1 ){
                  if( bond1.type.equals( "double" ) ){
                     bond1.type = "deloc" ;
                     isChange = true ;
                  }
               }
            }
         }
      }

      else if( noNeighbours == 4 ){

         for( Bond bond1 : atom1 ){
            if( ! bond1.type.equals( "single" ) ){
               bond1.type = "single" ;
               isChange = true ;
            }
         }

         if( atom1.charge != 1.0f ){
            atom1.charge = 1.0f ;
            isChange = true ;
         }
      }
   }
   // --------------------------------------------------------------------------

   static void sortPhosphors( Atom atom1 ){

      int noNeighbours = atom1.size( ) ;

      if( noNeighbours == 1 ){
         // do nothing
      }

      else if( noNeighbours == 2 ){
         // do nothing
      }

      else if( noNeighbours == 3 ){

         for( Bond bond1 : atom1 ){
            if( ! bond1.type.equals( "single" ) ){
               bond1.type = "single" ;
               isChange = true ;
            }
         }
      }

      else if( noNeighbours == 4 ){

         int noLoneOxygens = 0 ;

         for( Bond bond1 : atom1 ){
            Atom atom2 = bond1.another( atom1 ) ;

            if( atom2.symbol.equals( "O" ) && atom2.size() == 1 ){
               noLoneOxygens++ ;
            }
         }

         for( Bond bond1 : atom1 ){
            Atom atom2 = bond1.another( atom1 ) ;

            if( noLoneOxygens == 1 ){

               if( atom2.symbol.equals( "O" ) && atom2.size() == 1 ){

                  if( ! bond1.type.equals( "double" ) ){
                     bond1.type = "double" ;
                     isChange = true ;
                  }

                  if( atom2.charge != 0.0f ){
                     atom2.charge = 0.0f ;
                     isChange = true ;
                  }
               }

               else{

                  if( ! bond1.type.equals( "single" ) ){
                     bond1.type = "single" ;
                     isChange = true ;
                  }

                  if( atom2.charge != 0.0f ){
                     atom2.charge = 0.0f ;
                     isChange = true ;
                  }
               }
            }

            else if( noLoneOxygens == 2 ){

               if( atom2.symbol.equals( "O" ) && atom2.size() == 1 ){

//-- bug          if( ! bond1.type.equals( "single" ) ){
                  if( ! bond1.type.equals( "deloc" ) ){
//--
                     bond1.type = "deloc" ;
                     isChange = true ;
                  }

                  if( atom2.charge != -0.50f ){
                     atom2.charge = -0.50f ;
                     isChange = true ;
                  }
               }

               else{

                  if( ! bond1.type.equals( "single" ) ){
                     bond1.type = "single" ;
                     isChange = true ;
                  }

                  if( atom2.charge != 0.0f ){
                     atom2.charge = 0.0f ;
                     isChange = true ;
                  }
               }
            }

            else if( noLoneOxygens == 3 ){

               if( atom2.symbol.equals( "O" ) && atom2.size() == 1 ){

                  if( ! bond1.type.startsWith( "delo" ) ){
                     bond1.type = "deloc" ;
                     isChange = true ;
                  }

                  if( atom2.charge != -0.66f ){
                     atom2.charge = -0.66f ;
                     isChange = true ;
                  }
               }

               else{

                  if( ! bond1.type.equals( "single" ) ){
                     bond1.type = "single" ;
                     isChange = true ;
                  }

                  if( atom2.charge != 0.0f ){
                     atom2.charge = 0.0f ;
                     isChange = true ;
                  }
               }
            }

            else if( noLoneOxygens == 4 ){

               if( ! bond1.type.startsWith( "delo" ) ){
                  bond1.type = "deloc" ;
                  isChange = true ;
               }

               if( atom2.charge != -0.75f ){
                  atom2.charge = -0.75f ;
                  isChange = true ;
               }
            }
         }
      }
   }
   // --------------------------------------------------------------------------

   static void sortSulphurs( Atom atom1 ){

      int noNeighbours = atom1.size( ) ;

      if( noNeighbours == 1 ){
         // cannot sort out?
      }

      else if( noNeighbours == 2 ){

         for( Bond bond1 : atom1 ){
            if( ! bond1.type.equals( "single" ) ){
               bond1.type = "single" ;
               isChange = true ;
            }
         }
      }

      // If charge on S is zero then one of the bonds should be double.
      // If the charge is 1 then bond orders are single.
      else if( noNeighbours == 3 ){

         if( atom1.charge == 1.0f ){

            for( Bond bond1 : atom1 ){
               if( ! bond1.type.equals( "single" ) ){
                  bond1.type = "single" ;
                  isChange = true ;
               }
            }
         }

         // We should check bond orders. there should be two singles and one double. Otherwise we have a problem.
         else{
         }
      }

      else if( noNeighbours == 4 ){

         int noLoneOxygens = 0 ;

         for( Bond bond1 : atom1 ){
            Atom atom2 = bond1.another( atom1 ) ;

            if( atom2.symbol.equals( "O" ) && atom2.size() == 1 ){
               noLoneOxygens++ ;
            }
         }

         for( Bond bond1 : atom1 ){
            Atom atom2 = bond1.another( atom1 ) ;

            if( noLoneOxygens == 1 ){

               if( atom2.symbol.equals( "O" ) && atom2.size() == 1 ){

                  if( ! bond1.type.equals( "double" ) ){
                     bond1.type = "double" ;
                     isChange = true ;
                  }

                  if( atom2.charge != 0.0f ){
                     atom2.charge = 0.0f ;
                     isChange = true ;
                  }
               }

               else{

                  //bond1.type = "single" ;

                  if( atom2.charge != 0.0f ){
                     atom2.charge = 0.0f ;
                     isChange = true ;
                  }
               }
            }

            else if( noLoneOxygens == 2 ){

               if( atom2.symbol.equals( "O" ) && atom2.size() == 1 ){

                  if( ! bond1.type.equals( "double" ) ){
                     bond1.type = "double" ;
                     isChange = true ;
                  }

                  if( atom2.charge != 0.0f ){
                     atom2.charge = 0.0f ;
                     isChange = true ;
                  }
               }

               else{

                  if( ! bond1.type.equals( "single" ) ){
                     bond1.type = "single" ;
                     isChange = true ;
                  }

                  if( atom2.charge != 0.0f ){
                     atom2.charge = 0.0f ;
                     isChange = true ;
                  }
               }
            }

            else if( noLoneOxygens == 3 ){

               if( atom2.symbol.equals( "O" ) && atom2.size() == 1 ){

                  if( ! bond1.type.startsWith( "delo" ) ){
                     bond1.type = "deloc" ;
                     isChange = true ;
                  }

                  if( atom2.charge != -0.33f ){
                     atom2.charge = -0.33f ;
                     isChange = true ;
                  }

               }
//             else{
//                bond1.type = "single" ;
//                atom2.charge = 0.0f ;
//             }
            }

            else if( noLoneOxygens == 4 ){

               if( ! bond1.type.startsWith( "delo" ) ){
                  bond1.type = "deloc" ;
                  isChange = true ;
               }

               if( atom2.charge != -0.50f ){
                  atom2.charge = -0.50f ;
                  isChange = true ;
               }
            }
         }
      }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

