
import java.io.* ;
import java.util.* ;

// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class CifFile extends CifParser{

   final static CifTemplate template = new CifTemplate() ;
   private List<CifColumn> table = new ArrayList<CifColumn>() ;
   private CifColumn junk = new CifColumn() ;
   private CifLoop loop ;

   CifFile( CharArray cha ){
      parse( cha ) ;
   }

   protected void dataTag(){
      CifData d = template.get( item ) ;
      int i = item.lastIndexOf( "_" ) ;
      if( d == null && i > 0 ){
         d = template.get( item.substring( 0, 1 + i ) ) ;
         i = item.lastIndexOf( "_", i - 1 ) ;
         if( d == null && i > 0 ){
            d = template.get( item.substring( 0, 1 + i ) ) ;
         }
      }
      data = null ;
      if( d != null ) data = new CifData( d ) ;
      super.dataTag() ;
   }

   protected void loopTag(){
      super.loopTag() ;
      loop = data == null ? null : data.get( loopKey ) ;
   }

   protected void clmnTag(){
      super.clmnTag() ;
      CifColumn clmn = loop == null ? null : loop.get( clmnKey ) ;
      table.add( clmn == null ? junk : clmn ) ;
   }

   protected void value(){
      super.value() ;
      table.get( ic ).addString( item ) ;
   }

   protected void loopEnd(){
      super.loopEnd() ;
      table.clear() ;
      junk.clear() ;
      if( loop != null ){ loop.trim( nr ) ; loop = null ; }
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class CifTemplate extends CifParser{

   public CifTemplate(){
      InputStream is = getClass().getResourceAsStream( "/resources/template" ) ;
      parse( new CharArray( is ) ) ;
   }

   private List<String> clmnKeys = new ArrayList<String>() ;
   private List<String> clmnFmts = new ArrayList<String>() ;

   protected void dataTag(){
      data = new CifData() ;
      super.dataTag() ;
   }

   protected void loopTag(){
      super.loopTag() ;
   }

   protected void clmnTag(){
      super.clmnTag() ;
      clmnKeys.add( clmnKey ) ;
   }

   protected void value(){
      super.value() ;
      clmnFmts.add( item ) ;
   }

   protected void loopEnd(){
      super.loopEnd() ;
      data.put( loopKey, new CifLoop( loopKey, clmnKeys, clmnFmts ) ) ;
      clmnKeys.clear() ;
      clmnFmts.clear() ;
      loopKey = null ;
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class CifData extends LinkedHashMap<String,CifLoop> implements Iterable<CifLoop>{

   public int first, last ;

   CifData(){}

   CifData( CifData cd ){
      for( Map.Entry<String,CifLoop> me : cd.entrySet() ){
         put( me.getKey(), new CifLoop( me.getValue() ) ) ;
      }
   }

   public String toString(){
      CharArray aw = new CharArray() ;
      appendTo( aw ) ;
      return aw.toString() ;
   }

   public void appendTo( CharArray aw ){
      for( CifLoop loop : values() ) loop.appendTo( aw ) ;
   }

   public boolean isVoid(){
      for( CifLoop loop : values() ) if( ! loop.isVoid() ) return false ;
      return true ;
   }

   public void clear(){
      for( CifLoop loop : values() ) loop.clear() ;
   }

   public Iterator<CifLoop> iterator(){
      return values().iterator() ;
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class CifLoop extends LinkedHashMap<String,CifColumn>{
   private static final String bsn = System.getProperty( "line.separator" ) ;

   int[] blockSizes ;
   int[] blockRef ;

   List<CifColumn> columns ;
   int ic = 0 ;
   int jc = 0 ;
   int nc = 0 ;
   int ir = 0 ;
   int nr = 0 ;

   CifLoop( String key, List<String> keys, List<String> formats ){
      int n = keys.size() ;
      if( 2* n != formats.size() ) throw new NullPointerException() ;
      for( int i = 0 ; i < n ; i++ ){
         String s = keys.get( i ) ;
         put( s, new CifColumn( key + s, formats.get( i ) ) ) ;
      }

      blockRef = new int[n] ;
      Map<String,Integer> blockMap = new LinkedHashMap<String,Integer>() ;
      int cou = 0 ;
      for( int i = 0 ; i < n ; i++ ){
         String s = formats.get( n + i ) ;
         Integer ref = blockMap.get( s ) ;
         if( ref == null ){
            ref = new Integer( cou++ ) ;
            blockMap.put( s, ref ) ;
         }
         blockRef[i] = ref ;
      }
      blockSizes = new int[cou] ;

      columns = new ArrayList<CifColumn>( values() ) ;
      nc = columns.size() ;
   }

   CifLoop( CifLoop ct ){
      for( Map.Entry<String,CifColumn> me : ct.entrySet() ){
         put( me.getKey(), new CifColumn( me.getValue() ) ) ;
      }
      columns = new ArrayList<CifColumn>( values() ) ;
      nc = columns.size() ;
      blockSizes = ct.blockSizes ;
      blockRef = ct.blockRef ;
   }

   public void add( Object o ){
      if( ic != 0 || ir != 0 ) throw new NullPointerException() ;
      columns.get( jc++ ).add( o ) ;
      if( jc == nc ){ jc = 0 ; nr++ ; }
   }

//@@@ arrange as an iterator
   public boolean hasNextNoCycle(){
      if( jc != 0 ) throw new NullPointerException() ;
      if( ir < nr && ic < nc ) return true ;
      return false ;
   }

   public boolean hasNext(){
      if( jc != 0 ) throw new NullPointerException() ;
      if( ir < nr && ic < nc ) return true ;
      ir = ic = 0 ;
      return false ;
   }

   public Object next(){
      if( jc != 0 ) throw new NullPointerException() ;
      Object o = columns.get( ic++ ).get( ir ) ;
      if( ic == nc ){ ic = 0 ; ir++ ; }
      return o ;
   }

   public void trimBlocks(){
      for( int i = 0, n = blockSizes.length ; i < n ; i++ ) blockSizes[i] = 0 ;
      for( int i = 0, n = size() ; i < n ; i++ ){
         if( ! columns.get(i).isVoid() ) blockSizes[blockRef[i]]++ ;
      }
      for( int i = 0, n = size() ; i < n ; i++ ){
         if( blockSizes[blockRef[i]] == 0 ) columns.get(i).clear() ;
      }
   }

   void trim( int kr ){
      int kc = 0 ;
      for( CifColumn col : values() ){
         if( col.size() == kr ){
            if( col.isVoid() ) col.clear() ;
            else kc++ ;
         }
         else if( col.size() != 0 ){
            System.err.println( "kr = " + kr + ", nr = " + nr ) ;
            throw new NullPointerException() ;
         }
      }
      nr = kr ;
      if( kc == 0 ) nr = 0 ;
   }

   public boolean isVoid(){
      for( CifColumn col : values() ) if( ! col.isVoid() ) return false ;
      return true ;
   }

   public void clear(){
      for( CifColumn col : values() ) col.clear() ;
      nr = 0 ;
   }

   public String toString(){
      CharArray aw = new CharArray() ;
      appendTo( aw ) ;
      return aw.toString() ;
   }

   void appendTo( CharArray aw ){
      if( this.isVoid() ) return ;
      trimBlocks() ;

      aw.append( bsn ).append( "loop_" ).append( bsn ) ;
      for( CifColumn col : values() ) if( ! col.isEmpty() ) aw.append( col.asString() ).append( bsn ) ;
      for( int ir = 0 ; ir < nr ; ir++ ){
         for( CifColumn col : values() ) if( ! col.isEmpty() ) aw.append( col.asString( ir ) ) ;
         aw.append( bsn ) ;
      }
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class CifColumn extends ArrayList<Object>{

   String tag = "junk" ;
   String format = " %s" ;
   String NA = " ." ;
   int type = 0 ;

   CifColumn(){}

   CifColumn( String s, String f ){
      tag = s ;

      if( f.indexOf( "s" ) >= 0 ) type = 0 ;
      else if( f.indexOf( "f" ) >= 0 ) type = 1 ;
      else if( f.indexOf( "d" ) >= 0 ) type = 2 ;
      else throw new NullPointerException() ;

      int j = f.indexOf( ':' ) ;
      if( j >= 0 ){
         int i = f.indexOf( 'n' ) ;
         if( i < 1 && i + 1 < j && j < i + 4 == false ) throw new NullPointerException() ;
         char[] a = new char[Integer.parseInt( f.substring( i + 1, j ) )] ;
         Arrays.fill( a, ' ' ) ;
         format = new String( a ) + "%" + f.substring( j + 1 ) ;
         if( i == 0 ) format = System.getProperty( "line.separator" ) + format ;
      }
      else format = " %" + f ;

      NA = String.format( format.replace('d','s').replace('f','s'), "." ) ;
   }

   CifColumn( CifColumn ct ){
      format = ct.format ;
      type = ct.type ;
      tag = ct.tag ;
      NA = ct.NA ;
   }

   public Object get( int ir ){
      if( isEmpty() ) return null ;
      return super.get( ir ) ;
   }

   private int errCount = 0 ;
   void addString( String s ){
      Object o = s ;
      if( s.equals( "." ) || s.equals( "?" ) ){
         o = null ;
      }
      else if( type == 1 ){
         try{
            if( s.equals( "NaN" ) ) throw new Exception() ;
            o = Float.valueOf( s ) ;
         }
         catch( Exception e ){
            o = null ;
            if( errCount++ == 0 ) System.err.println( "WARNING: error parsing float: " + s ) ;
         }
      }
      else if( type == 2 ){
         try{
            o = Integer.valueOf( s ) ;
         }
         catch( Exception e ){
            o = null ;
            if( errCount++ == 0 ) System.err.println( "WARNING: error parsing int: " + s ) ;
         }
      }
      add( o ) ;
   }

   boolean isVoid(){
      for( Object o : this ) if( o != null ) return false ;
      return true ;
   }

   String asString(){
      return tag ;
   }

   String asString( int i ){
      Object o = get( i ) ;
      if( o != null && type == 0 ){
         if( ! ( o instanceof String ) ) throw new NullPointerException() ;
         String s = (String)o ;
         String q = null ;
         boolean b32 = s.indexOf( (char)32 ) >= 0 ;
         boolean b34 = s.indexOf( (char)34 ) >= 0 ;
         boolean b39 = s.indexOf( (char)39 ) >= 0 ;
         if( b34 && b39 ){
            q = String.valueOf( (char)39 ) ;
         }
         else if( b34 ){
            q = String.valueOf( (char)39 ) ;
         }
         else if( b39 ){
            q = String.valueOf( (char)34 ) ;
         }
         else if( b32 ){
            q = String.valueOf( (char)39 ) ;
         }
         if( q != null ) o = q + s + q ;
      }
      return o == null ? NA : String.format( format, o ) ;
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class CifParser extends LinkedHashMap<String,CifData>{

   protected CifData data ;
   protected String dataKey ;
   protected String loopKey ;
   protected String clmnKey ;
   protected String item ;
   protected int ic, nc, nr ;
   protected int current ;

   protected void dataTag(){
      dataKey = item ;
      if( data != null ){
         data.first = current ;
      }
   }

   protected void loopTag(){
      int i = item.indexOf( "." ) ;
      if( i < 2 ) throw new NullPointerException() ;
      loopKey = item.substring( 0, i + 1 ) ;
      nc = nr = 0 ;
   }

   protected void clmnTag(){
      if( ! item.startsWith( loopKey ) ) throw new NullPointerException() ;
      clmnKey = item.substring( loopKey.length() ) ;
      ic = nc++ ;
   }

   protected void value(){
      if( ++ic == nc ){ ic = 0 ; nr++ ; }
      clmnKey = null ;
   }

   protected void loopEnd(){
      if( ic != nc - 1 ) throw new NullPointerException() ;
   }

   protected void dataEnd(){
      if( data != null ){
         data.last = current ;
         put( dataKey, data ) ;
         data = null ;
      }
      dataKey = null ;
      loopKey = null ;
   }

   //-------------------------------------------------------------------------------------------
   //               \         |            |            |            |            |            |
   //   waiting for: \ read:  | data_XXX   |   loop_    |   _XXX     |    XXX     |    null    |
   //                 \       |            |            |            |            |            |
   //-------------------------------------------------------------------------------------------
   //   data          0       |          1 |          - |          0 |          0 |          - |
   //                         | Dt         |            |            |            |            |
   //-------------------------------------------------------------------------------------------
   //   loop          1       |          - |          2 |          - |          - |          - |
   //                         |            |            |            |            |            |
   //-------------------------------------------------------------------------------------------
   //   loop tag      2       |          - |          - |          3 |          - |          - |
   //                         |            |            | Lt, Ct     |            |            |
   //-------------------------------------------------------------------------------------------
   //   loop value    3       |          - |          - |          3 |          4 |          - |
   //                         |            |            | Ct         | Ce, Lv     |            |
   //-------------------------------------------------------------------------------------------
   //   loop end      4       |          1 |          2 |          - |          4 |          R |
   //                         | Le, De, Dt | Le         |            | Lv         | Le, De     |
   //-------------------------------------------------------------------------------------------

   public void parse( CharArray cha ){
      CifTokenizer ct = new CifTokenizer( cha ) ;
      int i = 0 ;
      while( true ){

         item = ct.nextToken() ;
         current = item == null ? ct.j : ct.j - item.length() ;

         if( item == null ){

            switch( i ){
               case 0:                     throw new NullPointerException() ;
               case 1:                     throw new NullPointerException() ;
               case 2:                     throw new NullPointerException() ;
               case 3:                     throw new NullPointerException() ;
               case 4:         loopEnd() ;
                               dataEnd() ; return ;
            }
         }
         else if( item.startsWith("data_") ){

            switch( i ){
               case 0: i = 1 ; dataTag() ; break ;
               case 1: i = 1 ; dataTag() ; break ;
//::           case 1:                     throw new NullPointerException() ;
               case 2:                     throw new NullPointerException() ;
               case 3:                     throw new NullPointerException() ;
               case 4: i = 1 ; loopEnd() ;
                               dataEnd() ;
                               dataTag() ; break ;
            }
         }
         else if( item.equals("loop_") ){

            switch( i ){
               case 0: i = 2             ; break ;
               case 1: i = 2             ; break ;
               case 2:                     throw new NullPointerException() ;
//Libmol       case 3:                     throw new NullPointerException() ;
               case 3: i = 2 ; loopEnd() ; break ;
               case 4: i = 2 ; loopEnd() ; break ;
            }
         }
         else if( item.startsWith("_") ){

            switch( i ){
               case 0:                     break ;
               case 1:                     throw new NullPointerException() ;
               case 2: i = 3 ; loopTag() ;
                               clmnTag() ; break ;
               case 3:         clmnTag() ; break ;
               case 4:                     throw new NullPointerException() ;
            }
         }
         else{

            switch( i ){
               case 0:                     break ;
               case 1:                     throw new NullPointerException() ;
               case 2:                     throw new NullPointerException() ;
               case 3: i = 4 ; value()   ; break ;
               case 4:         value()   ; break ;
            }
         }
      }
   }
}
// -----------------------------------------------------------------------------

class CifTokenizer{
   char[] cha ;
   int j = - 1 ;
   int n = - 1 ;

   public CifTokenizer( CharArray charArray ){
      n = charArray.length() ;
      cha = new char[n+1] ;
      for( int i = 0 ; i < n ; i++ ){
         cha[i] = charArray.charAt(i) ;
         if( cha[i] == (char) 9 ) cha[i] = ' ' ;
         if( cha[i] == (char)13 ) cha[i] = (char)10 ;
      }
      cha[n] = (char)10 ;
   }

   protected String nextToken(){
      if( j == n ) return null ;
      while( cha[++j] == ' ' || cha[j] == (char)10 || cha[j] == '#' ){
         if( cha[j] == '#' ) while( cha[++j] != (char)10 ){}
         if( j == n ) return null ;
      }

      char quo = ' ' ;
      if( cha[j] == (char)34 ) quo = (char)34 ;
      if( cha[j] == (char)39 ) quo = (char)39 ;
      int i = j ;

      if( quo == ' ' ){
         while( cha[++j] != ' ' && cha[j] != (char)10 ){}
         return new String( cha, i, j - i ) ;
      }
      else{
         while( cha[++j] == ' ' ){} ;
         while( ! ( cha[j] == quo & cha[++j] == ' ' ) && cha[j] != (char)10 ){}
         return new String( cha, i + 1, j - i - 2 ) ;
      }
   }
}
// -----------------------------------------------------------------------------

class CharArray implements CharSequence{
   char[] buf ;
   int count = 0 ;

   private char[] ArraysCopyOf( char[] cha1, int length ){
      char[] cha2 = new char[length] ;
      for( int i = 0, n = Math.min( cha1.length, length ) ; i < n ; i++ ) cha2[i] = cha1[i] ;
      return cha2 ;
   }

   public CharArray(){
      buf = new char[0] ;
   }

   public CharArray( CharSequence csq ){
      int n = csq.length() ;
      buf = new char[n] ;
      append( csq ) ;
   }

   public CharArray append( CharSequence csq ){
      int n = csq.length() ;
      if( count + n > buf.length ) buf = ArraysCopyOf( buf, count + n + buf.length ) ;
      for( int i = 0 ; i < n ; i++ ) buf[ count + i ] = csq.charAt( i ) ;
      count += n ;
      return this ;
   }

   public CharArray( File file ){
      int n = (int)file.length() ;
      buf = new char[n] ;
      append( file ) ;
   }

   public CharArray append( File file ){
      int n = (int)file.length() ;
      if( count + n > buf.length ) buf = ArraysCopyOf( buf, count + n ) ;
      try{
         Reader r = new FileReader( file ) ;
         int j = r.read( buf, count, n ) ;
         if( j != n ) throw new NullPointerException() ;
         r.close() ;
      }
      catch( Exception e ){ throw new NullPointerException() ; }
      count += n ;
      return this ;
   }

   public CharArray( InputStream is ){
      buf = new char[512] ;
      append( is ) ;
   }

   public CharArray append( InputStream is ){
      InputStreamReader isr = new InputStreamReader( is ) ;
      try{
         for( int n = 0 ; n >= 0 ; n = isr.read( buf, count, buf.length - count ) ){
            count += n ;
            if( count > 1000000 ) throw new NullPointerException() ;
            if( count == buf.length ) buf = ArraysCopyOf( buf, buf.length* 2 ) ;
         }
      }
      catch( Exception e ){ throw new NullPointerException() ; }
      return this ;
   }

   public void writeTo( File file ){
      try{
         Writer w = new FileWriter( file ) ;
         w.write( new String( buf, 0, count ) ) ;
         w.close() ;
      }
      catch( Exception e ){ throw new NullPointerException() ; }
   }

   public int length(){ return count ; }

   public char charAt( int index ){ return buf[index] ; }

   public String toString(){ return new String( buf, 0, count ) ; }

// int cou = 0 ;
   public CharSequence subSequence( int start, int end ){
//    System.err.println( ++cou + "   subSequence" ) ;
      return new String( buf, start, end - start ) ;
   }
}
// -----------------------------------------------------------------------------

