
import java.awt.AWTEvent ;
import java.awt.Toolkit ;
import java.awt.Color ;
import java.awt.Container ;
import java.awt.Dimension ;
import java.awt.FlowLayout ;
import java.awt.BorderLayout ;
import java.awt.Component ;
import java.awt.Font ;
import java.awt.GridLayout ;
import java.awt.Insets ;
import java.awt.Graphics ;
import java.awt.Graphics2D ;
import java.awt.BasicStroke ;
import java.awt.font.* ;
import java.awt.event.* ;
import java.awt.geom.* ;

import javax.swing.* ;
import javax.swing.event.* ;
import javax.swing.table.* ;
//::import javax.swing.filechooser.FileNameExtensionFilter ;
import javax.swing.filechooser.FileFilter ;
//import java.text.* ;
import java.io.* ;
import java.util.* ;

// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class JLigand extends JFrame implements ActionListener, ItemListener{
   public static void main( String[] args ){
      Locale.setDefault(new Locale("en", "GB"));
      Env.init( args ) ;
      new JLigand() ;
   }

   public static JLPanelCentre panelCentre ;
   public static JToggleButton defaultOptionButton ;
   public static JToggleButton choiceOptionButton ;
   public static CootListener cootListener ;

   public JLigand() {

      JOptionPane pane = new JOptionPane(
         Versions.welcomeInfo(),
         JOptionPane.INFORMATION_MESSAGE
      ) ;

      final JDialog dialog = pane.createDialog( null, Versions.programInfo() ) ;

      SwingUtilities.invokeLater( new Runnable(){
         public void run(){ dialog.setVisible( true ) ; }
      } ) ;

//    the same result as above:
//    new Thread(){
//       public void run(){ dialog.setVisible( true ) ; }
//    }.start() ;

      JLMain.mainFrame = this ;
      if( Env.talkToCoot ) cootListener = new CootListener() ;

      Toolkit toolkit = this.getToolkit() ;
      new MouseTracer( toolkit ) ;
      new KeyTracer( toolkit ) ;

      DialogWindows.initialise( toolkit.getScreenSize() ) ;

      try{ Scratch.init() ; }
      catch( Exception e ){
        if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         String msg = "Cannot create scratch directory" ;
         DialogWindows.showErrorDialog( msg ) ;
         System.exit( 1 ) ;
      }

      try{ StdLib.init() ; }
      catch( Exception e ){
        if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         String msg = "Cannot parse the standard monomer library - please notify your system administrator" ;
         DialogWindows.showErrorDialog( msg ) ;
         System.exit( 1 ) ;
      }

      try{ Versions.init() ; }
      catch( Exception e ){
        if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         DialogWindows.showErrorDialog( e.getMessage() ) ;
         System.exit( 1 ) ;
      }

      try{ Libmol.init() ; }
      catch( Exception e ){
        if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         DialogWindows.showErrorDialog( e.getMessage() ) ;
         System.exit( 1 ) ;
      }

      Runtime.getRuntime().addShutdownHook( new Thread( new Runnable(){
         public void run(){
            if( Libmol.isAvailable() ) Libmol.finish() ;
            if( ! Env.vbScratch ) Scratch.clear() ;
            if( Env.vbPrint ) System.err.println( "FINISHED" + Env.bsn ) ;
         }
      } ) ) ;

      this.addWindowListener( new WindowAdapter(){
         @Override public void windowClosing( WindowEvent e ){ System.exit( 0 ) ; }
      } ) ;

//    The short library needs to contain something to be ready for reading PDB files etc.
      try{
         String key8 = "LYS" ;
         boolean lcoorNo = false ;
//       Monomer comp = StdLib.get( key8, lcoorNo ) ;
         StdLib.get( key8, lcoorNo ) ;
      }
      catch( Exception e ){
        if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         DialogWindows.showErrorDialog( e.getMessage() ) ;
         System.exit( 1 ) ;
      }

      this.setSize( DialogWindows.windowWidth,  DialogWindows.windowHeight ) ;
      this.setLocation( DialogWindows.locationLeft, DialogWindows.locationTop ) ;

      this.setJMenuBar( new JLMenuBar( this ) ) ;

      JPanel panelNorth = new JLPanelNorth( this ) ;
      JPanel panelWest = new JLPanelWest( this );
      panelCentre = new JLPanelCentre() ;
      panelCentre.addTab() ;

      Container contentPane = this.getContentPane();
      contentPane.setLayout(new BorderLayout());
      contentPane.add(panelCentre, BorderLayout.CENTER);
      contentPane.add(panelNorth, BorderLayout.NORTH);
      contentPane.add(panelWest, BorderLayout.WEST);

      this.setTitle( Versions.programInfo() ) ;
      this.setVisible( true );

      if( dialog.isVisible() ) dialog.setVisible( true ) ;

      JLMain.currentPanel.resetPanel() ;

      if( Env.talkToCoot ) cootListener.run() ;
//    if( Env.talkToCoot ) new CootListener().run() ;
   }

   void actionAbout(){
      DialogWindows.showInformationDialog( Versions.aboutInfo() ) ;
   }

   void actionQuit(){
      int i = JOptionPane.showConfirmDialog( this, "Really Quit?", "Quit JLigand", JOptionPane.YES_NO_OPTION ) ;
      if( i == JOptionPane.YES_OPTION ) System.exit( 0 ) ;
   }

   public void actionPerformed( ActionEvent e ){
      String[] words = e.getActionCommand().trim().split(" +") ; ;
      String command = words[0] ;
      String target = null ;
      if( words.length == 2 ) target = words[1] ;

      if( command.equals( "OPTION" ) ) JLMain.option = target ;

      else if( command.equals( "ABOUT"            ) ) actionAbout() ;
      else if( command.equals( "QUIT"             ) ) actionQuit() ;

      else if( command.equals( "TAB_NEW"          ) ) panelCentre.addTab() ;
      else if( command.equals( "TAB_DELETE"       ) ) panelCentre.deleteTab() ;
      else if( command.equals( "TAB_CLEAR"        ) ) JLMain.actionClear() ;

//::  FILE_LIB: take care of "L" and "CIF" as these are different

      else if( command.equals( "FILE_LIB"         ) ) JLMain.actionReadLib( true ) ;
      else if( command.equals( "FILE_PDB"         ) ) JLMain.actionReadPDB( true ) ;
      else if( command.equals( "FILE_SDF"         ) ) JLMain.actionReadSDF( true ) ;
      else if( command.equals( "FILE_MOL"         ) ) JLMain.actionReadMOL( true ) ;
      else if( command.equals( "ENTER_SMILE"      ) ) JLMain.actionEnterSmile() ;
      else if( command.equals( "FILE_LIB_ASIS"    ) ) JLMain.actionReadLib( false ) ;
      else if( command.equals( "FILE_PDB_ASIS"    ) ) JLMain.actionReadPDB( false ) ;
      else if( command.equals( "FILE_SDF_ASIS"    ) ) JLMain.actionReadSDF( false ) ;
      else if( command.equals( "FILE_MOL_ASIS"    ) ) JLMain.actionReadMOL( false ) ;
      else if( command.equals( "SAVE_COMP"        ) ) JLMain.actionSave( "SAVE", "COMP", target, false ) ;
      else if( command.equals( "SAVE_LINK"        ) ) JLMain.actionSave( "SAVE", "LINK", target, false ) ;
      else if( command.equals( "APPEND_COMP"      ) ) JLMain.actionSave( "APPEND", "COMP", target, false ) ;
      else if( command.equals( "APPEND_LINK"      ) ) JLMain.actionSave( "APPEND", "LINK", target, false ) ;
      else if( command.equals( "SAVE_COOT_LINK"   ) ) JLMain.actionSave( "SAVE", "LINK", target, true ) ;
      else if( command.equals( "APPEND_COOT_LINK" ) ) JLMain.actionSave( "APPEND", "LINK", target, true ) ;
      else if( command.equals( "SAVE_PDB"         ) ) JLMain.actionSavePdb( target ) ;

      else if( command.equals( "LIBMOL"           ) ) JLMain.actionRegularise( target, true ) ;
      else if( command.equals( "REGULARISE"       ) ) JLMain.actionRegularise( target, false ) ;
      else if( command.equals( "DELETE"           ) ) JLMain.actionDelete( target ) ;
      else if( command.equals( "FIND_COMP"        ) ) JLMain.actionFind( target ) ;

      else if( command.equals( "ATOMS"            ) ) JLMain.actionViewTable( command, target ) ;
      else if( command.equals( "BONDS"            ) ) JLMain.actionViewTable( command, target ) ;
      else if( command.equals( "ANGLES"           ) ) JLMain.actionViewTable( command, target ) ;
      else if( command.equals( "TORSIONS"         ) ) JLMain.actionViewTable( command, target ) ;
      else if( command.equals( "CHIRALITIES"      ) ) JLMain.actionViewTable( command, target ) ;
      else if( command.equals( "PLANES"           ) ) JLMain.actionViewTable( command, target ) ;
      else if( command.equals( "VIEW_LINK"        ) ) JLMain.actionViewLinkFile( target ) ;

      else if( command.equals( "PERIODIC_TABLE"   ) ) DialogWindows.periodicTableDialog.showDialog() ;
      else if( command.equals( "VIEW_FILE"        ) ) JLMain.actionViewFile() ;

      else if( command.equals( "HELP_JLIGAND"     ) ) DialogWindows.helpJLigand.showDialog() ;
      else if( command.equals( "HELP_MOUSE"       ) ) DialogWindows.helpJLigandMouse.showDialog() ;
      else if( command.equals( "HELP_KEYPAD"      ) ) DialogWindows.helpJLigandKeypad.showDialog() ;

      else if( command.equals( "LIGAND_NEW"       ) ) JLMain.actionNewLigand() ;
      else if( command.equals( "LIGAND_ID"        ) ) JLMain.actionLoadLigand( target ) ;
      else if( command.equals( "LIGAND_SEARCH"    ) ) JLMain.actionLoadLigand() ;
      else if( command.equals( "CHECK_CHEM"       ) ) JLMain.actionCheckChemistry() ;
      else if( command.equals( "CHECK_BONDS"      ) ) JLMain.actionCorrectBonds() ;
      else if( command.equals( "UNDO"             ) ) JLMain.actionUndo() ;
      else if( command.equals( "REDO"             ) ) JLMain.actionRedo() ;
      else throw new NullPointerException() ;

      JLMain.currentPanel.resetPanel() ;
   }

   public void itemStateChanged( ItemEvent e ){
      AbstractButton button = (AbstractButton)e.getItem() ;
      String command = button.getActionCommand() ;
      if(      command.equals( "SHOW_IDS"         ) ) JLMain.showNames     = button.isSelected() ;
      else if( command.equals( "SHOW_HYDROGENS"   ) ) JLMain.showHydrogens = button.isSelected() ;
      else if( command.equals( "SHOW_CHARGES"     ) ) JLMain.showCharges   = button.isSelected() ;

      else throw new NullPointerException() ;

      defaultOptionButton.doClick() ;
      JLMain.currentPanel.requestFocusInWindow() ;
   }
}
// -----------------------------------------------------------------------------

class MouseTracer extends ButtonTracer{
   MouseTracer( Toolkit toolkit ){
      pressed = MouseEvent.MOUSE_PRESSED ;
      released = MouseEvent.MOUSE_RELEASED ;
      mask = InputEvent.BUTTON1_DOWN_MASK | InputEvent.BUTTON3_DOWN_MASK ;
      toolkit.addAWTEventListener( this, AWTEvent.MOUSE_EVENT_MASK ) ;
   }
}
// -----------------------------------------------------------------------------

class KeyTracer extends ButtonTracer{
   KeyTracer( Toolkit toolkit ){
      pressed = KeyEvent.KEY_PRESSED ;
      released = KeyEvent.KEY_RELEASED ;
      mask = InputEvent.CTRL_DOWN_MASK | InputEvent.SHIFT_DOWN_MASK | InputEvent.ALT_DOWN_MASK ;
      toolkit.addAWTEventListener( this, AWTEvent.KEY_EVENT_MASK ) ;
   }
}
// -----------------------------------------------------------------------------

abstract class ButtonTracer implements AWTEventListener{
   int pressed, released, mask ;
   public void eventDispatched( AWTEvent awte ){
      InputEvent ie = (InputEvent)awte ;
      int id = ie.getID() ;
      if( id != pressed && id != released ) return ;
      JLMain.buttonsMask = ( JLMain.buttonsMask & ~mask ) | ( ie.getModifiersEx() & mask ) ;
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class JCheckedFileChooser extends JFileChooser{
   JCheckedFileChooser( File openDir ){
      super( openDir ) ;
   }
   public void approveSelection(){
      File selection = getSelectedFile() ;
      if( selection.isFile() ){
         int n = JOptionPane.showOptionDialog( this,
            "'" + selection.getName() + "' already exists. Do you want to replace it?",
            "File exists",
            JOptionPane.OK_CANCEL_OPTION,
            JOptionPane.WARNING_MESSAGE,
            null, new String[]{ "Save", "Cancel" }, "Save"
         ) ;
         if( n != JOptionPane.OK_OPTION ) return ;
      }
      super.approveSelection() ;
   }
}
// -----------------------------------------------------------------------------

//:: this is not defined in java 1.5
class FileNameExtensionFilter extends FileFilter{
   private final String description ;
   private final String[] extensions ;
   public FileNameExtensionFilter( String d, String ... e ){
      description = d ;
      extensions = e ;
   }
   public boolean accept( File f ){
      String s = f.getName() ;
      boolean b = f.isDirectory() ;
      for( int i = 0 ; i < extensions.length ; i++ ) b |= s.endsWith( "." + extensions[i]  ) ;
      return b ;
   }
   public String getDescription(){ return description ; }
}
// -----------------------------------------------------------------------------

class OneLigandInfo{
   String key3 ;
   String chainId ;
   String residueNo ;
   String file ;

   public boolean consumed( String[] words ){
      if( words.length != 4 || ! words[0].equals( "CODE" ) ) return true ;
      if( key3 == null ) key3 = words[1] ;
      if( words[1].equals( key3 ) ){
         if( words[2].equals( "FILE" ) ){
            file = words[3] ;
         }
         else{
            chainId = words[2] ;
            residueNo = words[3] ;
         }
         return true ;
      }
      return false ;
   }

   public boolean isValid(){
      return key3 != null && chainId != null && residueNo != null ;
   }

   public String toString(){
      String s = key3 + " " + chainId + " " + residueNo ;
      if( file != null ) s += " " + file ;
      return s ;
   }
}
// -----------------------------------------------------------------------------

class CootInfoReader extends ArrayList<OneLigandInfo>{

   CootInfoReader( File cootInfoFile ) throws Exception{
      super(2) ;
      add( new OneLigandInfo() ) ;
      add( new OneLigandInfo() ) ;
      BufferedReader br = new BufferedReader( new FileReader( cootInfoFile ) ) ;
      for ( String line = "" ; line != null ; line = br.readLine() ){
         String[] words = line.trim().split( "\\s+" ) ;
         if( ! get(0).consumed( words ) && ! get(1).consumed( words ) ) throw new NullPointerException() ;
      }
      if( ! get(0).isValid() || ! get(1).isValid() ) throw new NullPointerException() ;
   }
}
// -----------------------------------------------------------------------------


class CootListener extends Thread{
   private final long timeInterval = 500 ;
   private long modificationTime ;
   private static int cootId = 0 ;

   CootListener(){
      try{
         FileWriter fw = new FileWriter( JLMain.toCoot ) ;
         fw.write( "READY" + Env.bsn ) ;
         fw.close() ;
         modificationTime = JLMain.toCoot.lastModified() ;
//       modificationTime = new Date().getTime() ;
      }
      catch( Exception e ){
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         DialogWindows.showErrorDialog( "Error writing message to Coot" ) ;
      }
      modificationTime -= 5 ;
   }

// CootListener(){
//    modificationTime = JLMain.fromCoot.lastModified() ;
//    if( new Date().getTime() - modificationTime < 5000 ) modificationTime = 0 ;
// }

   public void run(){
      while( true ){
         if( JLMain.fromCoot.isFile() ){
            long time = JLMain.fromCoot.lastModified() ;
            if( modificationTime < time ){
//             System.err.println( "AAAAAAAA " + modificationTime + " " + time ) ;
               modificationTime = time ;
               try{
                  final CootInfoReader cir = new CootInfoReader( JLMain.fromCoot ) ;
                  cootId++ ;
                  for( final OneLigandInfo info : cir ){
                     System.err.println( info ) ;
                     SwingUtilities.invokeAndWait( new Runnable(){ public void run(){
   
                        Molecule mol = null ;
                        if( info.file == null ){
                           mol = JLMain.actionLoadLigand( info.key3 ) ;
                        }
                        else{
                           mol = JLMain.actionReadFromLib( info.file, info.key3 ) ;
                        }
                        if( mol == null ) throw new NullPointerException() ;
                        Residue rsd = mol.getResidue() ;
                        if( rsd == null ) throw new NullPointerException() ;
                        rsd.setCootInfo( info.chainId, info.residueNo, cootId ) ;
                        JLMain.currentPanel.resetPanel() ;
                     } } ) ;
                  }
                  System.err.println() ;
               }
               catch( Exception e ){
                  if( Env.vbPrint ) e.printStackTrace( System.err ) ;
                  DialogWindows.showErrorDialog( "Error reading message from Coot" ) ;
               }
            }
            try{
               Thread.sleep( timeInterval ) ;
            }
            catch( InterruptedException ie ){
               System.err.println("x3") ;
            }
         }
      }
   }
}
// -----------------------------------------------------------------------------

class JLMain{
   static JLGPanel currentPanel ;
   static String option ;
   static int buttonsMask ;
   static boolean showNames = false ;
   static boolean showHydrogens = false ;
   static boolean showCharges = false ;

   static File openDir = Env.cwd ;
   static String bsn = Env.bsn ;

   static JFrame mainFrame ;

// static final File toCoot = new File( Env.cwd, "toCoot" ) ;
// static final File fromCoot = new File( Env.cwd, "fromCoot" ) ;

   static final File toCoot = new File( Env.cwd, ".jligand-to-coot-link" ) ;
   static final File fromCoot = new File( Env.cwd, ".coot-to-jligand-8lcs" ) ;

   static void actionNewLigand(){
      NewLigandDialog dialog = new NewLigandDialog( currentPanel.gLib ) ;
      dialog.showDialog() ;
      String ligandId = dialog.getLigandId() ;
      String atomType = dialog.getAtomType() ;
      if( ligandId == null || atomType == null ) return ;
      currentPanel.gLib.undoableAddMolecule( ligandId, atomType ) ;
   }

   static void actionLoadLigand(){
      SearchLibraryDialog dialog = new SearchLibraryDialog() ;
      dialog.showDialog() ;
      actionLoadLigand( dialog.getLigandId() ) ;
   }

   static Molecule actionLoadLigand( String target ){
      if( target == null || target.length() == 0 ) return null ;

      Monomer comp = null ;
      if( ! StdLib.hasReady( target ) ){
         currentPanel.drawMessage( "Please wait, " + target + " is being loaded" ) ;
      }
      boolean lcoorNo = false ;
      try{ comp = StdLib.get( target, lcoorNo ) ; }
      catch( Exception e ){
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
      }

      if( comp != null ) comp.recentre() ;

      Molecule mol = null ;
      if( comp != null ){
         String key8 = target ;
         String msg = "The ligand '" + key8 + "' has been loaded" + bsn + "but the" ;
         CorrectLinkIdDialog dialog = new CorrectLinkIdDialog( msg, currentPanel.gLib, key8 ) ;
         key8 = dialog.newKey8 ;
         if( key8 != null ) mol = currentPanel.gLib.undoableAddMolecule( key8, comp ) ;
      }
      else{
         String msg = "Code '" + target + "' does not exist in the library - please try again" ;
         DialogWindows.showErrorDialog( msg ) ;
      }
      return mol ;
   }

   static void actionSave( String toDo, String type, String target, boolean sendToCoot ){

      Molecule molecule = null ;
      try{ molecule = currentPanel.gLib.get( target ) ; }
      catch( Exception e ){
//       if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         DialogWindows.showErrorDialog( "Internal error") ;
         return ;
      }

      if( molecule == null ){
         DialogWindows.showErrorDialog( "Internal error") ;
         return ;
      }

      if( molecule.restraints == null ){
         String msg = "Ligand " + target + " is not regularised - please regularise before saving" ;
         DialogWindows.showWarningDialog( msg ) ;
         return ;
      }

      File file = null ;
      CifLib cifLib = null ;

      if( toDo.equals( "APPEND" ) ){
         JFileChooser chooser = new JFileChooser( openDir ) ;
         chooser.setDialogTitle( "Append to CIF-library" ) ;
         chooser.setFileFilter( new FileNameExtensionFilter( "CIF Library Files", "cif", "lib" ) ) ;
         chooser.setApproveButtonText( "Append" ) ;
         if( chooser.showOpenDialog( mainFrame ) != JFileChooser.APPROVE_OPTION ) return ;

         file = chooser.getSelectedFile() ;
         openDir = file.getParentFile() ;
         try{
            currentPanel.drawMessage( "Please wait, " + file.getName() + " is being read and validated" ) ;
            cifLib = new CifLib( file ) ;
         }
         catch( Exception e ){
//          if( Env.vbPrint ) e.printStackTrace( System.err ) ;
            DialogWindows.showErrorDialog( "File " + file + " does not seem to be a valid CIF-library" ) ;
            return ;
         }
      }
      else if( toDo.equals( "SAVE" ) ){
         JFileChooser chooser = new JCheckedFileChooser( openDir ) ;
         chooser.setDialogTitle( "Save as CIF-library" ) ;
         chooser.setSelectedFile( new File( openDir, target + ".cif" ) ) ;
         chooser.setFileFilter( new FileNameExtensionFilter( "CIF Library Files", "cif", "lib" ) ) ;
         if( chooser.showDialog( mainFrame, "Save" ) != JFileChooser.APPROVE_OPTION ) return ;

         file = chooser.getSelectedFile() ;
         openDir = file.getParentFile() ;
         cifLib = new CifLib() ;
      }
      else{
//       if( Env.vbPrint ) System.err.println( toDo ) ;
         DialogWindows.showErrorDialog( "Internal error") ;
         return ;
      }

      String ligandKind = null ;
      String info = "" ;
      try{
         if( type.equals( "COMP" ) ){
            Monomer comp = null ;
            try{
               comp = new Monomer( molecule ) ;
            }
            catch( Exception e ){
//             if( Env.vbPrint ) e.printStackTrace( System.err ) ;
               DialogWindows.showErrorDialog( "Internal error") ;
               return ;
            }
            ligandKind = "monomer" ;
            cifLib.add( comp ) ;
         }
         else if( type.equals( "LINK" ) ){
            if( molecule.getResidue() != null ){
               DialogWindows.showErrorDialog( "Internal error") ;
               return ;
            }
            ligandKind = "link" ;
            info = cifLib.add( molecule ) ;
         }
         else{
            DialogWindows.showErrorDialog( "Internal error") ;
            return ;
         }
      }
      catch( Exception e ){
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         DialogWindows.showWarningDialog(
            "The " + ligandKind + " '" + target +
            "' cannot be appended to the CIF-library " +
            file.getName() + " because of name conflict(s)"
         ) ;
         return ;
      }

      try{
         cifLib.toBuffer().writeTo( file ) ;
      }
      catch( Exception e ){
//       if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         DialogWindows.showErrorDialog( "Error writing file " + file.getName() ) ;
         return ;
      }

      String msg = "The " + ligandKind + " '" + target ;
      if( toDo.equals( "APPEND" ) ){
         msg += "' has been appended to the CIF-library " + file.getName() ;
      }
      else if( toDo.equals( "SAVE" ) ){
         msg += "' has been saved to the library file " + file.getName() ;
      }
      else{
//       if( Env.vbPrint ) System.err.println( toDo ) ;
         DialogWindows.showErrorDialog( "Internal error") ;
         return ;
      }
      if( sendToCoot ){
         msg += Env.bsn + "and will be sent to Coot" ;
      }
      msg += info ;
      DialogWindows.showInformationDialog( msg ) ;

      if( type.equals( "LINK" ) ){
         String linkRecord = molecule.linkRecordPDB() ;
         if( Env.vbPrint ) System.err.println( linkRecord ) ;

         File linkRecordFile = toCoot ;
         if( ! sendToCoot ){
            String s = "To use the link '" + target + "'" ;
            s += Env.bsn + "1) Add the following template LINK record to the input PDB-file " +
                               "(just before the CRYST1 record)" ;
            s += Env.bsn + "2) Change the chain IDs and residue numbers as appropriate" ;
            s += Env.bsn + "3) In REFMAC task of CCP4I, define '" + file.getName() + "' as additional library" ;
            s += Env.bsn ;
            s += Env.bsn + "NB. DO NOT change the order of residues in the LINK record!" ;
            s += Env.bsn ;
            s += Env.bsn + "..." ;
            s += Env.bsn + linkRecord ;
            s += Env.bsn + "CRYST1 ..." ;
            s += Env.bsn ;
            s += Env.bsn + "Do you want to save this template in a separate file?" ;
            int i = JOptionPane.showConfirmDialog( mainFrame, s, "Save Link Record", JOptionPane.YES_NO_OPTION ) ;
            if( i != JOptionPane.YES_OPTION ) return ;

            JFileChooser chooser = new JCheckedFileChooser( openDir ) ;
            chooser.setDialogTitle( "Save PDB LINK record" ) ;
            chooser.setSelectedFile( new File( openDir, "linkrecord.pdb" ) ) ;
            chooser.setFileFilter( new FileNameExtensionFilter( "PDB Files", "pdb" ) ) ;
            if( chooser.showDialog( mainFrame, "Save" ) != JFileChooser.APPROVE_OPTION ) return ;

            linkRecordFile = chooser.getSelectedFile() ;
            openDir = linkRecordFile.getParentFile() ;
         }

         try{
            FileWriter fw = new FileWriter( linkRecordFile ) ;
            fw.write( file.getCanonicalFile().toString() + Env.bsn ) ;
            fw.write( linkRecord + Env.bsn ) ;
            fw.close() ;
         }
         catch( Exception e ){
            msg = "Error writing file " + linkRecordFile.getName() ;
            DialogWindows.showErrorDialog( msg ) ;
         }

         if( ! sendToCoot ){
            msg = "File " + linkRecordFile.getName() + " was successfully written" ;
            DialogWindows.showInformationDialog( msg ) ;
         }
      }
   }

   static void actionRegularise( String target, boolean useLibmol ){
      Molecule mol = currentPanel.gLib.get( target ) ;
      if( mol.restraints != null ){
         DialogWindows.showWarningDialog( "This ligand is already regularised" ) ;
         return ;
      }
      currentPanel.drawMessage( "Please wait, " + target + " is being regularised" ) ;
      regularise( mol, useLibmol ) ;
   }

   static void actionDelete( String target ){
      String s = "Are you sure you want to delete ligand " + target + "?" ;
      int i = JOptionPane.showConfirmDialog( mainFrame, s, "Delete Ligand", JOptionPane.YES_NO_OPTION ) ;
      if( i != JOptionPane.YES_OPTION ) return ;
      currentPanel.gLib.undoableRemoveMolecule( currentPanel.gLib.get( target ) ) ;
   }

   static void actionFind( String target ){
      try{
         Molecule molecule = currentPanel.gLib.get( target ) ;
         if( molecule.restraints == null ){
            String msg = "Ligand " + target + " is not regularised - please regularise before searching" ;
            DialogWindows.showWarningDialog( msg ) ;
            return ;
         }

         String message = "Please wait, the library is being searched for " + target + " matches" ;
         currentPanel.drawMessage( message ) ;
         List<String> list = new Libcheck().getMatches( molecule ) ;
         currentPanel.resetPanel() ;
         if( list == null || list.size() == 0 ){
            String s = "Ligand " + target + " has no exact matches in the standard library" ;
            DialogWindows.showInformationDialog( s ) ;
            return ;
         }

         SearchLibraryDialog dialog = new SearchLibraryDialog( list ) ;
         dialog.showDialog() ;
         actionLoadLigand( dialog.getLigandId() ) ;
      }
      catch( Exception e ){
//       if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         DialogWindows.showErrorDialog( "Internal error") ;
         return ;
      }
   }

   static void actionEnterSmile(){
      SmileStringDialog smileDialog = new SmileStringDialog( currentPanel.gLib ) ;
      smileDialog.showDialog() ;
      if( smileDialog.canceled() ) return ;

      String key8 = smileDialog.getLigandId() ;
      String smileString = smileDialog.getSmileString() ;

      Monomer mono = null ;
      boolean coorNo = true ;
      boolean lcoorNo = true ;
      try{
         currentPanel.drawMessage( "Please wait, your smile string is being processed" ) ;
         CharArray buffer = new CharArray( smileString + bsn ) ;
         File libFile = new Libcheck().runLibcheckOnly( buffer, "SMILE", key8, lcoorNo, coorNo ) ;
         CifLib cifLib = new CifLib( libFile ) ;
         mono = cifLib.getMonomer() ;
         mono.recentre() ;
         if( coorNo ){
            Molecule mol = new Molecule( mono ) ;
            mol.restraints = null ;
            boolean useLibmol = false ;
            mono = new Monomer( mol.regularise( true, lcoorNo, useLibmol ) ) ;
         }
      }
      catch( Exception e ){
         DialogWindows.showErrorDialog( "Cannot read the smile string" ) ;
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         return ;
      }
      currentPanel.gLib.undoableAddMolecule( mono.key8, mono ) ;
   }

   static Molecule actionReadFromLib( String fileName, String ligandId ){
      boolean lcoorNo = true ;
      Monomer mono = null ;
      File inputFile = new File( fileName ) ;
      try{
         currentPanel.drawMessage( "Please wait, " + ligandId + " is being read from " + inputFile.getName() + " and regularised" ) ;
         openDir = inputFile.getParentFile() ;
         mono = new CifLib( inputFile ).getMonomer( ligandId ) ;
      }
      catch( Exception e ){
         DialogWindows.showErrorDialog( "Could not read library-file '" + inputFile.getName() + "'" ) ;
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         return null ;
      }
      if( mono == null ) return null ;

      try{
         mono.recentre() ;
         Molecule mol = new Molecule( mono ) ;
         mol.restraints = null ;
         boolean useLibmol = false ;
         mono = new Monomer( mol.regularise( true, lcoorNo, useLibmol ) ) ;
      }
      catch( Exception e ){
         DialogWindows.showErrorDialog( "Regularisation failed" ) ;
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         return null ;
      }

      String msg = "The ligand '" + mono.key8 + "' has been imported" + bsn + "but the" ;
      CorrectLinkIdDialog dialog = new CorrectLinkIdDialog( msg, currentPanel.gLib, mono.key8 ) ;
      mono.key8 = dialog.newKey8 ;
      if( mono.key8 == null ) return null ;
      return currentPanel.gLib.undoableAddMolecule( mono.key8, mono ) ;
   }

   static void actionReadLib( boolean lcoorNo ){
      JFileChooser chooser = new JFileChooser( openDir ) ;
      chooser.setDialogTitle( "Import Ligand from CIF-library" ) ;
      chooser.setFileFilter( new FileNameExtensionFilter( "CIF Library Files", "cif", "lib" ) ) ;
      if( chooser.showOpenDialog( mainFrame ) != JFileChooser.APPROVE_OPTION ) return ;

      File inputFile = chooser.getSelectedFile() ;
      openDir = inputFile.getParentFile() ;

      CifLib cifLib = null ;
      try{
         currentPanel.drawMessage( "Please wait, " + inputFile.getName() + " is being read and validated" ) ;
         cifLib = new CifLib( inputFile ) ;
      }
      catch( Exception e ){
         DialogWindows.showErrorDialog( "Could not read library-file '" + inputFile.getName() + "'" ) ;
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         return ;
      }

      Map<String,Monomer> map = new LinkedHashMap<String,Monomer>() ;
      for( Monomer comp : cifLib.monomerView() ) map.put( comp.key8, comp ) ;
      if( map.size() == 0 ){
         DialogWindows.showErrorDialog( "Could not find Ligands in file '" + inputFile.getName() + "'" ) ;
         return ;
      }

      Object[] keys = map.keySet().toArray() ;
      Object key = keys[0] ;
      if( keys.length > 1 ){
         ChooseLigandFromFile dialog = new ChooseLigandFromFile() ;
         dialog.showDialog( keys ) ;
         key = dialog.getLigandId() ;
         if( ! map.keySet().contains( key ) ) return ;
      }

      Monomer mono = map.get( key ) ;
      try{
         if( keys.length > 1 ){
            currentPanel.drawMessage( "Please wait, " + mono.key8 + " is being regularised" ) ;
         }
         mono.recentre() ;
         Molecule mol = new Molecule( mono ) ;
         mol.restraints = null ;
//       boolean lcoorNo = true ;
         boolean useLibmol = false ;
         mono = new Monomer( mol.regularise( true, lcoorNo, useLibmol ) ) ;
      }
      catch( Exception e ){
         DialogWindows.showErrorDialog( "Regularisation failed" ) ;
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         return ;
      }

      String msg = "The ligand '" + mono.key8 + "' has been imported" + bsn + "but the" ;
      CorrectLinkIdDialog dialog = new CorrectLinkIdDialog( msg, currentPanel.gLib, mono.key8 ) ;
      mono.key8 = dialog.newKey8 ;
      if( mono.key8 == null ) return ;
      currentPanel.gLib.undoableAddMolecule( mono.key8, mono ) ;
   }

   static void actionReadMOL( boolean coorNo ){
      JFileChooser chooser = new JFileChooser( openDir ) ;
      chooser.setDialogTitle( "Import Ligand from MOL(2)-file" ) ;
      chooser.setFileFilter( new FileNameExtensionFilter( "MOL(2) Files", "mol", "mol2" ) ) ;
      if( chooser.showOpenDialog( mainFrame ) != JFileChooser.APPROVE_OPTION ) return ;

      File inputFile = chooser.getSelectedFile() ;
      openDir = inputFile.getParentFile() ;

      Monomer mono = null ;
      boolean lcoorNo = true ;
      try{
         currentPanel.drawMessage( "Please wait, file " + inputFile.getName() + " is being processed" ) ;
         CharArray buffer = new CharArray( inputFile ) ;
         File libFile = new Libcheck().runLibcheckOnly( buffer, "MOL", (String)null, lcoorNo, coorNo ) ;
         CifLib cifLib = new CifLib( libFile ) ;
         mono = cifLib.getMonomer() ;
         mono.recentre() ;
         if( coorNo ){
            Molecule mol = new Molecule( mono ) ;
            mol.restraints = null ;
            boolean useLibmol = false ;
            mono = new Monomer( mol.regularise( true, lcoorNo, useLibmol ) ) ;
         }
      }
      catch( Exception e ){
         DialogWindows.showErrorDialog( "Could not read MOL(2)-file '" + inputFile.getName() + "'" ) ;
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         return ;
      }

      String msg = "The ligand '" + mono.key8 + "' has been imported" + bsn + "but the" ;
      CorrectLinkIdDialog dialog = new CorrectLinkIdDialog( msg, currentPanel.gLib, mono.key8 ) ;
      mono.key8 = dialog.newKey8 ;
      if( mono.key8 == null ) return ;
      currentPanel.gLib.undoableAddMolecule( mono.key8, mono ) ;
   }

   static void actionReadSDF( boolean coorNo ){
      JFileChooser chooser = new JFileChooser( openDir ) ;
      chooser.setDialogTitle( "Import Ligand from SDF-file" ) ;
      chooser.setFileFilter( new FileNameExtensionFilter( "SDF Files", "sdf" ) ) ;
      if( chooser.showOpenDialog( mainFrame ) != JFileChooser.APPROVE_OPTION ) return ;

      File inputFile = chooser.getSelectedFile() ;
      openDir = inputFile.getParentFile() ;

      Monomer mono = null ;
      boolean lcoorNo = true ;
      try{
         currentPanel.drawMessage( "Please wait, file " + inputFile.getName() + " is being processed" ) ;
         CharArray buffer = new CharArray( inputFile ) ;
         File libFile = new Libcheck().runLibcheckOnly( buffer, "SDF", (String)null, lcoorNo, coorNo ) ;
         CifLib cifLib = new CifLib( libFile ) ;
         mono = cifLib.getMonomer() ;
         mono.recentre() ;
         if( coorNo ){
            Molecule mol = new Molecule( mono ) ;
            mol.restraints = null ;
            boolean useLibmol = false ;
            mono = new Monomer( mol.regularise( true, lcoorNo, useLibmol ) ) ;
         }
      }
      catch( Exception e ){
         DialogWindows.showErrorDialog( "Could not read SDF-file '" + inputFile.getName() + "'" ) ;
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         return ;
      }

      String msg = "The ligand '" + mono.key8 + "' has been imported" + bsn + "but the" ;
      CorrectLinkIdDialog dialog = new CorrectLinkIdDialog( msg, currentPanel.gLib, mono.key8 ) ;
      mono.key8 = dialog.newKey8 ;
      if( mono.key8 == null ) return ;
      currentPanel.gLib.undoableAddMolecule( mono.key8, mono ) ;
   }

   static void actionReadPDB( boolean coorNo ){
      JFileChooser chooser = new JFileChooser( openDir ) ;
       chooser.setDialogTitle( "Import Ligand from PDB-file" ) ;
       chooser.setFileFilter( new FileNameExtensionFilter( "PDB Files", "pdb" ) ) ;
      if( chooser.showOpenDialog( mainFrame ) != JFileChooser.APPROVE_OPTION ) return ;

      File inputFile = chooser.getSelectedFile() ;
      openDir = inputFile.getParentFile() ;

      CifLib cifLib = null ;
      try{
         currentPanel.drawMessage( "Please wait, file " + inputFile.getName() + " is being processed" ) ;
         CharArray buffer = new CharArray( inputFile ) ;
         boolean lcoorNo = true ;
         File libFile = new Libcheck().runLibcheckOnly( buffer, "PDB", (String)null, lcoorNo, coorNo ) ;
         cifLib = new CifLib( libFile ) ;
      }
      catch( Exception e ){
         DialogWindows.showErrorDialog( "Could not read PDB-file '" + inputFile.getName() + "'" ) ;
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         return ;
      }

      Map<String,Monomer> map = new LinkedHashMap<String,Monomer>() ;
      for( Monomer comp : cifLib.monomerView() ) map.put( comp.key8, comp ) ;
      if( map.size() == 0 ){
         DialogWindows.showErrorDialog( "Could not find Ligands in file '" + inputFile.getName() + "'" ) ;
         return ;
      }

      Object[] keys = map.keySet().toArray() ;
      Object key = keys[0] ;
      if( keys.length > 1 ){
         ChooseLigandFromFile dialog = new ChooseLigandFromFile() ;
         dialog.showDialog( keys ) ;
         key = dialog.getLigandId() ;
         if( ! map.keySet().contains( key ) ) return ;
      }

      Monomer mono = map.get( key ) ;
      try{
         mono.recentre() ;
         if( coorNo ){
            if( keys.length > 1 ){
               currentPanel.drawMessage( "Please wait, " + mono.key8 + " is being regularised" ) ;
            }
            Molecule mol = new Molecule( mono ) ;
            mol.restraints = null ;
            boolean lcoorNo = true ;
            boolean useLibmol = false ;
            mono = new Monomer( mol.regularise( true, lcoorNo, useLibmol ) ) ;
         }
      }
      catch( Exception e ){
         DialogWindows.showErrorDialog( "Regularisation failed" ) ;
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         return ;
      }

      String msg = "The ligand '" + mono.key8 + "' has been imported" + bsn + "but the" ;
      CorrectLinkIdDialog dialog = new CorrectLinkIdDialog( msg, currentPanel.gLib, mono.key8 ) ;
      mono.key8 = dialog.newKey8 ;
      if( mono.key8 == null ) return ;
      currentPanel.gLib.undoableAddMolecule( mono.key8, mono ) ;
   }

   static void actionRead( String fileType ){
      JFileChooser chooser = new JFileChooser( openDir ) ;
      if( fileType.equals( "L" ) ){
         chooser.setDialogTitle( "Import Ligand from CIF-library" ) ;
         chooser.setFileFilter( new FileNameExtensionFilter( "CIF Library Files", "cif", "lib" ) ) ;
      }
      else if( fileType.equals( "PDB" ) ){
         chooser.setDialogTitle( "Import Ligand from PDB-file" ) ;
         chooser.setFileFilter( new FileNameExtensionFilter( "PDB Files", "pdb" ) ) ;
      }
      else if( fileType.equals( "SDF" ) ){
         chooser.setDialogTitle( "Import Ligand from SDF-file" ) ;
         chooser.setFileFilter( new FileNameExtensionFilter( "SDF Files", "sdf" ) ) ;
      }
      if( chooser.showOpenDialog( mainFrame ) != JFileChooser.APPROVE_OPTION ) return ;

      File inputFile = chooser.getSelectedFile() ;
      openDir = inputFile.getParentFile() ;

      File libFile = null ;
      if( fileType.equals( "L" ) ){
         libFile = inputFile ;
      }
      else{
         currentPanel.drawMessage( "Please wait, file " + inputFile.getName() + " is being processed" ) ;

         CharArray buffer = new CharArray( inputFile ) ;
         boolean lcoorNo = true ;
         boolean coorNo = true ;
         try{ libFile = new Libcheck().runLibcheckOnly( buffer, fileType, (String)null, lcoorNo, coorNo ) ; }
         catch( NullPointerException npe ){}
      }

      CifLib cifLib = null ;
      try{
         cifLib = new CifLib( libFile ) ;
      }
      catch( NullPointerException npe ){
         DialogWindows.showErrorDialog( "Could not read " + fileType + "-file '" + inputFile.getName() + "'" ) ;
         return ;
      }

      Map<String,Monomer> map = new LinkedHashMap<String,Monomer>() ;
      for( Monomer comp : cifLib.monomerView() ) map.put( comp.key8, comp ) ;
      if( map.size() == 0 ){
         DialogWindows.showErrorDialog( "Could not find Ligands in file '" + inputFile.getName() + "'" ) ;
         return ;
      }

      Object[] keys = map.keySet().toArray() ;
      Object key = keys[0] ;
      if( keys.length > 1 ){
         ChooseLigandFromFile dialog = new ChooseLigandFromFile() ;
         dialog.showDialog( keys ) ;
         key = dialog.getLigandId() ;
         if( ! map.keySet().contains( key ) ) return ;
      }

      String key8 = (String)key ;
      String msg = "The ligand '" + key8 + "' has been imported" + bsn + "but the" ;
      CorrectLinkIdDialog dialog = new CorrectLinkIdDialog( msg, currentPanel.gLib, key8 ) ;
      key8 = dialog.newKey8 ;
      if( key8 == null ) return ;
      currentPanel.gLib.undoableAddMolecule( key8, map.get( key ) ) ;
   }

   static void actionSavePdb( String target ){
      Molecule mol = currentPanel.gLib.get( target ) ;

      if( mol.getResidue() == null ){
         String s = "Something wrong: this should not be a link" ;
         DialogWindows.showErrorDialog( s ) ;
         return ;
      }
      if( mol.restraints == null ){
         String s = "Ligand " + target + " is not regularised - please regularise before saving" ;
         DialogWindows.showInformationDialog( s ) ;
         return ;
      }
      Monomer mono = new Monomer( mol ) ;

      SavePdbDialog dialog = new SavePdbDialog() ;
      dialog.showDialog( target ) ;
      if( ! dialog.savePdb ) return ;
      boolean includeH = dialog.includeH ;

      JFileChooser chooser = new JCheckedFileChooser( openDir ) ;
      chooser.setDialogTitle( "Save Coordinates in PFB-file" ) ;
      chooser.setSelectedFile( new File( openDir, target + ".pdb" ) ) ;
      chooser.setFileFilter( new FileNameExtensionFilter( "PDB Files", "pdb" ) ) ;
      if( chooser.showDialog( mainFrame, "Save Coordinates" ) != JFileChooser.APPROVE_OPTION ) return ;

      File pdbFile = chooser.getSelectedFile() ;
      openDir = pdbFile.getParentFile() ;

      try{
         new PdbWriter().write( mono, pdbFile, includeH ) ;
         String s = "File " + pdbFile.getName() + " was successfully written" ;
         DialogWindows.showInformationDialog( s ) ;
      }
      catch( Exception e ){
         String s = "Error writing PDB-file " + pdbFile.getName() ;
         DialogWindows.showErrorDialog( s ) ;
      }
   }

   static void actionViewFile() {
      JFileChooser chooser = new JFileChooser( openDir ) ;
      chooser.setDialogTitle( "View File" ) ;
      chooser.addChoosableFileFilter( new FileNameExtensionFilter( "CIF Library Files", "cif", "lib" ) ) ;
      chooser.addChoosableFileFilter( new FileNameExtensionFilter( "PDB Files", "pdb" ) ) ;
      chooser.setFileFilter( chooser.getAcceptAllFileFilter() ) ;
      chooser.setApproveButtonText( "View" ) ;
      if( chooser.showOpenDialog( mainFrame ) != JFileChooser.APPROVE_OPTION ) return ;

      File file = chooser.getSelectedFile() ;
      openDir = file.getParentFile() ;

      if( ! file.canRead() ){
         DialogWindows.showErrorDialog( "Cannot read file '" + file.getName() + "'" ) ;
         return ;
      }
      if( file.length() > 2000000 ){
         DialogWindows.showWarningDialog( "File size for '" + file.getName() + "' is too large( > 2MB )" ) ;
         return ;
      }

      new ViewFileDialog().show( file.getName(), new CharArray( file ).toString() ) ;
   }

   static void actionViewLinkFile( String target ){
      Molecule link = currentPanel.gLib.get( target ) ;
      if( link.restraints == null ){
         String s = "This ligand is not regularised - please regularise before viewing link data" ;
         DialogWindows.showWarningDialog( s ) ;
         return ;
      }
      try{
         CifLib cifLib = new CifLib() ;
//       String info = cifLib.add( link ) ;
         cifLib.add( link ) ;
         new ViewFileDialog().show( link.key8, cifLib.toBuffer().toString() ) ;
      }
      catch( Exception e ){
         DialogWindows.showWarningDialog(
            "The link '" + target + "' cannot be viewed because of unexpected error"
         ) ;
      }
   }

   static void actionViewTable( String mode, String target ){
      Molecule molecule = currentPanel.gLib.get( target ) ;
      if( molecule.restraints == null ){
         int i = JOptionPane.showConfirmDialog(
            mainFrame,
            "This ligand needs to be regularised before viewing table data - " + Env.bsn +
            "would you like to regularise now?",
            "Ligand is not regularised",
            JOptionPane.YES_NO_OPTION
         ) ;
         if( i != JOptionPane.YES_OPTION ) return ;
         currentPanel.drawMessage( "Please wait, " + target + " is being regularised" ) ;
         if( ! regularise( molecule, false ) ) return ;
      }
      molecule = currentPanel.gLib.get( target ) ;
      new ViewTable().showTable( molecule, mode ) ;
   }

   static void actionClear(){
      int i = JOptionPane.showConfirmDialog(
         mainFrame,
         "Are you sure you want to remove all the ligands from this Tab ?",
         "Clear Tab",
         JOptionPane.YES_NO_OPTION
      ) ;
      if( i != JOptionPane.YES_OPTION ) return ;
      currentPanel.gLib.undoableClear() ;
   }

   static void actionCheckChemistry(){
      GLib gLib = currentPanel.gLib ;
      if( gLib.isEmpty() ) return ;
      Molecule molecule = gLib.get( 0 ) ;
      if( gLib.size() > 1 ){                                            //:::  a bit too messy access to mol
         List<String> list = new ArrayList<String>( gLib.size() ) ;
         for( Molecule mol : gLib ) list.add( mol.key8 ) ;
         SelectMoleculeDialog dialog = new SelectMoleculeDialog() ;
         dialog.showDialog( list, "Check Valencies" ) ;
         molecule = gLib.get( dialog.ligandId ) ;
         if( molecule == null ) return ;
      }

      String errMessage = CheckChemistry.checkValencies( new Molecule( molecule, false ).atoms, "restrictive" ) ;
      for( Atom orig : molecule.atoms ) orig.coupled = orig ;

      String message = "Check Valencies for ligand " + molecule.key8 ;
      String message2 = Env.bsn ;
      message2 += Env.bsn + "Missing / excessive hydrogen atoms, if any, will be" ;
      message2 += Env.bsn + "handled during regularisation (Ligand > Regularise)" ;
      String message3 = Env.bsn ;
      message3 += Env.bsn + "However, missing or excessive hydrogen atoms will be" ;
      message3 += Env.bsn + "handled during regularisation (Ligand > Regularise)" ;
      if( errMessage == null ){
         DialogWindows.showInformationDialog( message + ": OK" + message2 ) ;
      }
      else{
         DialogWindows.showErrorDialog( message + ": " + errMessage + message3 ) ;
      }
   }

   static void actionCorrectBonds(){
      GLib gLib = currentPanel.gLib ;
      if( gLib.isEmpty() ) return ;
      Molecule molecule1 = gLib.get( 0 ) ;
      if( gLib.size() > 1 ){
         List<String> list = new ArrayList<String>( gLib.size() ) ;
         for( Molecule mol : gLib ) list.add( mol.key8 ) ;
         SelectMoleculeDialog dialog = new SelectMoleculeDialog() ;
         dialog.showDialog( list, "Correct Bonds" ) ;
         molecule1 = gLib.get( dialog.ligandId ) ;
         if( molecule1 == null ) return ;
      }

      Molecule molecule2 = null ;
      String diff = "" ;

      if( molecule1.restraints == null ){
         boolean lcoorNo = true ;
         boolean useLibmol = false ;
         molecule2 = molecule1.regularise( false, lcoorNo, useLibmol ) ;
         diff = CorrectBonds.compare( molecule1.atoms, molecule2.atoms ) ;
         for( Atom orig : molecule1.atoms ) orig.coupled = orig ;
      }

      String message = "Correct Bonds for ligand " + molecule1.key8 ;
      String message2 = Env.bsn ;
      message2 += Env.bsn + "Missing / excessive hydrogen atoms, if any, will be" ;
      message2 += Env.bsn + "handled during regularisation (Ligand > Regularise)" ;

      if( diff.equals( "" ) ){
         DialogWindows.showInformationDialog( message + " - no changes to make" + message2 ) ;
      }
      else{
         currentPanel.gLib.undoableReplaceMolecule( molecule1, molecule2 ) ;
         message2 = Env.bsn + Env.bsn + diff + message2 ;
         DialogWindows.showWarningDialog( message + " - changes have been made" + message2 ) ;
      }
   }

   static void actionUndo(){
//    String msg = currentPanel.gLib.undoRedo.undo() ;
      currentPanel.gLib.undoRedo.undo() ;
   }

   static void actionRedo(){
//    String msg = currentPanel.gLib.undoRedo.redo() ;
      currentPanel.gLib.undoRedo.redo() ;
   }

   static boolean regularise( Molecule molecule1, boolean useLibmol ){
      try{
         currentPanel.gLib.undoableRegularise( molecule1, useLibmol ) ;
         return true ;
      }
      catch( Exception e ){
         DialogWindows.showErrorDialog( "Regularisation failed" ) ;
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
      }
      return false ;
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class JLGPanel
   extends
      JPanel
   implements
      MouseListener,
      MouseMotionListener,
      MouseWheelListener
{
   GLib gLib = new GLib() ;

   Molecule selectedLink, storedLink ;
   Atom selectedAtom, storedAtom ;
   Bond selectedBond ;
   Molecule selectedBoxLink ;

   boolean selectLink = true ;
   boolean selectAtom = true ;
   boolean selectBond = true ;
   boolean dialogBlocked = false ;

   JLGPanel( Dimension dim, MouseListener listener ){
      setName( "JLigand Panel" ) ;
      setBackground( Color.WHITE ) ;
      setFocusable( true ) ;
      setPreferredSize( dim ) ;
      centreX = dim.getWidth()/ 2 ;
      centreY = dim.getHeight()/ 2 ;

      addMouseMotionListener( this ) ;
      addMouseWheelListener( this ) ;
      addMouseListener( listener ) ;
      addMouseListener( this ) ;

      if( Env.showLys ){
         String key8 = "LYS" ;
         boolean lcoorNo = false ;
         Monomer comp = StdLib.get( key8, lcoorNo ) ;
         comp.recentre() ;
         gLib.undoableAddMolecule( key8, comp ) ;
      }
   }

   void drawMessage( String message ){
      getGraphics().drawString( message, ( getWidth() - getRootPane().getWidth() )/ 2 + 44, 20 ) ;
   }

   MouseAdapter currentMode ;

   public void resetPanel(){
      storedLink = null ;
      storedAtom = null ;
      selectedLink = null ;
      selectedAtom = null ;
      selectedBond = null ;
      selectedBoxLink = null ;

      selectLink = true ;
      selectAtom = true ;
      selectBond = true ;
      dialogBlocked = false ;

      removeMouseListener( currentMode ) ;

      if( JLMain.option.equals( "DELETE" ) ){
         currentMode = new ModeDelete() ;
      }
      else if( JLMain.option.equals( "LINK" ) ){
         currentMode = new ModeLink() ;
      }
      else{
         currentMode = new ModeAdd() ;
      }

      addMouseListener( currentMode ) ;
      requestFocusInWindow() ;
      repaint() ;
   }

   // --------------------------------------------------------------------------

   class ModeDelete extends MouseAdapter{
      public void mouseClicked( MouseEvent e ){
//       if( e.getButton() == e.BUTTON1 ){
         if( e.getButton() == MouseEvent.BUTTON1 ){

            if( selectedAtom == null && selectedBond == null ) return ;
            if( selectedLink == null ) throw new NullPointerException() ;

            if( selectedLink.getResidue() == null ){
               String s = "Cannot edit a linked ligand (except to change atom and bond types) - " ;
// (1?)        String s = "Cannot edit a linked ligand (except to change the link bond type) - " ;
               s += "please undo the link, then edit" ;
               DialogWindows.showWarningDialog( s ) ;
            }
            else if( selectedAtom != null ){
               gLib.undoableRemoveAtom( selectedLink, selectedAtom ) ;
            }
            else if( selectedBond != null ){
               gLib.undoableRemoveBond( selectedLink, selectedBond ) ;
            }
            repaint() ;
         }
      }
   }
   // --------------------------------------------------------------------------

   abstract class ModeDialogs extends MouseAdapter{
      public void mouseClicked( MouseEvent e ){
//       if( e.getClickCount() != 2 || e.getButton() != e.BUTTON1 ) return ;
         if( e.getClickCount() != 2 || e.getButton() != MouseEvent.BUTTON1 ) return ;

         if( selectedAtom == null && selectedBoxLink == null && selectedBond == null ) return ;
         if( selectedLink == null ) throw new NullPointerException() ;

         storedLink = null ;
         storedAtom = null ;
         repaint() ;

         boolean isLink = selectedLink.getResidue() == null ;

         if( selectedAtom != null ){
// (1?)     if( isLink ){
// (1?)        cannotDialog() ;
// (1?)     }
// (1?)     else{
               new ChangeAtomDialog( selectedLink, selectedAtom ) ;
// (1?)     }
         }
         else if( selectedBond != null ){
// (1?)     if( isLink && selectedBond.atom1.residue == selectedBond.atom2.residue ){
// (1?)        cannotDialog() ;
// (1?)     }
// (1?)     else{
               new ChangeBondDialog( selectedLink, selectedBond ).showDialog() ;
// (1?)     }
         }
         else if( selectedBoxLink != null ){
            ChangeLigandIdDialog dialog = new ChangeLigandIdDialog( gLib, selectedBoxLink.key8 ) ;
            dialog.showDialog() ;
            if( dialog.newKey8 != null ){
               gLib.undoableRenameMolecule( selectedBoxLink, dialog.newKey8, dialog.newKey8 ) ;
            }
         }
         else throw new NullPointerException() ;
         selectedLink = null ;
         selectedAtom = null ;
         selectedBond = null ;
         repaint() ;
      }

      void cannotDialog(){
         String s = "Cannot edit a linked ligand (except to change the link bond type) - " ;
         s += "please undo the link, then edit" ;
         DialogWindows.showWarningDialog( s ) ;
      }
   }
   // --------------------------------------------------------------------------

   class ModeLink extends ModeDialogs{
      public void mouseClicked( MouseEvent e ){

         if( e.getClickCount() == 2 ){
            if( ! dialogBlocked ) super.mouseClicked( e ) ;
            return ;
         }
         dialogBlocked = false ;

         if( selectedAtom == storedAtom ) return ;

         if( selectedAtom == null && storedAtom != null ){
            storedLink = null ;
            storedAtom = null ;
         }
         else if( selectedAtom != null && storedAtom == null ){
            storedLink = selectedLink ;
            storedAtom = selectedAtom ;
         }
         else if( storedAtom != null ){
            boolean joined = false ;
            for( Bond bond : storedAtom ) joined |= selectedAtom == bond.another( storedAtom ) ;
            if( joined ){
               String msg = "Error - cannot join atoms which are already joined" ;
               DialogWindows.showErrorDialog( msg ) ;
            }
            else if( selectedLink == storedLink ){
               if( selectedLink.getResidue() == null ){
                  String s = "Cannot edit a linked ligand (except to change the link bond type) - " ;
                  s += "please undo the link, then edit" ;
                  DialogWindows.showWarningDialog( s ) ;
               }
               else{
                  gLib.undoableAddBond( selectedLink, storedAtom, selectedAtom, "single" ) ;
               }
            }
            else{
               Residue storedRsd = storedLink.getResidue() ;
               Residue selectedRsd = selectedLink.getResidue() ;
               if( storedRsd == null || selectedRsd == null ){
                  String s1 = storedRsd == null ? storedLink.key8 : null ;
                  String s2 = selectedRsd == null ? selectedLink.key8 : null ;
                  cannot2Dialog( "Cannot link linked ligand", s1, s2, "" ) ;
               }
               else if( storedRsd.cootId != selectedRsd.cootId ){
                  String s1 = storedLink.key8 ;
                  String s2 = selectedLink.key8 ;
                  cannot2Dialog( "Link between", s1, s2, " is not the one requested by Coot" ) ;
               }
               else if( storedRsd.mono == null || selectedRsd.mono == null ){
                  String s1 = storedRsd.mono == null ? storedRsd.key8 : null ;
                  String s2 = selectedRsd.mono == null ? selectedRsd.key8 : null ;
                  cannot2Dialog( "Please regularise ligand", s1, s2, " before linking" ) ;
               }
               else{
                  String key8 = storedLink.key8 + "-" + selectedLink.key8 ;
                  CorrectLinkIdDialog dialog = new CorrectLinkIdDialog( "The", gLib, key8 ) ;
                  key8 = dialog.newKey8 ;
                  if( key8 != null ){
                     gLib.undoableAddBond( key8, storedLink, storedAtom, selectedLink, selectedAtom, "single" ) ;
                  }
               }
            }
            selectedLink = null ;
            selectedAtom = null ;
            storedLink = null ;
            storedAtom = null ;
            dialogBlocked = true ;
         }
         repaint() ;
      }

      void cannot2Dialog( String s0, String s1, String s2, String s3 ){
         String s = s0 ;
         if( s1 == null && s2 == null ) throw new NullPointerException() ;
         else if( s1 == null ) s += " " + s2 + s3 ;
         else if( s2 == null ) s += " " + s1 + s3 ;
         else s += "s " + s1 + " and " + s2 + s3 ;
         DialogWindows.showWarningDialog( s ) ;
      }
   }
   // --------------------------------------------------------------------------

   class ModeAdd extends ModeDialogs{
      public void mouseClicked( MouseEvent e ){
         super.mouseClicked( e ) ;

         boolean shiftPressed = JLMain.buttonsMask == InputEvent.SHIFT_DOWN_MASK ;

         if( storedAtom != null && selectedAtom == null && shiftPressed ){

            if( selectedLink == null ) throw new NullPointerException() ;

            if( selectedLink.getResidue() == null ){
               String s = "Cannot edit a linked ligand (except to change the link bond type) - " ;
               s += "please undo the link, then edit" ;
               DialogWindows.showWarningDialog( s ) ;
            }
            else{
               storedAtom = gLib.undoableAddAtom( storedLink, storedAtom, JLMain.option, rAn ) ;
            }
         }
         else if( selectedAtom != null ){
            storedLink = selectedLink ;
            storedAtom = selectedAtom ;
         }
//::     else{
//::        storedLink = null ;
//::        storedAtom = null ;
//::     }
         repaint() ;
      }
   }
   // --------------------------------------------------------------------------

   Point2D cursorNew = new Point2D.Double( 0f, 0f ) ;
   Point2D cursorOld = new Point2D.Double( 0f, 0f ) ;
   double centreX ;
   double centreY ;
   double scale = 25 ;
   double[] rAo = new double[2] ;
   double[] rAn = new double[2] ;
   double[] rAd = new double[2] ;
   double coef = 0.08f;

   public void setCoordinates( Point2D p ){
      cursorOld.setLocation( cursorNew ) ;
      cursorNew.setLocation( p ) ;
      rAo[0] = ( cursorOld.getX() - centreX )/ scale ;
      rAo[1] = ( cursorOld.getY() - centreY )/ scale ;
      rAn[0] = ( cursorNew.getX() - centreX )/ scale ;
      rAn[1] = ( cursorNew.getY() - centreY )/ scale ;
      rAd[0] = ( cursorNew.getX() - cursorOld.getX() )/ scale ;
      rAd[1] = ( cursorNew.getY() - cursorOld.getY() )/ scale ;
   }

   Atom movingAtom = null ;
   Molecule movingLink = null ;

   public void mousePressed( MouseEvent e ){
      setCoordinates( e.getPoint() ) ;
      setCoordinates( e.getPoint() ) ;
      selectBond = selectAtom = selectLink = false ;
      stopMoving() ;

      if( JLMain.buttonsMask == ( InputEvent.BUTTON1_DOWN_MASK | InputEvent.SHIFT_DOWN_MASK ) ){
         if( selectedLink != null && selectedAtom != null ){
            selectedAtom = gLib.undoableReplaceAtom( selectedLink, selectedAtom ) ;
            movingAtom = selectedAtom ;
            movingLink = selectedLink ;
         }
      }
      repaint() ;
   }
   public void mouseReleased( MouseEvent e ){
      setCoordinates( e.getPoint() ) ;
      setCoordinates( e.getPoint() ) ;
      selectBond = selectAtom = selectLink = true ;
      stopMoving() ;
      repaint() ;
   }
   public void mouseClicked( MouseEvent e ){
      stopMoving() ;
   }
   public void mouseEntered( MouseEvent e ){
      setCoordinates( e.getPoint() ) ;
      setCoordinates( e.getPoint() ) ;
      selectBond = selectAtom = selectLink = true ;
      stopMoving() ;
   }
   public void mouseExited( MouseEvent e ){
      stopMoving() ;
   }

   public void stopMoving(){
      if( movingLink != null && movingAtom != null ){
         gLib.movingAtomStopped( movingLink, movingAtom ) ;
      }
      movingAtom = null ;
      movingLink = null ;
   }

   public void mouseMoved( MouseEvent e ){ mouseDragged( e ) ; }
   public void mouseDragged( MouseEvent e ){
      setCoordinates( e.getPoint() ) ;

      if( movingLink != null && movingAtom != null ){
         gLib.moveAtom( movingLink, movingAtom, rAd ) ;
      }
      else if( JLMain.buttonsMask == ( InputEvent.BUTTON1_DOWN_MASK | InputEvent.ALT_DOWN_MASK ) ){
         double yo = 2* centreY* coef ;
         double ycNew = Math.max( 0f, cursorNew.getY() ) ;
         double ycOld = Math.max( 0f, cursorOld.getY() ) ;
         double scaleToApply = ( yo + ycNew )/ ( yo + ycOld ) ;
         scale *= scaleToApply ;
      }
      else if( JLMain.buttonsMask == ( InputEvent.BUTTON1_DOWN_MASK | InputEvent.CTRL_DOWN_MASK ) ){
         if( selectedLink != null ){
            gLib.moveMol( selectedLink, rAd ) ;
         }
      }
      else if( JLMain.buttonsMask == InputEvent.BUTTON1_DOWN_MASK ){
         if( selectedLink != null ){
            gLib.rotateMol( selectedLink, rAo, rAn ) ;
         }
      }
      else if( JLMain.buttonsMask == InputEvent.BUTTON3_DOWN_MASK ){
         if( selectedLink != null ){
            gLib.moveMol( selectedLink, rAd ) ;
         }
      }
      repaint() ;
   }

   public void mouseWheelMoved( MouseWheelEvent e ){
      double y0 = centreY* coef ;
      double y1 = Math.max( y0/ 2, y0 + e.getWheelRotation() ) ;
      double scaleToApply = y1/ y0 ;
      scale *= scaleToApply ;
      repaint();
   }

   // --------------------------------------------------------------------------

   void checkVisited(){
      for( Molecule link : gLib ){
         for( Atom atom : link.atoms ){
            for( Bond bond : atom ) if( bond.visited ) throw new NullPointerException() ;
         }
      }
   }

// String message = "" ;

   public void paintComponent( Graphics g ){
      super.paintComponent( g ) ;

//    g.drawString( message, ( getWidth() - getRootPane().getWidth() )/ 2 + 44, 20) ;

      for( Molecule link : gLib ) Collections.sort( link.atoms, Atom.zComparator ) ;   //:: to remove?

      checkVisited() ;

      Graphics2D g2 = (Graphics2D) g ;
      Ellipse2D ellipse = new Ellipse2D.Double() ;
      Line2D bondLine = new Line2D.Double() ;
      Point2D atomPoint = new Point2D.Double() ;
      Point2D anotherPoint = new Point2D.Double() ;
//    double r = 4f ;

      Molecule focusedLink = null ;
      Atom focusedAtom = null ;
      Bond focusedBond = null ;
      double distBond = 4f ;
      double distAtom = 5f ;
      double dist = 10000000f ;
      for( Molecule link : gLib ){
         double distLink = dist* 2 ;
         for( Atom atom : link.atoms ){
            boolean showAtom = JLMain.showHydrogens || ! atom.symbol.equals( "H" ) ;
            atomPoint.setLocation(
               atom.xyz[0]* scale + centreX,
               atom.xyz[1]* scale + centreY
            ) ;
            for( Bond bond : atom ){
               if( bond.visited ){
                  bond.visited = false ;
               }
               else{
                  bond.visited = true ;
                  Atom anotherAtom = bond.another( atom ) ;
                  boolean showAnother = JLMain.showHydrogens || ! anotherAtom.symbol.equals( "H" ) ;
                  if( showAtom && showAnother ){
                     anotherPoint.setLocation(
                        anotherAtom.xyz[0]* scale + centreX,
                        anotherAtom.xyz[1]* scale + centreY
                     ) ;
                     bondLine.setLine( anotherPoint, atomPoint ) ;
                     double db = bondLine.ptSegDist( cursorNew ) ;
                     if( db < distBond ){ focusedAtom = null ; focusedBond = bond ; }
                     if( db < distLink ) distLink = db ;
                  }
               }
            }
            if( showAtom ){
               double da = atomPoint.distance( cursorNew ) ;
               if( da < distAtom ){ focusedAtom = atom ; focusedBond = null ; }
               if( da < distLink ) distLink = da ;
            }
         }
         if( focusedAtom != null || focusedBond != null ){ focusedLink = link ; break ; }
         if( distLink < dist ){ dist = distLink ; focusedLink = link ; }
      }
      checkVisited() ;

      if( selectLink ) selectedLink = focusedLink ;
      if( selectAtom ) selectedAtom = focusedAtom ;
      if( selectBond ) selectedBond = focusedBond ;

      selectedBoxLink = null ;
      for( Molecule link : gLib ){
         double xS = link.xyz[0]* scale + centreX ;
         double yS = link.xyz[1]* scale + centreY ;
         double rad = link.radius* scale ;

         g.setColor( Color.LIGHT_GRAY ) ;
         if( Env.vbGraphics ){
            ellipse.setFrame( xS - rad, yS - rad, 2* rad, 2* rad ) ;
            g2.draw( ellipse ) ;
            g2.drawLine( (int)xS, (int)yS, (int)xS, (int)yS + (int)rad ) ;
         }

         FontRenderContext frc = g2.getFontRenderContext() ;
         Rectangle2D rect = f1.getStringBounds( link.key8, frc ) ;
         double xC = xS ;
         double yC = yS + rad + 17 ;
         double xD = rect.getWidth()/ 2 ;
         double yD = rect.getHeight()/ 2 ;
         double xM = 3.0 ;
         double yM = 3.0 ;
         Point2D pT = new Point2D.Double( xC - xD     , yC + yD            ) ;
         Point2D p1 = new Point2D.Double( xC - xD - xM, yC - yD - yM + 5.0 ) ;
         Point2D p2 = new Point2D.Double( xC + xD + xM, yC + yD + yM + 0.0 ) ;
         Rectangle2D box = new Rectangle2D.Double() ;
         box.setFrameFromDiagonal( p1, p2 ) ;
         g2.setStroke( line ) ;
         if( Env.vbGraphics ){
            g2.draw( box ) ;
         }

         g.setFont( f1 ) ;
         g.setColor( Color.BLACK ) ;
         if( link.restraints == null ) g.setColor( Color.RED ) ;
         g2.drawString( link.key8, (float)pT.getX(), (float)pT.getY() ) ;
         g.setColor( Color.BLACK ) ;

         if( selectedAtom == null && selectedBond == null && selectedBoxLink == null ){
            if( box.contains( cursorNew ) ){
               selectedBoxLink = link ;
            }
         }
      }

      for( Molecule link : gLib ){
         for( Atom atom : link.atoms ){
            for( Bond bond : atom ){
               if( bond.visited ){
                  bond.visited = false ;
               }
               else{
                  bond.visited = true ;
                  drawBond( g, g2, bond ) ;
               }
            }
            drawAtom( g, g2, atom ) ;
         }
      }
      checkVisited() ;
   }
   // --------------------------------------------------------------------------

   Font f1 = new Font( "serif", Font.BOLD, 14 ) ;
   Font f2 = new Font( "serif", Font.BOLD, 10 ) ;
   float dash1[] = { 3.0f } ;
   BasicStroke line = new BasicStroke( 1.0f ) ;
   BasicStroke dashed = new BasicStroke( 1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, dash1, 0.0f ) ;
   BasicStroke line2 = new BasicStroke( 2.0f ) ;
   BasicStroke dashed2 = new BasicStroke( 2.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, dash1, 0.0f ) ;

   public void drawAtom( Graphics g, Graphics2D g2, Atom atom ){
      if( ! JLMain.showHydrogens && atom.symbol.equals( "H" ) ) return ;

      AtomDisplay atomDisplay = new AtomDisplay( atom.symbol ) ;

      g.setFont( f1 ) ;
      g.setColor( Color.BLACK ) ;
      g2.setStroke( line ) ;

      int x1 = (int)( atom.xyz[0]* scale + centreX ) ;
      int y1 = (int)( atom.xyz[1]* scale + centreY ) ;

      int xOffset = atomDisplay.xOffset ;
      int yOffset = atomDisplay.yOffset ;

      int rad = atomDisplay.ovalRadius ;
      g.setColor( Color.WHITE ) ;
      g.fillOval( x1 - rad, y1 - rad, rad * 2, rad * 2 ) ;
//==  if( Env.vbGraphics && atom.hasChanged() ){
      if( Env.vbGraphics && atom.isHidden() ){
         int rr = rad* 2 ;
//==     g.setColor( Color.BLUE ) ;
         g.setColor( Color.RED ) ;
         g.drawOval( x1 - rr, y1 - rr, rr * 2, rr * 2 ) ;
      }
      if( Env.vbGraphics && atom.hasChangedCharge() ){
         int rr = rad* 2 + 2 ;
//==     g.setColor( Color.MAGENTA ) ;
         g.setColor( Color.GREEN ) ;
         g.drawOval( x1 - rr, y1 - rr, rr * 2, rr * 2 ) ;
      }

      g.setColor( atomDisplay.color ) ;
      String atomSymbol = atom.symbol ;
      if( JLMain.showNames ) atomSymbol = atom.key4 ;
      g.drawString( atomSymbol, x1 + xOffset, y1 + yOffset ) ;

      if( JLMain.showCharges ){
         String s = "" ;
         if( atom.charge > 0 ) s = "+" + atom.charge ;
         else if( atom.charge < 0 ) s = "" + atom.charge ;

         g.setFont( f2 ) ;
         g.drawString( s, x1 - xOffset, y1 - yOffset ) ;
         g.setFont( f1 ) ;
      }

      if( atom == storedAtom || atom == selectedAtom ){
         if( JLMain.option.equals( "DELETE" ) ){
            g.setColor( Color.RED ) ;
         }
         else if( JLMain.option.equals( "LINK" ) ){
            g.setColor( Color.MAGENTA ) ;
         }
         else{
            g.setColor( Color.BLUE ) ;
         }
         int focusLineX = atomDisplay.focusLineX ;
         int focusLineY = yOffset + 2 ;
         g.drawRect( x1 - focusLineX, y1 + focusLineY, 2 * focusLineX, 1 ) ;
      }
   }
   // --------------------------------------------------------------------------

   public void drawBond( Graphics g, Graphics2D g2, Bond bond ){
      if( ! JLMain.showHydrogens && ( bond.atom1.symbol.equals( "H" ) || bond.atom2.symbol.equals( "H" ) ) ) return ;


      int x1 = (int)( bond.atom1.xyz[0]* scale + centreX ) ;
      int y1 = (int)( bond.atom1.xyz[1]* scale + centreY ) ;
      int x2 = (int)( bond.atom2.xyz[0]* scale + centreX ) ;
      int y2 = (int)( bond.atom2.xyz[1]* scale + centreY ) ;
      int dx = 0 ;
      int dy = 0 ;

      double grad = ( (double)y2 - (double)y1 )/ ( (double)x2 - (double)x1 ) ;
      double rads = Math.abs( Math.atan( grad ) );
      if( rads >= 0 && rads <= ( Math.PI/ 8.0 ) ){
         dx = 0 ; dy = 3 ;
      }
      else if( rads > ( Math.PI/ 8.0 ) && rads <= ( 3.0* Math.PI/ 8.0 ) ){
         dx = 2 ; dy = 2 ;
      }
      else if( rads > ( 3.0* Math.PI/ 8.0 ) && rads <= Math.PI/ 2.0 ){
         dx = 3 ; dy = 0 ;
      }
      if( dy == 2 && grad > 0 ) dy *= -1 ;

      g.setColor( Color.BLACK ) ;
      if( bond.atom1.residue != bond.atom2.residue ) g.setColor( Color.MAGENTA ) ;

      boolean bold = false ;
      g2.setStroke( line ) ;
      if( bond == selectedBond && JLMain.option.equals( "DELETE" ) ){
         bold = true ;
         g2.setStroke( line2 ) ;
         g.setColor( Color.RED ) ;
      }

      if( bond.type.equals( "single" ) ){
         g2.draw( new Line2D.Double( x1, y1, x2, y2 ) ) ;
      }
      else if( bond.type.equals( "double" ) ){
         g2.draw( new Line2D.Double( x1 - dx, y1 - dy, x2 - dx, y2 - dy ) ) ;
         g2.draw( new Line2D.Double( x1 + dx, y1 + dy, x2 + dx, y2 + dy ) ) ;
      }
      else if( bond.type.equals( "triple" ) ){
         g2.draw( new Line2D.Double( x1 - dx, y1 - dy, x2 - dx, y2 - dy ) ) ;
         g2.draw( new Line2D.Double( x1, y1, x2, y2 ) ) ;
         g2.draw( new Line2D.Double( x1 + dx, y1 + dy, x2 + dx, y2 + dy ) ) ;
      }
      else if( bond.type.equals( "aromatic" ) || bond.type.equals( "deloc" ) ){
         g2.setStroke( dashed ) ;
         if( bold ) g2.setStroke( dashed2 ) ;
         g2.draw( new Line2D.Double( x1 - dx, y1 - dy, x2 - dx, y2 - dy ) ) ;
         g2.draw( new Line2D.Double( x1 + dx, y1 + dy, x2 + dx, y2 + dy ) ) ;
      }
      else if( bond.type.equals( "metal" ) ){
         g2.setStroke( dashed ) ;
         if( bold ) g2.setStroke( dashed2 ) ;
         g2.draw( new Line2D.Double( x1, y1, x2, y2 ) ) ;
      }
   }
   // --------------------------------------------------------------------------

   public class AtomDisplay{

      private Color color = null ;
      private int xOffset = 0 ;
      private int yOffset = 0 ;
      private int ovalRadius = 0 ;
      private int focusLineX = 0 ;

      public AtomDisplay( String atomType ){

         int atomWidth = 0 ;
         int atomHeight = 0 ;

         if( atomType.equals( "C" ) ){
            color = Color.BLACK ;
            atomWidth = 8 ;
            atomHeight = 10 ;
            focusLineX = 6 ;
         }
         else if( atomType.equals( "O" ) ){
            color = Color.RED ;
            atomWidth = 9 + 1 ;
            atomHeight = 10 ;
            focusLineX = 6 ;
         }
         else if( atomType.equals( "N" ) ){
            color = Color.BLUE ;
            atomWidth = 11 ;
            atomHeight = 10 ;
            focusLineX = 6 ;
         }
         else if( atomType.equals( "S" ) ){
            color = Color.ORANGE ;
            atomWidth = 6 ;
            atomHeight = 10 ;
            focusLineX = 5 ;
         }
         else if( atomType.equals( "F" ) ){
            color = Color.MAGENTA ;
            atomWidth = 8 ;
            atomHeight = 10 ;
            focusLineX = 5 ;
         }
         else if( atomType.equals( "Cl" ) ){
            color = Color.MAGENTA ;
            atomWidth = 14 ; // is really 12 but focus line looked better if 14
            atomHeight = 11 ;
            focusLineX = 7 ;
         }
         else if( atomType.equals( "Br" ) ){
            color = Color.MAGENTA ;
            atomWidth = 15 ;
            atomHeight = 10 ;
            focusLineX = 7 ;
         }
         else if( atomType.equals( "I" ) ){
            color = Color.MAGENTA ;
            atomWidth = 4 ;
            atomHeight = 10 ;
            focusLineX = 5 ;
         }
         else if( atomType.equals( "P" ) ){
            color = Color.CYAN ;
            atomWidth = 8 ;
            atomHeight = 10 ;
            focusLineX = 5 ;
         }
         else if( atomType.equals( "H" ) ){
            color = Color.BLACK ;
            atomWidth = 11 ;
            atomHeight = 10 ;
            focusLineX = 6 ;
         }
         else{
            color = Color.BLUE ;
            atomWidth = 11 ;
            atomHeight = 10 ;
            focusLineX = 6 ;
         }

         ovalRadius = atomWidth > atomHeight ? atomWidth : atomHeight ;
         ovalRadius = ( ovalRadius + 2 )/ 2 ;
         xOffset = - atomWidth/ 2 ;
         yOffset = atomHeight / 2 ;
      }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

class DialogWindows{
   public static int panelWidth, panelHeight ;
   public static int screenWidth = 0 ;
   public static int screenHeight = 0 ;
   public static int windowWidth = 0 ;
   public static int windowHeight = 0 ;
   public static int locationTop = 0 ;
   public static int locationLeft = 0 ;
   public static int locationTop2 = 0 ;
   public static int dialogWidth = 500 ;
   public static int dialogHeight = 300 ;

   public static PeriodicTableDialog periodicTableDialog = new PeriodicTableDialog() ;

   public static HelpDialog helpDialog =
      new HelpDialog( "JLigand Help", "/resources/helpDialog.html" )
   ;
   public static HelpDialog helpJLigand =
      new HelpDialog( "JLigand Help", "/resources/helpJLigand.html" )
   ;
   public static HelpDialog helpJLigandMouse =
      new HelpDialog( "JLigand Mouse Help", "/resources/helpJLigandMouse.html" )
   ;
   public static HelpDialog helpJLigandKeypad =
      new HelpDialog( "JLigand Keypad Help", "/resources/helpJLigandKeypad.html" )
   ;

   public static void initialise( Dimension dim ){
      screenWidth =  dim.width ;
      screenHeight = dim.height ;
      windowWidth = 9 * screenWidth / 10 ;
      windowHeight = 7 * screenHeight / 8 ;
      panelWidth = windowWidth - 55 ;
      panelHeight = windowHeight - 145 ;
      locationTop = screenHeight/16 - 10 ;
      locationLeft = screenWidth/20 ;
      locationTop2 = locationTop + (windowHeight - panelHeight) ;
   }

   public static void showErrorDialog( String message ){
      JOptionPane.showMessageDialog( JLMain.mainFrame, message, "Error", JOptionPane.ERROR_MESSAGE ) ;
   }

   public static void showInformationDialog( String message ){
      JOptionPane.showMessageDialog( JLMain.mainFrame, message, "Info", JOptionPane.INFORMATION_MESSAGE ) ;
   }

   public static void showWarningDialog( String message ){
      JOptionPane.showMessageDialog( JLMain.mainFrame, message, "Warning", JOptionPane.WARNING_MESSAGE ) ;
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class JLMenuBar extends JMenuBar{
// private JMenu jLigandMenu, tabMenu, fileMenu, openMenu, viewMenu, ligandMenu;
   private JMenu jLigandMenu, tabMenu, fileMenu, openMenu, ligandMenu;
   private JMenu openAsIsMenu ;
   private JMenu libmolMenu;
   private JMenu regulariseMenu, saveCooMenu, ligandDataMenu, viewLigandDataMenu;
   private JMenu saveCompMenu, saveLinkMenu ;
   private JMenu cootMenu ;
   private JMenu saveCootLinkMenu ;
   private JMenu appendCootLinkMenu ;
   private JMenu appendCompMenu, appendLinkMenu ;
   private JMenu findMenu ;
// private JMenu viewLinkDataMenu, deleteMenu, optionsMenu, helpMenu;
   private JMenu viewLinkDataMenu, deleteMenu, helpMenu;
   private JMenuItem aboutItem, quitItem;
// private JMenuItem openCIFItem, openPDBItem, openCSDItem, openSybylMol2Item, openSDFItem, openSmilesItem;
   private JMenuItem openCIFItem, openPDBItem, openSDFItem, openMOLItem;
   private JMenuItem openAsIsCIFItem, openAsIsPDBItem, openAsIsSDFItem, openAsIsMOLItem;
   private JMenuItem enterSmileItem, addTabItem, deleteTabItem, clearTabItem;
// private JMenuItem periodicTableItem, useOwnLibraryItem, viewFileItem, energyLibraryItem;
   private JMenuItem periodicTableItem, viewFileItem;
// private JMenuItem helpItem, helpJLigandItem, helpJLigandMouseItem, helpJLigandKeypadItem;
   private JMenuItem helpJLigandItem, helpJLigandMouseItem, helpJLigandKeypadItem;
   private ActionListener listener ;

   JLMenuBar( ActionListener jLigand ){
      listener = jLigand ;

      aboutItem = new JMenuItem("About ...");
      aboutItem.addActionListener(listener);
      aboutItem.setActionCommand( "ABOUT" ) ;

      quitItem = new JMenuItem("Quit ...");
      quitItem.addActionListener(listener);
      quitItem.setActionCommand( "QUIT" ) ;

      addTabItem = new JMenuItem("New Tab");
      addTabItem.addActionListener(listener);
      addTabItem.setActionCommand( "TAB_NEW" ) ;

      openCIFItem = new JMenuItem("CIF File ...");
      openCIFItem.addActionListener(listener);
      openCIFItem.setActionCommand( "FILE_LIB" ) ;

      openPDBItem = new JMenuItem("PDB File ...");
      openPDBItem.addActionListener(listener);
      openPDBItem.setActionCommand( "FILE_PDB" ) ;

      openSDFItem = new JMenuItem("SDF File ...");
      openSDFItem.addActionListener(listener);
      openSDFItem.setActionCommand("FILE_SDF");

      openMOLItem = new JMenuItem("MOL(2) File ...");
      openMOLItem.addActionListener(listener);
      openMOLItem.setActionCommand("FILE_MOL");

      enterSmileItem = new JMenuItem("Enter Smile String ...");
      enterSmileItem.addActionListener(listener);
      enterSmileItem.setActionCommand("ENTER_SMILE");

      openAsIsCIFItem = new JMenuItem("CIF File ...");
      openAsIsCIFItem.addActionListener(listener);
      openAsIsCIFItem.setActionCommand( "FILE_LIB_ASIS" ) ;

      openAsIsPDBItem = new JMenuItem("PDB File ...");
      openAsIsPDBItem.addActionListener(listener);
      openAsIsPDBItem.setActionCommand( "FILE_PDB_ASIS" ) ;

      openAsIsSDFItem = new JMenuItem("SDF File ...");
      openAsIsSDFItem.addActionListener(listener);
      openAsIsSDFItem.setActionCommand("FILE_SDF_ASIS");

      openAsIsMOLItem = new JMenuItem("MOL(2) File ...");
      openAsIsMOLItem.addActionListener(listener);
      openAsIsMOLItem.setActionCommand("FILE_MOL_ASIS");

      deleteTabItem = new JMenuItem("Delete Current Tab ...");
      deleteTabItem.addActionListener(listener);
      deleteTabItem.setActionCommand( "TAB_DELETE" ) ;

      clearTabItem = new JMenuItem("Clear Tab ...");
      clearTabItem.addActionListener(listener);
      clearTabItem.setActionCommand( "TAB_CLEAR" ) ;

      openMenu = new JMenu("Open");
      openMenu.add(openCIFItem);
      openMenu.add(openPDBItem);
      boolean hideItem = false ;
//    hideItem = Env.asIs ;
      if( ! hideItem ) openMenu.add(openSDFItem);
      if( ! hideItem ) openMenu.add(openMOLItem);
      openMenu.add(enterSmileItem);

      openAsIsMenu = new JMenu("Open As Is");
      openAsIsMenu.add(openAsIsCIFItem);
      openAsIsMenu.add(openAsIsPDBItem);
      openAsIsMenu.add(openAsIsSDFItem);
      openAsIsMenu.add(openAsIsMOLItem);

      saveCompMenu = new JMenu("Save as Monomer");

      saveLinkMenu = new JMenu("Save as Link");

      appendCompMenu = new JMenu("Append as Monomer");

      appendLinkMenu = new JMenu("Append as Link");

      saveCooMenu = new JMenu("Save Coordinates");

      saveCootLinkMenu = new JMenu("Save as Link and Send to Coot");
      appendCootLinkMenu = new JMenu("Append as Link and Send to Coot");

      libmolMenu = new JMenu("LibMol");
      regulariseMenu = new JMenu("Regularise");

      deleteMenu = new JMenu("Delete");

      findMenu = new JMenu("Find in standard library");

      viewLigandDataMenu = new JMenu("View Ligand Data");

      viewLinkDataMenu = new JMenu("View Link File");

      periodicTableItem = new JMenuItem("Periodic Table ...");
      periodicTableItem.addActionListener(listener);
      periodicTableItem.setActionCommand("PERIODIC_TABLE");

      viewFileItem = new JMenuItem("View File ...");
      viewFileItem.addActionListener(listener);
      viewFileItem.setActionCommand("VIEW_FILE");

      helpJLigandItem = new JMenuItem("JLigand Help");
      helpJLigandItem.addActionListener(listener);
      helpJLigandItem.setActionCommand("HELP_JLIGAND");

      helpJLigandMouseItem = new JMenuItem("JLigand Mouse Help");
      helpJLigandMouseItem.addActionListener(listener);
      helpJLigandMouseItem.setActionCommand("HELP_MOUSE");

      helpJLigandKeypadItem = new JMenuItem("JLigand Keypad Help");
      helpJLigandKeypadItem.addActionListener(listener);
      helpJLigandKeypadItem.setActionCommand("HELP_KEYPAD");

//    boolean newLayout = false ;
      boolean newLayout = true ;

      jLigandMenu = new JMenu("JLigand");
      jLigandMenu.add(aboutItem);
      jLigandMenu.add(quitItem);

      String label = "File" ;
      if( newLayout && Env.talkToCoot ){
         label = "File / Coot" ;
      }
      fileMenu = new JMenu(label);
      fileMenu.add(openMenu);
      if( Env.asIs ) fileMenu.add(openAsIsMenu);
      fileMenu.addSeparator() ;
      fileMenu.add(saveCompMenu);
      fileMenu.add(saveLinkMenu);
      if( newLayout && Env.talkToCoot ){
         fileMenu.add(saveCootLinkMenu);
      }
      fileMenu.addSeparator() ;
      fileMenu.add(appendCompMenu);
      fileMenu.add(appendLinkMenu);
      if( newLayout && Env.talkToCoot ){
         fileMenu.add(appendCootLinkMenu);
      }
      fileMenu.addSeparator() ;
      fileMenu.add(saveCooMenu);

      cootMenu = new JMenu("Coot");
      if( ! newLayout && Env.talkToCoot ){
         cootMenu.add(saveCootLinkMenu);
         cootMenu.add(appendCootLinkMenu);
      }

      tabMenu = new JMenu("Tab");
      tabMenu.add(addTabItem);
      tabMenu.add(deleteTabItem);
      tabMenu.add(clearTabItem);

      ligandMenu = new JMenu("Ligand");
      if( Libmol.isAvailable() ) ligandMenu.add(libmolMenu);
      ligandMenu.add(regulariseMenu);
      ligandMenu.add(deleteMenu);
      if( Versions.enableSearch() ) ligandMenu.add(findMenu);

      ligandDataMenu = new JMenu("View");
      ligandDataMenu.add(viewLigandDataMenu);
      ligandDataMenu.add(viewLinkDataMenu);
      ligandDataMenu.add(viewFileItem);

      helpMenu = new JMenu("Help");
      helpMenu.add(helpJLigandItem);
      helpMenu.add(helpJLigandMouseItem);
      helpMenu.add(helpJLigandKeypadItem);

      int strutWidth = 25;
      this.add(Box.createHorizontalStrut(strutWidth));
      this.add(jLigandMenu);
      this.add(Box.createHorizontalStrut(strutWidth));
      this.add(fileMenu);
      this.add(Box.createHorizontalStrut(strutWidth));
      this.add(tabMenu);
      this.add(Box.createHorizontalStrut(strutWidth));
      this.add(ligandMenu);
      this.add(Box.createHorizontalStrut(strutWidth));
      this.add(ligandDataMenu);
      this.add(Box.createHorizontalStrut(strutWidth));
      this.add(helpMenu);
      if( ! newLayout && Env.talkToCoot ){
         this.add(Box.createHorizontalStrut(strutWidth));
         this.add(cootMenu);
      }

      fileMenu.addMouseListener( new MouseAdapter(){
         public void mouseEntered( MouseEvent e ){ resetMenus() ; }
      } ) ;

      cootMenu.addMouseListener( new MouseAdapter(){
         public void mouseEntered( MouseEvent e ){ resetMenus() ; }
      } ) ;

      ligandMenu.addMouseListener( new MouseAdapter(){
         public void mouseEntered( MouseEvent e ){ resetMenus() ; }
      } ) ;

      ligandDataMenu.addMouseListener( new MouseAdapter(){
         public void mouseEntered( MouseEvent e ){ resetMenus() ; }
      } ) ;
   }

   public void resetMenus() {
      saveCompMenu.removeAll();
      saveLinkMenu.removeAll();
      appendCompMenu.removeAll();
      appendLinkMenu.removeAll();
      saveCootLinkMenu.removeAll();
      appendCootLinkMenu.removeAll();
      saveCooMenu.removeAll();
      libmolMenu.removeAll();
      regulariseMenu.removeAll();
      viewLigandDataMenu.removeAll();
      deleteMenu.removeAll();
      findMenu.removeAll();
      viewLinkDataMenu.removeAll();

      GLib gLib = JLMain.currentPanel.gLib ;
      for( Molecule link : gLib ){
         String key8 = link.key8 ;
         if( link.getResidue() == null ){
            int cootId = link.atoms.get(0).residue.cootId ;
            if( cootId > 0 ){
               addCootLinkItem( key8 ) ;
            }
            else if( cootId == 0 ){
               addLinkItem( key8 ) ;
            }
            else throw new NullPointerException() ;
         }
         else addCooItem( key8 ) ;
         addLigandItem( key8 ) ;
      }
   }

   public void addLigandItem( String ligandId ){

      String[] itemName = { "Atoms", "Bonds", "Angles", "Torsions", "Chiralities", "Planes" } ;
      String[] itemCommand = { "ATOMS ", "BONDS ", "ANGLES ", "TORSIONS ", "CHIRALITIES ", "PLANES " } ;
      JMenu menu = new JMenu( ligandId ) ;
      JMenuItem item = null ;
      for( int i = 0 ; i < 6 ; i++ ){
         item = new JMenuItem( itemName[i] ) ;
         item.addActionListener( listener ) ;
         item.setActionCommand( itemCommand[i] + ligandId ) ;
         menu.add( item ) ;
      }
      viewLigandDataMenu.add(menu);

      JMenuItem saveCompItem = new JMenuItem(ligandId + " ...");
      saveCompItem.addActionListener(listener);
      saveCompItem.setActionCommand( "SAVE_COMP " + ligandId ) ;
      saveCompMenu.add(saveCompItem);

      JMenuItem appendCompItem = new JMenuItem(ligandId + " ...");
      appendCompItem.addActionListener(listener);
      appendCompItem.setActionCommand( "APPEND_COMP " + ligandId ) ;
      appendCompMenu.add(appendCompItem);

      JMenuItem libmolItem = new JMenuItem(ligandId);
      libmolItem.addActionListener(listener);
      libmolItem.setActionCommand( "LIBMOL " + ligandId ) ;
      libmolMenu.add(libmolItem);

      JMenuItem regulariseItem = new JMenuItem(ligandId);
      regulariseItem.addActionListener(listener);
      regulariseItem.setActionCommand( "REGULARISE " + ligandId ) ;
      regulariseMenu.add(regulariseItem);

      JMenuItem deleteItem = new JMenuItem(ligandId + " ...");
      deleteItem.addActionListener(listener);
      deleteItem.setActionCommand( "DELETE " + ligandId ) ;
      deleteMenu.add(deleteItem);

      JMenuItem findItem = new JMenuItem(ligandId + " ...");
      findItem.addActionListener(listener);
      findItem.setActionCommand( "FIND_COMP " + ligandId ) ;
      findMenu.add(findItem);
   }

   public void addCooItem(String ligandId) {
      JMenuItem saveCooItem = new JMenuItem(ligandId + " ...");
      saveCooMenu.add(saveCooItem);
      saveCooItem.addActionListener(listener);
      saveCooItem.setActionCommand( "SAVE_PDB " + ligandId ) ;
   }

   public void addLinkItem(String ligandId) {
      JMenuItem linkItem = new JMenuItem(ligandId);
      viewLinkDataMenu.add(linkItem);
      linkItem.addActionListener(listener);
      linkItem.setActionCommand( "VIEW_LINK " + ligandId ) ;

      JMenuItem saveLinkItem = new JMenuItem(ligandId + " ...");
      saveLinkItem.addActionListener(listener);
      saveLinkItem.setActionCommand( "SAVE_LINK " + ligandId ) ;
      saveLinkMenu.add(saveLinkItem);

      JMenuItem appendLinkItem = new JMenuItem(ligandId + " ...");
      appendLinkItem.addActionListener(listener);
      appendLinkItem.setActionCommand( "APPEND_LINK " + ligandId ) ;
      appendLinkMenu.add(appendLinkItem);
   }

   public void addCootLinkItem(String ligandId) {
      JMenuItem saveCootLinkItem = new JMenuItem(ligandId);
      saveCootLinkItem.addActionListener(listener);
      saveCootLinkItem.setActionCommand( "SAVE_COOT_LINK " + ligandId ) ;
      saveCootLinkMenu.add(saveCootLinkItem);

      JMenuItem appendCootLinkItem = new JMenuItem(ligandId);
      appendCootLinkItem.addActionListener(listener);
      appendCootLinkItem.setActionCommand( "APPEND_COOT_LINK " + ligandId ) ;
      appendCootLinkMenu.add(appendCootLinkItem);
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class JLPanelNorth extends JPanel{
   JLabel loadFromLibraryLabel ;
   JTextField loadFromLibrary ;
   private static JButton newLigand, searchLibrary, checkChemistry, correctBonds, undo, redo ;
   JToolBar toolBar = new JToolBar( "Toolbar", JToolBar.HORIZONTAL ) ;
   private ActionListener listener ;
   private ItemListener itemListener ;
   JButton loadFromLibraryButton ;
   JCheckBox atomIdCheckbox, hydrogenCheckbox, chargeCheckbox ;

   JLPanelNorth( JLigand jLigand ){
      listener = jLigand ;
      itemListener = jLigand ;

      toolBar.setFloatable( false ) ;
      this.setLayout( new FlowLayout( FlowLayout.LEFT ) ) ;
      this.add( toolBar ) ;

      toolBar.add(Box.createHorizontalStrut(5));

      newLigand = new JButton("New Ligand ...");
      toolBar.add(newLigand);
      newLigand.addActionListener(listener);
      newLigand.setActionCommand( "LIGAND_NEW" ) ;

      toolBar.add(Box.createHorizontalStrut(5));

      loadFromLibraryLabel = new JLabel("Load Ligand: ");
      toolBar.add(loadFromLibraryLabel);

      loadFromLibraryButton = new JButton("Load Ligand");
      loadFromLibraryButton.addActionListener(listener);

      loadFromLibrary = new JTextField(5);
      loadFromLibrary.setToolTipText("enter library ligand id");
      loadFromLibrary.addActionListener( new ActionListener(){
         public void actionPerformed( ActionEvent e ){
            loadFromLibraryButton.setActionCommand( "LIGAND_ID " + e.getActionCommand() ) ;
            loadFromLibraryButton.doClick() ;
         }
      } ) ;
      toolBar.add(loadFromLibrary);

      toolBar.add(Box.createHorizontalStrut(5));

      searchLibrary = new JButton("Search Library ...");
      searchLibrary.addActionListener(listener);
      searchLibrary.setActionCommand( "LIGAND_SEARCH" ) ;
      toolBar.add(searchLibrary);

      toolBar.add(Box.createHorizontalStrut(2));

      checkChemistry = new JButton("Check Valencies");
      checkChemistry.addActionListener(listener);
      checkChemistry.setActionCommand( "CHECK_CHEM" ) ;
      toolBar.add(checkChemistry);

      toolBar.add(Box.createHorizontalStrut(2));

      correctBonds = new JButton("Correct Bonds");
      correctBonds.addActionListener(listener);
      correctBonds.setActionCommand( "CHECK_BONDS" ) ;
      toolBar.add(correctBonds);

      toolBar.add(Box.createHorizontalStrut(2));

      undo = new JButton("Undo");
      undo.addActionListener(listener);
      undo.setActionCommand( "UNDO" ) ;
      toolBar.add(undo);

      toolBar.add(Box.createHorizontalStrut(2));

      redo = new JButton("Redo");
      redo.addActionListener(listener);
      redo.setActionCommand( "REDO" ) ;
      toolBar.add(redo);

      toolBar.add(Box.createHorizontalStrut(5));

      atomIdCheckbox = new JCheckBox("Atom Ids");
      atomIdCheckbox.addItemListener(itemListener);
      atomIdCheckbox.setActionCommand( "SHOW_IDS" ) ;
      toolBar.add(atomIdCheckbox);

      hydrogenCheckbox = new JCheckBox("Hydrogens");
      hydrogenCheckbox.addItemListener(itemListener);
      hydrogenCheckbox.setActionCommand( "SHOW_HYDROGENS" ) ;
      toolBar.add(hydrogenCheckbox);

      chargeCheckbox = new JCheckBox("Charges");
      chargeCheckbox.addItemListener(itemListener);
      chargeCheckbox.setActionCommand( "SHOW_CHARGES" ) ;
      toolBar.add(chargeCheckbox);

      if( Env.vbButtons ){
         toolBar.add(Box.createHorizontalStrut(15));

         JCheckBox jcb = new JCheckBox( "Hidden lines" ) ;
         toolBar.add( jcb ) ;
         jcb.addItemListener( new ItemListener(){
            boolean start = true ;
            public void itemStateChanged( ItemEvent e ){
               AbstractButton button = (AbstractButton)e.getItem() ;
               Env.vbGraphics = button.isSelected() ;
               if( ! start ) JLigand.defaultOptionButton.doClick() ;
               start = false ;
            }
         } ) ;
         if( Env.vbGraphics ) jcb.doClick() ;
      }
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class JLPanelCentre extends JPanel implements ChangeListener{
   JTabbedPane tabbedPane = new JTabbedPane() ;
   JLPanelCentre(){
      tabbedPane.setTabLayoutPolicy( JTabbedPane.SCROLL_TAB_LAYOUT ) ;
      tabbedPane.addChangeListener( this ) ;
      this.add( tabbedPane ) ;
   }

   public void stateChanged( ChangeEvent e ){
      JLMain.currentPanel = (JLGPanel)tabbedPane.getSelectedComponent() ;

      JLigand.defaultOptionButton.doClick() ;
   }

   public void addTab(){
      if( JLMain.currentPanel != tabbedPane.getSelectedComponent() ) throw new NullPointerException() ;
      int n = tabbedPane.getTabCount() ;
      if( ++n > 20 ){
         DialogWindows.showWarningDialog( "Cannot create more than 20 Tabs" ) ;
         return ;
      }
      JLGPanel tab = new JLGPanel(
         new Dimension( DialogWindows.panelWidth, DialogWindows.panelHeight ),
         new MouseAdapter(){
            public void mouseClicked( MouseEvent e ){
//             if( e.getButton() == e.BUTTON3 ) JLigand.defaultOptionButton.doClick() ;
               if( e.getButton() == MouseEvent.BUTTON3 ) JLigand.defaultOptionButton.doClick() ;
            }
         }
      ) ;
      tabbedPane.addTab( "Tab " + n, tab ) ;
   }

   public void deleteTab(){
      if( JLMain.currentPanel != tabbedPane.getSelectedComponent() ) throw new NullPointerException() ;
      int n = tabbedPane.getTabCount() ;
      if( --n == 0 ){
         DialogWindows.showWarningDialog( "Cannot delete the last Tab" ) ;
         return ;
      }
      int i = tabbedPane.getSelectedIndex() ;
      int j = JOptionPane.showConfirmDialog(
         JLMain.mainFrame,
         "Are you sure you want to delete " + tabbedPane.getTitleAt( i ) + " with all its ligands?",
         "Delete Tab",
         JOptionPane.YES_NO_OPTION
      ) ;
      if( j != JOptionPane.YES_OPTION ) return ;
      tabbedPane.remove( i ) ;
      tabbedPane.setSelectedIndex( Math.min( i, n - 1 ) ) ;
      for( i = 0 ; i < n ; i++ ) tabbedPane.setTitleAt( i, "Tab " + ( i + 1 ) ) ;
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class JLPanelWest extends JPanel implements ActionListener{

   JToggleButton linkButton, deleteButton;
   JButton periodicTableButton ;

   JToolBar toolBar = new JToolBar( "Toolbar", JToolBar.VERTICAL ) ;
   ButtonGroup group = new ButtonGroup();

   JButton modeButton = new JButton() ;

   JLPanelWest( ActionListener listener ){
      modeButton.addActionListener( listener ) ;

      toolBar.setFloatable( false ) ;
      this.add( toolBar ) ;
      toolBar.add( Box.createVerticalStrut( 15 ) ) ;

      JLigand.defaultOptionButton = new JToggleButton( "C" ) ;
      add( JLigand.defaultOptionButton, "add carbon atom and bond" ) ;

      add( new JToggleButton( "N"  ), "add nitrogen atom and bond"   ) ;
      add( new JToggleButton( "O"  ), "add oxygen atom and bond"     ) ;
      add( new JToggleButton( "H"  ), "add hydrogen atom and bond"   ) ;
      add( new JToggleButton( "S"  ), "add sulphur atom and bond"    ) ;
      add( new JToggleButton( "F"  ), "add flourine atom and bond"   ) ;
      add( new JToggleButton( "Cl" ), "add chlorine atom and bond"   ) ;
      add( new JToggleButton( "Br" ), "add bromine atom and bond"    ) ;
      add( new JToggleButton( "I"  ), "add iodine atom and bond"     ) ;
      add( new JToggleButton( "P"  ), "add phosphorus atom and bond" ) ;

      periodicTableButton = new JButton( getIcon( "periodicTableAtom" ) ) ;
      periodicTableButton.setMaximumSize( new Dimension( 50, 25 ) ) ;
      String tip = "choose an atom from the periodic table - see button below" ;
      setJButtonDetails( toolBar, periodicTableButton, tip ) ;

      JLigand.choiceOptionButton = new JToggleButton( "Se" ) ;
      add( JLigand.choiceOptionButton, "add this periodic table atom - see button above" ) ;

      add( new JToggleButton( getIcon( "n3"  ) ), "add n=3 ring" ) ;
      add( new JToggleButton( getIcon( "n4"  ) ), "add n=4 ring" ) ;
      add( new JToggleButton( getIcon( "n5"  ) ), "add n=5 ring" ) ;
      add( new JToggleButton( getIcon( "n6"  ) ), "add n=6 ring" ) ;
      add( new JToggleButton( getIcon( "n6B" ) ), "add n=6 benzene ring" ) ;
      add( new JToggleButton( getIcon( "n7"  ) ), "add n=7 ring" ) ;

      add( linkButton   = new JToggleButton( "Link" ), "link 2 atoms"        ) ;
      add( deleteButton = new JToggleButton( "Del"  ), "delete atom or bond" ) ;

      if( Env.vbButtons ){
         ButtonGroup groupLR = new ButtonGroup();
         toolBar.add( Box.createVerticalStrut( 20 ) ) ;

         JToggleButton jtb = new JToggleButton( "L" ) ;
         jtb.setMaximumSize( new Dimension( 50, 25 ) ) ;
         jtb.setToolTipText( "coordinates: libcheck -> graphics" ) ;
         groupLR.add( jtb ) ;
         toolBar.add( jtb ) ;
         jtb.addActionListener( new ActionListener(){ public void actionPerformed( ActionEvent e ){
            Env.coordinates = Env.GRAPH_LIBCHECK ;
         } } ) ;
         if( Env.coordinates == Env.GRAPH_LIBCHECK ) jtb.doClick() ;

         jtb = new JToggleButton( "L+R" ) ;
         jtb.setMaximumSize( new Dimension( 50, 25 ) ) ;
         jtb.setToolTipText( "coordinates: libcheck -> refmac -> graphics" ) ;
         groupLR.add( jtb ) ;
         toolBar.add( jtb ) ;
         jtb.addActionListener( new ActionListener(){ public void actionPerformed( ActionEvent e ){
            Env.coordinates = Env.GRAPH_LIBCHECK_REFMAC ;
         } } ) ;
         if( Env.coordinates == Env.GRAPH_LIBCHECK_REFMAC ) jtb.doClick() ;

         jtb = new JToggleButton( "R" ) ;
         jtb.setMaximumSize( new Dimension( 50, 25 ) ) ;
         jtb.setToolTipText( "coordinates: graphics -> refmac -> graphics" ) ;
         groupLR.add( jtb ) ;
         toolBar.add( jtb ) ;
         jtb.addActionListener( new ActionListener(){ public void actionPerformed( ActionEvent e ){
            Env.coordinates = Env.GRAPH_REFMAC ;
         } } ) ;
         if( Env.coordinates == Env.GRAPH_REFMAC ) jtb.doClick() ;
      }

      if( Env.vbLibs ){
         toolBar.add( Box.createVerticalStrut( 20 ) ) ;

         JButton jb = new JButton( "Libs" ) ;
         jb.setMaximumSize( new Dimension( 50, 25 ) ) ;
         jb.setToolTipText( "Print contents of graphics and standard libraries" ) ;
         toolBar.add( jb );
         jb.addActionListener( new ActionListener(){ public void actionPerformed( ActionEvent e ){
            System.out.println( JLMain.currentPanel.gLib.summary() ) ;
         } } ) ;
      }
   }

   void add( JToggleButton jtb, String tip ){
      jtb.setMaximumSize( new Dimension( 50, 25 ) ) ;
      jtb.setToolTipText( tip ) ;
      group.add( jtb ) ;
      toolBar.add( jtb ) ;
      jtb.addActionListener( this ) ;
   }

   void setJButtonDetails(JToolBar toolBar, JButton button, String toolTipText) {
      button.setToolTipText(toolTipText);
      toolBar.add(button);
      button.addActionListener(this);
   }

   public Icon getIcon( String text ){
      return new ImageIcon( getClass().getResource( "/resources/" + text + ".gif" ), text ) ;
   }

   public void actionPerformed( ActionEvent e ){
      if( e.getSource() == periodicTableButton ) DialogWindows.periodicTableDialog.showDialog() ;
      else{
         JToggleButton jtb = (JToggleButton)e.getSource() ;
         ImageIcon ii = (ImageIcon)jtb.getIcon() ;
         String option ;
         if( ii != null ) option = ii.getDescription() ;
         else if( jtb == deleteButton ) option = "DELETE" ;
         else if( jtb == linkButton ) option = "LINK" ;
         else option = jtb.getText() ;
         modeButton.setActionCommand( "OPTION " + option ) ;
         modeButton.doClick() ;
      }
   }
}
// -----------------------------------------------------------------------------

class HelpDialog{

   private static final String bsn = JLMain.bsn ;
   private String text, title ;

   public HelpDialog( String title, String resource ){
      try{
         this.text = new CharArray( getClass().getResourceAsStream( resource ) ).toString() ;
         this.title = title ;
      }
      catch( Exception e ){
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         String msg = resource + " not found." + bsn + title + " will not be available." ;
         DialogWindows.showErrorDialog( msg ) ;
      }
   }

   public void showDialog() {
      if( this.title == null || this.text == null ) return ;

      final JDialog dialog = new JDialog(JLMain.mainFrame, this.title);

      JLabel label = new JLabel(this.text);

      label.setHorizontalAlignment(JLabel.LEFT);
      label.setVerticalAlignment(JLabel.TOP);

//    Font font = label.getFont();
      Font f1 = new Font("", Font.PLAIN, 12);
      label.setFont(f1);

      JButton closeButton = new JButton("Close");
      closeButton.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            dialog.setVisible(false);
            dialog.dispose();
         }
      });

      JPanel closePanel = new JPanel();
      closePanel.setLayout(new FlowLayout(FlowLayout.CENTER));
      closePanel.add(closeButton);
      closePanel.setBorder(BorderFactory. createEmptyBorder(0,0,5,5));

      JPanel labelPanel = new JPanel(new BorderLayout());
      labelPanel.add(label, BorderLayout.CENTER);
      labelPanel.setBorder(BorderFactory. createEmptyBorder(10,10,5,5));

      JPanel contentPane = new JPanel(new BorderLayout());
      contentPane.add(labelPanel, BorderLayout.NORTH);
      contentPane.add(closePanel, BorderLayout.SOUTH);
      contentPane.setOpaque(true);
      dialog.setContentPane(contentPane);

      int w = 800, h = 600;
      dialog.setSize(new Dimension(w, h));
      dialog.setLocation(DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop);
      dialog.setVisible(true);
   }
}
// -----------------------------------------------------------------------------

class NewLigandDialog implements ActionListener {

   private JDialog dialog;
   private JLabel label1, label2;
   private JTextField textField1;
   private JButton submitButton, cancelButton;
   private JComboBox atomTypeList;
   private String[] atomTypeArray = {
      "C", "N", "O", "H", "S", "F", "Cl", "Br", "I", "P", "n=3", "n=4", "n=5", "n=6", "benzene", "n=7"
   } ;
   private GLib targetLib ;

   NewLigandDialog( GLib gLib ){ targetLib = gLib ; }

   public void showDialog() {

      dialog = new JDialog(JLMain.mainFrame, true);
      dialog.setTitle("Create New Ligand");

      label1 = new JLabel("Ligand Id");

      textField1 = new JTextField(10);
      textField1.addActionListener(this);

      label2 = new JLabel("Atom Type");

      submitButton = new JButton("Submit");
      submitButton.addActionListener(this);

      cancelButton = new JButton("Cancel");
      cancelButton.addActionListener(this);

      JButton closeButton = new JButton("Cancel");
      closeButton.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            dialog.setVisible(false);
            dialog.dispose();
         }
      });

      // label
      JPanel p0 = new JPanel();
      p0.setLayout(new BoxLayout(p0, BoxLayout.X_AXIS));
      p0.add(new JLabel("Please enter a Ligand Id (1 - 8 characters) and initial Atom Type:"));

      // ligand id
      JPanel p1 = new JPanel();
      p1.setLayout(new BoxLayout(p1, BoxLayout.X_AXIS));
      p1.add(label1);
      p1.add(Box.createRigidArea(new Dimension(20, 0)));
      setComponentSize(textField1, 150, 25);
      p1.add(textField1);

      // atom type
      JPanel p2 = new JPanel();
      p2.setLayout(new BoxLayout(p2, BoxLayout.LINE_AXIS));
      p2.add(label2);
      p2.add(Box.createRigidArea(new Dimension(10, 0)));
      atomTypeList = new JComboBox(atomTypeArray);
      atomTypeList.setSelectedIndex(0);
      setComponentSize(atomTypeList, 150, 30);
      p2.add(atomTypeList);

      // button
      JPanel p3 = new JPanel();
      p3.setLayout(new BoxLayout(p3, BoxLayout.LINE_AXIS));
      p3.add(submitButton);
      p3.add(Box.createRigidArea(new Dimension(10, 0)));
      p3.add(cancelButton);

      JPanel contentPane = new JPanel();
      contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.PAGE_AXIS));
      // spacer - creates rigid box at top with width 0 and height 15
      contentPane.add(Box.createRigidArea(new Dimension(0, 30)));
      contentPane.add(p0);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p1);
      contentPane.add(Box.createRigidArea(new Dimension(0, 5)));
      contentPane.add(p2);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p3);
      contentPane.setOpaque(true);

      dialog.setContentPane(contentPane);
      int w = 500, h = 400;
      dialog.setSize(new Dimension(w, h));
      dialog.setLocation(DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop2);
      dialog.setVisible(true);

      dialog.addWindowListener(new WindowAdapter(){
         public void windowOpened( WindowEvent we ){
         SwingUtilities.invokeLater(new Runnable(){
            public void run(){
               textField1.requestFocusInWindow() ;
            }
         });
      }
   });
   }

   public void setComponentSize(JComponent component, int width, int height) {
      Dimension size = new Dimension(width, height);
      component.setPreferredSize(size);
      component.setMinimumSize(size);
      component.setMaximumSize(size);
   }

   public void actionPerformed(ActionEvent e) {

      if (e.getSource() == cancelButton) {

         dialog.setVisible(false);
         dialog.dispose();

      } else if (e.getSource() == textField1 || e.getSource() == submitButton) {

         ligandId = textField1.getText();
         atomType = (String)atomTypeList.getSelectedItem();

         if (ligandId.length() == 0 || ligandId.length() > 8) {
            DialogWindows.showWarningDialog("Ligand Id must be 1 - 8 characters - please try again");

         } else if (!ligandId.matches("[a-zA-Z0-9-_]+")) {
            DialogWindows.showWarningDialog(
               "Ligand Id must consist of these characters: " +
               "A-Z, a-z, 1-9, -, _ (no spaces allowed) - please try again"
            ) ;

         } else if( targetLib.contains( ligandId ) ){

               DialogWindows.showWarningDialog("This Ligand Id already exists - please enter a new Ligand Id");

         } else {

            dialog.setVisible(false);
            dialog.dispose();
         }
      }
   }
   String ligandId, atomType ;
   String getLigandId(){ return ligandId ; }
   String getAtomType(){
      if( atomType == null ) return null ;
      else if( atomType.equals( "benzene" ) ) return "n6B" ;
      else if( atomType.startsWith( "n=" ) ) return "n" + atomType.charAt( 2 ) ;
      else return atomType ;
   }
}
// -----------------------------------------------------------------------------

class PeriodicTableDialog implements ActionListener {

   private JDialog dialog;
   private Font f1 = new Font("serif", Font.BOLD, 12);
   private JLabel b1, Lanthanoides, Actinoides;
   //private JButton n3, n4, n5, n6, n6B, n7;
   private JButton H, He;
   private JButton Li, Be, B, C, N, O, F, Ne;
   private JButton Na, Mg, Al, Si, P, S, Cl, Ar;
   private JButton K, Ca, Sc, Ti, V, Cr, Mn, Fe, Co, Ni, Cu, Zn, Ga, Ge, As, Se, Br, Kr;
   private JButton Rb, Sr, Y, Zr, Nb, Mo, Tc, Ru, Rh, Pd, Ag, Cd, In, Sn, Sb, Te, I, Xe;
   private JButton Cs, Ba, Hf, Ta, W, Re, Os, Ir, Pt, Au, Hg, Tl, Pb, Bi, Po, At, Rn;
   private JButton Fr, Ra;
   private JButton La, Ce, Pr, Nd, Pm, Sm, Eu, Gd, Tb, Dy, Ho, Er, Tm, Yb, Lu;
   private JButton Ac, Th, Pa, U, Np, Pu, Am, Cm, Bk, Cf, Es, Fm, Md, No, Lr;
   private final Color SILVER = new Color(192, 192, 192);
   private final Color GREEN = new Color(152, 251, 152);
   private final Color YELLOW = new Color(240, 230, 140);
   private final Color CYAN = new Color(135, 206, 235);
   private final Color ORANGE = new Color(244, 164, 96);
   private final Color MAGENTA = new Color(218, 160, 221);
   private final Color RED = new Color(240, 128, 128);

   public void setJButtonDetails(JPanel p, JButton button, Color color, String toolTipText) {
      p.add(button);
      button.setFont(f1);
      button.setToolTipText(toolTipText);
      button.setBackground(color);
      button.setMargin(new Insets(1, 1, 1, 1));
      button.addActionListener(this);
   }

   public String buttonText(String atomNo, String atomName) {
      String a = "<html><table border=0><tr><td align='center'><font color='gray'>";
      String b = "</font></td><tr><td align='center'>";
      String c =  "</td></table></html>";
      return a + atomNo + b + atomName + c;
   }

   public void showDialog() {

      dialog = new JDialog(JLMain.mainFrame, "Periodic Table - choose atom");

      GridLayout g1 = new GridLayout(9, 18);
      JPanel p1 = new JPanel(g1);

      // ROW 1
      H = new JButton(buttonText("1", "H")); setJButtonDetails(p1, H, SILVER, "");
      for (int i = 0; i < 16; i++) { b1 = new JLabel(); p1.add(b1, ""); }
      He = new JButton(buttonText("2", "He")); setJButtonDetails(p1, He, CYAN, "");

      // ROW 2
      Li = new JButton(buttonText("3", "Li")); setJButtonDetails(p1, Li, ORANGE, "");
      Be = new JButton(buttonText("4", "Be")); setJButtonDetails(p1, Be, ORANGE, "");
      for (int i = 0; i < 10; i++) { b1 = new JLabel(); p1.add(b1); }
      B = new JButton(buttonText("5", "B")); setJButtonDetails(p1, B, YELLOW, "");
      C = new JButton(buttonText("6", "C")); setJButtonDetails(p1, C, YELLOW, "");
      N = new JButton(buttonText("7", "N")); setJButtonDetails(p1, N, YELLOW, "");
      O = new JButton(buttonText("8", "O")); setJButtonDetails(p1, O, YELLOW, "");
      F = new JButton(buttonText("9", "F")); setJButtonDetails(p1, F, YELLOW, "");
      Ne = new JButton(buttonText("10", "Ne")); setJButtonDetails(p1, Ne, CYAN, "");

      // ROW 3
      Na = new JButton(buttonText("11", "Na")); setJButtonDetails(p1, Na, ORANGE, "");
      Mg = new JButton(buttonText("12", "Mg")); setJButtonDetails(p1, Mg, ORANGE, "");
      for (int i = 0; i < 10; i++) { b1 = new JLabel(); p1.add(b1); }
      Al = new JButton(buttonText("13", "Al")); setJButtonDetails(p1, Al, GREEN, "");
      Si = new JButton(buttonText("14", "Si")); setJButtonDetails(p1, Si, YELLOW, "");
      P = new JButton(buttonText("15", "P")); setJButtonDetails(p1, P, YELLOW, "");
      S = new JButton(buttonText("16", "S")); setJButtonDetails(p1, S, YELLOW, "");
      Cl = new JButton(buttonText("17", "Cl")); setJButtonDetails(p1, Cl, YELLOW, "");
      Ar = new JButton(buttonText("18", "Ar")); setJButtonDetails(p1, Ar, CYAN, "");

      // ROW 4
      K = new JButton(buttonText("19", "K")); setJButtonDetails(p1, K, ORANGE, "");
      Ca = new JButton(buttonText("20", "Ca")); setJButtonDetails(p1, Ca, ORANGE, "");
      Sc = new JButton(buttonText("21", "Sc")); setJButtonDetails(p1, Sc, Color.PINK, "");
      Ti = new JButton(buttonText("22", "Ti")); setJButtonDetails(p1, Ti, Color.PINK, "");
      V = new JButton(buttonText("23", "V")); setJButtonDetails(p1, V, Color.PINK, "");
      Cr = new JButton(buttonText("24", "Cr")); setJButtonDetails(p1, Cr, Color.PINK, "");
      Mn = new JButton(buttonText("25", "Mn")); setJButtonDetails(p1, Mn, Color.PINK, "");
      Fe = new JButton(buttonText("26", "Fe")); setJButtonDetails(p1, Fe, Color.PINK, "");
      Co = new JButton(buttonText("27", "Co")); setJButtonDetails(p1, Co, Color.PINK, "");
      Ni = new JButton(buttonText("28", "Ni")); setJButtonDetails(p1, Ni, Color.PINK, "");
      Cu = new JButton(buttonText("29", "Cu")); setJButtonDetails(p1, Cu, Color.PINK, "");
      Zn = new JButton(buttonText("30", "Zn")); setJButtonDetails(p1, Zn, Color.PINK, "");
      Ga = new JButton(buttonText("31", "Ga")); setJButtonDetails(p1, Ga, GREEN, "");
      Ge = new JButton(buttonText("32", "Ge")); setJButtonDetails(p1, Ge, GREEN, "");
      As = new JButton(buttonText("33", "As")); setJButtonDetails(p1, As, YELLOW, "");
      Se = new JButton(buttonText("34", "Se")); setJButtonDetails(p1, Se, YELLOW, "");
      Br = new JButton(buttonText("35", "Br")); setJButtonDetails(p1, Br, YELLOW, "");
      Kr = new JButton(buttonText("36", "Kr")); setJButtonDetails(p1, Kr, CYAN, "");

      // ROW 5
      Rb = new JButton(buttonText("37", "Rb")); setJButtonDetails(p1, Rb, ORANGE, "");
      Sr = new JButton(buttonText("38", "Sr")); setJButtonDetails(p1, Sr, ORANGE, "");
      Y = new JButton(buttonText("39", "Y")); setJButtonDetails(p1, Y, Color.PINK, "");
      Zr = new JButton(buttonText("40", "Zr")); setJButtonDetails(p1, Zr, Color.PINK, "");
      Nb = new JButton(buttonText("41", "Nb")); setJButtonDetails(p1, Nb, Color.PINK, "");
      Mo = new JButton(buttonText("42", "Mo")); setJButtonDetails(p1, Mo, Color.PINK, "");
      Tc = new JButton(buttonText("43", "Tc")); setJButtonDetails(p1, Tc, Color.PINK, "");
      Ru = new JButton(buttonText("44", "Ru"));
      setJButtonDetails(p1, Ru, Color.PINK,
         "<html>Ruthenium<br>Atom No: 44<br>Relative Atomic Mass: 101.07</html>"
      ) ;
      Rh = new JButton(buttonText("45", "Rh")); setJButtonDetails(p1, Rh, Color.PINK, "");
      Pd = new JButton(buttonText("46", "Pd")); setJButtonDetails(p1, Pd, Color.PINK, "");
      Ag = new JButton(buttonText("47", "Ag")); setJButtonDetails(p1, Ag, Color.PINK, "");
      Cd = new JButton(buttonText("48", "Cd")); setJButtonDetails(p1, Cd, Color.PINK, "");
      In = new JButton(buttonText("49", "In")); setJButtonDetails(p1, In, GREEN, "");
      Sn = new JButton(buttonText("50", "Sn")); setJButtonDetails(p1, Sn, GREEN, "");
      Sb = new JButton(buttonText("51", "Sb")); setJButtonDetails(p1, Sb, GREEN, "");
      Te = new JButton(buttonText("52", "Te")); setJButtonDetails(p1, Te, YELLOW, "");
      I = new JButton(buttonText("53", "I")); setJButtonDetails(p1, I, YELLOW, "");
      Xe = new JButton(buttonText("54", "Xe")); setJButtonDetails(p1, Xe, CYAN, "");

      // ROW 6
      Cs = new JButton(buttonText("55", "Cs")); setJButtonDetails(p1, Cs, ORANGE, "");
      Ba = new JButton(buttonText("56", "Ba")); setJButtonDetails(p1, Ba, ORANGE, "");
      Lanthanoides = new JLabel("*");
      p1.add(Lanthanoides);
//    Lanthanoides.setHorizontalAlignment(Lanthanoides.CENTER);
      Lanthanoides.setHorizontalAlignment(JLabel.CENTER);
      Lanthanoides.setBackground(MAGENTA);
      Lanthanoides.setOpaque(true);
      Hf = new JButton(buttonText("72", "Hf")); setJButtonDetails(p1, Hf, Color.PINK, "");
      Ta = new JButton(buttonText("73", "Ta")); setJButtonDetails(p1, Ta, Color.PINK, "");
      W = new JButton(buttonText("74", "W")); setJButtonDetails(p1, W, Color.PINK, "");
      Re = new JButton(buttonText("75", "Re")); setJButtonDetails(p1, Re, Color.PINK, "");
      Os = new JButton(buttonText("76", "Os")); setJButtonDetails(p1, Os, Color.PINK, "");
      Ir = new JButton(buttonText("77", "Ir")); setJButtonDetails(p1, Ir, Color.PINK, "");
      Pt = new JButton(buttonText("78", "Pt")); setJButtonDetails(p1, Pt, Color.PINK, "");
      Au = new JButton(buttonText("79", "Au")); setJButtonDetails(p1, Au, Color.PINK, "");
      Hg = new JButton(buttonText("80", "Hg")); setJButtonDetails(p1, Hg, Color.PINK, "");
      Tl = new JButton(buttonText("81", "Tl")); setJButtonDetails(p1, Tl, GREEN, "");
      Pb = new JButton(buttonText("82", "Pb")); setJButtonDetails(p1, Pb, GREEN, "");
      Bi = new JButton(buttonText("83", "Bi")); setJButtonDetails(p1, Bi, GREEN, "");
      Po = new JButton(buttonText("84", "Po")); setJButtonDetails(p1, Po, GREEN, "");
      At = new JButton(buttonText("85", "At")); setJButtonDetails(p1, At, YELLOW, "");
      Rn = new JButton(buttonText("86", "Rn")); setJButtonDetails(p1, Rn, CYAN, "");

      // ROW 7
      Fr = new JButton(buttonText("87", "Fr")); setJButtonDetails(p1, Fr, ORANGE, "");
      Ra = new JButton(buttonText("88", "Ra")); setJButtonDetails(p1, Ra, ORANGE, "");
      Actinoides = new JLabel("**");
      p1.add(Actinoides);
//    Actinoides.setHorizontalAlignment(Actinoides.CENTER);
      Actinoides.setHorizontalAlignment(JLabel.CENTER);
      Actinoides.setBackground(RED);
      Actinoides.setOpaque(true);
      for (int i = 0; i < 15; i++) { b1 = new JLabel(); p1.add(b1); }

      // ROW 8
      for (int i = 0; i < 3; i++) { b1 = new JLabel(); p1.add(b1); }
      La = new JButton(buttonText("57", "La")); setJButtonDetails(p1, La, MAGENTA, "");
      Ce = new JButton(buttonText("58", "Ce")); setJButtonDetails(p1, Ce, MAGENTA, "");
      Pr = new JButton(buttonText("59", "Pr")); setJButtonDetails(p1, Pr, MAGENTA, "");
      Nd = new JButton(buttonText("60", "Nd")); setJButtonDetails(p1, Nd, MAGENTA, "");
      Pm = new JButton(buttonText("61", "Pm")); setJButtonDetails(p1, Pm, MAGENTA, "");
      Sm = new JButton(buttonText("62", "Sm")); setJButtonDetails(p1, Sm, MAGENTA, "");
      Eu = new JButton(buttonText("63", "Eu")); setJButtonDetails(p1, Eu, MAGENTA, "");
      Gd = new JButton(buttonText("64", "Gd")); setJButtonDetails(p1, Gd, MAGENTA, "");
      Tb = new JButton(buttonText("65", "Tb")); setJButtonDetails(p1, Tb, MAGENTA, "");
      Dy = new JButton(buttonText("66", "Dy")); setJButtonDetails(p1, Dy, MAGENTA, "");
      Ho = new JButton(buttonText("67", "Ho")); setJButtonDetails(p1, Ho, MAGENTA, "");
      Er = new JButton(buttonText("68", "Er")); setJButtonDetails(p1, Er, MAGENTA, "");
      Tm = new JButton(buttonText("69", "Tm")); setJButtonDetails(p1, Tm, MAGENTA, "");
      Yb = new JButton(buttonText("70", "Yb")); setJButtonDetails(p1, Yb, MAGENTA, "");
      Lu = new JButton(buttonText("71", "Lu")); setJButtonDetails(p1, Lu, MAGENTA, "");

      // ROW 9
      for (int i = 0; i < 3; i++) { b1 = new JLabel(); p1.add(b1); }
      Ac = new JButton(buttonText("89", "Ac")); setJButtonDetails(p1, Ac, RED, "");
      Th = new JButton(buttonText("90", "Th")); setJButtonDetails(p1, Th, RED, "");
      Pa = new JButton(buttonText("91", "Pa")); setJButtonDetails(p1, Pa, RED, "");
      U = new JButton(buttonText("92", "U")); setJButtonDetails(p1, U, RED, "");
      Np = new JButton(buttonText("93", "Np")); setJButtonDetails(p1, Np, RED, "");
      Pu = new JButton(buttonText("94", "Pu")); setJButtonDetails(p1, Pu, RED, "");
      Am = new JButton(buttonText("95", "Am")); setJButtonDetails(p1, Am, RED, "");
      Cm = new JButton(buttonText("96", "Cm")); setJButtonDetails(p1, Cm, RED, "");
      Bk = new JButton(buttonText("97", "Bk")); setJButtonDetails(p1, Bk, RED, "");
      Cf = new JButton(buttonText("98", "Cf")); setJButtonDetails(p1, Cf, RED, "");
      Es = new JButton(buttonText("99", "Es")); setJButtonDetails(p1, Es, RED, "");
      Fm = new JButton(buttonText("100", "Fm")); setJButtonDetails(p1, Fm, RED, "");
      Md = new JButton(buttonText("101", "Md")); setJButtonDetails(p1, Md, RED, "");
      No = new JButton(buttonText("102", "No")); setJButtonDetails(p1, No, RED, "");
      Lr = new JButton(buttonText("103", "Lr")); setJButtonDetails(p1, Lr, MAGENTA, "");

      JPanel contentPane = new JPanel();
      contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.Y_AXIS));
      contentPane.add(p1);

      dialog.setContentPane(contentPane);
      int w = 900, h = 600;
      dialog.setSize(new Dimension(w, h));
      dialog.setLocation(DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop);
      dialog.setVisible(true);
   }

   public void actionPerformed(ActionEvent e) {

      String c = e.getActionCommand();
      String t1 = "<td align='center'>";
      String t2 = "</td>";
      int i1 = c.lastIndexOf(t1);
      int i2 = c.lastIndexOf(t2);
      String choice = c.substring(i1 + t1.length(), i2);

      JLigand.choiceOptionButton.setText(choice);
      JLigand.choiceOptionButton.setIcon(null);
      JLigand.choiceOptionButton.doClick();

      dialog.setVisible(false);
      dialog.dispose();
   }
}
// -----------------------------------------------------------------------------

class SearchLibraryDialog implements ActionListener, MouseListener {

   private JDialog dialog;
// private JTextField textField1, textField2, loadField;
   private JTextField textField1, textField2;
// private JButton go1, go2, clear, load, loadAndClose;
   private JButton go1, go2;
   private JTextArea textArea;
   private JScrollPane scrollPane;

// private final String fs = File.separator;
   private static final String bsn = JLMain.bsn ;

   SearchLibraryDialog(){

      dialog = new JDialog(JLMain.mainFrame, true);
      dialog.setTitle("Search Standard Library for Ligand");

      JPanel p0 = new JPanel();
      JTextArea jta = new JTextArea(6,40) ;
      jta.append(
         "Enter a search string (e.g. P, PL, PLP), click on Search, " +
         "then double-click on the Ligand Id to load the ligand." + bsn +
         "A blank search returns all ligands in the library. " + bsn +
         "Note: it may take a few seconds to load the ligand after double-clicking." + bsn +
         "Also note that the search is case-sensitive and that most monomers are upper-case, e.g. LYS"
      ) ;
      jta.setLineWrap( true ) ;
      jta.setWrapStyleWord( true ) ;
      jta.setBackground( p0.getBackground() ) ;
      jta.setEditable( false ) ;
      MouseListener[] mls = jta.getMouseListeners() ;
      for( int i = 0 ; i < mls.length ; i++ ) jta.removeMouseListener( mls[i] ) ;
      MouseMotionListener[] mmls = jta.getMouseMotionListeners() ;
      for( int i = 0 ; i < mmls.length ; i++ ) jta.removeMouseMotionListener( mmls[i] ) ;
      p0.add(jta);

      JPanel pA1 = new JPanel();
      JPanel pA2 = new JPanel();
      JPanel pB = new JPanel();
      JPanel pC = new JPanel();

      Font font = new Font("", Font.PLAIN, 12);

      JLabel label = new JLabel("Ligand ID");
      pA1.add(label);
      label.setFont(font);

      label = new JLabel("Keyword ");
      pA2.add(label);
      label.setFont(font);

      textField1 = new JTextField(10);
      pA1.add(textField1);
      textField1.addActionListener(this);

      textField2 = new JTextField(10);
      pA2.add(textField2);
      textField2.addActionListener(this);

      go1 = new JButton("Search");
      pA1.add(go1);
      go1.addActionListener(this);

      go2 = new JButton("Search");
      pA2.add(go2);
      go2.addActionListener(this);

      textArea = new JTextArea(20, 40);
      textArea.setEditable(false);
      scrollPane = new JScrollPane(textArea);
      pB.add(scrollPane);
      textArea.addMouseListener(this);

      JButton closeButton = new JButton("Cancel");
      closeButton.addActionListener(new ActionListener() {
         public void actionPerformed(ActionEvent e) {
            dialog.setVisible(false);
            dialog.dispose();
         }
      });

      pC.add(closeButton);

      JPanel contentPane = new JPanel();
      contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.PAGE_AXIS));
      contentPane.add(Box.createRigidArea(new Dimension(0, 15))); // creates rigid box - width, height
      p0.setAlignmentX(Component.CENTER_ALIGNMENT);
      pA1.setAlignmentX(Component.CENTER_ALIGNMENT);
      pA2.setAlignmentX(Component.CENTER_ALIGNMENT);
      pB.setAlignmentX(Component.CENTER_ALIGNMENT);
      pC.setAlignmentX(Component.CENTER_ALIGNMENT);

      contentPane.add(p0);
      contentPane.add(Box.createRigidArea(new Dimension(0, 5)));
      contentPane.add(pA1);
      contentPane.add(pA2);
      //contentPane.add(Box.createRigidArea(new Dimension(0, 5)));
      contentPane.add(pB);
      //contentPane.add(Box.createRigidArea(new Dimension(0, 5)));
      contentPane.add(pC);

      contentPane.setOpaque(true);

      dialog.setContentPane(contentPane);
      int w = 520, h = 600;
      dialog.setSize(new Dimension(w, h));
      dialog.setLocation(DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop);
   }

   public void showDialog(){
      dialog.setVisible(true);
      dialog.addWindowListener(new WindowAdapter(){
         public void windowOpened( WindowEvent we ){
            SwingUtilities.invokeLater(new Runnable(){
               public void run(){
                  textField1.requestFocusInWindow() ;
               }
            });
         }
      });
   }

   SearchLibraryDialog( List<String> initialList ){
      this() ;
      textArea.setText( "" ) ;
      for( String key8 : initialList ){
         MonomerHeader header = StdLib.compHeaderMap.get( key8 ) ;
         if( header != null && header.key8.equals( key8 ) ){
            textArea.append( " " + header.key8 + "\t" + header.name + bsn ) ;
         }
      }
      textArea.setCaretPosition( 0 ) ;
   }

   public void actionPerformed( ActionEvent e ){
      if( e.getSource() == textField1 || e.getSource() == go1 ){
         textArea.setText( "" ) ;
         String searchString = textField1.getText() ;
         for( MonomerHeader header : StdLib.compHeaderMap.values() ){
            if( header.key8.startsWith( searchString ) ){
               textArea.append( " " + header.key8 + "\t" + header.name + bsn ) ;
            }
         }
         textArea.setCaretPosition( 0 ) ;
      }
      else if( e.getSource() == textField2 || e.getSource() == go2 ){
         textArea.setText( "" ) ;
         String searchString = textField2.getText() ;
         for( MonomerHeader header : StdLib.compHeaderMap.values() ){
            if( header.name.contains( searchString ) ){
               textArea.append( " " + header.key8 + "\t" + header.name + bsn ) ;
            }
         }
         textArea.setCaretPosition( 0 ) ;
      }
   }

   public void mouseClicked(MouseEvent e) {

      if( textArea.getSelectedText() == null ) return ;

      String s = textArea.getText() ;
      int j0 = textArea.getSelectionStart() - 1 ;
      int j1 = s.lastIndexOf( bsn, j0 ) + 1 + bsn.length() ;
      if( j1 == 2 ) j1 = 1 ;
      int j2 = s.indexOf( "\t", j1 ) ;
      ligandId = s.substring( j1, j2 ) ;

//    if( Env.vbPrint ) String msg = "Library: @" + ligandId + "@" + " " + j0 + " " + j1 + " " + j2 ;
//    if( Env.vbPrint ) System.err.println( msg ) ;
//    if( Env.vbPrint ) { JLMain.currentPanel.message = msg ; JLMain.currentPanel.repaint() ; }

      dialog.setVisible(false);
      dialog.dispose();
   }

   String ligandId = "" ;
   String getLigandId(){ return ligandId ; }

   public void mouseEntered(MouseEvent e) {
   }

   public void mouseExited(MouseEvent e) {
   }

   public void mouseReleased(MouseEvent e) {
   }

   public void mousePressed(MouseEvent e) {
   }
}
// -----------------------------------------------------------------------------

class SmileStringDialog implements ActionListener {

// private JLGPanel currentPanel;
   private JDialog dialog;
   private JLabel label1, label2;
   private JTextField textField1, textField2;
   private JButton submitButton, cancelButton;
   private GLib targetLib ;

   SmileStringDialog( GLib gLib ){ targetLib = gLib ; }

   public void showDialog() {

      dialog = new JDialog(JLMain.mainFrame, true);
      dialog.setTitle("Enter Smile String");

      label1 = new JLabel("Smile String");

      textField1 = new JTextField(20);
      textField1.addActionListener(this);

      label2 = new JLabel("Ligand Id");

      textField2 = new JTextField(20);
      textField2.addActionListener(this);

      submitButton = new JButton("Submit");
      submitButton.addActionListener(this);

      cancelButton = new JButton("Cancel");
      cancelButton.addActionListener(this);

      // label
      JPanel p0 = new JPanel();
      p0.setLayout(new BoxLayout(p0, BoxLayout.X_AXIS));
      p0.add(new JLabel("Please enter a Ligand Id (1 - 8 characters) and Smile String:"));

      // ligand id
      JPanel p1 = new JPanel();
      p1.setLayout(new BoxLayout(p1, BoxLayout.X_AXIS));
      p1.add(label1);
      p1.add(Box.createRigidArea(new Dimension(5, 0)));
      setComponentSize(textField1, 300, 25);
      p1.add(textField1);

      // smile string
      JPanel p2 = new JPanel();
      p2.setLayout(new BoxLayout(p2, BoxLayout.X_AXIS));
      p2.add(label2);
      p2.add(Box.createRigidArea(new Dimension(21, 0)));
      setComponentSize(textField2, 300, 25);
      p2.add(textField2);

      // button
      JPanel p3 = new JPanel();
      p3.setLayout(new BoxLayout(p3, BoxLayout.LINE_AXIS));
      p3.add(submitButton);
      p3.add(Box.createRigidArea(new Dimension(10, 0)));
      p3.add(cancelButton);

      JPanel contentPane = new JPanel();
      contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.PAGE_AXIS));
      // spacer - creates rigid box at top with width 0 and height 15
      contentPane.add(Box.createRigidArea(new Dimension(0, 30)));
      contentPane.add(p0);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p1);
      contentPane.add(Box.createRigidArea(new Dimension(0, 5)));
      contentPane.add(p2);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p3);
      contentPane.setOpaque(true);

      dialog.setContentPane(contentPane);
      int w = 600, h = 400;
      dialog.setSize(new Dimension(w, h));
      dialog.setLocation(DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop2);
      dialog.setVisible(true);

      dialog.addWindowListener(new WindowAdapter() {
         public void windowOpened(WindowEvent we) {
            SwingUtilities.invokeLater(new Runnable() {
               public void run() {
                  textField1.requestFocusInWindow() ;
               }
            });
         }
      });
   }

   public void setComponentSize(JComponent component, int width, int height) {
      Dimension size = new Dimension(width, height);
      component.setPreferredSize(size);
      component.setMinimumSize(size);
      component.setMaximumSize(size);
   }

   public void actionPerformed(ActionEvent e) {

      if (e.getSource() == cancelButton) {

         canceled = true ;
         dialog.setVisible(false);
         dialog.dispose();

      } else if (e.getSource() == textField1 || e.getSource() == textField2 || e.getSource() == submitButton) {

         ligandId = textField2.getText();
         smileString = textField1.getText();

         if (ligandId.length() == 0 || ligandId.length() > 8) {
            DialogWindows.showWarningDialog("Ligand Id must be 1 - 8 characters - please try again");

         } else if (!ligandId.matches("[a-zA-Z0-9-_]+")) {
            DialogWindows.showWarningDialog(
               "Ligand Id must consist of these characters: " +
               "A-Z, a-z, 1-9, -, _ (no spaces allowed) - please try again"
            ) ;

         } else if( targetLib.contains( ligandId ) ){

            DialogWindows.showWarningDialog("This Ligand Id already exists - please enter a new Ligand Id");

         } else {
            dialog.setVisible(false);
            dialog.dispose();
         }
      }
   }

   private boolean canceled ;
   private String ligandId ;
   private String smileString ;
   boolean canceled(){ return canceled ; }
   String getLigandId(){ return ligandId ; }
   String getSmileString(){ return smileString ; }
}
// -----------------------------------------------------------------------------

class ViewFileDialog{

   public ViewFileDialog(){}

   public void show( String filename, String text ){
      final JDialog dialog = new JDialog( JLMain.mainFrame, filename ) ;

      JLabel label = new JLabel( "<html><body><p>View File</p></body></html>" ) ;

      label.setHorizontalAlignment( JLabel.CENTER ) ;
      label.setVerticalAlignment( JLabel.TOP ) ;

//    Font font = label.getFont() ;
      Font f1 = new Font( "", Font.PLAIN, 12 ) ;
      label.setFont( f1 ) ;

      JTextArea textArea = new JTextArea() ;
//    Font.MONOSPACED is not defined in 1.5, although works with -source 1.5
//    textArea.setFont( new Font( Font.MONOSPACED, Font.PLAIN, 14 ) ) ;
      textArea.setFont( new Font( "MONOSPACED", Font.PLAIN, 14 ) ) ;
      textArea.setEditable( false ) ;
      textArea.setText( text ) ;
      textArea.setCaretPosition( 0 ) ;
      JScrollPane scrollPane = new JScrollPane( textArea ) ;

      JButton closeButton = new JButton( "Close" ) ;
      closeButton.addActionListener( new ActionListener(){
         public void actionPerformed( ActionEvent e ){
            dialog.setVisible( false ) ;
            dialog.dispose() ;
         }
      } ) ;

      JPanel closePanel = new JPanel() ;
      closePanel.setLayout( new BoxLayout( closePanel, BoxLayout.LINE_AXIS ) ) ;
      closePanel.add( Box.createHorizontalGlue() ) ;
      closePanel.add( closeButton ) ;
      closePanel.setBorder( BorderFactory.createEmptyBorder( 0, 0, 5, 5 ) ) ;

      JPanel contentPane = new JPanel( new BorderLayout() ) ;
      contentPane.add( scrollPane, BorderLayout.CENTER ) ;
      contentPane.add( closePanel, BorderLayout.PAGE_END ) ;
      contentPane.setOpaque( true ) ;
      dialog.setContentPane( contentPane ) ;

      int w = 800, h = 600 ;
      dialog.setSize( new Dimension( w, h ) ) ;
      dialog.setLocation( DialogWindows.screenWidth/ 2 - w/ 2, DialogWindows.locationTop ) ;
      dialog.setVisible( true ) ;
   }
}
// -----------------------------------------------------------------------------

class ChangeBondDialog implements ActionListener {

   private Molecule ligand ;
   private JDialog dialog;
   private JLabel label1;
   private JComboBox bondTypeList;
   private JButton submitButton, cancelButton;

   public String[] bondTypeArray = { "single", "double", "triple", "deloc", "aromatic", "metal" } ;
   private Bond bond ;

   public ChangeBondDialog( Molecule ligand, Bond bond ){
      this.bond = bond ;
      this.ligand = ligand ;
   }

   public void showDialog() {

      dialog = new JDialog(JLMain.mainFrame, true);
      dialog.setTitle("Edit Bond Type");

      label1 = new JLabel("Bond Type");

      submitButton = new JButton("Submit");
      submitButton.addActionListener(this);

      cancelButton = new JButton("Cancel");
      cancelButton.addActionListener(this);

      // label panel
      JPanel p0 = new JPanel();
      p0.setLayout(new BoxLayout(p0, BoxLayout.X_AXIS));
      p0.add(new JLabel("Please edit your bond type and click on Submit:"));

      // get selected index
      int selectedIndex = 0;

      for( int i = 0 ; i < bondTypeArray.length ; i++ ){
         if( bond.type.startsWith( bondTypeArray[i] ) ){
            selectedIndex = i ;
            break ;
         }
      }

      // bondType panel
      JPanel p1 = new JPanel();
      p1.setLayout(new BoxLayout(p1, BoxLayout.LINE_AXIS));
      p1.add(label1);
      p1.add(Box.createRigidArea(new Dimension(23, 0)));
      bondTypeList = new JComboBox(bondTypeArray);
      bondTypeList.setSelectedIndex(selectedIndex);
      setComponentSize(bondTypeList, 150, 30);
      p1.add(bondTypeList);

      // button panel
      JPanel p2 = new JPanel();
      p2.setLayout(new BoxLayout(p2, BoxLayout.LINE_AXIS));
      p2.add(submitButton);
      p2.add(Box.createRigidArea(new Dimension(10, 0)));
      p2.add(cancelButton);

      JPanel contentPane = new JPanel();
      contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.PAGE_AXIS));
      contentPane.add(Box.createRigidArea(new Dimension(0, 30)));
      contentPane.add(p0);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p1);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p2);
      contentPane.setOpaque(true);

      dialog.setContentPane(contentPane);
      int w = DialogWindows.dialogWidth, h = DialogWindows.dialogHeight;
      dialog.setSize(new Dimension(w, h));
      dialog.setLocation(DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop2);
      dialog.setVisible(true);

      dialog.addWindowListener(new WindowAdapter() {
         public void windowOpened(WindowEvent we ) {
            SwingUtilities.invokeLater(new Runnable(){
               public void run() {
                  bondTypeList.requestFocusInWindow() ;
               }
            });
         }
      });
   }

   public void setComponentSize(JComponent component, int width, int height) {
      Dimension size = new Dimension(width, height);
      component.setPreferredSize(size);
      component.setMinimumSize(size);
      component.setMaximumSize(size);
   }

   public void actionPerformed(ActionEvent e) {

      if (e.getSource() == cancelButton) {
         dialog.setVisible(false);
         dialog.dispose();

      } else if (e.getSource() == submitButton) {

//       boolean isError = false;

         String newBondType = (String)bondTypeList.getSelectedItem();

         if( ! newBondType.equals( bond.type ) ){
            JLMain.currentPanel.gLib.undoableChangeBond( ligand, bond, newBondType ) ;
         }


         dialog.setVisible(false);
         dialog.dispose();
      }
   }
}
// -----------------------------------------------------------------------------

class ChangeAtomDialog implements ActionListener {

   private Map<String,Atom> atomMap ;
   private Molecule ligand;
   private JDialog dialog;
   private JLabel label1, label2, label3;
   private JLabel label5;
   private JTextField textField1;
   private JComboBox atomTypeList;
   private JComboBox atomChirList;
   private JTextField textField3;
   private JButton submitButton, cancelButton;

   public String[] atomTypeArray = {
      "H",  "He", "Li", "Be", "B",  "C",  "N",  "O",  "F",  "Ne",
      "Na", "Mg", "Al", "Si", "P",  "S",  "Cl", "Ar", "K",  "Ca",
      "Sc", "Ti", "V",  "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn",
      "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y",  "Zr",
      "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn",
      "Sb", "Te", "I",  "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd",
      "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb",
      "Lu", "Hf", "Ta", "W",  "Re", "Os", "Ir", "Pt", "Au", "Hg",
      "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th",
      "Pa", "U",  "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm",
      "Md", "No", "Lr",
   };

   private Atom atom = null;
   private String previousType ;

   public ChangeAtomDialog( Molecule ligand, Atom atom ){

      this.ligand = ligand;
      this.atom = atom;
      atomMap = ligand.atomView( atom ) ;
      if( atomMap.remove( atom.key4 ) == null ) throw new NullPointerException() ;

      dialog = new JDialog(JLMain.mainFrame, true);
      dialog.setTitle("Edit Atom Details");

      label1 = new JLabel("Atom Id");

      textField1 = new JTextField(10);
      textField1.setText(atom.key4);
      textField1.addActionListener(this);

      label2 = new JLabel("Atom Type");

      label3 = new JLabel("Atom Charge");

      label5 = new JLabel("Chirality");

      textField3 = new JTextField(10);
      textField3.setText(String.valueOf(atom.charge));
      textField3.addActionListener(this);

      submitButton = new JButton("Submit");
      submitButton.addActionListener(this);

      cancelButton = new JButton("Cancel");
      cancelButton.addActionListener(this);

      // label panel
      JPanel p0 = new JPanel();
      p0.setLayout(new BoxLayout(p0, BoxLayout.X_AXIS));
      p0.add(new JLabel("Please edit your atom details and click on Submit:"));

      // atomId panel
      JPanel p1 = new JPanel();
      p1.setLayout(new BoxLayout(p1, BoxLayout.X_AXIS));
      p1.add(label1);
      p1.add(Box.createRigidArea(new Dimension(40, 0)));
      setComponentSize(textField1, 150, 25);
      p1.add(textField1);

      // get selected index
      int selectedIndex = 0;

      for (int i = 0; i < atomTypeArray.length; i++) {
         if (atom.symbol.equals(atomTypeArray[i])) {
            selectedIndex = i;
            break;
         }
      }

      // atomTypeSymbol panel
      JPanel p2 = new JPanel();
      p2.setLayout(new BoxLayout(p2, BoxLayout.LINE_AXIS));
      p2.add(label2);
      p2.add(Box.createRigidArea(new Dimension(23, 0)));
      atomTypeList = new JComboBox(atomTypeArray);
      atomTypeList.setSelectedIndex(selectedIndex);
      setComponentSize(atomTypeList, 154, 30);
      p2.add(atomTypeList);
      atomTypeList.addActionListener( new ActionListener(){
         public void actionPerformed( ActionEvent e ){
            String name = textField1.getText() ;
            String type = (String)atomTypeList.getSelectedItem() ;
            if( name.toUpperCase().startsWith( previousType.toUpperCase() ) ){
               type = type.toUpperCase() ;
               textField1.setText( type + name.substring( previousType.length() ) ) ;
            }
            previousType = type ;
         }
      } ) ;
      previousType = (String)atomTypeList.getSelectedItem() ;

      // atomPartialCharge panel
      JPanel p3 = new JPanel();
      p3.setLayout(new BoxLayout(p3, BoxLayout.X_AXIS));
      p3.add(label3);
      p3.add(Box.createRigidArea(new Dimension(10, 0)));
      setComponentSize(textField3, 150, 25);
      p3.add(textField3);

      // atomChirSymbol panel
      JPanel p5 = new JPanel();
      p5.setLayout(new BoxLayout(p5, BoxLayout.LINE_AXIS));
      p5.add(label5);
      p5.add(Box.createRigidArea(new Dimension(40, 0)));
      int selectedChirIndex = - 1 ;
      if( atom.chir != null ) selectedChirIndex = atom.chir.getIndex() ;
      if( selectedChirIndex < 0 ){
         atomChirList = new JComboBox( new String[]{ "n/a" } ) ;
         atomChirList.setSelectedIndex( 0 ) ;
         atomChirList.setEnabled( false ) ;
         label5.setEnabled( false ) ;
      }
      else{
         atomChirList = new JComboBox( atom.chir.getOptions() ) ;
         atomChirList.setSelectedIndex( selectedChirIndex ) ;
         atomChirList.setEnabled( true ) ;
         label5.setEnabled( true ) ;
      }
      setComponentSize(atomChirList, 154, 30);
      p5.add(atomChirList);

      // button panel
      JPanel p4 = new JPanel();
      p4.setLayout(new BoxLayout(p4, BoxLayout.LINE_AXIS));
      p4.add(cancelButton);
      p4.add(Box.createRigidArea(new Dimension(72, 0)));
      p4.add(submitButton);

      JPanel contentPane = new JPanel();
      contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.PAGE_AXIS));
      contentPane.add(Box.createRigidArea(new Dimension(0, 30)));
      contentPane.add(p0);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p1);
      contentPane.add(Box.createRigidArea(new Dimension(0, 5)));
      contentPane.add(p2);
      contentPane.add(Box.createRigidArea(new Dimension(0, 5)));
      contentPane.add(p3);
      contentPane.add(Box.createRigidArea(new Dimension(0, 5)));
      contentPane.add(p5);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p4);
      contentPane.setOpaque(true);

      dialog.setContentPane(contentPane);
      int w = DialogWindows.dialogWidth, h = DialogWindows.dialogHeight;
      dialog.setSize(new Dimension(w, h));
      dialog.setLocation(DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop2);
      dialog.setVisible(true);

      dialog.addWindowListener(new WindowAdapter() {
         public void windowOpened(WindowEvent we) {
            SwingUtilities.invokeLater(new Runnable() {
               public void run() {
                  textField1.requestFocusInWindow() ;
               }
            });
         }
      });
   }

   public void setComponentSize(JComponent component, int width, int height) {
      Dimension size = new Dimension(width, height);
      component.setPreferredSize(size);
      component.setMinimumSize(size);
      component.setMaximumSize(size);
   }

   public void actionPerformed(ActionEvent e) {

      if (e.getSource() == cancelButton) {

         dialog.setVisible(false);
         dialog.dispose();

      } else if (e.getSource() == textField1 || e.getSource() == textField3 || e.getSource() == submitButton) {

         boolean isError = false;

         String newDisplayId = textField1.getText();
         String newAtomTypeSymbol = (String)atomTypeList.getSelectedItem();
         float newAtomPartialCharge = 0;

         boolean chargeException = false;

         try {
            newAtomPartialCharge = Float.parseFloat(textField3.getText());
         } catch (NumberFormatException ex) {
            chargeException = true;
         }

         if (newDisplayId.length() == 0 || newDisplayId.length() > 4) {
            isError = true;
            DialogWindows.showWarningDialog("Atom Id must be 1 - 4 characters - please try again");
         }

         else if (chargeException == true || newAtomPartialCharge < -3.0 || newAtomPartialCharge > 3.0) {
            isError = true;
            DialogWindows.showWarningDialog("Charge must be a number between -3.0 and +3.0 - please try again");
         }

         else if( atomMap.get( newDisplayId ) != null ){
            isError = true;
            DialogWindows.showWarningDialog("This Atom Id already exists - please enter a new Atom Id");
         }

         if (isError == false) {

            String type = GenericRestraint.VOID_TYPE ;
            if( atom.chir != null ) type = atom.chir.getType( atomChirList.getSelectedIndex() ) ;
            JLMain.currentPanel.gLib.undoableChangeAtom(
               type,
               ligand, atom, newDisplayId, newAtomTypeSymbol, newAtomPartialCharge
            ) ;
            dialog.setVisible(false);
            dialog.dispose();
         }
      }
   }
}
// -----------------------------------------------------------------------------

class CorrectLinkIdDialog extends ChangeLigandIdDialog{

   public CorrectLinkIdDialog( String prefix, GLib gLib, String key8 ){
      super( gLib, key8 ) ;

      if( gLib.get( key8 ) != null ){

         String msg = prefix + " ligand Id '" + key8 + "' is already in use. Please select a new Id." ;
         DialogWindows.showWarningDialog( msg ) ;

         showDialog() ;
      }

      else if( key8.length() > 8 ){

         String msg = prefix + " ligand Id '" + key8 + "' contains more than 8 characters. " ;
         msg += "Please select a new Id." ;
         DialogWindows.showWarningDialog( msg ) ;

         showDialog() ;
      }

      else{
         newKey8 = key8 ;
      }
   }
   boolean exists( String ligandId ){
      return targetLib.contains( ligandId ) ;
   }
}
// -----------------------------------------------------------------------------

class ChangeLigandIdDialog implements ActionListener {

// private JLGPanel panel;
   private JDialog dialog;
   private JLabel label1;
   private JTextField textField1;
   private JButton submitButton, cancelButton;
   private String oldLigandId = "";
// private static String fs = File.separator;
   GLib targetLib ;

   public ChangeLigandIdDialog( GLib gLib, String key8 ){
      targetLib = gLib ;
      oldLigandId = key8 ;
   }

   public void showDialog() {

      dialog = new JDialog(JLMain.mainFrame, true);
      dialog.setTitle("Edit Ligand Id");

      label1 = new JLabel("Ligand Id");

      textField1 = new JTextField(10);
      textField1.setText(oldLigandId);
      textField1.addActionListener(this);

      submitButton = new JButton("Submit");
      submitButton.addActionListener(this);

      cancelButton = new JButton("Cancel");
      cancelButton.addActionListener(this);

      // label
      JPanel p0 = new JPanel();
      p0.setLayout(new BoxLayout(p0, BoxLayout.X_AXIS));
      p0.add(new JLabel("Please edit your Ligand Id (1 - 8 characters) and click on Submit:"));

      // ligandId
      JPanel p1 = new JPanel();
      p1.setLayout(new BoxLayout(p1, BoxLayout.X_AXIS));
      p1.add(label1);
      p1.add(Box.createRigidArea(new Dimension(40, 0)));
      setComponentSize(textField1, 150, 25);
      p1.add(textField1);

      // button
      JPanel p2 = new JPanel();
      p2.setLayout(new BoxLayout(p2, BoxLayout.LINE_AXIS));
      p2.add(submitButton);
      p2.add(Box.createRigidArea(new Dimension(10, 0)));
      p2.add(cancelButton);

      JPanel contentPane = new JPanel();
      contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.PAGE_AXIS));
      contentPane.add( Box.createVerticalStrut( 10 ) ) ;
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p0);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p1);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p2);
      contentPane.setOpaque(true);

      dialog.setContentPane(contentPane);
      int w = DialogWindows.dialogWidth, h = DialogWindows.dialogHeight;
      dialog.setSize(new Dimension(w, h));
      dialog.setLocation(DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop2);
      dialog.setVisible(true);

      dialog.addWindowListener(new WindowAdapter(){
         public void windowOpened( WindowEvent we ){
            SwingUtilities.invokeLater(new Runnable(){
               public void run(){
                  textField1.requestFocusInWindow() ;
               }
            });
         }
      });
   }

   public void setComponentSize(JComponent component, int width, int height) {
      Dimension size = new Dimension(width, height);
      component.setPreferredSize(size);
      component.setMinimumSize(size);
      component.setMaximumSize(size);
   }

   String newKey8 = null ;

   public void actionPerformed(ActionEvent e) {

      if (e.getSource() == cancelButton) {
         dialog.setVisible(false);
         dialog.dispose();
      }
      else if (e.getSource() == textField1 || e.getSource() == submitButton) {

         String ligandId = textField1.getText();

         if (ligandId.length() == 0 || ligandId.length() > 8) {
            DialogWindows.showWarningDialog("Ligand Id must be 1 - 8 characters - please try again");

         }
         else if (!ligandId.matches("[a-zA-Z0-9-_]+")) {
            DialogWindows.showWarningDialog(
               "Ligand Id must consist of these characters: " +
               "A-Z, a-z, 1-9, -, _ (no spaces allowed) - please try again"
            ) ;

         } else if( exists( ligandId ) ){

            DialogWindows.showWarningDialog("This Ligand Id already exists - please enter a new Ligand Id");

         } else {

            if (!ligandId.equals(oldLigandId)) {

               newKey8 = ligandId ;

            }
            dialog.setVisible(false);
            dialog.dispose();
         }
      }
   }
   boolean exists( String ligandId ){
      return ! ligandId.equals( oldLigandId ) && targetLib.contains( ligandId ) ;
   }
}
// -----------------------------------------------------------------------------

class SaveFileDialog implements ActionListener {

   private String ligandId = "";
   private boolean canSaveLink = false;
   private boolean canAppendLink = false;
   private boolean canAppendComp = false;

   private JDialog dialog;
// private JLabel label1;
   private ButtonGroup radioButtonGroup;
   private JButton submitButton, cancelButton;
// private String oldLigandId = "";
// private static String fs = File.separator;
   public String fromFile = "";
   public boolean append = false;
   public boolean asLink = false;
   public boolean cancel = false;

   public SaveFileDialog( String ligandId, boolean canSaveLink, boolean canAppendLink, boolean canAppendComp ){
      this.ligandId = ligandId;
      this.canSaveLink = canSaveLink;
      this.canAppendLink  = canAppendLink ;
      this.canAppendComp  = canAppendComp ;
      fromFile = ligandId + ".cif";
   }

   public void showDialog() {

//    String f1 = ligandId;
//    String f2 = ligandId + "_LINK";

      dialog = new JDialog(JLMain.mainFrame, true);
      dialog.setTitle("Choose file");

      radioButtonGroup = new ButtonGroup();

      // ligandId
      JPanel p1 = new JPanel();
      p1.setLayout(new BoxLayout(p1, BoxLayout.X_AXIS));

      JRadioButton radioButton1 = new JRadioButton(ligandId);
      radioButton1.setSelected(true);
      radioButton1.addActionListener(this);
      radioButtonGroup.add(radioButton1);
      p1.add(radioButton1);
      p1.add(Box.createRigidArea(new Dimension(40, 0)));

      if (canAppendComp  == true) {
         JRadioButton radioButton3 = new JRadioButton(ligandId + " (append)");
         radioButton3.addActionListener(this);
         radioButtonGroup.add(radioButton3);
         p1.add(radioButton3);
         p1.add(Box.createRigidArea(new Dimension(40, 0)));
      }

      if (canSaveLink == true) {
         JRadioButton radioButton2 = new JRadioButton(ligandId + "_LINK");
         radioButton2.addActionListener(this);
         radioButtonGroup.add(radioButton2);
         p1.add(radioButton2);
         p1.add(Box.createRigidArea(new Dimension(40, 0)));
      }

      if (canAppendLink  == true && canSaveLink == true) {
         JRadioButton radioButton4 = new JRadioButton(ligandId + "_LINK (append)");
         radioButton4.addActionListener(this);
         radioButtonGroup.add(radioButton4);
         p1.add(radioButton4);
         p1.add(Box.createRigidArea(new Dimension(40, 0)));
      }

      submitButton = new JButton("Submit");
      submitButton.addActionListener(this);

      cancelButton = new JButton("Cancel");
      cancelButton.addActionListener(this);

      // label
      JPanel p0 = new JPanel();
      p0.setLayout(new BoxLayout(p0, BoxLayout.X_AXIS));

      if (canSaveLink == true && canAppendComp  == true) {
         p0.add( new JLabel(
            "Please choose how you wish to save the ligand: as monomer or link; " +
            "in a separate file or in the atached library:"
         ) ) ;
      } else if (canSaveLink == true) {
         p0.add(new JLabel("Please choose how you wish to save the ligand (as monomer or link):"));
      } else {
         p0.add(new JLabel(
            "Please choose whether you wish to save the ligand or append it to the attached library"
         ));
      }

      // button
      JPanel p2 = new JPanel();
      p2.setLayout(new BoxLayout(p2, BoxLayout.LINE_AXIS));
      p2.add(submitButton);
      p2.add(Box.createRigidArea(new Dimension(10, 0)));
      p2.add(cancelButton);

      JPanel contentPane = new JPanel();
      contentPane.setLayout(new BoxLayout(contentPane, BoxLayout.PAGE_AXIS));
      contentPane.add(Box.createRigidArea(new Dimension(0, 30)));
      contentPane.add(p0);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p1);
      contentPane.add(Box.createRigidArea(new Dimension(0, 20)));
      contentPane.add(p2);
      contentPane.setOpaque(true);

      dialog.setContentPane(contentPane);
      int w = 800, h = DialogWindows.dialogHeight;
      dialog.setSize(new Dimension(w, h));
      dialog.setLocation(DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop2);
      dialog.setVisible(true);

      dialog.addWindowListener(new WindowAdapter(){
         public void windowOpened( WindowEvent we ){
         SwingUtilities.invokeLater(new Runnable(){
            public void run(){
            }
         });
      }
   });
   }

   public void actionPerformed(ActionEvent e) {

      if (e.getActionCommand().equals(ligandId)) {
//       fromFile = ligandId + ".lib";
         append = false ;
         asLink = false ;
         cancel = false ;
      }
      else if (e.getActionCommand().equals(ligandId + "_LINK")) {
//       fromFile = "linkDir" + File.separator + ligandId + "_LINK.lib";
         append = false ;
         asLink = true ;
         cancel = false ;
      }
      else if (e.getActionCommand().equals(ligandId + " (append)")) {
//       fromFile = ligandId + ".lib";
         append = true ;
         asLink = false ;
         cancel = false ;
      }
      else if (e.getActionCommand().equals(ligandId + "_LINK (append)")) {
//       fromFile = "linkDir" + File.separator + ligandId + "_LINK.lib";
         append = true ;
         asLink = true ;
         cancel = false ;
      }
      else if (e.getSource() == cancelButton) {
//       fromFile = "";
         cancel = true ;

         dialog.setVisible(false);
         dialog.dispose();
      }
      else if (e.getSource() == submitButton) {
         dialog.setVisible(false);
         dialog.dispose();
      }
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class ViewTable extends JPanel{
// private boolean DEBUG = true ;

   public ViewTable(){
      super(new GridLayout( 1, 0 ) ) ;
   }

   public void showTable( Molecule molecule, String mode ){
      JFrame frame = new JFrame( molecule.key8 + ": " + mode ) ;
      this.setOpaque( true ) ;

      frame.setContentPane( this ) ;
      frame.pack() ; // pack must be before setSize
      int w = 800, h = 600 ;
      frame.setSize( w, h ) ;
      frame.setLocation( DialogWindows.screenWidth/ 2 - w/ 2, DialogWindows.locationTop ) ;
      frame.setVisible( true ) ;

      MyTableModel myTableModel = new MyTableModel( molecule ) ;
      myTableModel.showTable( mode ) ;
      JTable table = new JTable( myTableModel ) ;

      table.setForeground( Color.BLACK ) ;
      table.setGridColor( Color.LIGHT_GRAY ) ;
      table.setPreferredScrollableViewportSize( new Dimension( 1000, 700 ) ) ;
      JScrollPane scrollPane = new JScrollPane( table ) ;
      this.add( scrollPane ) ;
   }

   @SuppressWarnings("serial")
   class MyTableModel extends AbstractTableModel{
      private Molecule molecule = null ;
      private String[] columnNames = null ;
      private Object[][] data = null ;

      public MyTableModel( Molecule mol ){
         molecule = mol ;
      }

      public void showTable( String mode ){

         if( mode.equals( "ATOMS" ) ){

            columnNames = new String[7] ;

            columnNames[0] = "Atom Id" ;
            columnNames[1] = "Atom Type Symbol" ;
            columnNames[2] = "Atom Type Energy" ;
            columnNames[3] = "Atom Partial Charge" ;
            columnNames[4] = "x" ;
            columnNames[5] = "y" ;
            columnNames[6] = "z" ;

            data = new Object[molecule.atoms.size()][7] ;

            for( int i = 0, n = molecule.atoms.size() ; i < n ; i++ ){
               Atom atom = molecule.atoms.get( i ) ;
               data[i][0] = atom.key4 ;
               data[i][1] = atom.symbol ;
               data[i][2] = atom.energy ;
               data[i][3] = atom.charge ;
               data[i][4] = String.format( "%8.3f", atom.xyz[0] ) ;
               data[i][5] = String.format( "%8.3f", atom.xyz[1] ) ;
               data[i][6] = String.format( "%8.3f", atom.xyz[2] ) ;
            }
         }

         else if( mode.equals( "BONDS" ) ){

            columnNames = new String[5] ;

            columnNames[0] = "Atom Id1" ;
            columnNames[1] = "Atom Id2" ;
            columnNames[2] = "Bond Type" ;
            columnNames[3] = "Bond Dist" ;
            columnNames[4] = "Bond Dist Esd" ;

            List<GenericRestraint> rstList = molecule.restraints.get(0) ;
            data = new Object[rstList.size()][5] ;

            for( int i = 0, n = rstList.size() ; i < n ; i++ ){
               GenericRestraint bond = rstList.get( i ) ;
               data[i][0] = bond.atoms.get( 0 ).key4 ;
               data[i][1] = bond.atoms.get( 1 ).key4 ;
               data[i][2] = bond.type ;
               data[i][3] = bond.value.mean ;
               data[i][4] = bond.value.sd ;
            }
         }

         else if( mode.equals( "ANGLES" ) ){

            columnNames = new String[5] ;

            columnNames[0] = "Atom Id1" ;
            columnNames[1] = "Atom Id2" ;
            columnNames[2] = "Atom Id3" ;
            columnNames[3] = "Bond Angle" ;
            columnNames[4] = "Bond Angle Esd" ;

            List<GenericRestraint> rstList = molecule.restraints.get(1) ;
            data = new Object[rstList.size()][5] ;

            for( int i = 0, n = rstList.size() ; i < n ; i++ ){
               GenericRestraint angle = rstList.get( i ) ;
               data[i][0] = angle.atoms.get( 0 ).key4 ;
               data[i][1] = angle.atoms.get( 1 ).key4 ;
               data[i][2] = angle.atoms.get( 2 ).key4 ;
               data[i][3] = angle.value.mean ;
               data[i][4] = angle.value.sd ;
            }
         }

         else if( mode.equals( "TORSIONS" ) ){

            columnNames = new String[7] ;

            columnNames[0] = "Id" ;
            columnNames[1] = "Atom Id1" ;
            columnNames[2] = "Atom Id2" ;
            columnNames[3] = "Atom Id3" ;
            columnNames[4] = "Atom Id4" ;
            columnNames[5] = "Torsion Angle" ;
            columnNames[6] = "Torsion Angle Esd" ;

            List<GenericRestraint> rstList = molecule.restraints.get(2) ;
            data = new Object[rstList.size()][7] ;

            for( int i = 0, n = rstList.size() ; i < n ; i++ ){
               GenericRestraint tor = rstList.get( i ) ;
               data[i][0] = tor.key ;
               data[i][1] = tor.atoms.get( 0 ).key4 ;
               data[i][2] = tor.atoms.get( 1 ).key4 ;
               data[i][3] = tor.atoms.get( 2 ).key4 ;
               data[i][4] = tor.atoms.get( 3 ).key4 ;
               data[i][5] = tor.value.mean ;
               data[i][6] = tor.value.sd ;
            }
         }

         else if( mode.equals( "CHIRALITIES" ) ){

            columnNames = new String[11] ;

            columnNames[0] = "Id" ;
            columnNames[1] = "Id Centre" ;
            columnNames[2] = "Atom Id1" ;
            columnNames[3] = "Atom Id2" ;
            columnNames[4] = "Atom Id3" ;
            columnNames[5] = "Volume Sign" ;
            columnNames[6] = "Atom Id4" ;
            columnNames[7] = "Atom Id5" ;
            columnNames[8] = "Atom Id6" ;
            columnNames[9] = "Atom Id7" ;
            columnNames[10] = "Atom Id8" ;

            List<GenericRestraint> rstList = molecule.restraints.get(3) ;
            data = new Object[rstList.size()][11] ;

            for( int i = 0, n = rstList.size() ; i < n ; i++ ){
               GenericRestraint chir = rstList.get( i ) ;
               data[i][0] = chir.key ;
               data[i][1] = chir.atoms.get( 0 ).key4 ;
               data[i][2] = chir.atoms.get( 1 ).key4 ;
               data[i][3] = chir.atoms.get( 2 ).key4 ;
               data[i][4] = chir.atoms.get( 3 ).key4 ;
               data[i][5] = chir.type ;
               data[i][6] = chir.atoms.get( 4 ).key4 ;
               data[i][7] = chir.atoms.get( 5 ).key4 ;
               data[i][8] = chir.atoms.get( 6 ).key4 ;
               data[i][9] = chir.atoms.get( 7 ).key4 ;
               data[i][10] = chir.atoms.get( 8 ).key4 ;
            }
         }

         else if( mode.equals( "PLANES" ) ){

            columnNames = new String[3] ;

            columnNames[0] = "Plane Id" ;
            columnNames[1] = "Atom Id" ;
            columnNames[2] = "Dist Esd" ;

            List<GenericRestraint> rstList = molecule.restraints.get(4) ;
            int i = 0 ;
            for( GenericRestraint plane : rstList ) i += plane.atoms.size() ;
            data = new Object[i][3] ;

            i = 0 ;
            for( GenericRestraint plane : rstList ){
               for( Atom atom : plane.atoms ){
                  data[i][0] = plane.key ;
                  data[i][1] = atom.key4 ;
                  data[i][2] = plane.sigma.sd ;
                  i++ ;
               }
            }
         }
      }

      public int getColumnCount(){
         return columnNames.length ;
      }

      public int getRowCount(){
         return data.length ;
      }

      public String getColumnName( int col ){
         return columnNames[col] ;
      }

      public Object getValueAt( int row, int col ){
         return data[row][col] ;
      }

      public boolean isCellEditable( int row, int col ){
         return false ;
      }
   }
}
// -----------------------------------------------------------------------------

class ChooseLigandFromFile implements ActionListener{

   private JDialog dialog ;
   private JLabel label1 ;
   private JComboBox ligandIdList ;
   private JButton submitButton, cancelButton ;
   public String ligandId = "" ;

// public void showDialog( Vector listVector ){
   public void showDialog( Object[] keys ){

      dialog = new JDialog( JLMain.mainFrame, true ) ;
      dialog.setTitle( "Choose Ligand Id" ) ;

      label1 = new JLabel( "Ligand Id" ) ;

      submitButton = new JButton( "Submit" ) ;
      submitButton.addActionListener( this ) ;

      cancelButton = new JButton( "Cancel" ) ;
      cancelButton.addActionListener( this ) ;

      // label panel
      JPanel p0 = new JPanel() ;
      p0.setLayout( new BoxLayout( p0, BoxLayout.X_AXIS ) ) ;
      p0.add( new JLabel( "This file contains more than 1 ligand - please choose the ligand you wish to load:" ) ) ;

      // ligandId panel
      JPanel p1 = new JPanel() ;
      p1.setLayout( new BoxLayout( p1, BoxLayout.LINE_AXIS ) ) ;
      p1.add( label1 ) ;
      p1.add( Box.createRigidArea( new Dimension( 23, 0 ) ) ) ;
//    ligandIdList = new JComboBox( listVector.toArray() ) ;
      ligandIdList = new JComboBox( keys ) ;
      ligandIdList.setSelectedIndex( 0 ) ;
      setComponentSize( ligandIdList, 150, 30 ) ;
      p1.add( ligandIdList ) ;

      // button panel
      JPanel p2 = new JPanel() ;
      p2.setLayout( new BoxLayout( p2, BoxLayout.LINE_AXIS ) ) ;
      p2.add( submitButton ) ;
      p2.add( Box.createRigidArea( new Dimension( 10, 0 ) ) ) ;
      p2.add( cancelButton ) ;

      JPanel contentPane = new JPanel() ;
      contentPane.setLayout( new BoxLayout( contentPane, BoxLayout.PAGE_AXIS ) ) ;
      contentPane.add( Box.createRigidArea( new Dimension( 0, 30 ) ) ) ;
      contentPane.add( p0 ) ;
      contentPane.add( Box.createRigidArea( new Dimension( 0, 20 ) ) ) ;
      contentPane.add( p1 ) ;
      contentPane.add( Box.createRigidArea( new Dimension( 0, 20 ) ) ) ;
      contentPane.add( p2 ) ;
      contentPane.setOpaque( true ) ;

      dialog.setContentPane( contentPane ) ;
      int w = 600, h = DialogWindows.dialogHeight ;
      dialog.setSize( new Dimension( w, h ) ) ;
      dialog.setLocation( DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop2 ) ;
      dialog.setVisible( true ) ;

      dialog.addWindowListener( new WindowAdapter(){
         public void windowOpened( WindowEvent we ){
            SwingUtilities.invokeLater( new Runnable(){
               public void run(){
                  ligandIdList.requestFocusInWindow() ;
               }
            }) ;
         }
      }) ;
   }

   public void setComponentSize( JComponent component, int width, int height ){
      Dimension size = new Dimension( width, height ) ;
      component.setPreferredSize( size ) ;
      component.setMinimumSize( size ) ;
      component.setMaximumSize( size ) ;
   }

   public void actionPerformed( ActionEvent e ){

      ligandId = "" ;

      if( e.getSource() == cancelButton ){
         dialog.setVisible( false ) ;
         dialog.dispose() ;

      }
      else if( e.getSource() == submitButton ){

//       boolean isError = false ;

         ligandId =( String )ligandIdList.getSelectedItem() ;

         dialog.setVisible( false ) ;
         dialog.dispose() ;
      }
   }

   public String getLigandId(){
      return ligandId ;
   }
}
// -----------------------------------------------------------------------------

class SavePdbDialog implements ActionListener{

   public boolean includeH = true ;
   public boolean savePdb = false ;

   private String text1 = "Include Hydrogens" ;
   private String text2 = "Exclude Hydrogens" ;

   private JDialog dialog ;
   private JButton submitButton, cancelButton ;

   public void showDialog( String ligandId ){

      dialog = new JDialog( JLMain.mainFrame, true ) ;
      dialog.setTitle( "Save coordinates" ) ;

      ButtonGroup radioButtonGroup = new ButtonGroup() ;

      JPanel p1 = new JPanel() ;
      p1.setLayout( new BoxLayout( p1, BoxLayout.X_AXIS ) ) ;

      JRadioButton radioButton1 = new JRadioButton( text1 ) ;
      radioButton1.setSelected( true ) ;
      radioButton1.addActionListener( this ) ;
      radioButtonGroup.add( radioButton1 ) ;
      p1.add( radioButton1 ) ;
      p1.add( Box.createRigidArea( new Dimension( 40, 0 ) ) ) ;

      JRadioButton radioButton2 = new JRadioButton( text2 ) ;
      radioButton2.addActionListener( this ) ;
      radioButtonGroup.add( radioButton2 ) ;
      p1.add( radioButton2 ) ;
      p1.add( Box.createRigidArea( new Dimension( 40, 0 ) ) ) ;

      submitButton = new JButton( "Submit" ) ;
      submitButton.addActionListener( this ) ;

      cancelButton = new JButton( "Cancel" ) ;
      cancelButton.addActionListener( this ) ;

      // label panel
      JPanel p0 = new JPanel() ;
      p0.setLayout( new BoxLayout( p0, BoxLayout.X_AXIS ) ) ;

      p0.add( new JLabel( "Saving coordinates of " + ligandId ) ) ;

      // button panel
      JPanel p2 = new JPanel() ;
      p2.setLayout( new BoxLayout( p2, BoxLayout.LINE_AXIS ) ) ;
      p2.add( submitButton ) ;
      p2.add( Box.createRigidArea( new Dimension( 10, 0 ) ) ) ;
      p2.add( cancelButton ) ;

      JPanel contentPane = new JPanel() ;
      contentPane.setLayout( new BoxLayout( contentPane, BoxLayout.PAGE_AXIS ) ) ;
      contentPane.add( Box.createRigidArea( new Dimension( 0, 30 ) ) ) ;
      contentPane.add( p0 ) ;
      contentPane.add( Box.createRigidArea( new Dimension( 0, 20 ) ) ) ;
      contentPane.add( p1 ) ;
      contentPane.add( Box.createRigidArea( new Dimension( 0, 20 ) ) ) ;
      contentPane.add( p2 ) ;
      contentPane.setOpaque( true ) ;

      dialog.setContentPane( contentPane ) ;
      int w = 800, h = DialogWindows.dialogHeight ;
      dialog.setSize( new Dimension( w, h ) ) ;
      dialog.setLocation( DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop2 ) ;
      dialog.setVisible( true ) ;

      dialog.addWindowListener( new WindowAdapter(){
         public void windowOpened( WindowEvent we ){
            SwingUtilities.invokeLater( new Runnable(){ public void run(){} } ) ;
         }
      } ) ;
   }

   public void actionPerformed( ActionEvent e ){

      if( e.getActionCommand().equals( text1 ) ){
         includeH = true ;

      }
      else if( e.getActionCommand().equals( text2 ) ){
         includeH = false ;

      }
      else if( e.getSource() == cancelButton ){
         dialog.setVisible( false ) ;
         dialog.dispose() ;

      }
      else if( e.getSource() == submitButton ){
         savePdb = true ;
         dialog.setVisible( false ) ;
         dialog.dispose() ;
      }
   }
}
// -----------------------------------------------------------------------------

class SelectMoleculeDialog implements ActionListener{

   public String ligandId = null ;
   private JDialog dialog ;
   private JLabel label1 ;
   private JComboBox ligandList ;
   private JButton submitButton, cancelButton ;

   public SelectMoleculeDialog(){}

   public void showDialog( List<String> keyList, String title ){

      dialog = new JDialog( JLMain.mainFrame, true ) ;
      dialog.setTitle( title + " Ligand" ) ;

      label1 = new JLabel( "Choose Ligand" ) ;

      submitButton = new JButton( "Submit" ) ;
      submitButton.addActionListener( this ) ;

      cancelButton = new JButton( "Cancel" ) ;
      cancelButton.addActionListener( this ) ;

      // label panel
      JPanel p0 = new JPanel() ;
      p0.setLayout( new BoxLayout( p0, BoxLayout.X_AXIS ) ) ;
      p0.add( new JLabel( "Please choose the ligand which you would like to " + title + ":" ) ) ;

      // ligandIdArray panel
      JPanel p1 = new JPanel() ;
      p1.setLayout( new BoxLayout( p1, BoxLayout.LINE_AXIS ) ) ;
      p1.add( label1 ) ;
      p1.add( Box.createRigidArea( new Dimension( 23, 0 ) ) ) ;
      ligandList = new JComboBox( keyList.toArray() ) ;
      setComponentSize( ligandList, 150, 30 ) ;
      p1.add( ligandList ) ;

      // button panel
      JPanel p2 = new JPanel() ;
      p2.setLayout( new BoxLayout( p2, BoxLayout.LINE_AXIS ) ) ;
      p2.add( submitButton ) ;
      p2.add( Box.createRigidArea( new Dimension( 10, 0 ) ) ) ;
      p2.add( cancelButton ) ;

      JPanel contentPane = new JPanel() ;
      contentPane.setLayout( new BoxLayout( contentPane, BoxLayout.PAGE_AXIS ) ) ;
      contentPane.add( Box.createRigidArea( new Dimension( 0, 30 ) ) ) ;
      contentPane.add( p0 ) ;
      contentPane.add( Box.createRigidArea( new Dimension( 0, 20 ) ) ) ;
      contentPane.add( p1 ) ;
      contentPane.add( Box.createRigidArea( new Dimension( 0, 20 ) ) ) ;
      contentPane.add( p2 ) ;
      contentPane.setOpaque( true ) ;

      dialog.setContentPane( contentPane ) ;
      int w = DialogWindows.dialogWidth, h = DialogWindows.dialogHeight ;
      dialog.setSize( new Dimension( w, h ) ) ;
      dialog.setLocation( DialogWindows.screenWidth/2 - w/2, DialogWindows.locationTop2 ) ;
      dialog.setVisible( true ) ;

      dialog.addWindowListener( new WindowAdapter(){
         public void windowOpened( WindowEvent we ){
            SwingUtilities.invokeLater( new Runnable(){
               public void run(){
                  ligandList.requestFocusInWindow() ;
               }
            } ) ;
         }
      } ) ;
   }

   public void setComponentSize( JComponent component, int width, int height ){
      Dimension size = new Dimension( width, height ) ;
      component.setPreferredSize( size ) ;
      component.setMinimumSize( size ) ;
      component.setMaximumSize( size ) ;
   }

   public void actionPerformed( ActionEvent e ){
      if( e.getSource() == submitButton ) ligandId = (String) ligandList.getSelectedItem() ;
      dialog.setVisible( false ) ;
      dialog.dispose() ;
   }
}
// -----------------------------------------------------------------------------

