
import java.util.* ;
import java.io.* ;
//import java.awt.geom.* ;

// -----------------------------------------------------------------------------
// class TestLib{

class T{
   public static void main( String[] args ){
      Env.init( args ) ;
      new T() ;
   }

   GLib gLib = new GLib() ;

   String atomType = "Po" ;
   String bondType = "single" ;

   Finder finder1 ;
   Finder finder2 ;

   T(){
      File testFile = new File( "test" ) ;
      if( ! testFile.isFile() ) throw new NullPointerException() ;

      Scratch.init() ;
      StdLib.init() ;
      Versions.init() ;

      try{ new Rings() ; }
      catch( ExceptionInInitializerError e ){
         Throwable th = e.getCause() ;
         th.printStackTrace( System.err ) ;
         System.err.println( "AAAAA " + th.getMessage() ) ;
         System.err.println( "BBBB " ) ;
      }
      catch( Throwable e ){
         System.err.println( "CCCCC " + e.getMessage() ) ;
      }

      try{
         BufferedReader bf = new BufferedReader( new FileReader( testFile ) ) ;
         String line = bf.readLine() ;
         while( line != null ){ test( line ) ; line = bf.readLine() ; }
         System.out.println() ;
      }
      catch( FileNotFoundException e ){ throw new NullPointerException() ; }
      catch( IOException e ){ throw new NullPointerException() ; }

      if( ! Env.vbScratch ) Scratch.clear() ;
   }

   void test( String line ){
      String s = line.trim() ;
      if( s.length() == 0 ){ System.out.println() ; return ; }
      if( s.startsWith( "#" ) ){ System.out.println( s ) ; return ; }
      System.out.println() ;
      System.out.println( ">  " + s ) ;

      String[] words = s.split("\\s+") ;
      if(      words[0].equals( "details"    ) ) details    (          ) ;
      else if( words[0].equals( "set"        ) ) set        ( words    ) ;
      else if( words[0].equals( "edit"       ) ) edit       ( words    ) ;
      else if( words[0].equals( "remove"     ) ) remove     ( words[1] ) ;
      else if( words[0].equals( "find1"      ) ) find1      ( words    ) ;
      else if( words[0].equals( "find2"      ) ) find2      ( words    ) ;
      else if( words[0].equals( "read"       ) ) read       ( words[1] ) ;
      else if( words[0].equals( "write"      ) ) write      ( words[1] ) ;
      else if( words[0].equals( "summary"    ) ) summary    (          ) ;
      else if( words[0].equals( "add"        ) ) add        ( words    ) ;
      else if( words[0].equals( "clear"      ) ) clear      (          ) ;
      else if( words[0].equals( "regularise" ) ) regularise ( words[1] ) ;
      else if( words[0].equals( "convert"    ) ) convert    ( words[1] ) ;
      else throw new NullPointerException() ;
   }

   void set( String[] words ){
      String what = words[1] ;
      if( words.length < 3 ) throw new NullPointerException() ;
      if( what.equals( "bond" ) ){
         bondType = words[2] ;
      }
      else if( what.equals( "atom" ) ){
         atomType = words[2] ;
      }
   }

   void read( String inputFileName ){
      CifLib cifLib = new CifLib( new File( inputFileName ) ) ;
      gLib = new GLib() ;
      gLib.undoableAddMols( cifLib.polymerView() ) ;
   }

   void write( String outputFileName ){
      CifLib cifLib = new CifLib() ;
      for( Molecule mol : gLib ){
         Residue rsd = mol.getResidue() ;
         if( rsd == null ){
//          String info = cifLib.add( mol ) ;
            cifLib.add( mol ) ;
         }
         else{
            cifLib.add( rsd.mono ) ;
         }
      }
      cifLib.toBuffer().writeTo( new File( outputFileName ) ) ;
   }

   void summary(){
      System.out.println( gLib.summary() ) ;
   }

   void details(){
      finder1.details() ;
   }

   void clear(){
      StdLib.clear() ;
      gLib.undoableClear() ;
   }

   void find1( String[] words ){
      finder1 = new Finder() ;
      finder1.find( words ) ;
   }

   void find2( String[] words ){
      finder2 = new Finder() ;
      finder2.find( words ) ;
   }

   void remove( String key ){
      System.out.println( finder1 ) ;
      if( finder1 == null ) throw new NullPointerException() ;
      else if( key.equals( "bond" ) && finder1.bond != null ){
         gLib.undoableRemoveBond( finder1.link, finder1.bond ) ;
      }
      else if( key.equals( "atom" ) && finder1.atom != null ){
         gLib.undoableRemoveAtom( finder1.link, finder1.atom ) ;
      }
      else if( key.equals( "link" ) && finder1.link != null ){
         gLib.undoableRemoveMolecule( finder1.link ) ;
      }
      else throw new NullPointerException() ;
   }

   void edit( String[] words ){
      System.out.println( finder1 ) ;
      String what = words[1] ;
      String key = words[2] ;
      String value = words[3] ;
      if( finder1 == null ) throw new NullPointerException() ;
      else if( what.equals( "bond" ) && finder1.bond != null ){
         if( key.equals( "type"   ) ) gLib.undoableChangeBond( finder1.link, finder1.bond, value ) ;
      }
      else if( what.equals( "atom" ) && finder1.atom != null ){
         String key4New =  finder1.atom.key4 ;
         String symbolNew =  finder1.atom.symbol ;
         float chargeNew =  finder1.atom.charge ;
         String chirTypeNew = GenericRestraint.VOID_TYPE ;
         if( key.equals( "key4"   ) ) key4New = value ;
         if( key.equals( "symbol" ) ) symbolNew = value ;
         if( key.equals( "charge" ) ) chargeNew = Float.parseFloat( value ) ;
         if( key.equals( "chir"   ) ) chirTypeNew = GenericRestraint.getChirType( value ) ;
         gLib.undoableChangeAtom(
            chirTypeNew,
            finder1.link, finder1.atom, key4New, symbolNew, chargeNew
         ) ;
      }
      else if( what.equals( "link" ) && finder1.link != null ){
         String key8New =  finder1.link.key8 ;
         String nameNew =  finder1.link.name ;
         if( key.equals( "key8"   ) ) key8New  = value ;
         if( key.equals( "name"   ) ) nameNew  = value ;
         gLib.undoableRenameMolecule( finder1.link, key8New, nameNew ) ;
      }
      else throw new NullPointerException() ;
   }

   void add( String[] words ){
      String what = words[1] ;
      if( what.equals( "bond" ) ){
         if( finder1 == null || finder1.atom == null ) throw new NullPointerException() ;
         if( finder2 == null || finder2.atom == null ) throw new NullPointerException() ;
         System.out.println( finder1 ) ;
         System.out.println( finder2 ) ;
         if( finder1.link == finder2.link ){
            gLib.undoableAddBond( finder1.link, finder1.atom, finder2.atom, bondType ) ;
         }
         else{
            String key8 = finder1.link.getResidue().key8 + "-" + finder2.link.getResidue().key8 ;
            gLib.undoableAddBond( key8, finder1.link, finder1.atom, finder2.link, finder2.atom, bondType ) ;
         }
         finder2 = null ;
         finder1 = new Finder() ;
      }
      else if( what.equals( "atom" ) ){
         if( finder1 == null || finder1.atom == null ) throw new NullPointerException() ;
         System.out.println( finder1 ) ;
         gLib.undoableAddAtom( finder1.link, finder1.atom, atomType, new double[]{ 0f, 0f } ) ;
      }
      else if( what.equals( "link" ) ){
         if( words.length < 3 ) throw new NullPointerException() ;
         String key8 = words[2] ;
         System.out.println( key8 ) ;
//       boolean lcoorNo = true ;
         boolean lcoorNo = false ;
         Monomer comp = StdLib.get( key8, lcoorNo ) ;
         if( comp == null ) throw new NullPointerException() ;
         gLib.undoableAddMolecule( key8, comp ) ;
      }
      else if( what.equals( "new" ) ){
         if( words.length < 3 ) throw new NullPointerException() ;
         System.out.println( words[2] ) ;
         gLib.undoableAddMolecule( words[2], atomType ) ;
      }
      else throw new NullPointerException() ;
   }

   void regularise( String key ){
      boolean useLibmol = false ;
      gLib.undoableRegularise( gLib.get( key ), useLibmol ) ;
   }

   void convert( String key ){
      gLib.undoableConvertMolecule( gLib.get( key ) ) ;
   }

   // --------------------------------------------------------------------------

   class Finder{
      Molecule link ;
      Atom atom ;
      Bond bond ;

      public String toString(){
         if( atom != null && bond != null ) throw new NullPointerException() ;
         String s = "   Focused: " ;
         if( link == null ){
            if( bond != null || atom != null ) throw new NullPointerException() ;
            return s + "nothing" ;
         }
         s += link.key8 ;
         if( atom != null ) s += " " + atom ;
         if( bond != null ) s += " " + bond ;
         return s ;
      }

      void find( String[] words ){
         link = null ;
         atom = null ;
         bond = null ;
         if( words.length == 1 ) throw new NullPointerException() ;
         int k = 1 ;

         String key = words[k] ;
         for( Molecule mol : gLib ){
            link = mol ;
            if( link.key8.equals( key ) ) break ;
            link = null ;
         }
         if( link == null ) return ;
         if( ++k == words.length ) return ;

         String key8 = words[k] ;
         if( k == words.length ) throw new NullPointerException() ;
         key = words[++k] ;
//       System.err.println( key8 + " " + key ) ;
         for( int i = 0, n = link.atoms.size() ; i < n ; i++ ){
            atom = link.atoms.get(i) ;
//          System.err.println( atom.residue.key8 + " " + atom.key4 ) ;
            if( atom.key4.equals( key ) && atom.residue.key8.equals( key8 ) ) break ;
            atom = null ;
         }
         if( atom == null ) return ;
         if( ++k == words.length ) return ;

         key8 = words[k] ;
         if( k == words.length ) throw new NullPointerException() ;
         key = words[++k] ;
         for( int i = 0, n = atom.size() ; i < n ; i++ ){
            bond = atom.get(i) ;
            Atom atom2 = bond.another( atom ) ;
            if( atom2.key4.equals( key ) && atom2.residue.key8.equals( key8 ) ) break ;
            bond = null ;
         }
         if( bond == null ) return ;
         atom = null ;
      }

      void details(){
         if( link == null ){
            System.out.println( "   Focused: nothing" ) ;
            return ;
         }
         System.out.println( "   Focused: " + link.key8 ) ;
         System.out.println( "   Details:" ) ;
         for( Residue rsd : link.residueList() ){
            System.out.println( rsd.mono ) ;
            System.out.println( "---" ) ;
         }
      }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

class Env{
   static boolean windows = System.getProperty( "os.name" ).toLowerCase().startsWith( "windows" ) ;
   static String bsn = System.getProperty( "line.separator" ) ;
   static String version = new CharArray( bsn.getClass().getResourceAsStream( "/resources/version" ) ).toString().trim() ;
   static String versionDate = new CharArray( bsn.getClass().getResourceAsStream( "/resources/version_date" ) ).toString().trim() ;
   static String monDirName = System.getenv( "CLIBD_MON" ) ;
   static File cwd = new File( System.getProperty( "user.dir" ) ) ;

   static boolean vbScratch = false ;
   static boolean asIs = false ;
   static boolean keepTor = false ;
   static boolean enableSearch = false ;
   static boolean vbOpt = false ;
   static boolean vbLibs = false ;
   static boolean vbPrint = false ;
   static boolean vbUndo = false ;
   static boolean vbGraphics = false ;
   static boolean showLys = false ;
   static boolean talkToCoot = false ;
   static File inFile ;
   static File outFile ;

   static final int GRAPH_LIBCHECK = 1 ;
   static final int GRAPH_REFMAC = 2 ;
   static final int GRAPH_LIBCHECK_REFMAC = 3 ;
   static int coordinates ;
   static boolean vbButtons = false ;

   static void init( String[] args ){

      boolean showHelp = false ;
      boolean showVersion = false ;

      boolean noRefmac = false ;
      boolean noLibcheck = false ;

      String errMessage = null ;

      int i = 0 ;
      while( i < args.length ){
         String word = args[i++] ;
         if( word.equals( "-cwd" ) ){
            if( i < args.length ){
               String path = args[i++] ;
               try{ cwd = new File( path ).getCanonicalFile() ; }
               catch( Exception e ){
                  errMessage = path + ": cannot canonicalise" ;
               }
               if( ! cwd.isDirectory() ){
                  errMessage = cwd.getName() + ": no such directory" ;
               }
               else if( ! ( cwd.canRead() && cwd.canWrite() ) ){
                  errMessage = cwd.getName() + ": no permission for reading or writeing" ;
               }
            }
            else{
               errMessage = word + ": value required" ;
               showHelp = true ;
            }
         }
         else if( word.equals( "-h" ) ){
            showHelp = true ;
         }
         else if( word.equals( "-version" ) ){
            showVersion = true ;
         }

         else if( word.equals( "-v_scratch" ) ){
            vbScratch = true ;
         }
         else if( word.equals( "-as_is" ) ){
            asIs = true ;
         }
         else if( word.equals( "-dihedral" ) ){
            keepTor = true ;
         }
         else if( word.equals( "-search" ) ){
            enableSearch = true ;
         }
         else if( word.equals( "-v_opt" ) ){
            vbOpt = true ;
         }
         else if( word.equals( "-v_libs" ) ){
            vbLibs = true ;
         }
         else if( word.equals( "-v_print" ) ){
            vbPrint = true ;
         }
         else if( word.equals( "-v_undo" ) ){
            vbUndo = true ;
         }
         else if( word.equals( "-v_graphics" ) ){
            vbGraphics = true ;
         }
         else if( word.equals( "-lys" ) ){
            showLys = true ;
         }
         else if( word.equals( "-coot" ) ){
            talkToCoot = true ;
         }
         else if( word.equals( "-in" ) ){
            if( i < args.length ){
               String path = args[i++] ;
               try{ inFile = new File( path ).getCanonicalFile() ; }
               catch( Exception e ){
                  errMessage = path + ": cannot canonicalise" ;
               }
            }
            else{
               errMessage = word + ": value required" ;
               showHelp = true ;
            }
         }
         else if( word.equals( "-out" ) ){
            if( i < args.length ){
               String path = args[i++] ;
               try{ outFile = new File( path ).getCanonicalFile() ; }
               catch( Exception e ){
                  errMessage = path + ": cannot canonicalise" ;
               }
            }
            else{
               errMessage = word + ": value required" ;
               showHelp = true ;
            }
         }

         else if( word.equals( "-no_refmac" ) ){
            noRefmac = true ;
         }
         else if( word.equals( "-no_libcheck" ) ){
            noLibcheck = true ;
         }

         else{
            errMessage = word + ": invalid keyword" ;
            showHelp = true ;
         }
      }

      vbButtons = false ;

      if( noLibcheck ){
         if( noRefmac ){
            coordinates = GRAPH_LIBCHECK_REFMAC ;
            vbButtons = true ;
         }
         else{
            coordinates = GRAPH_REFMAC ;
            vbButtons = true ;
         }
      }
      else{
         if( noRefmac ){
            coordinates = GRAPH_LIBCHECK ;
            vbButtons = true ;
         }
         else{
            coordinates = GRAPH_LIBCHECK_REFMAC ;
         }
      }

      if( vbOpt ){
         System.out.println() ;
         System.out.println( "monDirName:  " + monDirName ) ;
         System.out.println( "cwd:         " + cwd ) ;
         System.out.println( "showHelp:    " + showHelp ) ;

         System.out.println( "vbScratch:   " + vbScratch ) ;
         System.out.println( "asIs:        " + asIs    ) ;
         System.out.println( "keepTor:     " + keepTor ) ;
         System.out.println( "enableSearch " + enableSearch ) ;
         System.out.println( "vbOpt:       " + vbOpt ) ;
         System.out.println( "vbLibs:      " + vbLibs ) ;
         System.out.println( "vbPrint:     " + vbPrint ) ;
         System.out.println( "vbUndo:      " + vbUndo ) ;
         System.out.println( "vbGraphics:  " + vbGraphics ) ;
         System.out.println( "showLys:     " + showLys ) ;
         System.out.println( "talkToCoot:  " + talkToCoot ) ;
         System.out.println( "inFile:      " + inFile ) ;
         System.out.println( "outFile:     " + outFile ) ;

         System.out.println( "coordinates: " + coordinates ) ;
         System.out.println( "vbButtons:   " + vbButtons ) ;
         System.out.println() ;
      }

      if( showVersion ){
         System.out.println( version ) ;
      }

      if( showHelp ){
         if( errMessage != null ) System.err.println( bsn + "ERROR: " + errMessage ) ;
         System.out.println() ;
         System.out.println( "Usage: java -jar jligand.jar <options>" ) ;
         System.out.println() ;
         System.out.println( "Options:" ) ;
         System.out.println( "-cwd <path>   use <path> as a working directory" ) ;
         System.out.println( "-h            show this help" ) ;
         System.out.println( "-version      show version" ) ;
         System.out.println() ;
         System.out.println( "Debuging options:" ) ;
         System.out.println( "-v_scratch    do not remove scratch file" ) ;
         System.out.println( "-dihedral     output dihedral angles for modifications and links" ) ;
         System.out.println( "-search       enable search for exact matches in standard library" ) ;
         System.out.println( "-v_opt        print options" ) ;
         System.out.println( "-v_libs       print contents of standard and local libraries" ) ;
         System.out.println( "-v_print      print various messages (and stack traces)" ) ;
         System.out.println( "-v_undo       print undo" ) ;
         System.out.println( "-v_graphics   verbose graphics" ) ;
         System.out.println( "-lys          show LYS on start" ) ;
         System.out.println( "-coot         enable exchange with coot" ) ;
         System.out.println( "-in <file>    an input test file" ) ;
         System.out.println( "-out <file>   an output test file" ) ;
         System.out.println() ;
         System.out.println( "Regularisation options:" ) ;
         System.out.println( "-as_is         enables Open As Is menu" ) ;
         System.out.println( "(to derive resraint targets directly from input coordinates)" ) ;
         System.out.println( "-no_refmac    coordinates: libcheck > graphics" ) ;
         System.out.println( "-no_libcheck  coordinates: graphics > refmac > graphics" ) ;
         System.out.println( "Neither -no_refmac nor -no_libcheck:  libcheck > refmac > graphics" ) ;
         System.out.println( "Either -no_refmac or -no_libcheck also enables buttons L, R, LR" ) ;
         System.out.println( "Both -no_refmac and -no_libcheck only enables buttons L, R, LR" ) ;
         System.out.println() ;
      }

      if( showVersion || showHelp ){
         System.exit( 0 ) ;
      }
   }
}
// -----------------------------------------------------------------------------

abstract class Versions{
   private static String bsn = System.getProperty( "line.separator" ) ;
   private static String authors = "Andrey Lebedev, Paul Young, Alexei Vagin, Garib Murshudov" ;
   private static String eMail = "andrey.lebedev@stfc.ac.uk" ;
// private static String version = new CharArray( eMail.getClass().getResourceAsStream( "/resources/version" ) ).toString().trim() ;
// private static String versionDate = new CharArray( eMail.getClass().getResourceAsStream( "/resources/version_date" ) ).toString().trim() ;
   private static String version = Env.version ;
   private static String versionDate = Env.versionDate ;

   private static Version requiredLibrary ;
   private static Version requiredLibcheck ;
   private static Version requiredRefmac ;

   private static Version currentLibrary ;
   private static Version currentLibcheck ;
   private static Version currentRefmac ;

   static void init(){
      requiredLibrary = new Version( "5.28" ) ;
      requiredLibcheck = new Version( "5.1.14" ) ;
      requiredRefmac = new Version( "5.6.0116" ) ;

      currentLibrary = StdLib.getVersion() ;
      currentLibcheck = Libcheck.getVersion() ;
      currentRefmac = Refmac.getVersion() ;
   }

   static String programInfo(){
      return "JLigand " + version ;
   }

   static String welcomeInfo(){
//    return bsn
//       + "Welcome to " + programInfo() + bsn + bsn
//       + "Note that the loading and regularising of ligands" + bsn
//       + "can take several seconds to complete, so please be patient." + bsn + bsn
//    ;
      return bsn
         + "Please wait, Jligand is reading library" + bsn + bsn
         + "Please also note that the loading and regularisation" + bsn
         + "of ligands can take several seconds to complete." + bsn + bsn ;
   }

   static String aboutInfo(){
      return bsn
         + "JLigand " + version + " - " + versionDate + bsn + bsn
         + "Recommended external components:" + bsn
         + "libcheck "   + requiredLibcheck.getVersion() + ", "
         + "refmac "     + requiredRefmac.getVersion()   + ", "
         + "dictionary " + requiredLibrary.getVersion()  + " or newer." + bsn + bsn
         + "Found:" + bsn
         + "libcheck "   + currentLibcheck.getVersion() + ", "
         + "refmac "     + currentRefmac.getVersion()   + ", "
         + "dictionary " + currentLibrary.getVersion()  + bsn + bsn
      ;
   }

   static private String versionAndDate( String what, Version current ){
      String fmt = "%-11s %-7s ( %10s )" ;
      return String.format( fmt, what, current.getVersion(), current.getDate() ) ;
   }
   static String cifHeaderInfo(){
      String fmt = "%-17s %s" ;
      return
         "# Created using " + programInfo() + bsn +
         "# and" + bsn  +
         "# " + versionAndDate( "dictionary", currentLibrary  ) + bsn +
         "# " + versionAndDate( "libcheck"  , currentLibcheck ) + bsn +
         "# " + versionAndDate( "refmac"    , currentRefmac   ) + bsn +
         "#" + bsn +
         "# Authors: " + authors + bsn +
         "# E-mail: " + eMail + bsn +
         "#" + bsn + bsn +
         "global_" + bsn +
         String.format( fmt, "_lib_name", "mon_lib" ) + bsn +
         String.format( fmt, "_lib_version", currentLibrary.getVersion() ) + bsn +
         String.format( fmt, "_lib_update", currentLibrary.getDate() ) + bsn + bsn
      ;
   }

   static private String currentAndRequired( String what, Version current, Version required ){
      String fmt = " + " ;
      if( current.isOlderThan( required ) ) fmt = " - " ;
      fmt += "%-16s  %-11s  %s" ;
      return String.format( fmt, what, current.getVersion(), required.getVersion() ) ;
   }
   static String printedInfo(){
      return
         "   component         current      required" + bsn + bsn +
         currentAndRequired( "dictionary", currentLibrary , requiredLibrary  ) + bsn +
         currentAndRequired( "libcheck"  , currentLibcheck, requiredLibcheck ) + bsn +
         currentAndRequired( "refmac"    , currentRefmac  , requiredRefmac   ) + bsn
      ;
   }

   static boolean enableSearch(){
      if( Env.enableSearch ) return true ;
      if( currentLibrary.isOlderThan( new Version("5.28") ) ) return false ;
      if( currentLibcheck.isOlderThan( new Version("5.1.14") ) ) return false ;
      return true ;
   }
}
// -----------------------------------------------------------------------------

class Version{
   private final int[] ver = new int[]{ -1, -1, -1 } ;
   private final int[] date = new int[]{ -1, -1, -1 } ;

   Version( String s ){
      int i = s.toLowerCase().indexOf( "ver" ) ;
      if( i > 0 ) s = s.substring( i ) ;
      char[] chars = s.toCharArray() ;

      for( int k = 0 ; k < chars.length ; k++ ){
         char ch = chars[k] ;
         if( ( '0' > ch || ch > '9' ) && ch != '.' && ch != '/' ) chars[k] = ' ' ;
      }

      String[] sa = new String( chars ).replaceAll( " +", " " ).trim().split( " " ) ;
      if( sa.length > 0 ) setVersion( sa[0] ) ;
      if( sa.length > 1 ) setDate( sa[1].replace( '.', '/' ) ) ;
   }

   private void setVersion( String s ){
      String[] sa = s.replace( '.', ' ' ).split( " " ) ;
      for( int i = 0, n = Math.min( 3, sa.length ) ; i < n ; i++ ) ver[i] = Integer.parseInt( sa[i] ) ;
   }

   private void setDate( String s ){
      String[] sa = s.split( "/" ) ;
      for( int i = 0 ; i < 3 ; i++ ) date[i] = Integer.parseInt( sa[i] ) ;
      if( date[2] < 2000 ) date[2] += 2000 ;
   }

   String getVersion(){
      String s = "" ;
      if( ver[0] >= 0 ) s += ver[0] ;
      if( ver[1] >= 0 ) s += "." + ver[1] ;
      if( ver[2] >= 0 ) s += "." + ver[2] ;
      return s ;
   }

   String getDate(){
      return String.format( "%02d/%02d/%02d", date[0], date[1], date[2] ) ;
   }

   boolean isOlderThan( Version v ){
      for( int i = 0 ; i < 3 ; i++ ) if( ver[i] != v.ver[i] ) return ver[i] < v.ver[i] ;
      return false ;
   }
}
// -----------------------------------------------------------------------------

abstract class StdLib{
   static final Map<String,Monomer> compMap = new LinkedHashMap<String,Monomer>() ;
   static MonLib compHeaderMap ;
   static MonLib shortHeaderMap ;
   static private String freeKey3 ;

   static void init(){
      try{
         compHeaderMap = new MonLib() ;
         shortHeaderMap = new MonLib( compHeaderMap ) ;
      }
      catch( Exception e ){
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         throw new NullPointerException( "cannot read monomer dictionary" ) ;
      }
      freeKey3 = setFreeKey3() ;
   }

   static private String setFreeKey3(){
      String s = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789" ;
      int n = s.length() ;
//    int cou = 0 ;
      for( int i = 0 ; i < n ; i++ ){
         String si = s.substring(i,i+1) ;
         for( int j = 0 ; j < n ; j++ ){
            String sj = si + s.substring(j,j+1) ;
            for( int k = 0 ; k < n ; k++ ){
               String sk = sj + s.substring(k,k+1) ;
               if( compHeaderMap.get( sk ) == null ) return sk ;
            }
         }
      }
      throw new NullPointerException( "no free names in the library" ) ;
   }

   static String getWrkKey(){ return freeKey3 ; }

   static Version getVersion(){ return compHeaderMap.getVersion() ; }

   static boolean hasReady( String key8 ){
      return compMap.get( key8 ) != null ;
   }

// for tests n Z
   static void clear(){
      compMap.clear() ;
      shortHeaderMap.clear() ;
   }

   static Monomer get( String key8, boolean lcoorNo ){

      Monomer comp = compMap.get( key8 ) ;
      if( comp != null ) return comp ;

      MonomerHeader header = compHeaderMap.get( key8 ) ;
      if( header == null ) return null ;

      header = shortHeaderMap.put( key8, header ) ;
      if( header != null ) throw new NullPointerException() ;

      File in = new Libcheck().getEntry( key8, lcoorNo ) ;

      comp = new CifLib( in ).getMonomer() ;
      if( ! ( comp != null && comp.key8.equals( key8 ) ) ) throw new NullPointerException() ;

//--
      Molecule mol = new Molecule( comp ) ;
      mol.restraints = null ;
      boolean useLibmol = false ;
      mol = mol.regularise( true, lcoorNo, useLibmol ) ;
      comp = new Monomer( mol ) ;
//--

      compMap.put( key8, comp ) ;
      return comp ;
   }

// for a test n X
   static CharArray toBuffer(){
      CharArray buffer = new CharArray() ;
      compHeaderMap.toBuffer( buffer ) ;
      return buffer ;
   }

// for a test n X
   static File getFile( String key8 ){
      return compHeaderMap.getFile( key8 ) ;
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class MonLib extends LinkedHashMap<String,MonomerHeader>{
   private final File srcDir ;
   private final File monDir ;
   private final File monList ;
   private final String listHead ;
   private final String listTail ;
   public final String CLIBD_MON ;

   MonLib() throws Exception{
      monDir = new File( Env.monDirName ).getCanonicalFile() ;
      srcDir = null ;
      CLIBD_MON = new File( monDir, " " ).toString().trim() ;

      File monListDir = new File( monDir, "list" ) ;
      monList = new File( monListDir, "mon_lib_list.cif" ) ;

      CharArray charArray = new CharArray( monList ) ;

      CifFile file = new CifFile( charArray ) ;
      CifData data = file.get( "data_comp_list" ) ;
      CifLoop loop = data.iterator().next() ;
      while( loop.hasNext() ){
         MonomerHeader comp = new MonomerHeader( loop ) ;
         if( super.put( comp.key8, comp ) != null ) throw new NullPointerException() ;
      }

      listHead = charArray.subSequence( 0, data.first ).toString() ;

      CharArray tailArray = new CharArray() ;
      for( Map.Entry<String,CifData> me : file.entrySet() ){
         data = me.getValue() ;
         if( me.getKey().startsWith( "data_mod_" ) ){
            tailArray.append( charArray.subSequence( data.first, data.last ) ) ;
         }
      }
      listTail = tailArray.toString() ;
   }

   MonLib( MonLib monLib ) throws Exception{
      monDir = Scratch.getWrkDir() ;
      srcDir = monLib.monDir ;
      CLIBD_MON = new File( monDir, " " ).toString().trim() ;

      File monListDir = new File( monDir, "list" ) ;
      monListDir.mkdir() ;
      monList = new File( monListDir, "mon_lib_list.cif" ) ;

      listHead = monLib.listHead.toString() ;
      listTail = monLib.listTail.toString() ;

      InputStream is = getClass().getResourceAsStream( "resources/mon_lib_ind.cif" ) ;
      new CharArray( is ).writeTo( new File( monDir, "mon_lib_ind.cif" ) ) ;

      is = getClass().getResourceAsStream( "resources/mon_lib_ind2.cif" ) ;
      new CharArray( is ).writeTo( new File( monDir, "mon_lib_ind2.cif" ) ) ;

      File f = new File( srcDir, "ener_lib.cif" ) ;
      new CharArray( f ).writeTo( new File( monDir, "ener_lib.cif" ) ) ;

      writeMonList() ;
   }

   Version getVersion(){
      return new Version( listHead ) ;
   }

   void toBuffer( CharArray buffer ){
      CifData data = CifFile.template.get( "data_comp_list" ) ;
      CifLoop loop = data.iterator().next() ;
      for( MonomerHeader comp : values() ) comp.toCif( loop ) ;

      buffer.append( "data_comp_list" ) ;
      data.appendTo( buffer ) ;
      data.clear() ;
      buffer.append( Env.bsn ) ;
   }

   private void writeMonList(){
      CharArray buffer = new CharArray( listHead ) ;
      toBuffer( buffer ) ;
      buffer.append( listTail ) ;
      buffer.writeTo( monList ) ;
   }

// for tests in Z:
   public void clear(){
      if( srcDir == null ) throw new NullPointerException() ;
      super.clear() ;
      try{
         writeMonList() ;
      }
      catch( Exception e ){
         throw new NullPointerException() ;
      }
   }

// for a test in X:
   File getFile( String key8 ){
      return new File( new File( monDir, key8.substring( 0, 1 ).toLowerCase() ), key8 + ".cif" ) ;
   }

   private void copyFile( String key8 ) throws Exception{
      String monChar = key8.substring( 0, 1 ).toLowerCase() ;
      File srcFile = new File( new File( srcDir, monChar ), key8 + ".cif" ) ;
      File copyDir = new File( monDir, monChar ) ;
      if( ! copyDir.exists() ) copyDir.mkdir() ;
      File copyFile = new File( copyDir, key8 + ".cif" ) ;
      new CharArray( srcFile ).writeTo( copyFile ) ;
   }

   public MonomerHeader put( String key8, MonomerHeader header ){
      if( srcDir == null ) throw new NullPointerException() ;
      header = super.put( key8, header ) ;
      try{
         writeMonList() ;
         copyFile( key8 ) ;
      }
      catch( Exception e ){
         super.remove( key8 ) ;
         throw new NullPointerException() ;
      }
      return header ;
   }
}
// -----------------------------------------------------------------------------

class MonomerHeader{
   String  key3 ;
   String  key8 ;
   String  name ;
   String  group ;
   Integer na ;
   Integer nh ;
   String  level ;

   MonomerHeader( CifLoop header ){
      key8  = (String)  header.next() ;
      key3  = (String)  header.next() ;
      name  = (String)  header.next() ;
      group = (String)  header.next() ;
      na    = (Integer) header.next() ;
      nh    = (Integer) header.next() ;
      level = (String)  header.next() ;
   }

   void toCif( CifLoop header ){
      header.add( key8  ) ;
      header.add( key3  ) ;
      header.add( name  ) ;
      header.add( group ) ;
      header.add( na    ) ;
      header.add( nh    ) ;
      header.add( level ) ;
   }
}
// -----------------------------------------------------------------------------

class Libmol{
   private static boolean available = false ;

// static private String prgName = new File( Env.cwd, "libmol" ).toString() ;
   static private String prgName = "libmol" ;
   static private String comName = "com" ;
   static private String logName = "log" ;
   static private String inpName = "inp.lib" ;
   static private String outName = "out.lib" ;
   static private String pdbName = "out.pdb" ;

   static private String cmd ;
   static private OutputStreamWriter osw ;
   static private BufferedReader br ;
   static private String fmt ;

   public static void init(){
      cmd = Env.windows ? prgName + ".exe" : prgName ;

      available = false ;
      try{
         Process p = new ProcessBuilder( cmd, "-jligand" ).start() ;
         osw = new OutputStreamWriter( p.getOutputStream() ) ;
         br = new BufferedReader( new InputStreamReader( p.getInputStream() ) ) ;
      }
      catch( Exception ioe ){
         return ;
      }
      available = true ;

      fmt = Env.bsn + "RUN_BEGIN" ;
      fmt += Env.bsn + "LIBIN %s" ;
      fmt += Env.bsn + "MON %s" ;
      fmt += Env.bsn + "LIBOUT %s" ;
      fmt += Env.bsn + "XYZOUT %s" ;
      fmt += Env.bsn + "RUN_END" + Env.bsn ;
   }

   public static boolean isAvailable(){
      return available;
   }

   public static File execute( CharArray buffer, String mon ){
      File wrkDir = Scratch.getWrkDir() ;
      File inpFile = new File( wrkDir, inpName ) ;
      File outFile = new File( wrkDir, outName ) ;
      File pdbFile = new File( wrkDir, pdbName ) ;
      try{
         buffer.writeTo( inpFile ) ;
         osw.write( String.format( fmt, inpFile.getPath(), mon, outFile.getPath(), pdbFile.getPath() ) ) ;
         osw.flush() ;

         FileWriter fw = new FileWriter( new File( wrkDir, logName ) ) ;
         String s = "" ;
         while( ! s.equals( "RUN_END" ) ) fw.write( ( s = br.readLine() ) + Env.bsn ) ;
         fw.close() ;

         fw = new FileWriter( new File( wrkDir, comName ) ) ;
         fw.write( "#!/bin/tcsh -f" + Env.bsn + Env.bsn ) ;
         fw.write( cmd + " >! " + logName + " <<+" + Env.bsn ) ;
         fw.write( String.format( fmt, inpName, mon, outName, pdbName ) ) ;
         fw.write( "END" + Env.bsn ) ;
         fw.write( "+" + Env.bsn + Env.bsn ) ;
         fw.close() ;

         if( ! outFile.isFile() ) throw new NullPointerException() ;
         return outFile ;
      }
      catch( Exception e ){
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         throw new NullPointerException() ;
      }
   }

   public static void finish(){
      try{
         osw.write( Env.bsn + "END" + Env.bsn ) ;
         osw.close() ;
         br.close() ;
      }
      catch( Exception e ){
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         throw new NullPointerException() ;
      }
   }
}
// -----------------------------------------------------------------------------

class Libcheck{
   private static final String libout = "libcheck.lib" ;

   private static final String comName = "com" ;
   private static final String logName = "libcheck.log" ;

   static Version getVersion(){
      File wrkdir = Scratch.getWrkDir() ;
      ProcessWrapper pw = new ProcessWrapper( "libcheck" ) ;

      pw.args( "-i" ) ;

      try{
         pw.execute( wrkdir, comName, logName ) ;
         return new Version( new CharArray( new File( wrkdir, logName ) ).toString() ) ;
      }
      catch( Exception e ){
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         throw new NullPointerException( "cannot run libcheck" ) ;
      }
   }

   Monomer regularise( Monomer comp, boolean lcoorNo, boolean useLibmol ){
      CifLib cifLib = new CifLib() ;
      cifLib.add( comp ) ;

      File file ;
      if( useLibmol ){
         file = Libmol.execute( cifLib.toBuffer(), comp.key8 ) ;
      }
      else{
         boolean coorNo = true ;
         file = runLibcheckOnly( cifLib.toBuffer(), "L", comp.key8, lcoorNo, coorNo ) ;
      }
      cifLib = new CifLib( file ) ;

      return cifLib.getMonomer() ;
   }

   File runLibcheckOnly( CharArray buffer, String type, String key8, boolean lcoorNo, boolean coorNo ){
      File wrkdir = Scratch.getWrkDir() ;
      ProcessWrapper pw = new ProcessWrapper( "libcheck" ) ;

      String inputName = "input." + ( type.equals( "L" ) ? "lib" : type.toLowerCase() ) ;
      buffer.writeTo( new File( wrkdir, inputName ) ) ;

      pw.keywords( "N" ) ;
      if( ! type.equals( "PDB" ) ) pw.keywords( "NODIST", "Y" ) ;
      pw.keywords( "REF", "0" ) ;
      if( ! coorNo ) pw.keywords( "COOR", "Y" ) ;
      if( lcoorNo ) pw.keywords( "LCOOR", "N" ) ;
      pw.keywords( "FILE_" + type.toUpperCase(), inputName ) ;
      if( key8 != null ) pw.keywords( "MON", key8 ) ;
      pw.keywords( "" ) ;

      pw.setenv( "CLIBD_MON", StdLib.shortHeaderMap.CLIBD_MON ) ;

      String fmt = "   Libcheck: %-4s %-11s ...  %-3s  ...  " ;
      pw.setInfo( fmt, wrkdir.getName(), inputName, key8 == null ? "" : key8 ) ;

      pw.execute( wrkdir, comName, logName ) ;
      cleanWrkdir( wrkdir ) ;

      File lib = new File( wrkdir, libout ) ;
      if( ! lib.isFile() ) throw new NullPointerException() ;
      return lib ;
   }

   File getEntry( String key8, boolean lcoorNo ){
      File wrkdir = Scratch.getWrkDir() ;
      ProcessWrapper pw = new ProcessWrapper( "libcheck" ) ;

      pw.keywords( "N" ) ;
      pw.keywords( "REF", "0" ) ;
      if( lcoorNo ) pw.keywords( "LCOOR", "N" ) ;
      pw.keywords( "MON", key8 ) ;
      pw.keywords( "" ) ;

      pw.setenv( "CLIBD_MON", StdLib.shortHeaderMap.CLIBD_MON ) ;
      pw.setInfo( "   Libcheck: %-4s %-11s ...  %-3s  ...  ", wrkdir.getName(), "STD_LIB", key8 ) ;
      pw.execute( wrkdir, comName, logName ) ;
      cleanWrkdir( wrkdir ) ;

      File lib = new File( wrkdir, libout ) ;
      if( ! lib.isFile() ) throw new NullPointerException() ;
      return lib ;
   }

   List<String> getMatches( Molecule molecule ){
      File wrkdir = Scratch.getWrkDir() ;
      ProcessWrapper pw = new ProcessWrapper( "libcheck" ) ;

      Monomer comp = new Monomer( molecule ) ;
      String wrkKey = StdLib.getWrkKey() ;
      comp.key3 = wrkKey ;
      comp.key8 = wrkKey ;
      comp.name = wrkKey ;

      String inputName = "input.pdb" ;
      new PdbWriter().write( comp, new File( wrkdir, inputName ), true ) ;

      pw.keywords( "N" ) ;
      pw.keywords( "FILE_PDB", inputName ) ;
      pw.keywords( "SRCH", "Y" ) ;
      pw.keywords( "MON", wrkKey ) ;
      pw.keywords( "" ) ;

      pw.setenv( "CLIBD_MON", StdLib.compHeaderMap.CLIBD_MON ) ;
      pw.setInfo( "   Libcheck: %-4s %-11s ...  %-3s  ...  ", wrkdir.getName(), inputName, wrkKey ) ;
      pw.execute( wrkdir, comName, logName ) ;
      cleanWrkdir( wrkdir ) ;

      List<String> list = new ArrayList<String>() ;
      try{
         BufferedReader reader = new BufferedReader( new FileReader( new File( wrkdir, logName ) ) ) ;
         String line = reader.readLine() ;
         while( line != null ){
            if( line.startsWith( "  WARNING : monomer looks like" ) ){
               list.add( line.substring(30,39).trim() ) ;
            }
            line = reader.readLine() ;
         }
         reader.close() ;
      }
      catch( IOException ioe ){ throw new NullPointerException() ; }
      return list ;
   }

   void cleanWrkdir( File wrkdir ){
      List<File> others = Arrays.asList( wrkdir.listFiles( new FilenameFilter(){
         public boolean accept( File dir, String name ){
            boolean selected = name.startsWith( "libcheck_" ) && ! name.endsWith(".pdb" ) ;
            selected |= name.equals( "libcheck.odb" ) ;
            return selected ;
         }
      } ) ) ;
      for( File file : others ) file.delete() ;
   }
}
// -----------------------------------------------------------------------------

class Refmac{
   private static final String xyzin = "input.pdb" ;
   private static final String libin = "input.lib" ;
   private static final String xyzout = "output.pdb" ;

   private static final String comName = "refmac.com" ;
   private static final String logName = "refmac.log" ;

   static Version getVersion(){
      File wrkdir = Scratch.getWrkDir() ;
      ProcessWrapper pw = new ProcessWrapper( "refmac5" ) ;

      pw.keywords( "end" ) ;

      try{
         pw.execute( wrkdir, comName, logName ) ;
         return new Version( new CharArray( new File( wrkdir, logName ) ).toString() ) ;
      }
      catch( Exception e ){
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         throw new NullPointerException( "cannot run refmac" ) ;
      }
   }

   void refine( Monomer comp ){
      File wrkdir = Scratch.getWrkDir() ;
      ProcessWrapper pw = new ProcessWrapper( "refmac5" ) ;

      pw.args( "libin", libin, "xyzin", xyzin, "xyzout", xyzout ) ;
      pw.keywords( "make", "hout", "yes" ) ;
      pw.keywords( "mode", "newentry" ) ;
      pw.keywords( "end" ) ;

      CifLib cifLib = new CifLib() ;
      cifLib.add( comp ) ;
      cifLib.toBuffer().writeTo( new File( wrkdir, libin ) ) ;

      new PdbWriter().write( comp, new File( wrkdir, xyzin ), true ) ;

      pw.setenv( "CLIBD_MON", StdLib.shortHeaderMap.CLIBD_MON ) ;
      pw.setInfo( "   Refmac:   %-4s %-11s ...  %-3s  ...  ", wrkdir.getName(), libin, comp.key8 ) ;
      pw.execute( wrkdir, comName, logName ) ;

      new PdbReader().read( comp, new File( wrkdir, xyzout ) ) ;
   }
}
// -----------------------------------------------------------------------------

class ProcessWrapper{
   private ProcessBuilder pb ;
   private StringBuilder sb = new StringBuilder() ;
   private String info ;

   ProcessWrapper( String cmd ){
      pb = new ProcessBuilder( Env.windows ? cmd + ".exe" : cmd ) ;
   }

   void args( String ... words ){
      List<String> list = pb.command() ;
      for( String word : Arrays.asList( words ) ) list.add( word ) ;
   }

   void keywords( String ... words ){
      for( String word : Arrays.asList( words ) ) sb.append( word + " " ) ;
      int n = sb.length() ;
      sb.replace( n - 1, n, Env.bsn ) ;
   }

   void setenv( String key, String value ){
      if( Env.vbOpt ) System.err.println( "ProcessWrapper.setenv: " + key + " = " + value ) ;
      pb.environment().put( key, value ) ;
   }

   void setInfo( String fmt, Object ... s ){
      info = String.format( fmt, s ) ;
   }

   void execute( File wrkdir, String comName, String logName ){
      boolean printInfo = Env.vbPrint && info != null ;
      try{
         if( printInfo ) System.err.print( info ) ;

         String kwd = sb.toString() ;

         FileWriter fw = new FileWriter( new File( wrkdir, comName ) ) ;
         fw.write( "#!/bin/tcsh -f" + Env.bsn + Env.bsn ) ;
         for( String s : pb.command() ) fw.write( s + " " ) ;
         fw.write( ">! " + logName + " <<+" + Env.bsn ) ;
         fw.write( kwd ) ;
         fw.write( "+" + Env.bsn + Env.bsn ) ;
         fw.close() ;

         Process p = pb.directory( wrkdir ).start() ;

         OutputStreamWriter osw = new OutputStreamWriter( p.getOutputStream() ) ;
         osw.write( kwd ) ;
         osw.close() ;

         BufferedReader br = new BufferedReader( new InputStreamReader( p.getInputStream() ) ) ;
         fw = new FileWriter( new File( wrkdir, logName ) ) ;
         String s = null ;
         while( ( s = br.readLine() ) != null ) fw.write( s + Env.bsn ) ;
         br.close() ;
         fw.close() ;

         if( printInfo ) System.err.println( "done" ) ;
         if( printInfo ) System.out.println( info + "done" ) ;
      }
      catch( Exception e ){
         if( Env.vbPrint ) e.printStackTrace( System.err ) ;
         throw new NullPointerException() ;
      }
   }
}
// -----------------------------------------------------------------------------

class PdbReader{
   void read( Monomer mono, File file ){
      Map<String,Atom> map = new LinkedHashMap<String,Atom>( mono.atoms.size() ) ;
      for( Atom atom : mono.atoms ){
         if( map.put( atom.key4, atom ) != null ) throw new NullPointerException() ;
         if( atom.residue != null ) throw new NullPointerException() ;
      }
      try{
         BufferedReader reader = new BufferedReader( new FileReader( file ) ) ;
         String resId = null ;
         String line = reader.readLine() ;
         while( line != null && ! ( line.startsWith( "ATOM " ) || line.startsWith( "HETATM " ) ) ){
            if( line.startsWith( "MODRES" ) ){
               if( resId != null ) throw new NullPointerException() ;
               if( ! line.substring( 12, 15 ).trim().equals( mono.key3 ) ) throw new NullPointerException() ;
               if( ! line.substring( 24, 32 ).trim().equals( mono.key8 ) ) throw new NullPointerException() ;
               if( mono.key3.equals( mono.key8 ) ) throw new NullPointerException() ;
               resId = line.substring( 12, 17 ) + line.substring( 18, 22 ) ;
            }
            line = reader.readLine() ;
         }
         if( resId == null ){
            if( ! line.substring( 17, 20 ).trim().equals( mono.key3 ) ) throw new NullPointerException() ;
            if( ! mono.key3.equals( mono.key8 ) ) throw new NullPointerException() ;
            resId = line.substring( 17, 26 ) ;
         }
         while( line != null && ( line.startsWith( "ATOM " ) || line.startsWith( "HETATM " ) ) ){
            if( ! line.substring( 17, 26 ).equals( resId ) ) throw new NullPointerException() ;

            Atom atom = map.remove( line.substring( 12, 16 ).trim() ) ;
            if( atom == null ){
               String key = line.substring( 13, 16 ) + line.substring( 12, 13 ) ;
               if( ! key.startsWith( "H" ) ) throw new NullPointerException() ;
               atom = map.remove( key ) ;
            }
            if( atom == null ) throw new NullPointerException() ;

            String s = line.substring( 76 ).trim() ;
            s = s.substring( 0, 1 ) + s.substring( 1 ).toLowerCase() ;
            if( ! atom.symbol.equals( s ) ) throw new NullPointerException() ;

            atom.xyz[0] = Double.parseDouble( line.substring( 30, 38 ) ) ;
            atom.xyz[1] = Double.parseDouble( line.substring( 39, 46 ) ) ;
            atom.xyz[2] = Double.parseDouble( line.substring( 47, 54 ) ) ;
            line = reader.readLine() ;
         }
         reader.close() ;
      }
      catch( IOException ioe ){ throw new NullPointerException() ; }
      if( ! map.isEmpty() ) throw new NullPointerException() ;
   }
}
// -----------------------------------------------------------------------------

class PdbWriter{
   String bsn = Env.bsn ;
   String atomFormat = "ATOM  %5d %-4s %-3s A   1    %8.3f%8.3f%8.3f  1.00 20.00 %11s" + bsn ;
   String crystFormat = "CRYST1  100.000  100.000  100.000  90.00  90.00  90.00 P 1" + bsn ;
   String modresFormat = "MODRES      %-3s A    1  %-8s                                        RENAME" + bsn ;
   void write( Monomer mono, File file, boolean includeH ){
      try{
         Writer writer = new FileWriter( file ) ;
         if( ! mono.key3.equals( mono.key8 ) ) writer.write( String.format( modresFormat, mono.key3, mono.key8 ) ) ;
         writer.write( crystFormat ) ;
         int cou = 0 ;
         for( Atom atom : mono.atoms ){
            if( includeH || ! atom.symbol.equals( "H" ) ){
               String symbol = atom.symbol.toUpperCase() ;
               String key4 = atom.key4 ;
               if( ! ( symbol.length() == 2 || key4.length() == 4 ) ) key4 = " " + key4 ;
               writer.write( String.format( atomFormat,
                  ++cou , key4, mono.key3, atom.xyz[0], atom.xyz[1], atom.xyz[2], symbol
               ) ) ;
            }
         }
         writer.close() ;
      }
      catch( Exception e ){ throw new NullPointerException() ; }
   }
}
// -----------------------------------------------------------------------------

abstract class Scratch{
   static private int cou = 0 ;
   static private ScratchFile scratchDir ;
   static private boolean initialised = false ;

   static void init(){
      if( initialised ) throw new NullPointerException() ;
      initialised = true ;
      scratchDir = new ScratchFile() ;
   }

   static File getWrkDir(){
      File wrkdir = new File( scratchDir, String.format( "%03d", cou++ ) ) ;
      if( wrkdir.exists() ) throw new NullPointerException() ;
      if( ! wrkdir.mkdir() ) throw new NullPointerException() ;
      return wrkdir ;
   }

   static void clear(){
      scratchDir.rmMinusR() ;
   }

   // --------------------------------------------------------------------------

   @SuppressWarnings("serial")
   static private class ScratchFile extends File{

      private ScratchFile(){
//       super( Env.cwd, "scratch" ) ;
         super( Env.cwd, String.format( "scratch:%tT", Calendar.getInstance() ).replace( ':', '_' ) ) ;
         if( exists() ) rmMinusR() ;
         if( exists() ) throw new NullPointerException() ;
         if( ! mkdir() ) throw new NullPointerException() ;
      }

      private ScratchFile( ScratchFile scr, String s ){
         super( scr, s ) ;
      }

      private void rmMinusR(){
         if( isDirectory() ){
            String[] names = list() ;
            for( int i = 0 ; i < names.length ; i++ ) new ScratchFile( this, names[i] ).rmMinusR() ;
         }
         else if( ! isFile() ) throw new NullPointerException() ;
         delete() ;
      }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

//class Molecule extends ArrayList<Atom>{
//class Molecule implements Iterable<Atom>{
class Molecule{
   List<Atom> atoms = new ArrayList<Atom>() ;
   //:: to avoid corresponding methods from super:
   private static int counter = 0 ;
   private final int cou = counter++ ;
   public int hashCode(){ return cou ; }
   public boolean equals( Object o ){ return compareTo( (Molecule)o ) == 0 ; }
   public int compareTo( Molecule mol ){ return cou - mol.cou ; }

   String key8 ;
   String name ;
   Restraints restraints ;

   static double minRadius = 0.5f ;
   double radius ;
   double[] xyz = new double[]{ 0f, 0f, 0f } ;

   public Iterator<Atom> iterator(){
      return atoms.iterator() ;
   }

   void updateSphere(){

      this.xyz = new double[]{ 0f, 0f, 0f } ;
      for( Atom atom : atoms ){
         for( int i = 0 ; i < 3 ; i++ ) this.xyz[i] += atom.xyz[i] ;
      }
      for( int i = 0 ; i < 3 ; i++ ) this.xyz[i] = this.xyz[i]/ atoms.size() ;

//    double rro = this.minRadius* this.minRadius ;
      double rro = minRadius* minRadius ;
      for( Atom atom : atoms ){
         double rrn = 0f ;
         for( int i = 0 ; i < 3 ; i++ ){ double rn = atom.xyz[i] - this.xyz[i] ; rrn += rn* rn ; }
         if( rro < rrn ) rro = rrn ;
      }
      this.radius = Math.sqrt( rro ) ;

//    if( Env.vbPrint ) System.out.println( this.key8 + " " + this.radius ) ;
   }

   void saveBaseline( double[] base ){
      base[0] = xyz[0] ;
      base[1] = xyz[1] + radius ;
      base[2] = xyz[2] ;
   }

   void restoreBaseline( double[] base ){
      double[] shift = new double[3] ;
      shift[0] = base[0] - xyz[0] ;
      shift[1] = base[1] - xyz[1] - radius ;
      shift[2] = base[2] - xyz[2] ;
      xyz[0] = base[0] ;
      xyz[1] = base[1] - radius ;
      xyz[2] = base[2] ;
      for( Atom atom : atoms ){
         for( int i = 0 ; i < 3 ; i++ ) atom.xyz[i] += shift[i] ;
      }
   }

   Molecule(){}

   Molecule( Molecule mol, boolean duplicateResidues ){
      Collections.sort( mol.atoms, Atom.originComparator ) ;

      Map<Residue,Residue> map = null ;
      if( duplicateResidues ){
         map = new LinkedHashMap<Residue,Residue>() ;
         for( Residue rsd : mol.residueList() ) map.put( rsd, rsd.duplicate() ) ;
      }

      this.key8 = mol.key8 ;
      this.name = mol.name ;
      this.xyz = new double[]{ mol.xyz[0], mol.xyz[1], mol.xyz[2] } ;
      this.radius = mol.radius ;
      this.restraints = null ;

      for( Atom orig : mol.atoms ){
         Residue rsd = orig.residue ;

         if( duplicateResidues ) rsd = map.get( rsd ) ;

         Atom atom = new Atom( rsd, orig ) ;
         orig.coupled = atom ;
         atoms.add( atom ) ;
      }
      for( Atom orig1 : mol.atoms ){
         for( Bond bondOrig : orig1 ){
            Atom orig2 = bondOrig.another( orig1 ) ;
            if( Atom.timeComparator.compare( orig1, orig2 ) < 0 ){
               Atom atom1 = orig1.coupled ;
               Atom atom2 = orig2.coupled ;
               Bond bond = new Bond( atom1, atom2, bondOrig ) ;
               atom1.add( bond ) ;
               atom2.add( bond ) ;
            }
         }
      }
      for( Atom orig : mol.atoms ){
         if( orig.chir != null ){
            GenericRestraint chir = orig.chir.duplicate() ;
            chir.atoms.get(0).chir = chir ;
         }
      }
//==  for( Atom orig : mol.atoms ) orig.coupled = orig ;
   }

   Molecule( String key ){
      name = key ;
      key8 = key ;
   }

   Molecule( String key, String atomType ){
      name = key ;
      key8 = key ;
      Residue rsd = new Residue() ;
      rsd.name = key ;
      rsd.key8 = key ;
      atoms.add( new Atom( rsd, atomType, atomType + "1", 0f, 0f, 0f ) ) ;
   }

   Molecule( Monomer comp ){
      if( comp.restraints == null ) throw new NullPointerException() ;
      name = comp.name ;
      key8 = comp.key8 ;
      Residue rsd = new Residue( comp ) ;

      for( Atom orig : comp.atoms ){
         Atom atom = new Atom( rsd, orig ) ;
         atoms.add( atom ) ;
         orig.coupled = atom ;
      }
      restraints = new Restraints( comp.restraints ) ;
      for( GenericRestraint rst : restraints.get(0) ){
         Bond bond = new Bond( rst ) ;
         bond.atom1.add( bond ) ;
         bond.atom2.add( bond ) ;
      }
      for( GenericRestraint chir : restraints.get(3) ){
         chir.atoms.get(0).chir = chir ;
      }
      for( Atom orig : comp.atoms ) orig.coupled = orig ;
   }

   Residue getResidue(){
      Residue rsd = atoms.get( 0 ).residue ;
      for( Atom atom : atoms ) if( atom.residue != rsd ) return null ;
      return rsd ;
   }

   List<Residue> residueList(){
      Set<Residue> set = new LinkedHashSet<Residue>() ;
      for( Atom atom : atoms ) set.add( atom.residue ) ;
      List<Residue> list = new ArrayList<Residue>( set ) ;
      Collections.sort( list ) ;
      return list ;
   }

/*
   String linkRecordPDB(){
      List<Atom> list = new ArrayList<Atom>() ;
      for( GenericRestraint rst : restraints.get(0) ){
         if( rst.atoms.size() != 2 ) throw new NullPointerException() ;
         Atom atom0 = rst.atoms.get(0) ;
         Atom atom1 = rst.atoms.get(1) ;
         Residue rsd0 = atom0.residue ;
         Residue rsd1 = atom1.residue ;
         if( rsd0 != rsd1 ){
            if( rsd0.isFirst ){
               if( rsd1.isFirst ) throw new NullPointerException() ;
               list.add( atom0 ) ;
               list.add( atom1 ) ;
            }
            else{
               if( rsd0.isFirst ) throw new NullPointerException() ;
               list.add( atom1 ) ;
               list.add( atom0 ) ;
            }
         }
      }
      if( list.size() != 2 ) throw new NullPointerException() ;
      Atom at0 = list.get(0) ;
      Atom at1 = list.get(1) ;
      String monoKey0 = at0.residue.mono.key3 ;
      String monoKey1 = at1.residue.mono.key3 ;
      String atomKey0 = at0.symbol.length() > 1 || at0.key4.length() > 3 ? at0.key4 : " " + at0.key4 ;
      String atomKey1 = at1.symbol.length() > 1 || at1.key4.length() > 3 ? at1.key4 : " " + at1.key4 ;
      String resiKey0 = "A 111" ;
      String resiKey1 = "B 222" ;
      if( at0.residue.cootId != at1.residue.cootId ) throw new NullPointerException() ;
      if( at0.residue.cootId > 0 ){
         resiKey0 = at0.residue.residueId ;
         resiKey1 = at1.residue.residueId ;
      }

//    String fmt = "LINK         NZ  LYS B 258                 C4A PLP D   1                LYS-PLP"
//    String fmt = "LINK        %-4s %-3s A 111                %-4s %-3s B 222                %s" ;
      String fmt = "LINK        %-4s %-3s %s5                %-4s %-3s %s5                %s" ;
      return String.format( fmt, atomKey0, monoKey0, resiKey0, atomKey1, monoKey1, resiKey1, key8 ) ;
   }
*/

   String linkRecordPDB(){
      List<Atom> list = new ArrayList<Atom>() ;
      for( Atom atom0 : atoms ){
         for( Bond bond : atom0 ){
            Atom atom1 = bond.another( atom0 ) ;
            Residue rsd0 = atom0.residue ;
            Residue rsd1 = atom1.residue ;
            if( rsd0.compareTo( rsd1 ) < 0 ){
               if( rsd0.isFirst == rsd1.isFirst ) throw new NullPointerException() ;
               list.add( atom0 ) ;
               list.add( atom1 ) ;
            }
         }
      }
      if( list.size() != 2 ) throw new NullPointerException() ;
      Atom at0 = list.get(0) ;
      Atom at1 = list.get(1) ;
      String monoKey0 = at0.residue.mono.key3 ;
      String monoKey1 = at1.residue.mono.key3 ;
      String atomKey0 = at0.symbol.length() > 1 || at0.key4.length() > 3 ? at0.key4 : " " + at0.key4 ;
      String atomKey1 = at1.symbol.length() > 1 || at1.key4.length() > 3 ? at1.key4 : " " + at1.key4 ;
      String resiKey0 = "A 111" ;
      String resiKey1 = "B 222" ;
      if( at0.residue.cootId != at1.residue.cootId ) throw new NullPointerException() ;
      if( at0.residue.cootId > 0 ){
         resiKey0 = at0.residue.residueId ;
         resiKey1 = at1.residue.residueId ;
      }

//    String fmt = "LINK         NZ  LYS B 258                 C4A PLP D   1                LYS-PLP"
//    String fmt = "LINK        %-4s %-3s A 111                %-4s %-3s B 222                %s" ;
      String fmt = "LINK        %-4s %-3s %5s                %-4s %-3s %5s                %s" ;
      return String.format( fmt, atomKey0, monoKey0, resiKey0, atomKey1, monoKey1, resiKey1, key8 ) ;
   }

   Map<String,Atom> atomView( Atom atom ){
      Map<String,Atom> map = new LinkedHashMap<String,Atom>() ;
      for( Atom a : atoms ){
         if( a.residue == atom.residue && map.put( a.key4, a ) != null ) throw new NullPointerException() ;
      }
      return map ;
   }

   Map<Residue,Molecule> residueView(){
      Map<Residue,Molecule> map = new TreeMap<Residue,Molecule>() ;
      for( Atom atom : atoms ){
         Residue residue = atom.residue ;
         Molecule molecule = map.get( residue ) ;
         if( molecule == null ){
            molecule = new Molecule() ;
            molecule.key8 = residue.key8 ;
            molecule.name = residue.name ;
            if( restraints != null ){
               molecule.restraints = new Restraints() ;
               for( int i = 0 ; i < 5 ; i++ ){
                  List<GenericRestraint> listAll = this.restraints.get( i ) ;
                  List<GenericRestraint> list = molecule.restraints.get( i ) ;
                  for( GenericRestraint rst : listAll ) if( rst.getResidue() == residue ) list.add( rst ) ;
                  Collections.sort( list ) ;
               }
            }
            map.put( residue, molecule ) ;
         }
         molecule.atoms.add( atom ) ;
      }
      return map ;
   }

   void restoreHydrogenKeys(){
      Set<String> set = new LinkedHashSet<String>() ;
      for( Atom atom : atoms ){
         if( atom.key4 != null && ! set.add( atom.key4 ) ) throw new NullPointerException() ;
      }
      for( Atom atom : atoms ){
         if( atom.key4 == null ){
            if( ! ( atom.symbol.equals( "H" ) && atom.size() == 1 ) ) throw new NullPointerException() ;
            Atom another = atom.get( 0 ).another( atom ) ;

            atom.key4 = "H" ;
            if( another.symbol.equals( "C" ) && another.key4.toUpperCase().startsWith( "C" ) ){
               atom.key4 += another.key4.substring( 1 ) ;
            }
            else{
               atom.key4 += another.key4 ;
            }
            if( atom.key4.length() > 3 ){
               atom.key4 = atom.key4.substring( 0, 4 ) ;
            }
            atom.resetKey( set ) ;
         }
      }
   }

   private List<Atom> hiddenAtoms = new ArrayList<Atom>() ;

   void excludeHidden(){
      List<Atom> shownAtoms = new ArrayList<Atom>() ;
      for( Atom atom : atoms ){
         if( atom.isHidden() ){
            hiddenAtoms.add( atom ) ;
            for( Bond bond : atom ) bond.another( atom ).remove( bond ) ;
         }
         else{
            shownAtoms.add( atom ) ;
         }
      }
      atoms.clear() ;
      atoms.addAll( shownAtoms ) ;
   }

   void includeHidden(){
      for( int i = hiddenAtoms.size() ; i-- > 0 ; ){
         Atom atom = hiddenAtoms.get( i ) ;
         for( Bond bond : atom ) bond.another( atom ).add( bond ) ;
      }
      atoms.addAll( hiddenAtoms ) ;
   }

   // The last two methods do not affect chirality as these only relate to H-atoms.

   // --------------------------------------------------------------------------

   Molecule regularise( boolean doRegularise, boolean lcoorNo, boolean useLibmol ){

      if( doRegularise && this.restraints != null ) throw new NullPointerException() ;

      if( useLibmol ) for( Atom atom : atoms ) atom.resetHidden() ;

      double[] base = new double[3] ;
      this.updateSphere() ;
      this.saveBaseline( base ) ;

      Molecule copy = new Molecule( this, doRegularise ) ;
      if( ! useLibmol ) for( Atom atom : copy.atoms ) atom.resetCharge() ;

//1
      copy.excludeHidden() ;
//1

      boolean changed = CorrectBonds.correctBondOrders( copy.atoms ) ;
      if( Env.vbPrint ) System.out.println( "   ----- changed " + changed ) ;

      if( ! doRegularise ){
//1
         copy.includeHidden() ;
//1
         return changed ? copy : null ;
      }
//2
//2   copy.excludeHidden() ;
//2

      for( Atom orig : atoms ) orig.coupled = orig ;

      Map<String,Atom> atoms = new LinkedHashMap<String,Atom>() ;

      String wrkKey = "JNK" ;
      Monomer comp = new Monomer() ;
      comp.key3 = wrkKey ;
      comp.key8 = wrkKey ;
      comp.name = wrkKey ;

      int cou = 0 ;
      for( Atom atom : copy.atoms ){
         Atom orig = new Atom( (Residue)null, atom ) ;
         comp.atoms.add( orig ) ;
         atom.coupled = orig ;
         while( atoms.get( orig.key4 ) != null ){
            if( cou > 999 ) throw new NullPointerException() ;
            orig.key4 = orig.symbol + ++cou ;
         }
         atoms.put( orig.key4, atom ) ;
      }

      for( Atom atom1 : copy.atoms ){
         for( Bond bond : atom1 ){
            Atom atom2 = bond.another( atom1 ) ;
            if( Atom.timeComparator.compare( atom1, atom2 ) < 0 ){
               Atom orig1 = atom1.coupled ;
               Atom orig2 = atom2.coupled ;
               Bond bondOrig = new Bond( orig1, orig2, bond ) ;
               orig1.add( bondOrig ) ;
               orig2.add( bondOrig ) ;
            }
         }
      }
      for( Atom atom : copy.atoms ){
         if( atom.chir != null ){
            GenericRestraint chir = atom.chir.duplicate() ;
            chir.atoms.get(0).chir = chir ;
         }
      }

      copy.includeHidden() ;
      if( ! useLibmol ) for( Atom atom : copy.atoms ) atom.resetChanged() ;
      for( Atom atom : copy.atoms ) atom.coupled = atom ;

      comp.restraints = new Restraints() ;
      List<GenericRestraint> list = comp.restraints.get(0) ;
      for( Atom atom1 : comp.atoms ){
         for( Bond bond : atom1 ){
            Atom atom2 = bond.another( atom1 ) ;
            if( Atom.timeComparator.compare( atom1, atom2 ) < 0 ){
               list.add( new GenericRestraint( bond ) ) ;
            }
         }
      }
      list = comp.restraints.get(3) ;
      for( Atom atom : comp.atoms ){
         if( atom.chir != null ){
            list.add( atom.chir ) ;
         }
      }

//--  Corrector corrector = new Corrector( comp ) ;
//--  if( Env.vbPrint ) System.err.println( "-----" ) ;
//--  if( Env.vbPrint ) System.err.println( "AAAAA Corrector converged? " + corrector.converged() ) ;
//--  if( Env.vbPrint ) System.err.println( "AAAAA Corrector made changes? " + corrector.madeChanges() ) ;
//--  if( Env.vbPrint ) System.err.print( corrector ) ;
//--  if( Env.vbPrint ) System.err.println( "AAAAA" ) ;
//--  if( Env.vbPrint ) System.err.println( "-----" ) ;
//--  comp = corrector.getMonomer() ;

      comp = new Libcheck().regularise( comp, lcoorNo, useLibmol ) ;
      if( ! comp.key8.equals( wrkKey ) ) throw new NullPointerException() ;

      if( comp.atoms.size() == 1 && copy.atoms.size() == 1 ) return copy ;

      for( Atom orig : comp.atoms ){
         orig.coupled = orig ;
         Atom atom = atoms.get( orig.key4 ) ;
         if( ! orig.symbol.equals("H") ){
            if( atom == null ) throw new NullPointerException() ;
            orig.coupled = atom ;
            atom.coupled = orig ;
         }
      }

      for( Atom orig1 : comp.atoms ){               //::: loop over reference molecule !!!
         if( orig1.coupled == orig1 ){

            if( ! orig1.symbol.equals("H") ) throw new NullPointerException() ;
            Atom orig2 = orig1.get(0).another( orig1 ) ;
            Atom atom2 = orig2.coupled ;
            if( atom2 == orig2 ) throw new NullPointerException() ;
            Atom atom1 = null ;
            for( Bond bond : atom2 ){
               atom1 = bond.another( atom2 ) ;
               if( atom1.coupled == atom1 ) break ;
               atom1 = null ;
            }
            if( atom1 == null ){
               atom1 = new Atom( atom2.residue, orig1 ) ;
               atom1.key4 = null ;
               copy.atoms.add( atom1 ) ;
            }
            if( ! atom1.symbol.equals("H") ) throw new NullPointerException() ;

            orig1.coupled = atom1 ;
            atom1.coupled = orig1 ;
         }
      }

      // The last method do not affect chirality as it only relate to H-atoms.

      for( Atom orig : comp.atoms ) if( orig.coupled == orig ) throw new NullPointerException() ;

      Iterator<Atom> iterator = copy.iterator() ;
      while( iterator.hasNext() ){
         Atom atom1 = iterator.next() ;
         if( atom1.coupled == atom1 ){
            if( ! atom1.symbol.equals("H") ) throw new NullPointerException() ;
            Atom atom2 = atom1.get(0).another( atom1 ) ;
            if( atom2.coupled == atom2 ) throw new NullPointerException() ;
            iterator.remove() ;
         }
      }

      for( Atom orig : comp.atoms ) if( orig.coupled == orig ) throw new NullPointerException() ;
      for( Atom atom : copy.atoms ) if( atom.coupled == atom ) throw new NullPointerException() ;
      for( Atom atom : copy.atoms ) atom.coupled = atom ;

      for( Atom atom : copy.atoms ) atom.clear() ;
      Restraints restraints = new Restraints( comp.restraints ) ;
      for( GenericRestraint rst : restraints.get(0) ){
         Bond bond = new Bond( rst ) ;
         bond.atom1.add( bond ) ;
         bond.atom2.add( bond ) ;
      }
      for( GenericRestraint chir : restraints.get(3) ){
         Atom atom = chir.atoms.get(0) ;
         if( atom.chir != null ){
            if( ! atom.chir.equals( chir ) ) throw new NullPointerException() ;
         }
         atom.chir = chir ;
      }

      copy.restraints = null ;
      for( Molecule mol : copy.residueView().values() ) mol.restoreHydrogenKeys() ;
      copy.restraints = restraints ;

      for( Atom orig : comp.atoms ){
         Atom atom = orig.coupled ;
         atom.symbol = orig.symbol ;
         atom.energy = orig.energy ;
         atom.charge = orig.charge ;
      }

      if( Env.coordinates == Env.GRAPH_REFMAC || useLibmol ){
         for( Atom orig : comp.atoms ) orig.xyz = orig.coupled.xyz ;
      }

      if( Env.coordinates != Env.GRAPH_LIBCHECK ){
         new Refmac().refine( comp ) ;
      }

      for( Atom orig : comp.atoms ) orig.coupled.xyz = orig.xyz ;
      for( Atom orig : comp.atoms ) orig.coupled = orig ;

      for( GenericRestraint chir : copy.restraints.get(3) ) chir.setSign().checkType() ;

      Collections.sort( copy.atoms, Atom.zComparator ) ;

      copy.updateSphere() ;

      if( Env.coordinates != Env.GRAPH_REFMAC ){
         copy.restoreBaseline( base ) ;
      }

      return copy ;
   }

   // --------------------------------------------------------------------------

   class UndoableEditName extends Backup{
      String valueOld, valueNew ;
      UndoableEditName( String nameNew ){
         valueOld = name ;
         valueNew = nameNew ;
         name = nameNew ;
         if( Env.vbPrint ) System.out.println( ":::: " + key8 + " name " + name ) ;
      }
      void undo(){ name = valueOld ; }
      void redo(){ name = valueNew ; }
   }
   // --------------------------------------------------------------------------

   class UndoableEditKey8 extends Backup{
      String valueOld, valueNew ;
      UndoableEditKey8( String key8New ){
         valueOld = key8 ;
         valueNew = key8New ;
         key8 = key8New ;
         if( Env.vbPrint ) System.out.println( ":::: " + key8 + " key8" ) ;
      }
      void undo(){ key8 = valueOld ; }
      void redo(){ key8 = valueNew ; }
   }
   // --------------------------------------------------------------------------

   class UndoableClearRestraints extends Backup{
      Restraints valueOld ;
      UndoableClearRestraints(){
         valueOld = restraints ;
         restraints = null ;
         if( Env.vbPrint ) System.out.println( ":::: " + key8 + " clear restraints" ) ;
      }
      void undo(){ restraints = valueOld ; }
      void redo(){ restraints = null ; }
   }
   // --------------------------------------------------------------------------

   class UndoableAddAtom extends Backup{
      Atom atom ;
      UndoableAddAtom( Atom atomNew ){
         atom = atomNew ;
         atoms.add( atom ) ;
         updateSphere() ;
         if( Env.vbPrint ) System.out.println( ":::: " + key8 + " add atom " + atom.key4 ) ;
      }
      void undo(){
         atoms.remove( atom ) ;
         atom.residue.transform.backward( atom.xyz ) ;
         updateSphere() ;
      }
      void redo(){
         atoms.add( atom ) ;
         atom.residue.transform.forward( atom.xyz ) ;
         updateSphere() ;
      }
   }
   // --------------------------------------------------------------------------

   class UndoableRemoveAtom extends Backup{
      Atom atom ;
      UndoableRemoveAtom( Atom atomOld ){
         atom = atomOld ;
         atoms.remove( atom ) ;
         atom.residue.transform.backward( atom.xyz ) ;
         updateSphere() ;
         if( Env.vbPrint ) System.out.println( ":::: " + key8 + " remove atom " + atom.key4 ) ;
      }
      void undo(){
         atoms.add( atom ) ;
         atom.residue.transform.forward( atom.xyz ) ;
         updateSphere() ;
      }
      void redo(){
         atoms.remove( atom ) ;
         atom.residue.transform.backward( atom.xyz ) ;
         updateSphere() ;
      }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

class Residue implements Comparable<Residue>{

//                                                                   coot begin
   String residueId = "" ;
   int cootId = 0 ;

   public void setCootInfo( String chainId, String residueNo, int id ){
      residueId = String.format( "%s%4s", chainId, residueNo ) ;
      cootId = id ;
   }
//                                                                   coot end

   boolean isFirst = false ;

   private static int counter = 0 ;
   private final int cou = counter++ ;
   Transform transform = new Transform() ;

   public int hashCode(){ return cou ; }
   public boolean equals( Object o ){ return cou == o.hashCode() ; }
   public int compareTo( Residue o ){
      if( isFirst && ! o.isFirst ) return -1 ;
      if( ! isFirst && o.isFirst ) return +1 ;
      return cou - o.cou ;
   }

   public String toString(){ return key8 + "-" + String.valueOf( cou ) ; }

   String key8 ;
   String name ;
   String group ;
   Monomer mono ;

   int newAtomCount ;

   Residue(){}

   Residue duplicate(){
      Residue rsd = new Residue() ;
//                                                                   coot begin
      rsd.residueId = residueId ;
      rsd.cootId = cootId ;
//                                                                   coot end
      rsd.isFirst = isFirst ;
      rsd.key8 = key8 ;
      rsd.name = name ;
      rsd.group = group ;
      rsd.mono = mono ;
      return rsd ;
   }

   Residue( Monomer monomer ){
      key8 = monomer.key8 ;
      name = monomer.name ;
      group = monomer.group ;
      mono = monomer ;
   }

   // --------------------------------------------------------------------------

   class UndoableEditName extends Backup{
      String valueOld, valueNew ;
      UndoableEditName( String nameNew ){
         valueOld = name ;
         valueNew = nameNew ;
         name = nameNew ;
         if( Env.vbPrint ) System.out.println( ":::: " + key8 + " name " + name ) ;
      }
      void undo(){ name = valueOld ; }
      void redo(){ name = valueNew ; }
   }
   // --------------------------------------------------------------------------

   class UndoableEditKey8 extends Backup{
      String valueOld, valueNew ;
      UndoableEditKey8( String key8New ){
         valueOld = key8 ;
         valueNew = key8New ;
         key8 = key8New ;
         if( Env.vbPrint ) System.out.println( ":::: " + key8 + " key8" ) ;
      }
      void undo(){ key8 = valueOld ; }
      void redo(){ key8 = valueNew ; }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

//class Monomer extends ArrayList<Atom>{
class Monomer{
   List<Atom> atoms = new ArrayList<Atom>() ;
   String key3 ;
   String key8 ;
   String name ;
   String group ;
   Restraints restraints ;

   public String toString(){ return atoms.toString() ; }

   boolean isStandard(){
      boolean lcoorNo = true ;
      return StdLib.hasReady( this.key8 ) && this == StdLib.get( this.key8, lcoorNo ) ;
   }

   void recentre(){
      double[] xyz = new double[]{ 0f, 0f, 0f } ;
      for( Atom atom : this.atoms ){
         for( int i = 0 ; i < 3 ; i++ ) xyz[i] += atom.xyz[i] ;
      }
      for( int i = 0 ; i < 3 ; i++ ) xyz[i] = xyz[i]/ this.atoms.size() ;
      for( Atom atom : this.atoms ){
         for( int i = 0 ; i < 3 ; i++ ) atom.xyz[i] -= xyz[i] ;
      }
   }

   Monomer(){}

   Monomer( Molecule molecule ){
      key8 = molecule.key8 ;
      key3 = key8.substring( 0, Math.min( 3, key8.length() ) ) ;
      name = molecule.name ;
      group = null ;

      Collections.sort( molecule.atoms, Atom.originComparator ) ;

      for( Atom atom : molecule.atoms ){
         Atom orig = new Atom( (Residue)null, atom ) ;
         this.atoms.add( orig ) ;
         atom.coupled = orig ;
      }

      if( molecule.restraints != null ){
         this.restraints = new Restraints( molecule.restraints ) ;
      }
      else{
         this.restraints = new Restraints() ;
         List<GenericRestraint> list = this.restraints.get(0) ;
         for( Atom atom1 : molecule.atoms ){
            for( Bond bond : atom1 ){
               Atom atom2 = bond.another( atom1 ) ;
//             if( atom1.compareTo( atom2 ) < 0 ){
               if( Atom.timeComparator.compare( atom1, atom2 ) < 0 ){
                  list.add( new GenericRestraint( new Bond( atom1.coupled, atom2.coupled, bond ) ) ) ;
               }
            }
         }
         list = this.restraints.get(3) ;
         for( Atom atom : molecule.atoms ){
            if( atom.chir != null ){
               GenericRestraint chir = atom.chir.duplicate() ;
               chir.atoms.get(0).chir = chir ;
               list.add( chir ) ;
            }
         }
      }

      for( Atom atom : molecule.atoms ) atom.coupled = atom ;
      Collections.sort( molecule.atoms, Atom.zComparator ) ;

      Set<String> set = new LinkedHashSet<String>() ;
      for( Atom orig : this.atoms ) orig.resetKey( set ) ;
   }

   // --------------------------------------------------------------------------

   class UndoableEditName extends Backup{
      String valueOld, valueNew ;
      UndoableEditName( String nameNew ){
         valueOld = name ;
         valueNew = nameNew ;
         name = nameNew ;
         if( Env.vbPrint ) System.out.println( ":::: " + key8 + " name " + name ) ;
      }
      void undo(){ name = valueOld ; }
      void redo(){ name = valueNew ; }
   }
   // --------------------------------------------------------------------------

   class UndoableEditKey3 extends Backup{
      String valueOld, valueNew ;
      UndoableEditKey3( String key3New ){
         valueOld = key3 ;
         valueNew = key3New ;
         key3 = key3New ;
         if( Env.vbPrint ) System.out.println( ":::: " + key3 + " key3" ) ;
      }
      void undo(){ key3 = valueOld ; }
      void redo(){ key3 = valueNew ; }
   }
   // --------------------------------------------------------------------------

   class UndoableEditKey8 extends Backup{
      String valueOld, valueNew ;
      UndoableEditKey8( String key8New ){
         valueOld = key8 ;
         valueNew = key8New ;
         key8 = key8New ;
         if( Env.vbPrint ) System.out.println( ":::: " + key8 + " key8" ) ;
      }
      void undo(){ key8 = valueOld ; }
      void redo(){ key8 = valueNew ; }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

class Rings{
   static private final RingMap ringMap ;

   static{
      ringMap = new RingMap() ;
   }

   static Ring get( String key ){
      return ringMap.get( key ) ;
   }

   // --------------------------------------------------------------------------

   @SuppressWarnings("serial")
   private static class RingMap extends LinkedHashMap<String,Ring>{
      private RingMap(){
         super( 6 ) ;
         put( "n3" , new Ring( 3, "C", new Bond( "single"  , 1.52 ) ) ) ;
         put( "n4" , new Ring( 4, "C", new Bond( "single"  , 1.52 ) ) ) ;
         put( "n5" , new Ring( 5, "C", new Bond( "single"  , 1.52 ) ) ) ;
         put( "n6" , new Ring( 6, "C", new Bond( "single"  , 1.52 ) ) ) ;
         put( "n6B", new Ring( 6, "C", new Bond( "aromatic", 1.39 ) ) ) ;
         put( "n7" , new Ring( 7, "C", new Bond( "single"  , 1.52 ) ) ) ;
      }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

class Ring{
   double[] xo, yo ;
   String atomType ;
   Bond templateBond ;

   Ring( int n, String s, Bond bond ){
      atomType = s ;
      templateBond = bond ;
      xo = new double[n] ;
      yo = new double[n] ;
      double bondLength = templateBond.value.mean ;
      double ang = Math.PI/ n ;
      double del = 2* ang ;
      xo[0] = yo[0] = 0f ;
      for( int i = 1 ; i < n ; i++ ){
         xo[i] = xo[i-1] + Math.sin( ang )* bondLength ;
         yo[i] = yo[i-1] + Math.cos( ang )* bondLength ;
         ang += del ;
      }
   }

   void undoableAddTo( UndoRedo undoRedo, Molecule mol, Atom orig, double[] xy ){
      int n = xo.length ;
      Atom[] atoms = new Atom[n] ;
      double[] xyzn = new double[3] ;
      double[] xyzo = new double[3] ;

      Residue rsd = null ;
      Set<String> set = null ;
      if( orig == null ){
         set = new LinkedHashSet<String>( xo.length ) ;
         rsd = new Residue() ;
         rsd.key8 = mol.key8 ;
         rsd.name = mol.name ;
         for( int i = 0 ; i < 3 ; i++ ) xyzo[i] = 0f ;
      }
      else{
//       set = mol.atomView( orig ).keySet() ;                                //:: wrong: immuable set
         set = new LinkedHashSet<String>( mol.atomView( orig ).keySet() ) ;   //:: not nice
         rsd = orig.residue ;
         for( int i = 0 ; i < 3 ; i++ ) xyzo[i] = orig.xyz[i] ;
      }

      double co = xy[0] - xyzo[0] ;
      double si = xy[1] - xyzo[1] ;
      double rr = co* co + si* si ;
      if( rr < 0.00001 ){
         co = 1f ;
         si = 0f ;
      }
      else{
         rr = Math.sqrt( rr ) ;
         co = co/ rr ;
         si = si/ rr ;
      }
      for( int i = 0 ; i < n ; i++ ){
         xyzn[0] = co* xo[i] - si* yo[i] + xy[0] ;
         xyzn[1] = si* xo[i] + co* yo[i] + xy[1] ;
         xyzn[2] = xyzo[2] ;
         Atom atom = new Atom( rsd, atomType, atomType, xyzn[0], xyzn[1], xyzn[2] ) ;
         atom.generateKey( set ) ;
         atoms[i] = atom ;
         undoRedo.backup( mol.new UndoableAddAtom( atom ) ) ;
      }

      if( orig != null ){
         Atom atom = atoms[0] ;
         Bond bond = new Bond( orig, atom, "single" ) ;
         undoRedo.backup( orig.new UndoableAddBond( bond ) ) ;
         undoRedo.backup( atom.new UndoableAddBond( bond ) ) ;
      }

      Atom atom1 = atoms[n-1] ;
      for( int i = 0 ; i < n ; i++ ){
         Atom atom2 = atoms[i] ;
         Bond bond = new Bond( atom1, atom2, templateBond ) ;
         undoRedo.backup( atom1.new UndoableAddBond( bond ) ) ;
         undoRedo.backup( atom2.new UndoableAddBond( bond ) ) ;
         atom1 = atom2 ;
      }
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class Atom extends ArrayList<Bond>{

   final Residue residue ;
   GenericRestraint chir ;

   String key4 ;
   String symbol ;
   String energy ;
   Float charge ;
   double[] xyz = new double[]{ 0f, 0f, 0f } ;

   Atom coupled = this ;
   private boolean changed = false ;
   private boolean changedCharge = false ;

   boolean hasChanged(){ return changed ; }

   boolean hasChangedCharge(){ return changedCharge ; }

   void resetChanged(){ changed = changedCharge ; }

   void resetCharge(){
      if( ! changedCharge ) charge = 0f ;
   }

   void resetHidden(){ changed = false ; }

   boolean isHidden(){
      if( ! symbol.equals( "H" ) ) return false ;
      if( size() != 1 ) return true ;
      return ! changed && get( 0 ).another( this ).changed ;
   }

   Atom( Residue rsd, String atomType, String key, double x, double y, double z ){
      cou1 = counter1++ ;
      residue = rsd ;
      key4 = key ;
      symbol = atomType ;
      energy = atomType ;
      charge = 0f ;
      xyz[0] = x ;
      xyz[1] = y ;
      xyz[2] = z ;
      changed = true ;
   }

   Atom( Residue r ){
      cou1 = counter1++ ;
      residue = r ;
   }

   Atom( Residue r, Atom a ){
      cou1 = a.cou1 ;
      residue = r ;
      key4   = a.key4   ;
      symbol = a.symbol ;
      energy = a.energy ;
      charge = a.charge ;
      xyz[0] = a.xyz[0] ;
      xyz[1] = a.xyz[1] ;
      xyz[2] = a.xyz[2] ;
      this.changed = a.changed ;
      this.changedCharge = a.changedCharge ;
   }

   void resetKey( Set<String> set ){
      if( set.add( key4 ) ) return ;
      String basename = key4.substring( 0, Math.min( 3, key4.length() ) ) ;
      for( int i = 1 ; i < 10 ; i++ ) if( set.add( key4 = basename + i ) ) return ;
      generateKey( set ) ;
   }

   void generateKey( Set<String> set ){
//::  for( int i = 1 ; i < 1000 ; ++i ) if( set.add( key4 = symbol + i ) ) return ;
      while( ++residue.newAtomCount < 1000 ){
         if( set.add( key4 = symbol + residue.newAtomCount ) ) return ;
      }
      throw new NullPointerException() ;
   }

   public String toString(){
      return key4 + "-" + String.valueOf( cou2 ) + "-" + String.valueOf( cou1 ) ;
   }

   static int sign( List<Atom> list ){
      double[] xyzA = list.get(0).xyz ;
//    if( xyzA == null || xyzA.length != 3 ) return 0 ;

      double[][] mat = new double[3][] ;
      for( int i = 0 ; i < 3 ; i++ ){
         double[] xyzB = list.get(i+1).xyz ;
//       if( xyzB == null || xyzB.length != 3 ) return 0 ;

         double[] vec = new double[3] ;
         for( int j = 0 ; j < 3 ; j++ ){
            vec[j] = xyzB[j] - xyzA[j] ;
         }
         mat[i] = vec ;
      }
      double det = mat[0][0]* mat[1][1]* mat[2][2] ;
      det -= mat[0][0]* mat[1][2]* mat[2][1] ;
      det -= mat[1][1]* mat[2][0]* mat[0][2] ;
      det -= mat[2][2]* mat[0][1]* mat[1][0] ;
      det += mat[0][2]* mat[1][0]* mat[2][1] ;
      det += mat[2][0]* mat[0][1]* mat[1][2] ;
      if( det > 0 ) return 1 ;
      return - 1 ;
   }

   // --------------------------------------------------------------------------

   final static Atom NULL = new Atom() ;

   private Atom(){
      cou1 = Integer.MAX_VALUE ;
      residue = null ;
   }

   // --------------------------------------------------------------------------

   private static int counter1 = 0 ;
   private final int cou1 ;

   public static final Comparator<Atom> originComparator = new Comparator<Atom>(){
      public int compare( Atom atom1, Atom atom2 ){
         return atom1.cou1 - atom2.cou1 ;
      }
   } ;

   // --------------------------------------------------------------------------

   private static int counter2 = 0 ;
   private final int cou2 = counter2++ ;

   public static final Comparator<Atom> timeComparator = new Comparator<Atom>(){
      public int compare( Atom atom1, Atom atom2 ){
         return atom1.cou2 - atom2.cou2 ;
      }
   } ;

   public int hashCode(){ return cou2 ; }
   public boolean equals( Object o ){ return o.hashCode() == cou2 ; }
   public int compareTo( Object o ){ throw new NullPointerException() ; }

   // --------------------------------------------------------------------------

   public static final Comparator<Atom> zComparator = new Comparator<Atom>(){
      public int compare( Atom atom1, Atom atom2 ){
         if( atom1.xyz[2] < atom2.xyz[2] ) return + 1 ;
         if( atom1.xyz[2] > atom2.xyz[2] ) return - 1 ;
         return atom1.cou2 - atom2.cou2 ;
      }
   } ;

   // --------------------------------------------------------------------------

   class UndoableEditKey4 extends Backup{
      String valueOld, valueNew ;
      UndoableEditKey4( String key4New ){
         valueOld = key4 ;
         valueNew = key4New ;
         key4 = key4New ;
         if( Env.vbPrint ) System.out.println( ":::: " + key4 + " key4 " + key4 ) ;
      }
      void undo(){ key4 = valueOld ; }
      void redo(){ key4 = valueNew ; }
   }
   // --------------------------------------------------------------------------

   class UndoableEditSymbol extends Backup{
      String energyOld, energyNew ;
      String valueOld, valueNew ;
      boolean changedOld ;
      UndoableEditSymbol( String symbolNew ){

         valueOld = symbol ;
         valueNew = symbolNew ;
         symbol = symbolNew ;

         energyOld = energy ;
         energyNew = symbolNew ;
         energy = symbolNew ;

         changedOld = changed ;
         changed = true ;
         if( Env.vbPrint ) System.out.println( ":::: " + key4 + " symbol " + symbol ) ;
      }
      void undo(){
         symbol = valueOld ;
         energy = energyOld ;
         changed = changedOld ;
      }
      void redo(){
         symbol = valueNew ;
         energy = energyNew ;
         changed = true ;
      }
   }
   // --------------------------------------------------------------------------

   class UndoableEditCharge extends Backup{
      Float valueOld, valueNew ;
      boolean changedOld ;
      boolean changedChargeOld ;
      UndoableEditCharge( Float chargeNew ){
         valueOld = charge ;
         valueNew = chargeNew ;
         charge = chargeNew ;

         changedChargeOld = changedCharge ;
         changedCharge = true ;

         changedOld = changed ;
         changed = true ;
         if( Env.vbPrint ) System.out.println( ":::: " + key4 + " charge " + charge ) ;
      }
      void undo(){
         charge = valueOld ;
         changedCharge = changedChargeOld ;
         changed = changedOld ;
      }
      void redo(){
         charge = valueNew ;
         changedCharge = true ;
         changed = true ;
      }
   }
   // --------------------------------------------------------------------------

   class UndoableAddBond extends Backup{
      Bond bond ;
      GenericRestraint chirOld ;
      boolean changedOld ;
      UndoableAddBond( Bond bondNew ){
         chirOld = chir ;
         chir = null ;
         bond = bondNew ;
         add( bond ) ;
         changedOld = changed ;
         changed = true ;
         if( Env.vbPrint ) System.out.println( ":::: " + key4 + " add bond" ) ;
      }
      void undo(){
         chir = chirOld ;
         remove( bond ) ;
         changed = changedOld ;
      }
      void redo(){
         chir = null ;
         add( bond ) ;
         changed = true ;
      }
   }
   // --------------------------------------------------------------------------

   class UndoableRemoveBond extends Backup{
      Bond bond ;
      GenericRestraint chirOld ;
      boolean changedOld ;
      UndoableRemoveBond( Bond bondOld ){
         chirOld = chir ;
         chir = null ;
         bond = bondOld ;
         remove( bond ) ;
         changedOld = changed ;
         changed = true ;
         if( Env.vbPrint ) System.out.println( ":::: " + key4 + " remove bond" ) ;
      }
      void undo(){
         chir = chirOld ;
         add( bond ) ;
         changed = changedOld ;
      }
      void redo(){
         chir = null ;
         remove( bond ) ;
         changed = true ;
      }
   }
   // --------------------------------------------------------------------------

   class UndoableReplaceBond extends Backup{
      Bond bondOld, bondNew ;
      GenericRestraint chirOld, chirNew ;
      UndoableReplaceBond( Bond bondOld, Bond bondNew ){
         chirOld = chir ;
         chirNew = chir == null ? null : chir.duplicatePart() ;
         this.bondOld = bondOld ;
         this.bondNew = bondNew ;
         remove( bondOld ) ;
         add( bondNew ) ;
         chir = chirNew ;
         if( Env.vbPrint ) System.out.println( ":::: " + key4 + " replace bond" ) ;
      }
      void undo(){
         remove( bondNew ) ;
         add( bondOld ) ;
         chir = chirOld ;
      }
      void redo(){
         remove( bondOld ) ;
         add( bondNew ) ;
         chir = chirNew ;
      }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

class Bond{
   public String toString(){
      return "(" + atom1 + ( type.equals( "double" ) ? " == " : " -- " ) + atom2 + ")" ;
   }

   boolean visited = false ;

   Atom atom1 ;
   Atom atom2 ;
   String type ;
   Value value ;

   Bond(){}

   Bond( String s, double d ){
      type  = s ;
      value = new Value( new Float( d ), new Float( 0.3f ) ) ;
   }

   Bond( Atom a1, Atom a2, String s ){
      type  = s ;
      value = new Value( new Float( 1.5f ), new Float( 0.3f ) ) ;
      int diff = Atom.timeComparator.compare( a1, a2 ) ;
      if( diff < 0 ){ atom1 = a1 ; atom2 = a2 ; return ; }
      if( diff > 0 ){ atom1 = a2 ; atom2 = a1 ; return ; }
      throw new NullPointerException() ;
   }

   Bond( Atom a1, Atom a2, Bond b ){
      type  = b.type ;
      value = b.value ;
      int diff = Atom.timeComparator.compare( a1, a2 ) ;
      if( diff < 0 ){ atom1 = a1 ; atom2 = a2 ; return ; }
      if( diff > 0 ){ atom1 = a2 ; atom2 = a1 ; return ; }
      throw new NullPointerException() ;
   }

   Bond( GenericRestraint rst ){
      if( rst.atoms.size() != 2 ) throw new NullPointerException() ;
      atom1 = rst.atoms.get( 0 ) ;
      atom2 = rst.atoms.get( 1 ) ;
      type  = rst.type ;
      value = rst.value ;
   }

   Atom another( Atom a ){
      if( a == atom1 ) return atom2 ;
      if( a == atom2 ) return atom1 ;
      throw new NullPointerException() ;
   }
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class Restraints extends ArrayList<List<GenericRestraint>>{

   Restraints(){
      super( 5 ) ;
      for( int i = 0 ; i < 5 ; i++ ) this.add( new ArrayList<GenericRestraint>() ) ;
   }

   Restraints( Restraints restraints ){
      super( 5 ) ;
      for( int i = 0 ; i < 5 ; i++ ){
         List<GenericRestraint> list = restraints.get(i) ;
         List<GenericRestraint> listNew = new ArrayList<GenericRestraint>( list.size() ) ;
         for( GenericRestraint rst : list ) listNew.add( rst.duplicate() ) ;
         this.add( listNew ) ;
      }
   }
}
// -----------------------------------------------------------------------------

class GenericRestraint implements Comparable<GenericRestraint>{

   int sign = 0 ;
   final static String VOID_TYPE = "/ void /" ;
   final static List<String> typeList =
      Arrays.asList( new String[]{ "positiv", "both", "negativ" } ) ;

   static String getChirType( String si ){
      String sl = si.toLowerCase() ;
      for( String t : typeList ){
         if( t.startsWith( sl ) || sl.startsWith( t ) ) return t ;
      }
      return VOID_TYPE ;
   }

   void initBothAndSign(){
      if( typeList.contains( type ) ) sign = 1 ;
   }

   GenericRestraint setSign(){
      if( sign != 0 ){
         int signFromXYZ = Atom.sign( atoms ) ;
         if( Math.abs( signFromXYZ ) != 1 ) throw new NullPointerException() ;
         sign = signFromXYZ ;
      }
      return this ;
   }

   boolean isConsistent(){
      return sign == 0 || type.equals( "both" ) || type.equals( typeList.get( 1 - sign ) ) ;
   }

   void checkType(){
//    if( ! isConsistent() ) throw new NullPointerException() ;
      if( ! isConsistent() && Env.vbPrint ){
         String s0 = "WARNING: actual (" + sign + ") and intended (" + type + ") chirality differ for" ;
         String s1 = " " ;
         for( int i = 0 ; i < 4 ; i++ ){
            s0 += s1 + atoms.get(i).key4 ;
            s1 = "-" ;
         }
         System.err.println( s0 ) ;
      }
   }

   void setType(){
      if( sign == 0 || type.equals( "both" ) ) return ;
      type = typeList.get( 1 - sign ) ;
   }

   String getType( int index ){
      if( 0 > index || index > 2 ) return VOID_TYPE ;
      return typeList.get( index ) ;
   }

   int getIndex(){
      if( sign == 0 ) return - 1 ;
      return typeList.indexOf( type ) ;
   }

   String[] getOptions(){
      String[] options = new String[3] ;
      if( sign == 0 ){
         throw new NullPointerException() ;
      }
      else if( sign > 0 ){
         options[0] = "keep \"positive\"" ;
         options[2] = "make \"negative\"" ;
      }
      else if( sign < 0 ){
         options[0] = "make \"positive\"" ;
         options[2] = "keep \"negative\"" ;
      }
      options[1] = "allow \"both\"" ;
      return options ;
   }

   class UndoableEditType extends Backup{
      String valueOld, valueNew ;
      UndoableEditType( String typeNew ){
         valueOld = type ;
         valueNew = typeNew ;
         type = typeNew ;
         if( Env.vbPrint ) System.out.println( ":::: " + type + " chir.type" ) ;
      }
      void undo(){ type = valueOld ; }
      void redo(){ type = valueNew ; }
   }
   // --------------------------------------------------------------------------

// Oh-oh-oh continued
// Special treatment of bond: must be: "1 SMTH 2 SMTH", otherwise refmac crashes.
// Used to restore original order, just to make sure nothing goes wrong:

   boolean reordered = false ;

// end of Oh-oh-oh

   final static PermGrp sortBond  = new PermGrp( 2, new int[][]{ { 1, 0 } } ) ;

   final static PermGrp sortAngle = new PermGrp( 2, new int[][]{ { 2, 1, 0 } } ) ;

   final static PermGrp sortTor   = new PermGrp( 2, new int[][]{ { 3, 2, 1, 0 } } ) ;

   final static PermGrp sortChir  = new PermGrp( 3, new int[][]{ { 0, 2, 3, 1 } } ) ;

   final static PermGrp sortPlane = null ;

   final static PermGrp sortCross0 =
      new PermGrp(  2, new int[][]{ { 0, 2, 1 } } ) ;

   final static PermGrp sortCross1 =
      new PermGrp(  2, new int[][]{ { 0, 2, 1 } } ) ;

   final static PermGrp sortCross2 =
      new PermGrp(  8, new int[][]{ { 0, 4, 3, 2, 1 }, { 0, 1, 2, 4, 3 } } ) ;

   final static PermGrp sortCross3 =
      new PermGrp(  6, new int[][]{ { 0, 1, 2, 4, 5, 3 }, { 0, 2, 1, 5, 4, 3 } } ) ;

// final static PermGrp sortCross4 =
//    new PermGrp(  8, new int[][]{ { 0, 1, 2, 4, 5, 6, 3 }, { 0, 2, 1, 6, 5, 4, 3 } } ) ;

   final static PermGrp sortCross4 =
      new PermGrp( 24, new int[][]{ { 0, 1, 2, 4, 5, 6, 3 }, { 0, 2, 1, 6, 5, 4, 3 }, { 0, 3, 5, 1, 6, 2, 4 } } ) ;

   final static PermGrp sortCross5 =
      new PermGrp( 10, new int[][]{ { 0, 1, 2, 4, 5, 6, 7, 3 }, { 0, 2, 1, 7, 6, 5, 4, 3 } } ) ;

   final static PermGrp sortCross6 =
      new PermGrp( 12, new int[][]{ { 0, 1, 2, 4, 5, 6, 7, 8, 3 }, { 0, 2, 1, 8, 7, 6, 5, 4, 3 } } ) ;

   List<Atom> atoms ;
   String key ;
   String type ;
   Value value ;
   Sigma sigma ;
   Integer period ;
   PermGrp sorter ;

   public boolean equals( Object o ){ return compareTo( (GenericRestraint)o ) == 0 ; }

   public String toString(){
      return atoms.toString() ;
   }

   public int compareTo( GenericRestraint rst ){
      int n = atoms.size() ;
      int diff = n - rst.atoms.size() ;
      for( int i = 0 ; i < n ; i++ ){
         if( diff != 0 ) return diff ;
//       diff = atoms.get( i ).compareTo( rst.atoms.get( i ) ) ;
         diff = Atom.originComparator.compare( atoms.get( i ), rst.atoms.get( i ) ) ;
      }
      return diff ;
   }

   GenericRestraint( int i ){
      atoms = new ArrayList<Atom>( i ) ;
   }

   GenericRestraint( Bond bond ){
      atoms = new ArrayList<Atom>( 2 ) ;
      atoms.add( bond.atom1 ) ;
      atoms.add( bond.atom2 ) ;
      type = bond.type ;
      value = bond.value ;
      sorter = sortBond ;
      sort() ;
   }

   GenericRestraint emptyInstance(){
      GenericRestraint rst = new GenericRestraint( this.atoms.size() ) ;
      rst.sorter = sorter ;
      return rst ;
   }

   void sort(){
      if( sorter == null ) Collections.sort( atoms, Atom.originComparator ) ;
      else sorter.sort( atoms, Atom.originComparator ) ;
   }

   GenericRestraint duplicatePart(){
      return duplicate( false ) ;
   }

   GenericRestraint duplicate(){
      return duplicate( true ) ;
   }

   private GenericRestraint duplicate( boolean strict ){
      GenericRestraint rstn = this.emptyInstance() ;
      rstn.atoms = new ArrayList<Atom>( this.atoms.size() ) ;
      for( Atom atom : this.atoms ){
         if( atom == Atom.NULL ){
            rstn.atoms.add( Atom.NULL ) ;
         }
         else if( atom == atom.coupled ){
            if( strict ) throw new NullPointerException() ;
            rstn.atoms.add( atom.coupled ) ;
         }
         else{
            rstn.atoms.add( atom.coupled ) ;
         }
      }
      rstn.sort() ;
      rstn.key    = key    ;
      rstn.type   = type   ;
      rstn.value  = value  ;
      rstn.sigma  = sigma  ;
      rstn.period = period ;
      rstn.sign = sign ;
      return rstn ;
   }

   GenericRestraint applyDiff( GenericRestraint diff ){
      GenericRestraint rstn = diff.emptyInstance() ;
      rstn.atoms = new ArrayList<Atom>( diff.atoms ) ;
      rstn.key    = this.key ;
      rstn.type   = diff.type    == null ? this.type   : diff.type    ;
      rstn.value  = diff.value   == null ? this.value  : diff.value   ;
      rstn.sigma  = diff.sigma   == null ? this.sigma  : diff.sigma   ;
      rstn.period = diff.period  == null ? this.period : diff.period  ;
      return rstn ;
   }

   boolean diff( GenericRestraint rstn ){
      boolean equal = this.equals( rstn ) ;
      equal &= rstn.type   == null ? this.type   == null : rstn.type  .equals ( this.type   ) ;
      equal &= rstn.value  == null ? this.value  == null : rstn.value .similar( this.value  ) ;
      equal &= rstn.sigma  == null ? this.sigma  == null : rstn.sigma .similar( this.sigma  ) ;
      equal &= rstn.period == null ? this.period == null : rstn.period.equals ( this.period ) ;
      return ! equal ;
   }

   GenericRestraint deriveDiff( GenericRestraint rstn ){
      GenericRestraint diff = rstn.emptyInstance() ;
      diff.atoms = new ArrayList<Atom>( rstn.atoms ) ;
      diff.key    = this.key ;
      diff.type   = rstn.type   != null && rstn.type  .equals ( this.type   ) ? null : rstn.type   ;
      diff.value  = rstn.value  != null && rstn.value .similar( this.value  ) ? null : rstn.value  ;
      diff.sigma  = rstn.sigma  != null && rstn.sigma .similar( this.sigma  ) ? null : rstn.sigma  ;
      diff.sigma  = rstn.sigma  != null && rstn.sigma .similar( this.sigma  ) ? null : rstn.sigma  ;
      diff.period = rstn.period != null && rstn.period.equals ( this.period ) ? null : rstn.period ;
      return diff ;
   }

   GenericRestraint asDeleted(){
      GenericRestraint rstn = this.emptyInstance() ;
      rstn.atoms = new ArrayList<Atom>( this.atoms ) ;
      rstn.key = this.key ;
      rstn.type = this.type ;
      return rstn ;
   }

   Residue getResidue(){
      Residue rsd = atoms.get(0).residue ;
      for( Atom atom : atoms ){
         if( atom != Atom.NULL && atom.residue != rsd ) return null ;
      }
      return rsd ;
   }

   void setValue( Object o1, Object o2 ){
      if( ( o1 == null ) != ( o2 == null ) ) throw new NullPointerException() ;
      value = o1 == null ? null : new Value( (Float)o1, (Float)o2 ) ;
   }
}
// -----------------------------------------------------------------------------

class Value{
   final float mean, sd ;
   Value( float f1, float f2 ){ mean = f1 ; sd = f2 ; }
   boolean similar( Value r ){ return Math.abs( r.mean - mean )* 3f < r.sd + sd + 0.00001f ; }
}
// -----------------------------------------------------------------------------

class Sigma{
   final float sd ;
   Sigma( float f ){ sd = f ; }
   boolean similar( Sigma r ){ return Math.abs( r.sd - sd )* 3f < r.sd + sd ; }
}
// -----------------------------------------------------------------------------

class PermGrp{
   List<PermOp> ops ;
   int len ;

   PermGrp( int size, int[][] generators ){
      len = generators[0].length ;
      ops = new ArrayList<PermOp>( size ) ;
      ops.add( new PermOp() ) ;
      for( int i = 0, n = generators.length ; i < n ; i++ ) add( size, generators[i] ) ;
      if( ops.size() != size ) throw new NullPointerException() ;
   }

// <T extends Comparable<? super T>> void sort( List<T> list ){
   <Y> void sort( List<Y> list, Comparator<Y> comparator ){
      PermOp op1 = ops.get( 0 ) ;
      for( int i = 1, n = ops.size() ; i < n ; i++ ){
         PermOp op2 = ops.get( i ) ;
         for( int j = 0 ; j < len ; j++ ){
            Y s1 = list.get( op1.perm[j] ) ;
            Y s2 = list.get( op2.perm[j] ) ;
//          int diff = s2.compareTo( s1 ) ;
            int diff = comparator.compare( s2, s1 ) ;
            if( diff < 0 ) op1 = op2 ;
            if( diff != 0 ) break ;
         }
      }
      List<Y> tmp = new ArrayList<Y>( list ) ;
      for( int j = 0 ; j < len ; j++ ) list.set( j, tmp.get( op1.perm[j] ) ) ;
   }

   void add( int size, int[] perm ){

      PermOp op = new PermOp( perm ) ;
      if( ops.contains( op ) ) throw new NullPointerException() ;
      ops.add( op ) ;

      int k = ops.size() ;
      if( k > size ) throw new NullPointerException() ;

      for( int j = k - 1 ; j < k ; j++ ){
         for( int i = 0 ; i < j ; i++ ){
            if( ! ops.contains( op = new PermOp( ops.get(i), ops.get(j) ) ) ) ops.add( op ) ;
            if( ! ops.contains( op = new PermOp( ops.get(j), ops.get(i) ) ) ) ops.add( op ) ;
            if( ! ops.contains( op = new PermOp( ops.get(j), ops.get(j) ) ) ) ops.add( op ) ;
         }
         k = ops.size() ;
         if( k > size ) throw new NullPointerException() ;
      }
   }

   public String toString(){
      String bsn = Env.bsn ;
      String s = "" ;
      int n = ops.size() ;
      for( int i = 0 ; i < n ; i++ ) s += String.format( " %3d %s%s", i+1, ops.get(i), bsn ) ;
      return s ;
   }

   // --------------------------------------------------------------------------

   class PermOp{
      private int[] perm ;

      PermOp(){
         this.perm = new int[len] ;
         for( int i = 0 ; i < len ; i++ ) this.perm[i] = i ;
      }

      PermOp( int[] perm ){
         this.perm = new int[len] ;
         for( int i = 0 ; i < len ; i++ ) this.perm[i] = perm[i] ;
      }

      PermOp( PermOp op1, PermOp op2 ){
         this.perm = new int[len] ;
         for( int i = 0 ; i < len ; i++ ) this.perm[i] = op2.perm[ op1.perm[i] ] ;
      }

      public boolean equals( Object o ){
         PermOp op = (PermOp)o ;
         for( int i = 0 ; i < len ; i++ ) if( this.perm[i] != op.perm[i] ) return false ;
         return true ;
      }

      public String toString(){
         String s = "" ;
         for( int i = 0 ; i < len ; i++ ) s += String.format( " %2d", this.perm[i] + 1 ) ;
         return s ;
      }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

class UndoRedo{
   final int max = 50 ;
   Edit[] edits = new Edit[max] ;
   int next = 0 ;

   // --------------------------------------------------------------------------

   void check( String s ){
      if( 0 > next || next > max ){
         if( Env.vbPrint ) System.err.println( s + ": next = " + next + ", max = " + max ) ;
         throw new NullPointerException() ;
      }
      if( Env.vbUndo ) System.err.println( s + ": next = " + next + ", max = " + max ) ;
   }
   // --------------------------------------------------------------------------

   void backupStarts( String s ){
      if( Env.vbUndo ) System.err.println( "--------" ) ;
      check( "before backupStarts()" ) ;

      if( next == max ){
         for( int i = 1 ; i < max ; i++ ){ edits[i-1] = edits[i] ; }
         next-- ;
      }
      for( int i = next ; i < max ; i++ ){ edits[i] = null ; }
      edits[next] = new Edit( s ) ;

      check( "after backupStarts()" ) ;
      if( Env.vbUndo ) System.err.println( "--------" ) ;
   }
   // --------------------------------------------------------------------------

   void backup( Backup  backup ){
      check( "before backup()" ) ;

      edits[next].add( backup ) ;

      check( "after backup()" ) ;
      if( Env.vbUndo ) System.err.println( "--------" ) ;
   }
   // --------------------------------------------------------------------------

   void backupEnds(){
      check( "before backupEnds()" ) ;

      Edit edit = edits[next] ;
      if( edit.size() > 0 ) next++ ;
      else edit = null ;

      check( "after backupEnds()" ) ;
      if( Env.vbUndo ) System.err.println( "--------" ) ;
   }
   // --------------------------------------------------------------------------

   String undo(){
      check( "before undo()" ) ;

      if( next == 0 ){
         if( Env.vbUndo ) System.out.println( "Cannot undo" ) ;
         if( Env.vbUndo ) System.err.println( "Cannot undo" ) ;
         return "Cannot undo" ;
      }
      Edit edit = edits[--next] ;
      if( Env.vbUndo ) System.err.println( "Undo " + edit.what ) ;
      for( int j = edit.size() - 1 ; j >= 0 ; j-- ) edit.get(j).undo() ;

      check( "after undo()" ) ;
      return edit.what ;
   }
   // --------------------------------------------------------------------------

   String redo(){
      check( "before redo()" ) ;

      Edit edit = null ;
      if( next < max ) edit = edits[next] ;
      if( edit == null ){
         if( Env.vbUndo ) System.out.println( "Cannot redo" ) ;
         if( Env.vbUndo ) System.err.println( "Cannot redo" ) ;
         return "Cannot redo" ;
      }
      if( Env.vbUndo ) System.err.println( "Redo " + edit.what ) ;
      for( int j = 0, n = edit.size() ; j < n ; j++ ) edit.get(j).redo() ;
      next++ ;

      check( "after redo()" ) ;
      return edit.what ;
   }
   // --------------------------------------------------------------------------

   @SuppressWarnings("serial")
   private class Edit extends ArrayList<Backup>{
      String what ;
      Edit( String s ){ what = s ; }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

abstract class Backup{
   abstract void undo() ;
   abstract void redo() ;
}
// -----------------------------------------------------------------------------

@SuppressWarnings("serial")
class GLib extends ArrayList<Molecule>{
   final UndoRedo undoRedo = new UndoRedo() ;

   String summary(){
      String sA = "   GLib:" ;
      for( Molecule link : this ){

         String sB = " [" + link.key8 + ": " ;
         for( Residue rsd : link.residueList() ){
            sA += sB + rsd.key8 ;

            String sC = ",new" ;
            if( rsd.mono != null ){
               sA += "," + rsd.mono.key8 + "(" + rsd.mono.key3 + ")" ;

               if( rsd.mono.isStandard() ) sC = ",std" ; else sC = ",own" ;
            }
            sA += sC ;

            sB = "; " ;
         }
         sA += "]" ;
      }
      for( Molecule link : this ){
         if( link.getResidue() == null ){
            sA += Env.bsn + link.linkRecordPDB() ;
         }
      }
      return sA ;
   }

   void undoableAddMols( List<Molecule> list ){
      undoRedo.backupStarts( "add molecules" ) ;
      for( Molecule mol : list ) undoRedo.backup( this.new UndoableAddMolecule( mol ) ) ;
      undoRedo.backupEnds() ;
   }

   Molecule get( String s ){
      for( Molecule link : this ) if( link.key8.equals( s ) ) return link ;
      return null ;
   }

   boolean contains( String s ){
      return get( s ) != null ;
   }

   public void moveAtom( Molecule mol, Atom atom, double[] xyd ){
      for( int i = 0 ; i < 2 ; i++ ) atom.xyz[i] += xyd[i] ;                //:: goes into Atom.move( xyd ) ;
      mol.updateSphere() ;
   }

   public void moveMol( Molecule mol, double[] xyd ){
      for( Residue rsd : mol.residueList() ) rsd.transform.translate( xyd ) ;
      for( int i = 0 ; i < 2 ; i++ ) mol.xyz[i] += xyd[i] ;
      for( Atom atom : mol.atoms ){
         for( int i = 0 ; i < 2 ; i++ ) atom.xyz[i] += xyd[i] ;                //:: goes into Atom.move( xyd ) ;
      }
   }

   public void rotateMol( Molecule mol, double[] xyo, double[] xyn ){
      Rotation rotation = new Rotation( mol.radius, mol.xyz, xyo, xyn ) ;
      for( Residue rsd : mol.residueList() ) rsd.transform.rotate( rotation ) ;
      for( Atom atom : mol.atoms ) rotation.forward( atom.xyz ) ;
      Collections.sort( mol.atoms, Atom.zComparator ) ;
   }

   void undoableClear(){
      undoRedo.backupStarts( "Clear" ) ;
      undoRedo.backup( this.new UndoableClear() ) ;
      undoRedo.backupEnds() ;
   }

   void undoableAddMolecule( String key8, String symbol ){
      undoRedo.backupStarts( "new ligand" ) ;
      Molecule mol = null ;

      Ring ring = Rings.get( symbol ) ;
      if( ring == null ){
         mol = new Molecule( key8, symbol ) ;
      }
      else{
         mol = new Molecule( key8 ) ;
         ring.undoableAddTo( undoRedo, mol, null, new double[]{ 0f, 0f } ) ;   //:: ring.new ....
      }
      undoRedo.backup( new UndoableAddMolecule( mol ) ) ;
      undoRedo.backupEnds() ;
   }

   Molecule undoableAddMolecule( String key8, Monomer comp ){
      Molecule mol = new Molecule( comp ) ;
      mol.key8 = key8 ;
      undoRedo.backupStarts( "add monmomer " + key8 ) ;
      undoRedo.backup( this.new UndoableAddMolecule( mol ) ) ;
      undoRedo.backupEnds() ;
      return mol ;
   }

   void undoableRegularise( Molecule mol1, boolean useLibmol ){
      if( ! this.contains( mol1 ) ) throw new NullPointerException() ;
      if( mol1.restraints != null ) throw new NullPointerException() ;

      boolean lcoorNo = true ;
      Molecule mol2 = mol1.regularise( true, lcoorNo, useLibmol ) ;

      if( mol2.getResidue() != null ) mol2 = new Molecule( new Monomer( mol2 ) ) ;

      undoRedo.backupStarts( "regularise" ) ;
      undoRedo.backup( this.new UndoableRemoveMolecule( mol1 ) ) ;
      undoRedo.backup( this.new UndoableAddMolecule( mol2 ) ) ;
      undoRedo.backupEnds() ;
   }

   void undoableConvertMolecule( Molecule mol1 ){
      if( ! this.contains( mol1 ) ) throw new NullPointerException() ;

      Molecule mol2 = new Molecule( new Monomer( mol1 ) ) ;

      undoRedo.backupStarts( "convert" ) ;
      undoRedo.backup( this.new UndoableRemoveMolecule( mol1 ) ) ;
      undoRedo.backup( this.new UndoableAddMolecule( mol2 ) ) ;
      undoRedo.backupEnds() ;
   }

   void undoableReplaceMolecule( Molecule mol1, Molecule mol2 ){
      if( mol1 == null || mol2 == null || ! this.contains( mol1 ) ) throw new NullPointerException() ;
      undoRedo.backupStarts( "replace " + mol1.key8 + " with " + mol2.key8 ) ;
      undoRedo.backup( this.new UndoableReplaceMolecule( mol1, mol2 ) ) ;
      undoRedo.backupEnds() ;
   }

   void undoableRemoveMolecule( Molecule mol ){
      if( mol == null || ! this.contains( mol ) ) throw new NullPointerException() ;
      undoRedo.backupStarts( "Remove Molecule " + mol.key8 ) ;
      undoRedo.backup( this.new UndoableRemoveMolecule( mol ) ) ;
      undoRedo.backupEnds() ;
   }

   void undoableRenameMolecule( Molecule mol, String key8New, String nameNew ){
      if( mol == null || ! this.contains( mol ) ) throw new NullPointerException() ;

      boolean key8Retained = mol.key8.equals( key8New ) ;
      boolean nameRetained = mol.name.equals( nameNew ) ;
      if( key8Retained && nameRetained ) return ;

      undoRedo.backupStarts( "Rename Molecule " + mol.key8 + " to " + key8New + " " + nameNew ) ;

      if( ! key8Retained ) undoRedo.backup( mol.new UndoableEditKey8( key8New ) ) ;
      if( ! nameRetained ) undoRedo.backup( mol.new UndoableEditName( nameNew ) ) ;

//::  undoRedo.backup( mol.new UndoableClearRestraints() ) ;
//::
      Residue rsd = mol.getResidue() ;
      if( rsd != null ){
         Monomer mono = rsd.mono ;
         if( ! ( mono == null || mono.isStandard() || mol.restraints == null ) ){
            if( ! nameRetained ) undoRedo.backup( rsd.new UndoableEditName( nameNew ) ) ;
            if( ! key8Retained ) undoRedo.backup( rsd.new UndoableEditKey8( key8New ) ) ;
            if( ! nameRetained ) undoRedo.backup( mono.new UndoableEditName( nameNew ) ) ;
            if( ! key8Retained ) undoRedo.backup( mono.new UndoableEditKey8( key8New ) ) ;
            String key3New = key8New.substring( 0, Math.min( 3, key8New.length() ) ) ;
            if( ! key8Retained ) undoRedo.backup( mono.new UndoableEditKey3( key3New ) ) ;
         }
      }
//::
      undoRedo.backupEnds() ;
   }

   Atom undoableReplaceAtom( Molecule mol, Atom atomOld ){
      if( mol == null || ! this.contains( mol ) ) throw new NullPointerException() ;
      if( atomOld == null || ! mol.atoms.contains( atomOld ) ) throw new NullPointerException() ;
      for( Bond bondOld : atomOld ){
         Atom atom = bondOld.another( atomOld ) ;
         if( atom.coupled != atom ) throw new NullPointerException() ;
      }

      undoRedo.backupStarts( "replace atom" ) ;

      Atom atomNew = new Atom( atomOld.residue, atomOld ) ;
      atomOld.coupled = atomNew ;
      if( atomOld.chir != null ) atomNew.chir = atomOld.chir.duplicatePart() ;
      for( Bond bondOld : atomOld ){
         Atom atom = bondOld.another( atomOld ) ;
         Bond bondNew = new Bond( atom, atomNew, bondOld ) ;
         undoRedo.backup( atom.new UndoableReplaceBond( bondOld, bondNew ) ) ;
         atomNew.add( bondNew ) ;
      }
      atomOld.coupled = atomOld ;
      undoRedo.backup( mol.new UndoableRemoveAtom( atomOld ) ) ;
      undoRedo.backup( mol.new UndoableAddAtom( atomNew ) ) ;

      undoRedo.backup( mol.new UndoableClearRestraints() ) ;
      undoRedo.backupEnds() ;
      return atomNew ;
   }

   void movingAtomStopped( Molecule mol, Atom atomNew ){
      if( mol == null || ! this.contains( mol ) ) throw new NullPointerException() ;
      if( atomNew == null || ! mol.atoms.contains( atomNew ) ) throw new NullPointerException() ;
      if( atomNew.chir != null ){
         for( Atom atom : atomNew.chir.atoms ){
            if( ! mol.atoms.contains( atom ) && atom != Atom.NULL ) throw new NullPointerException() ;
         }
         atomNew.chir.setSign().setType() ;
      }
      for( Bond bondNew : atomNew ){
         Atom atomOld = bondNew.another( atomNew ) ;
         if( atomOld.chir != null ){
            for( Atom atom : atomOld.chir.atoms ){
               if( ! mol.atoms.contains( atom ) && atom != Atom.NULL ) throw new NullPointerException() ;
            }
            atomOld.chir.setSign().setType() ;
         }
      }
   }

   void undoableRemoveAtom( Molecule mol, Atom atom1 ){
      undoRedo.backupStarts( "remove atom" ) ;

      if( mol.atoms.indexOf( atom1 ) < 0 ) throw new NullPointerException() ;
      for( Bond bond : atom1 ){
         Atom atom2 = bond.another( atom1 ) ;
         undoRedo.backup( atom2.new UndoableRemoveBond( bond ) ) ;
         if( atom2.size() == 0 && atom2.symbol.equals( "H" ) ){
            undoRedo.backup( mol.new UndoableRemoveAtom( atom2 ) ) ;
         }
      }
      undoRedo.backup( mol.new UndoableRemoveAtom( atom1 ) ) ;
      if( mol.atoms.size() == 0 ) undoRedo.backup( this.new UndoableRemoveMolecule( mol ) ) ;

      undoRedo.backup( mol.new UndoableClearRestraints() ) ;
      undoRedo.backupEnds() ;
   }

   Atom undoableAddAtom( Molecule mol, Atom atom1, String symbol, double[] xy ){
      Ring ring = Rings.get( symbol ) ;
      if( ring == null ){
         undoRedo.backupStarts( "add atom" ) ;

         Atom atom2 = new Atom( atom1.residue, symbol, symbol, xy[0], xy[1], atom1.xyz[2] ) ;
         Set<String> set = new LinkedHashSet<String>( mol.atomView( atom1 ).keySet() ) ;            //:: not nice
         atom2.generateKey( set ) ;

         Bond bond = new Bond( atom1, atom2, "single" ) ;
         undoRedo.backup( mol.new UndoableAddAtom( atom2 ) ) ;
         undoRedo.backup( atom1.new UndoableAddBond( bond ) ) ;
         undoRedo.backup( atom2.new UndoableAddBond( bond ) ) ;

         undoRedo.backup( mol.new UndoableClearRestraints() ) ;
         undoRedo.backupEnds() ;
         return atom2 ;
      }
      else{
         undoRedo.backupStarts( "add ring" ) ;

         ring.undoableAddTo( undoRedo, mol, atom1, xy ) ;                 //:: ring.new ....

         undoRedo.backup( mol.new UndoableClearRestraints() ) ;
         undoRedo.backupEnds() ;
         return atom1 ;
      }
   }

   void undoableChangeAtom(
      String chirTypeNew,
      Molecule mol, Atom atom, String key4New, String symbolNew, float chargeNew
   ){
      if( mol == null || ! this.contains( mol ) ) throw new NullPointerException() ;
      if( atom == null || ! mol.atoms.contains( atom ) ) throw new NullPointerException() ;

      boolean key4Retained = atom.key4.equals( key4New ) ;
      boolean typeRetained = atom.symbol.equals( symbolNew ) ;
      boolean chargeRetained = atom.charge.equals( chargeNew ) ;
      boolean chirRetained = true ;
      if( atom.chir != null ) chirRetained = atom.chir.type.equals( chirTypeNew ) ;
      if( key4Retained && typeRetained && chargeRetained && chirRetained ) return ;

      undoRedo.backupStarts( "Edit Atom" ) ;

      if( ! key4Retained ) undoRedo.backup( atom.new UndoableEditKey4( key4New ) ) ;
      if( ! typeRetained ) undoRedo.backup( atom.new UndoableEditSymbol( symbolNew ) ) ;
      if( ! chargeRetained ) undoRedo.backup( atom.new UndoableEditCharge( chargeNew ) ) ;
      if( ! chirRetained ) undoRedo.backup( atom.chir.new UndoableEditType( chirTypeNew ) ) ;

      boolean chirSignRetained = true ;
      if( atom.chir != null ) chirSignRetained = atom.chir.isConsistent() ;
      if( ! ( typeRetained && chargeRetained && chirSignRetained ) ){
         undoRedo.backup( mol.new UndoableClearRestraints() ) ;
      }
      undoRedo.backupEnds() ;
   }

   void undoableRemoveBond( Molecule mol, Bond bond ){
      undoRedo.backupStarts( "remove bond" ) ;

      Atom atom1 = bond.atom1 ;
      Atom atom2 = bond.atom2 ;
      if( mol.atoms.indexOf( atom1 ) < 0 || mol.atoms.indexOf( atom2 ) < 0 ) throw new NullPointerException() ;
      undoRedo.backup( atom1.new UndoableRemoveBond( bond ) ) ;
      undoRedo.backup( atom2.new UndoableRemoveBond( bond ) ) ;
      if( atom1.size() == 0 && atom1.symbol.equals( "H" ) ){
         undoRedo.backup( mol.new UndoableRemoveAtom( atom1 ) ) ;
      }
      if( atom2.size() == 0 && atom2.symbol.equals( "H" ) ){
         undoRedo.backup( mol.new UndoableRemoveAtom( atom2 ) ) ;
      }
      if( mol.atoms.size() == 0 ) undoRedo.backup( this.new UndoableRemoveMolecule( mol ) ) ;

      undoRedo.backup( mol.new UndoableClearRestraints() ) ;
      undoRedo.backupEnds() ;
   }

   void undoableAddBond( Molecule mol, Atom atom1, Atom atom2, String type ){

      Bond bond = new Bond( atom1, atom2, type ) ;

      undoRedo.backupStarts( "add bond" ) ;
      undoRedo.backup( atom1.new UndoableAddBond( bond ) ) ;
      undoRedo.backup( atom2.new UndoableAddBond( bond ) ) ;

      undoRedo.backup( mol.new UndoableClearRestraints() ) ;
      undoRedo.backupEnds() ;
   }

   void undoableAddBond( String key8, Molecule mol1, Atom atom1, Molecule mol2, Atom atom2, String type ){
      if( mol1 == mol2 ) throw new NullPointerException() ;
      if( atom1.residue.cootId != atom2.residue.cootId ) throw new NullPointerException() ;

      atom1.residue.isFirst = true ;
      atom2.residue.isFirst = false ;

      Bond bond = new Bond( atom1, atom2, type ) ;
      Molecule link = new Molecule() ;
      link.atoms.addAll( mol1.atoms ) ;
      link.atoms.addAll( mol2.atoms ) ;
      link.key8 = key8 ;
      link.name = mol1.getResidue().key8 + "-" + mol2.getResidue().key8 ;

      double z1 = atom1.xyz[2] ;
      double z2 = atom2.xyz[2] ;
      for( Atom atom : mol1.atoms ) atom.xyz[2] -= z1 ;
      for( Atom atom : mol2.atoms ) atom.xyz[2] -= z2 ;
      atom1.residue.transform.translate( - z1 ) ;
      atom2.residue.transform.translate( - z2 ) ;

      undoRedo.backupStarts( "add link" ) ;
      undoRedo.backup( this.new UndoableRemoveMolecule( mol1 ) ) ;
      undoRedo.backup( this.new UndoableRemoveMolecule( mol2 ) ) ;
      undoRedo.backup( this.new UndoableAddMolecule( link ) ) ;
      undoRedo.backup( atom1.new UndoableAddBond( bond ) ) ;
      undoRedo.backup( atom2.new UndoableAddBond( bond ) ) ;

      undoRedo.backup( link.new UndoableClearRestraints() ) ;           //:: ???
      undoRedo.backupEnds() ;
   }

   void undoableChangeBond( Molecule mol, Bond bond, String typeNew ){

      Atom atom1 = bond.atom1 ;
      Atom atom2 = bond.atom2 ;
      Bond bondNew = new Bond( atom1, atom2, typeNew ) ;

      undoRedo.backupStarts( "Edit Bond" ) ;
      undoRedo.backup( atom1.new UndoableRemoveBond( bond ) ) ;
      undoRedo.backup( atom2.new UndoableRemoveBond( bond ) ) ;
      undoRedo.backup( atom1.new UndoableAddBond( bondNew ) ) ;
      undoRedo.backup( atom2.new UndoableAddBond( bondNew ) ) ;
      undoRedo.backup( mol.new UndoableClearRestraints() ) ;
      undoRedo.backupEnds() ;
   }

   // --------------------------------------------------------------------------

   class UndoableClear extends Backup{
      List<Molecule> list = new ArrayList<Molecule>( size() ) ;
      UndoableClear(){
         Iterator<Molecule> i = iterator() ;
         while( i.hasNext() ) list.add( i.next() ) ;
         clear() ;
         if( Env.vbPrint ) System.out.println( ":::: GLib clear" ) ;
      }
      void undo(){
         addAll( list ) ;
      }
      void redo(){
         clear() ;
      }
   }
   // --------------------------------------------------------------------------

   class UndoableAddMolecule extends Backup{
      Molecule mol ;
      UndoableAddMolecule( Molecule molNew ){
         mol = molNew ;
         add( mol ) ;
         mol.updateSphere() ;
         if( Env.vbPrint ) System.out.println( ":::: GLib add " + mol.key8 ) ;
      }
      void undo(){
         remove( mol ) ;
      }
      void redo(){
         add( mol ) ;
         mol.updateSphere() ;
      }
   }
   // --------------------------------------------------------------------------

   class UndoableRemoveMolecule extends Backup{
      Molecule mol ;
      UndoableRemoveMolecule( Molecule molOld ){
         mol = molOld ;
         remove( mol ) ;
         if( Env.vbPrint ) System.out.println( ":::: GLib remove molecule " + mol.key8 ) ;
      }
      void undo(){
         add( mol ) ;
         mol.updateSphere() ;
      }
      void redo(){
         remove( mol ) ;
      }
   }
   // --------------------------------------------------------------------------

   class UndoableReplaceMolecule extends Backup{
      Molecule molOld, molNew ;
      UndoableReplaceMolecule( Molecule molOld, Molecule molNew ){
         this.molOld = molOld ;
         this.molNew = molNew ;
         for( Atom atom : molOld.atoms ) atom.residue.transform.backward( atom.xyz ) ;
         remove( molOld ) ;
         add( molNew ) ;
         molNew.updateSphere() ;
         if( Env.vbPrint ) System.out.println( ":::: GLib replace " + molOld.key8 + " " + molNew.key8 ) ;
      }
      void undo(){
         for( Atom atom : molOld.atoms ) atom.residue.transform.forward( atom.xyz ) ;
         for( Atom atom : molNew.atoms ) atom.residue.transform.backward( atom.xyz ) ;
         remove( molNew ) ;
         add( molOld ) ;
         molOld.updateSphere() ;
      }
      void redo(){
         for( Atom atom : molOld.atoms ) atom.residue.transform.backward( atom.xyz ) ;
         for( Atom atom : molNew.atoms ) atom.residue.transform.forward( atom.xyz ) ;
         remove( molOld ) ;
         add( molNew ) ;
         molNew.updateSphere() ;
      }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

class Transform{
   protected double[][] rot = new double[][]{ { 1f, 0f, 0f }, { 0f, 1f, 0f }, { 0f, 0f, 1f }, } ;
   protected double[] tra = new double[]{ 0f, 0f, 0f } ;
   private double[] rw = new double[3] ;

   void forward( double[] ru ){
      for( int j = 0 ; j < 3 ; j++ ) rw[j] = ru[j] ;
      for( int i = 0 ; i < 3 ; i++ ) ru[i] = tra[i] ;
      for( int i = 0 ; i < 3 ; i++ ){
         for( int j = 0 ; j < 3 ; j++ ) ru[i] += rot[i][j]* rw[j] ;
      }
   }

   void backward( double[] ru ){
      for( int j = 0 ; j < 3 ; j++ ) rw[j] = ru[j] - tra[j] ;
      for( int i = 0 ; i < 3 ; i++ ) ru[i] = 0f ;
      for( int i = 0 ; i < 3 ; i++ ){
         for( int j = 0 ; j < 3 ; j++ ) ru[i] += rot[j][i]* rw[j] ;
      }
   }

   void translate( double[] xy ){
      for( int i = 0 ; i < 2 ; i++ ) tra[i] += xy[i] ;
   }

   void translate( double z ){
      tra[2] += z ;
   }

   void rotate( Rotation rr ){
      for( int j = 0 ; j < 3 ; j++ ) rw[j] = tra[j] ;
      for( int i = 0 ; i < 3 ; i++ ) tra[i] = rr.tra[i] ;
      for( int i = 0 ; i < 3 ; i++ ){
         for( int j = 0 ; j < 3 ; j++ ) tra[i] += rr.rot[i][j]* rw[j] ;
      }
      for( int i = 0 ; i < 3 ; i++ ){
         for( int k = 0 ; k < 3 ; k++ ) rw[k] = rot[k][i] ;
         for( int j = 0 ; j < 3 ; j++ ) rot[j][i] = 0f ;
         for( int j = 0 ; j < 3 ; j++ ){
            for( int k = 0 ; k < 3 ; k++ ) rot[j][i] += rr.rot[j][k]* rw[k] ;
         }
      }
   }
}

// -----------------------------------------------------------------------------

class Rotation extends Transform{
   private double[] ro = new double[2] ;
   private double[] rn = new double[2] ;
   private double roro, roror ;
   private double rnrn, rnrnr ;
   private double so1, so2, so3 ;
   private double sn1, sn2, sn3 ;
   private double sp1, sp2, sp3, spsp, spspr ;
   private double sm1, sm2, sm3, smsm, spsmr ;
   private double u1, u2, u3, a, b ;

// void newRotation( double radius, double[] centre, double[] xyo, double[] xyn ){
   Rotation( double radius, double[] centre, double[] xyo, double[] xyn ){

      for( int i = 0 ; i < 2 ; i++ ){
         ro[i] = ( xyo[i] - centre[i] )/ radius ;
         rn[i] = ( xyn[i] - centre[i] )/ radius ;
      }

//    old coordinates on the sphere:

      roro = ( ro[0]* ro[0] ) + ( ro[1]* ro[1] ) ;
      roror = 1/ ( roro + 1 ) ;

      so1 = 2* ro[0]* roror ;
      so2 = 2* ro[1]* roror ;
      so3 = ( roro - 1 )* roror ;

//    new coordinates on the sphere:

      rnrn = ( rn[0]* rn[0] ) + ( rn[1]* rn[1] ) ;
      rnrnr = 1/ ( rnrn + 1 ) ;

      sn1 = 2* rn[0]* rnrnr ;
      sn2 = 2* rn[1]* rnrnr ;
      sn3 = ( rnrn - 1 )* rnrnr ;

//    matrix:

      sp1 = ( sn1 + so1 )/ 2 ;
      sp2 = ( sn2 + so2 )/ 2 ;
      sp3 = ( sn3 + so3 )/ 2 ;

      sm1 = ( sn1 - so1 )/ 2 ;
      sm2 = ( sn2 - so2 )/ 2 ;
      sm3 = ( sn3 - so3 )/ 2 ;

      spsp = ( sp1* sp1 ) + ( sp2* sp2 ) + ( sp3* sp3 ) ;
      smsm = ( sm1* sm1 ) + ( sm2* sm2 ) + ( sm3* sm3 ) ;
      spspr = 1/ spsp;
      spsmr = 1/ ( spsp + smsm ) ;

      u1 = ( sp2* sm3 - sp3* sm2 )* spspr ;
      u2 = ( sp3* sm1 - sp1* sm3 )* spspr ;
      u3 = ( sp1* sm2 - sp2* sm1 )* spspr ;

      a = ( spsp - smsm )* spsmr ;
      b = 2* spsp* spsmr ;

      rot[0][0] = u1* u1* b + a ;
      rot[1][1] = u2* u2* b + a ;
      rot[2][2] = u3* u3* b + a ;

      rot[0][1] = ( u1* u2 - u3 )* b ;
      rot[1][0] = ( u2* u1 + u3 )* b ;

      rot[1][2] = ( u2* u3 - u1 )* b ;
      rot[2][1] = ( u3* u2 + u1 )* b ;

      rot[2][0] = ( u3* u1 - u2 )* b ;
      rot[0][2] = ( u1* u3 + u2 )* b ;

      for( int i = 0 ; i < 3 ; i++ ) tra[i] = centre[i] ;
      for( int i = 0 ; i < 3 ; i++ ){
         for( int j = 0 ; j < 3 ; j++ ) tra[i] -= rot[i][j]* centre[j] ;
      }
   }
}
// -----------------------------------------------------------------------------

class CifLib{
   private Set<String> usedCompSet = new LinkedHashSet<String>() ;
   private Set<String> usedModSet = new LinkedHashSet<String>() ;
   private Set<String> stdCompSet = new LinkedHashSet<String>() ;
   private Map<String,Monomer> compMap = new LinkedHashMap<String,Monomer>() ;
   private Map<String,Molecule>  modMap = new LinkedHashMap<String,Molecule>() ;
   private Map<String,Molecule> linkMap = new LinkedHashMap<String,Molecule>() ;

   private AtomMethods atomMethods = new AtomMethods() ;
   private List<CifRestraint> rstTemplates = new ArrayList<CifRestraint>(5) ;

   public String toString(){ throw new NullPointerException() ; }

   public CifLib(){
      rstTemplates.add( new CifBond  () ) ;
      rstTemplates.add( new CifAngle () ) ;
      rstTemplates.add( new CifTor   () ) ;
      rstTemplates.add( new CifChir  () ) ;
      rstTemplates.add( new CifPlane () ) ;
   }

   public CifLib( File file ){
      rstTemplates.add( new CifBond  () ) ;
      rstTemplates.add( new CifAngle () ) ;
      rstTemplates.add( new CifTor   () ) ;
      rstTemplates.add( new CifChir  () ) ;
      rstTemplates.add( new CifPlane () ) ;

      CifFile cif = new CifFile( new CharArray( file ) ) ;
      CifLibReader clr = new CifLibReader() ;
      clr.fromCif( cif, "comp", new CifCompReader() ) ;
      clr.fromCif( cif, "mod" , new CifModReader () ) ;
      clr.fromCif( cif, "link", new CifLinkReader() ) ;
      if( ! usedModSet.equals( modMap.keySet() ) ) throw new NullPointerException() ;

/*
      if( Env.vbPrint ) System.err.println( "   stdCompSet: " + stdCompSet ) ;
      if( Env.vbPrint ) System.err.println( "      compMap: " + compMap.keySet() ) ;
      if( Env.vbPrint ) System.err.println( "       modMap: " +  modMap.keySet() ) ;
      if( Env.vbPrint ) System.err.println( "      linkMap: " + linkMap.keySet() ) ;
      if( Env.vbPrint ) System.err.println( "   ----------- " ) ;

      if( Env.vbPrint ) System.out.println( "   stdCompSet: " + stdCompSet ) ;
      if( Env.vbPrint ) System.out.println( "      compMap: " + compMap.keySet() ) ;
      if( Env.vbPrint ) System.out.println( "       modMap: " +  modMap.keySet() ) ;
      if( Env.vbPrint ) System.out.println( "      linkMap: " + linkMap.keySet() ) ;
      if( Env.vbPrint ) System.out.println( "   ----------- " ) ;
*/
   }

   public List<Molecule> polymerView(){
      List<Molecule> list = new ArrayList<Molecule>() ;
      for( Monomer comp : compMap.values() ){
         if( ! usedCompSet.contains( comp.key8 ) ) list.add( new Molecule( comp ) ) ;
      }
      list.addAll( linkMap.values() ) ;
      return list ;
   }

   public List<Monomer> monomerView(){
      return new ArrayList<Monomer>( compMap.values() ) ;
   }

   public Monomer getMonomer(){
      int cou0 = modMap.size() + linkMap.size() ;
      int cou1 = compMap.size() ;
      if( ! ( cou0 == 0 && cou1 == 1 ) ) throw new NullPointerException() ;
      return compMap.values().iterator().next() ;
   }

   public Monomer getMonomer( String key ){                                ///::: to be removed
      return compMap.get( key ) ;
   }

   public CharArray toBuffer(){
      return new CifLibWriter().headerBuffer ;
   }

   public void add( Monomer monomer ){
      if( compMap.get( monomer.key8 ) != null ) throw new NullPointerException() ;
      if( stdCompSet.contains( monomer.key8 ) ) throw new NullPointerException() ;

      compMap.put( monomer.key8, monomer ) ;

//::  class Restraints to be rearranged to include sort
      for( int i = 0 ; i < 5 ; i++ ){
         Collections.sort( monomer.restraints.get( i ) ) ;
      }
   }

   public String add( Molecule molecule ){
      if( molecule.restraints == null ) throw new NullPointerException() ;
      String info = "" ;
      String warn = "" ;

      Collections.sort( molecule.atoms, Atom.originComparator ) ;

      Map<Residue,Molecule> map = molecule.residueView() ;

      if( map.size() == 2 ){
         if( linkMap.put( molecule.key8, molecule ) != null ) throw new NullPointerException() ;

         for( Map.Entry<Residue,Molecule> entry : map.entrySet() ){

            Residue rsd = entry.getKey() ;
            Monomer comp = rsd.mono ;

            if( compMap.get( comp.key8 ) != null ) throw new NullPointerException() ;

            if( comp.isStandard() ){
               stdCompSet.add( comp.key8 ) ;
            }
            else{
               if( stdCompSet.contains( comp.key8 ) ) throw new NullPointerException() ;
               if( compMap.get( comp.key8 ) != null ) throw new NullPointerException() ;
//             if( StdLib.hasReady( comp.key8 ) ) throw new NullPointerException() ;
               compMap.put( comp.key8, comp ) ;
               info += Env.bsn + "new monomer '" + comp.key8 + "' has been added to the library file" ;
//             if( StdLib.hasReady( comp.key8 ) ){
               if( StdLib.compHeaderMap.get( comp.key8 ) != null ){
                  warn += Env.bsn + "'" + comp.key8 ;
                  warn += "' will override the standard library entry with the same ID" ;
               }
            }

            int cou = 0 ;
            do{ rsd.key8 = rsd.mono.key3 + "mod" + ++cou ; } while( cou < 100 && modMap.get( rsd.key8 ) != null ) ;
            if( cou > 99 ) throw new NullPointerException() ;

            Molecule mod = entry.getValue() ;
            mod.name = rsd.name ;
            mod.key8 = rsd.key8 ;

            modMap.put( mod.key8, mod ) ;
            usedModSet.add( mod.key8 ) ;
         }
      }
/*
      else if( map.size() == 1 ){
         if( compMap.get( molecule.key8 ) != null ) throw new NullPointerException() ;
         if( stdCompSet.contains( molecule.key8 ) ) throw new NullPointerException() ;
         Residue rsd = new Residue( molecule ) ;
         compMap.put( molecule.key8, rsd.mono ) ;
      }
*/
      else throw new NullPointerException() ;

      Collections.sort( molecule.atoms, Atom.zComparator ) ;

      if( ! info.equals( "" ) ){
         info = Env.bsn + Env.bsn + "In addition:" + info ;
         if( ! warn.equals( "" ) ){
            info += Env.bsn + Env.bsn + "WARNING:" + Env.bsn ;
            info += Env.bsn + "If this library will be used then:" + warn + Env.bsn ;
         }
      }
//    if( Env.vbPrint ) System.err.println( info ) ;
      return info ;
   }

   // --------------------------------------------------------------------------

   private class CifLibReader{
      void fromCif( CifFile cif, String type, CifEntryReader cer ){
         String tag = "data_" + type + "_" ;

         CifData header = cif.get( tag + "list" ) ;
         if( header == null ) return ;

         Iterator<CifLoop> loopIterator = header.iterator() ;
         if( ! loopIterator.hasNext() ) return ;
         CifLoop loop = loopIterator.next() ;

         while( loop.hasNext() ){
            String key = (String)loop.next() ;
            CifData data = cif.get( tag + key ) ;

            cer.fromCif( key, loop, data ) ;
         }
      }
   }
   // --------------------------------------------------------------------------

   private abstract class CifEntryReader{
      abstract void fromCif( String key, CifLoop loop, CifData data ) ;
   }
   // --------------------------------------------------------------------------

/*
   private class CifLibWriter1{
      private final String bsn = Env.bsn ;
      private CharArray headerBuffer = new CharArray() ;

      CifLibWriter1(){
         headerBuffer.append( Versions.cifHeaderInfo() ) ;
         this.append( "comp", new CifCompWriter() ) ;
         this.append( "mod" , new CifModWriter () ) ;
         this.append( "link", new CifLinkWriter() ) ;
      }

      void append( String type, CifEntryWriter cew ){
         String tag = "data_" + type + "_" ;

         CifData header = CifFile.template.get( tag + "list" ) ;
         CifLoop loop = header.iterator().next() ;

         CifData data = CifFile.template.get( tag ) ;
         CharArray dataBuffer = new CharArray() ;

         int cou = 0 ;
         while( cew.hasNext() ){
            String key = cew.toCif( loop, data ) ;
            if( key != null ){
               cou ++ ;
               dataBuffer.append( tag + key + bsn ) ;
               data.appendTo( dataBuffer ) ;
               data.clear() ;
               dataBuffer.append( bsn ) ;
            }
         }

         if( cou > 0 ){
            headerBuffer.append( tag + "list" + bsn ) ;
            header.appendTo( headerBuffer ) ;
            header.clear() ;
            headerBuffer.append( bsn ) ;
            headerBuffer.append( dataBuffer ) ;
         }
      }

      public String toString(){
         return headerBuffer.toString() ;
      }
   }
   // --------------------------------------------------------------------------
*/

   private class CifLibWriter{
      private final String bsn = Env.bsn ;
      private CharArray headerBuffer = new CharArray() ;
//2
      private CharArray dataBuffer = new CharArray() ;
//

      CifLibWriter(){
         headerBuffer.append( Versions.cifHeaderInfo() ) ;
         this.append( "comp", new CifCompWriter() ) ;
         this.append( "mod" , new CifModWriter () ) ;
         this.append( "link", new CifLinkWriter() ) ;
//2
         headerBuffer.append( dataBuffer ) ;
//2
      }

      void append( String type, CifEntryWriter cew ){
         String tag = "data_" + type + "_" ;

         CifData header = CifFile.template.get( tag + "list" ) ;
         CifLoop loop = header.iterator().next() ;

         CifData data = CifFile.template.get( tag ) ;
//1      CharArray dataBuffer = new CharArray() ;

         int cou = 0 ;
         while( cew.hasNext() ){
            String key = cew.toCif( loop, data ) ;
            if( key != null ){
               cou ++ ;
               dataBuffer.append( tag + key + bsn ) ;
               data.appendTo( dataBuffer ) ;
               data.clear() ;
               dataBuffer.append( bsn ) ;
            }
         }

         if( cou > 0 ){
            headerBuffer.append( tag + "list" + bsn ) ;
            header.appendTo( headerBuffer ) ;
            header.clear() ;
            headerBuffer.append( bsn ) ;
//1         headerBuffer.append( dataBuffer ) ;
         }
      }

      public String toString(){
         return headerBuffer.toString() ;
      }
   }
   // --------------------------------------------------------------------------

   private abstract class CifEntryWriter{
      abstract boolean hasNext() ;
      abstract String toCif( CifLoop loop, CifData data ) ;
   }
   // --------------------------------------------------------------------------

   private class CifCompReader extends CifEntryReader{

      void fromCif( String key, CifLoop header, CifData data ){
         if( compMap.containsKey( key ) ) throw new NullPointerException() ;

         final Monomer comp = new Monomer() ;
         comp.key8 = key ;
         comp.key3 = (String)header.next() ;
         comp.name = (String)header.next() ;
         comp.group = (String)header.next() ;

//       Integer na = (Integer)header.next() ;
//       Integer nh = (Integer)header.next() ;
//       String level = (String)header.next() ;
         header.next() ;
         header.next() ;
         header.next() ;

         compMap.put( key, comp ) ;

         Iterator<CifLoop> dataIterator = data.iterator() ;
         CifLoop loop = dataIterator.next() ;
         final AtomMap atoms = new AtomMap() ;
         while( loop.hasNext() ){
            if( comp.key8.equals( loop.next() ) == false ) throw new NullPointerException() ;
            Atom ga = new Atom( (Residue)null ) ;              //::     null !!!
//          Atom atom = ga ;
            atomMethods.fromCif( loop, ga, false ) ;
            atoms.put( ga.key4, ga ) ;
            comp.atoms.add( ga ) ;
         }

         RestraintAtomReader rar = new RestraintAtomReader(){
            void readFun( CifLoop loop ){
               if( ! comp.key8.equals( (String)loop.next() ) ) throw new NullPointerException() ;
            }
            Atom readAtom( CifLoop loop, boolean flagChir ){
               Object o = loop.next() ;
               Atom atom = null ;
               if( flagChir && o == null ) atom = Atom.NULL ;
               else atom = atoms.get( o ) ;
               if( atom == null ) throw new NullPointerException( o.toString() ) ;
               return atom ;
            }
         } ;

         comp.restraints = new Restraints() ;
         for( int i = 0 ; i < 5 ; i++ ){
            loop = dataIterator.next() ;
            List<GenericRestraint> list = comp.restraints.get(i) ;
            CifRestraint template = rstTemplates.get(i) ;
            while( loop.hasNext() ) list.add( template.fromCif( loop, rar ) ) ;
         }
         for( GenericRestraint rst : comp.restraints.get(0) ){
            Bond bond = new Bond( rst ) ;
            bond.atom1.add( bond ) ;
            bond.atom2.add( bond ) ;
         }
         for( GenericRestraint chir : comp.restraints.get(3) ){
            chir.atoms.get(0).chir = chir ;
         }
      }
   }
   // --------------------------------------------------------------------------

   private class CifCompWriter extends CifEntryWriter{
      private Iterator<Monomer> iterator = compMap.values().iterator() ;

      boolean hasNext(){ return iterator.hasNext() ; }

      String toCif( CifLoop header, CifData data ){
         final Monomer comp = iterator.next() ;
         for( Atom atom : comp.atoms ) if( atom.residue != null ) throw new NullPointerException() ;
         header.add( comp.key8 ) ;
         header.add( comp.key3 ) ;
         header.add( comp.name ) ;
//xx     header.add( comp.group ) ;
         header.add( null ) ;
//xx
         int nh = 0 ;
         for( Atom atom : comp.atoms ) if( ! atom.symbol.equals("H") ) nh++ ;
         header.add( new Integer( comp.atoms.size() ) ) ;
         header.add( new Integer( nh ) ) ;

//++     header.add( comp.restraints == null ? "M" : "C" ) ;
         header.add( comp.restraints.get(1).size() == 0 ? "M" : null ) ;
//++

         Iterator<CifLoop> dataIterator = data.iterator() ;
         CifLoop loop = dataIterator.next() ;
         for( Atom atom : comp.atoms ){
            loop.add( comp.key8 ) ;
            atomMethods.toCif( loop, atom, false ) ;
         }
         RestraintAtomWriter raw = new RestraintAtomWriter(){
            void writeFun( CifLoop loop ){
               loop.add( comp.key8 ) ;
            }
            void writeAtom( Atom atom, CifLoop loop ){
               loop.add( atom.key4 ) ;
            }
         } ;

         for( int i = 0 ; i < 5 ; i++ ){
            loop = dataIterator.next() ;
            List<GenericRestraint> list = comp.restraints.get(i) ;
            CifRestraint template = rstTemplates.get(i) ;
            for( GenericRestraint rst : list ) template.toCif( loop, raw, rst ) ;
         }
         return comp.key8 ;
      }
   }
   // --------------------------------------------------------------------------

   private class CifModReader extends CifEntryReader{

      void fromCif( String key, CifLoop header, CifData data ){
         if( modMap.containsKey( key ) ) throw new NullPointerException() ;

         final Residue rsd = new Residue() ;
         rsd.key8 = key ;
         rsd.name = (String)header.next() ;
         String monKey = (String)header.next() ;
         rsd.group = (String)header.next() ;

         rsd.mono = compMap.get( monKey ) ;
         if( rsd.mono == null ){
//          boolean lcoorNo = true ;
            boolean lcoorNo = false ;
            rsd.mono = StdLib.get( monKey, lcoorNo ) ;
            if( rsd.mono == null ) throw new NullPointerException() ;
            stdCompSet.add( monKey ) ;
         }
         else{
            usedCompSet.add( monKey ) ;
         }
         if( ! rsd.mono.key8.equals( monKey ) ) throw new NullPointerException() ;

         Molecule link = new Molecule() ;
         link.key8 = rsd.key8 ;
         link.name = rsd.name ;
         modMap.put( key, link ) ;

         Iterator<CifLoop> dataIterator = data.iterator() ;
         CifLoop loop = dataIterator.next() ;

         final AtomMap rawAtoms = new AtomMap( rsd.mono.atoms ) ;
         final AtomMap oldAtoms = new AtomMap() ;
         final AtomMap newAtoms = new AtomMap() ;

         while( loop.hasNext() ){
            if( rsd.key8.equals( loop.next() ) == false ) throw new NullPointerException() ;
            String fun = (String)loop.next() ;
            String key4 = (String)loop.next() ;

            if( fun.equals( "add" ) ){
               Atom gAtom = new Atom( rsd ) ;
//             Atom atom = gAtom ;
               atomMethods.fromCif( loop, gAtom, true ) ;

               newAtoms.put( gAtom.key4, gAtom ) ;
               link.atoms.add( gAtom ) ;
            }
            else if( fun.equals( "delete" ) ){
               Atom orig = rawAtoms.remove( key4 ) ;
               Atom gAtom = new Atom( null, orig ) ;
//             Atom atom = gAtom ;
               atomMethods.fromCif( loop, gAtom, true ) ;

                  orig.coupled = gAtom ;

               oldAtoms.put( key4, gAtom ) ;
            }
            else if( fun.equals( "change" ) ){
               Atom orig = rawAtoms.remove( key4 ) ;
               Atom gAtom = new Atom( rsd, orig ) ;
//             Atom atom = gAtom ;
               atomMethods.diffFromCif( loop, orig, gAtom ) ;

                  orig.coupled = gAtom ;

               oldAtoms.put( key4, gAtom ) ;
               newAtoms.put( gAtom.key4, gAtom ) ;
               link.atoms.add( gAtom ) ;
            }
         }
         for( Atom orig : rawAtoms.values() ){
            Atom gAtom = new Atom( rsd, orig ) ;
//          Atom atom = gAtom ;

               orig.coupled = gAtom ;

            oldAtoms.put( gAtom.key4, gAtom ) ;
            newAtoms.put( gAtom.key4, gAtom ) ;
            link.atoms.add( gAtom ) ;
         }

         RestraintAtomReader rar = new RestraintAtomReader(){
            void readFun( CifLoop loop ){
               if( ! rsd.key8.equals( (String)loop.next() ) ) throw new NullPointerException() ;
               function = (String)loop.next() ;
            }
            Atom readAtom( CifLoop loop, boolean flagChir ){
               if( function == null ) throw new NullPointerException() ;
               Object o = loop.next() ;
               Atom atom = null ;
               if( flagChir && o == null ) atom = Atom.NULL ;
               else if( function.equals( "add" ) ) atom = newAtoms.get(o) ;
               else if( function.equals( "delete" ) ) atom = oldAtoms.get(o) ;
               else if( function.equals( "change" ) ) atom = newAtoms.get(o) ;
               else throw new NullPointerException() ;
               if( atom == null ) throw new NullPointerException() ;
               return atom ;
            }
         } ;

         link.restraints = new Restraints( rsd.mono.restraints ) ;
         for( Atom atom : rsd.mono.atoms ) atom.coupled = atom ;

         for( int i = 0 ; i < 5 ; i++ ){
            loop = dataIterator.next() ;
            List<GenericRestraint> list = link.restraints.get(i) ;
            CifRestraint template = rstTemplates.get(i) ;

            while( loop.hasNext() ){
               GenericRestraint rst = template.fromCif( loop, rar ) ;
               int j = list.indexOf( rst ) ;
               if( rar.function.equals( "add" ) ){
                  if( j >= 0 ) throw new NullPointerException() ;
                  list.add( rst ) ;
               }
               else if( rar.function.equals( "delete" ) ){
                  if( j < 0 ) throw new NullPointerException() ;
                  list.remove( j ) ;
               }
               else if( rar.function.equals( "change" ) ){
                  if( j < 0 ) throw new NullPointerException() ;
                  GenericRestraint rstOld = list.get( j ) ;
                  GenericRestraint rstNew = rstOld.applyDiff( rst ) ;
                  list.set( j, rstNew ) ;
               }
            }
            Iterator<GenericRestraint> rstIterator = list.iterator() ;
            while( rstIterator.hasNext() ){
               GenericRestraint rst = rstIterator.next() ;
               if( rst.getResidue() == null ){
                  rstIterator.remove() ;
               }
            }
         }
         for( GenericRestraint rst : link.restraints.get(0) ){
            Bond bond = new Bond( rst ) ;
            bond.atom1.add( bond ) ;
            bond.atom2.add( bond ) ;
         }
         for( GenericRestraint chir : link.restraints.get(3) ){
            chir.atoms.get(0).chir = chir ;
         }
      }
   }
   // --------------------------------------------------------------------------

   private class CifModWriter extends CifEntryWriter{
      private Iterator<Molecule> iterator = modMap.values().iterator() ;

      boolean hasNext(){ return iterator.hasNext() ; }

      String toCif( CifLoop header, CifData data ){
         final Molecule link = iterator.next() ;
         final Residue rsd = link.atoms.get(0).residue ;
         for( Atom atom : link.atoms ) if( atom.residue != rsd ) throw new NullPointerException() ;

         Iterator<CifLoop> dataIterator = data.iterator() ;
         CifLoop loop = dataIterator.next() ;

         List<Atom> monAtoms = new ArrayList<Atom>() ;
         for( Atom orig : rsd.mono.atoms ) monAtoms.add( orig ) ;
         Collections.sort( monAtoms, Atom.originComparator ) ;
         for( Atom atom : link.atoms ){
            int i = Collections.binarySearch( monAtoms, atom, Atom.originComparator ) ;
            Atom orig = null ;
            if( i >= 0 ) orig = monAtoms.remove( i ) ;
            if( orig == null ){
               loop.add( rsd.key8 ) ;
               loop.add( "add" ) ;
               loop.add( null ) ;
               atomMethods.toCif( loop, atom, true ) ;
            }
            else{
               if( atomMethods.diff( orig, atom ) ){
                  loop.add( rsd.key8 ) ;
                  loop.add( "change" ) ;
                  loop.add( orig.key4 ) ;
                  atomMethods.diffToCif( loop, orig, atom ) ;
               }
            }
         }
         for( Atom orig : monAtoms ){
            loop.add( rsd.key8 ) ;
            loop.add( "delete" ) ;
            loop.add( orig.key4 ) ;
            Atom atom = new Atom( rsd ) ;
            atomMethods.toCif( loop, atom, true ) ;
         }

         RestraintAtomWriter raw = new RestraintAtomWriter(){
            void writeFun( CifLoop loop ){
               loop.add( rsd.key8 ) ;
               loop.add( function ) ;
            }
            void writeAtom( Atom oa, CifLoop loop ){
               loop.add( oa.key4 ) ;
            }
         } ;

         for( int i = 0 ; i < 5 ; i++ ){
            loop = dataIterator.next() ;

            if( i != 2 || Env.keepTor ){

               List<GenericRestraint> listOld = rsd.mono.restraints.get(i) ;
               List<GenericRestraint> list = link.restraints.get(i) ;
               CifRestraint template = rstTemplates.get(i) ;

               for( GenericRestraint rstOld : listOld ){
                  int j = list.indexOf( rstOld ) ;
                  if( j < 0 ){
                     raw.function = "delete" ;
                     GenericRestraint r = rstOld.asDeleted() ;
                     template.toCif( loop, raw, r ) ;
                  }
                  else{
                     GenericRestraint rst = list.remove(j) ;
                     if( rst.getResidue() != rsd ) throw new NullPointerException() ;
                     if( rstOld.diff( rst ) ){
                        raw.function = "change" ;
                        GenericRestraint r = rstOld.deriveDiff( rst ) ;
                        template.toCif( loop, raw, r ) ;
                     }
                  }
               }
               for( GenericRestraint rst : list ){
                  if( rst.getResidue() != rsd ) throw new NullPointerException() ;
                  int j = listOld.indexOf( rst ) ;
                  if( j < 0 ){
                     raw.function = "add" ;
                     template.toCif( loop, raw, rst ) ;
                  }
                  else throw new NullPointerException() ;
               }
            }
         }
         if( data.isVoid() ) return null ;

         header.add( rsd.key8 ) ;
         header.add( rsd.name ) ;
         header.add( rsd.mono.key8 ) ;
//xx     header.add( rsd.group ) ;
         header.add( null ) ;
//xx
         return rsd.key8 ;
      }
   }
   // --------------------------------------------------------------------------

   private class CifLinkReader extends CifEntryReader{

      void fromCif( final String key, CifLoop header, CifData data ){
         if( linkMap.containsKey( key ) ) throw new NullPointerException() ;

         Molecule link = new Molecule() ;
         link.key8 = key ;

         final List<AtomMap> atomMaps = new ArrayList<AtomMap>(2) ;
         List<Restraints> modRestraints = new ArrayList<Restraints>() ;

         for( int i = 0 ; i < 2 ; i++ ){
            String monoKey = (String)header.next() ;
            String modKey = (String)header.next() ;

            if( usedModSet.contains( modKey ) ) throw new NullPointerException() ;
            usedModSet.add( modKey ) ;
            Molecule lnk = modMap.get( modKey ) ;
            if( lnk == null ) throw new NullPointerException() ;

            Residue rsd = lnk.atoms.get(0).residue ;
            rsd.isFirst = i == 0 ;
            for( Atom atom : lnk.atoms ) if( atom.residue != rsd ) throw new NullPointerException() ;
            if( ! monoKey.equals( rsd.mono.key8 ) ) throw new NullPointerException() ;
            String s = (String)header.next() ;
            if( s != null && ! s.equals( rsd.group ) ) throw new NullPointerException() ;

            link.atoms.addAll( lnk.atoms ) ;
            atomMaps.add( new AtomMap( lnk.atoms ) ) ;
            modRestraints.add( lnk.restraints ) ;
         }
         link.name = (String)header.next() ;

         RestraintAtomReader rar = new RestraintAtomReader(){
            void readFun( CifLoop loop ){
               if( ! key.equals( (String)loop.next() ) ) throw new NullPointerException() ;
            }
            Atom readAtom( CifLoop loop, boolean flagChir ){
               Object o1 = loop.next() ;
               Object o2 = loop.next() ;
               Atom atom = null ;
               if( o1 != null && o2 != null ){
                  Integer i = (Integer)o1 ;
                  atom = atomMaps.get( i - 1 ).get( o2 ) ;
               }
               else if( o1 == null && o2 == null && flagChir ){
                  atom = Atom.NULL ;
               }
               if( atom == null ) throw new NullPointerException() ;
               return atom ;
            }
         } ;

         link.restraints = new Restraints() ;
         Iterator<CifLoop> dataIterator = data.iterator() ;
         for( int i = 0 ; i < 5 ; i++ ){
            CifLoop loop = dataIterator.next() ;
            List<GenericRestraint> list = link.restraints.get(i) ;
            CifRestraint template = rstTemplates.get(i) ;
            while( loop.hasNext() ) list.add( template.fromCif( loop, rar ) ) ;
         }
         for( GenericRestraint rst : link.restraints.get(0) ){
            Bond bond = new Bond( rst ) ;
            bond.atom1.add( bond ) ;
            bond.atom2.add( bond ) ;
         }
         for( GenericRestraint chir : link.restraints.get(3) ){
            chir.atoms.get(0).chir = chir ;
         }

         for( Restraints restraints : modRestraints ){
            for( int j = 0 ; j < 5 ; j++ ){
               link.restraints.get(j).addAll( restraints.get(j) ) ;
            }
         }

         linkMap.put( key, link ) ;
      }
   }
   // --------------------------------------------------------------------------

   private class CifLinkWriter extends CifEntryWriter{
      Iterator<Molecule> iterator ;
      CifLinkWriter(){ iterator = linkMap.values().iterator() ; }

      boolean hasNext(){ return iterator.hasNext() ; }

      String toCif( CifLoop header, CifData data ){
         final Molecule link = iterator.next() ;
         final List<Residue> rsds = link.residueList() ;

         header.add( link.key8 ) ;
         for( Residue rsd : rsds ){
            header.add( rsd.mono.key8 ) ;
            header.add( rsd.key8 ) ;
//xx        header.add( rsd.group ) ;
            header.add( null ) ;
//xx
         }
         header.add( link.name ) ;

         RestraintAtomWriter raw = new RestraintAtomWriter(){
            void writeFun( CifLoop loop ){
               loop.add( link.key8 ) ;
            }
            void writeAtom( Atom oa, CifLoop loop ){
               if( oa == Atom.NULL ){
                  loop.add( null ) ;
                  loop.add( null ) ;
               }
               else{
                  loop.add( new Integer( 1 + rsds.indexOf( oa.residue ) ) ) ;
                  loop.add( oa.key4 ) ;
               }
            }
         } ;

         Iterator<CifLoop> dataIterator = data.iterator() ;

// Oh-oh-oh
// Special treatment of bond: must be: "1 SMTH 2 SMTH", otherwise refmac crashes.

         for( GenericRestraint rst : link.restraints.get(0) ){
            if( rst.atoms.size() != 2 ) throw new NullPointerException() ;
            int i0 = rsds.indexOf( rst.atoms.get(0).residue ) ;
            int i1 = rsds.indexOf( rst.atoms.get(1).residue ) ;
            if( i1 < i0 ){
               rst.atoms.add( rst.atoms.remove(0) ) ;
               rst.reordered = true ;
            }
         }

// end of Oh-oh-oh

         for( int i = 0 ; i < 5 ; i++ ){
            CifLoop loop = dataIterator.next() ;

            if( i != 2 || Env.keepTor ){

               List<GenericRestraint> list = link.restraints.get(i) ;
               CifRestraint template = rstTemplates.get(i) ;
               for( GenericRestraint rst : list ){
                  if( rst.getResidue() == null ) template.toCif( loop, raw, rst ) ;
               }
            }
         }

// Oh-oh-oh continued
// Restoring original order, just to make sure nothing goes wrong

         for( GenericRestraint rst : link.restraints.get(0) ){
            if( rst.reordered ){
               rst.atoms.add( rst.atoms.remove(0) ) ;
               rst.reordered = false ;
            }
         }

// end of Oh-oh-oh

         return link.key8 ;
      }
   }
   // --------------------------------------------------------------------------

   @SuppressWarnings("serial")
   private class AtomMap extends LinkedHashMap<String,Atom>{
      AtomMap(){}
      AtomMap( List<Atom> list ){
         super( list.size() ) ;
         for( Atom atom : list ) put( atom.key4, atom ) ;
      }
      public Atom put( String key, Atom atom ){
         if( super.put( key, atom ) != null ) throw new NullPointerException() ;
         return null ;
      }
   }
   // --------------------------------------------------------------------------

   private class AtomMethods{

      void diffFromCif( CifLoop loop, Atom ao, Atom an ){
         Object o ;
         an.key4   = ( o = loop.next() ) == null ? ao.key4   : (String)o ;
         an.symbol = ( o = loop.next() ) == null ? ao.symbol : (String)o ;
         an.energy = ( o = loop.next() ) == null ? ao.energy : (String)o ;
         an.charge = ( o = loop.next() ) == null ? ao.charge : (Float )o ;
         String s = an.symbol ;
         if( s != null ) an.symbol = s.substring(0,1) + s.substring(1).toLowerCase() ;
      }

      boolean diff( Atom ao, Atom an ){
         boolean equal = true ;
         equal &= an.key4   == null ? ao.key4   == null : an.key4  .equals( ao.key4   ) ;
         equal &= an.symbol == null ? ao.symbol == null : an.symbol.equals( ao.symbol ) ;
         equal &= an.energy == null ? ao.energy == null : an.energy.equals( ao.energy ) ;
         equal &= an.charge == null ? ao.charge == null : an.charge.equals( ao.charge ) ;
         return ! equal ;
      }

      void diffToCif( CifLoop loop, Atom ao, Atom an ){
         loop.add( an.key4   != null && an.key4  .equals( ao.key4   ) ? null : an.key4   ) ;
         loop.add( an.symbol != null && an.symbol.equals( ao.symbol ) ? null : an.symbol ) ;
         loop.add( an.energy != null && an.energy.equals( ao.energy ) ? null : an.energy ) ;
         loop.add( an.charge != null && an.charge.equals( ao.charge ) ? null : an.charge ) ;
      }

      void fromCif( CifLoop loop, Atom an, boolean modification ){
         an.key4   = (String)loop.next() ;
         an.symbol = (String)loop.next() ;
         an.energy = (String)loop.next() ;
         an.charge = (Float )loop.next() ;
         String s = an.symbol ;
         if( s != null ) an.symbol = s.substring(0,1) + s.substring(1).toLowerCase() ;
         if( modification ) return ;

         Float x = (Float)loop.next() ;
         Float y = (Float)loop.next() ;
         Float z = (Float)loop.next() ;
         if( x != null && y != null && z != null ){
            an.xyz[0] = x.doubleValue() ; an.xyz[1] = y.doubleValue() ; an.xyz[2] = z.doubleValue() ;
         }
      }

      void toCif( CifLoop loop, Atom an, boolean modification ){
         loop.add( an.key4 ) ;
         loop.add( an.symbol ) ;
         loop.add( an.energy ) ;
         loop.add( an.charge ) ;
         if( modification ) return ;

         if( an.xyz == null ){               //:: to pass to libcheck for regularisation
            loop.add( (Float)null ) ;
            loop.add( (Float)null ) ;
            loop.add( (Float)null ) ;
            return ;
         }
         loop.add( new Float( an.xyz[0] ) ) ;
         loop.add( new Float( an.xyz[1] ) ) ;
         loop.add( new Float( an.xyz[2] ) ) ;
      }
   }
   // --------------------------------------------------------------------------

   private abstract class CifRestraint{
      abstract GenericRestraint fromCif( CifLoop loop, RestraintAtomReader rwr ) ;
      abstract void toCif( CifLoop loop, RestraintAtomWriter rwr, GenericRestraint rst ) ;
   }
   // --------------------------------------------------------------------------

   private abstract class RestraintAtomReader{
      String function = "" ;
      abstract void readFun( CifLoop loop ) ;
      abstract Atom readAtom( CifLoop loop, boolean flagChir ) ;
   }
   // --------------------------------------------------------------------------

   private abstract class RestraintAtomWriter{
      String function = "" ;
      abstract void writeFun( CifLoop loop ) ;
      abstract void writeAtom( Atom atom, CifLoop loop ) ;
   }
   // --------------------------------------------------------------------------

   private class CifBond extends CifRestraint{

      GenericRestraint fromCif( CifLoop loop, RestraintAtomReader rwr ){
         rwr.readFun( loop ) ;
         GenericRestraint rst = new GenericRestraint( 2 ) ;
         for( int i = 0 ; i < 2 ; i++ ) rst.atoms.add( rwr.readAtom( loop, false ) ) ;
         rst.type  = (String) loop.next() ;
         if( rst.type != null && "aromatic".startsWith( rst.type ) ) rst.type = "aromatic" ;
         Object o = loop.next() ;
         rst.setValue( o, loop.next() ) ;

         rst.sorter = GenericRestraint.sortBond ;
         rst.sort() ;
         return rst ;
      }

      void toCif( CifLoop loop, RestraintAtomWriter rwr, GenericRestraint rst ){
         rwr.writeFun( loop ) ;
         for( Atom atom : rst.atoms ) rwr.writeAtom( atom, loop ) ;
         loop.add( rst.type  ) ;
         if( rst.value == null ){
            loop.add( null ) ;
            loop.add( null ) ;
         }
         else{
            loop.add( rst.value.mean  ) ;
            loop.add( rst.value.sd    ) ;
         }
      }
   }
   // --------------------------------------------------------------------------

   private class CifAngle extends CifRestraint{

      GenericRestraint fromCif( CifLoop loop, RestraintAtomReader rwr ){
         rwr.readFun( loop ) ;
         GenericRestraint rst = new GenericRestraint( 3 ) ;
         for( int i = 0 ; i < 3 ; i++ ) rst.atoms.add( rwr.readAtom( loop, false ) ) ;
         Object o = loop.next() ;
         rst.setValue( o, loop.next() ) ;

         rst.sorter = GenericRestraint.sortAngle ;
         rst.sort() ;
         return rst ;
      }

      void toCif( CifLoop loop, RestraintAtomWriter rwr, GenericRestraint rst ){
         rwr.writeFun( loop ) ;
         for( Atom atom : rst.atoms ) rwr.writeAtom( atom, loop ) ;
         if( rst.value == null ){
            loop.add( null ) ;
            loop.add( null ) ;
         }
         else{
            loop.add( rst.value.mean  ) ;
            loop.add( rst.value.sd    ) ;
         }
      }
   }
   // --------------------------------------------------------------------------

   private class CifTor extends CifRestraint{

      GenericRestraint fromCif( CifLoop loop, RestraintAtomReader rwr ){
         rwr.readFun( loop ) ;
         GenericRestraint rst = new GenericRestraint( 4 ) ;
         rst.key = (String) loop.next() ;
         for( int i = 0 ; i < 4 ; i++ ) rst.atoms.add( rwr.readAtom( loop, false ) ) ;
         Object o = loop.next() ;
         rst.setValue( o, loop.next() ) ;
         rst.period = (Integer)loop.next() ;

         rst.sorter = GenericRestraint.sortTor ;
         rst.sort() ;
         return rst ;
      }

      void toCif( CifLoop loop, RestraintAtomWriter rwr, GenericRestraint rst ){
         rwr.writeFun( loop ) ;
         loop.add( rst.key    ) ;
         for( Atom atom : rst.atoms ) rwr.writeAtom( atom, loop ) ;
         if( rst.value == null ){
            loop.add( null ) ;
            loop.add( null ) ;
         }
         else{
            loop.add( rst.value.mean  ) ;
            loop.add( rst.value.sd    ) ;
         }
         loop.add( rst.period ) ;
      }
   }
   // --------------------------------------------------------------------------

   private class CifChir extends CifRestraint{

      GenericRestraint fromCif( CifLoop loop, RestraintAtomReader rwr ){
         rwr.readFun( loop ) ;
         GenericRestraint rst = new GenericRestraint( 9 ) ;
         rst.key = (String) loop.next() ;
         for( int i = 0 ; i < 4 ; i++ ) rst.atoms.add( rwr.readAtom( loop, true ) ) ;
         rst.type = (String) loop.next() ;
         for( int i = 0 ; i < 5 ; i++ ) rst.atoms.add( rwr.readAtom( loop, true ) ) ;

         String chirType = GenericRestraint.getChirType( rst.type ) ;
         if( ! chirType.equals( GenericRestraint.VOID_TYPE ) ){
            rst.type = chirType ;
            rst.sorter = GenericRestraint.sortChir ;
         }
         else if( rst.type.equals( "cross0" ) ) rst.sorter = GenericRestraint.sortCross0 ;
         else if( rst.type.equals( "cross1" ) ) rst.sorter = GenericRestraint.sortCross1 ;
         else if( rst.type.equals( "cross2" ) ) rst.sorter = GenericRestraint.sortCross2 ;
         else if( rst.type.equals( "cross3" ) ) rst.sorter = GenericRestraint.sortCross3 ;
         else if( rst.type.equals( "cross4" ) ) rst.sorter = GenericRestraint.sortCross4 ;
         else if( rst.type.equals( "cross5" ) ) rst.sorter = GenericRestraint.sortCross5 ;
         else if( rst.type.equals( "cross6" ) ) rst.sorter = GenericRestraint.sortCross6 ;
         else throw new NullPointerException( "rst.type " + rst.type ) ;
         if( rst.type.equals( "cross1" ) ) if( Env.vbPrint ) System.err.println( "WARNING: chir cross1" ) ;
         if( rst.type.equals( "cross5" ) ) if( Env.vbPrint ) System.err.println( "WARNING: chir cross5" ) ;
         rst.sort() ;
         rst.initBothAndSign() ;
         return rst ;
      }

      void toCif( CifLoop loop, RestraintAtomWriter rwr, GenericRestraint rst ){
         rwr.writeFun( loop ) ;
         loop.add( rst.key ) ;
         for( int i = 0 ; i < 4 ; i++ ) rwr.writeAtom( rst.atoms.get( i ), loop ) ;
         loop.add( rst.type ) ;
         for( int i = 4 ; i < 9 ; i++ ) rwr.writeAtom( rst.atoms.get( i ), loop ) ;
      }
   }
   // --------------------------------------------------------------------------

   private class CifPlane extends CifRestraint{
      String key_new ;
      String fun_new ;

      GenericRestraint fromCif( CifLoop loop, RestraintAtomReader rwr ){
         if( key_new == null ){
            rwr.readFun( loop ) ;
            fun_new = rwr.function ;
            key_new = (String) loop.next() ;
         }
         GenericRestraint rst = new GenericRestraint( 6 ) ;
         String fun = fun_new ;
         String key = key_new ;
         Float sd = new Float(0) ;
         while( fun.equals( fun_new ) && key.equals( key_new ) ){

            Atom atom = rwr.readAtom( loop, false ) ;
            if( rst.atoms.contains( atom ) ){
               if( Env.vbPrint ) System.err.println( "WARNING: duplicated atom in " + key ) ;
//             throw new NullPointerException( "WARNING: duplicated atom in" + key ) ;
            }
            else{
               rst.atoms.add( atom ) ;
            }

            Object o = loop.next() ;
            if( o == null ) sd = null ;
//::        if( sd != null ) sd += (Float)o ;
            if( sd != null ) sd = Math.max( sd, (Float)o ) ;
//::

            if( ! loop.hasNextNoCycle() ) break ;              //::  a proper Iterator<Object> is needed

            rwr.readFun( loop ) ;
            fun_new = rwr.function ;
            key_new = (String) loop.next() ;
         }
         if( ! loop.hasNextNoCycle() ) key_new = fun_new = null ;
         if( key.equals( key_new ) ){
            if( Env.vbPrint ) System.err.println( "WARNING: CifPlane: identical key_new " + key_new ) ;
         }
         rwr.function = fun ;
         if( rst.atoms.size() < 4 ){
            if( Env.vbPrint ) System.err.println( "WARNING: plane size < 4" ) ;
//          throw new NullPointerException( "WARNING: plane size < 4" ) ;
         }

         rst.key = key ;
//::     rst.sigma = sd == null ? null : new Sigma( sd/ rst.atoms.size() ) ;
         rst.sigma = sd == null ? null : new Sigma( sd ) ;
//::

         rst.sorter = GenericRestraint.sortPlane ;
         rst.sort() ;
         return rst ;
      }

      void toCif( CifLoop loop, RestraintAtomWriter rwr, GenericRestraint rst ){
         for( Atom atom : rst.atoms ){
            rwr.writeFun( loop ) ;
            loop.add( rst.key ) ;
            rwr.writeAtom( atom, loop ) ;
            if( rst.sigma == null ){
               loop.add( null ) ;
            }
            else{
               loop.add( rst.sigma.sd    ) ;
            }
         }
      }
   }
   // --------------------------------------------------------------------------
}
// -----------------------------------------------------------------------------

